//========================================================================================
//  PPI.java
//    en:8255 PPI -- It emulates joystick ports.
//    ja:8255 PPI -- ジョイスティックポートをエミュレートします。
//  Copyright (C) 2003-2019 Makoto Kamada
//
//  This file is part of the XEiJ (X68000 Emulator in Java).
//  You can use, modify and redistribute the XEiJ if the conditions are met.
//  Read the XEiJ License for more details.
//  https://stdkmd.net/xeij/
//========================================================================================

package xeij;

import java.awt.*;  //BasicStroke,BorderLayout,BoxLayout,Color,Component,Container,Cursor,Desktop,Dimension,Font,FlowLayout,Frame,Graphics,Graphics2D,GraphicsDevice,GraphicsEnvironment,GridLayout,Image,Insets,Paint,Point,Rectangle,RenderingHints,Robot,Shape,Stroke,TexturePaint,Toolkit
import java.awt.event.*;  //ActionEvent,ActionListener,ComponentAdapter,ComponentEvent,ComponentListener,FocusAdapter,FocusEvent,FocusListener,InputEvent,KeyAdapter,KeyEvent,KeyListener,MouseAdapter,MouseEvent,MouseListener,MouseMotionAdapter,MouseWheelEvent,WindowAdapter,WindowEvent,WindowListener,WindowStateListener
import java.awt.image.*;  //BufferedImage,DataBuffer,DataBufferByte,DataBufferInt,IndexColorModel
import java.util.*;  //ArrayList,Arrays,Calendar,GregorianCalendar,HashMap,Map,Map.Entry,LinkedList,TimeZone,Timer,TimerTask,TreeMap
import javax.swing.*;  //AbstractButton,AbstractSpinnerModel,Box,ButtonGroup,DefaultListModel,ImageIcon,JApplet,JButton,JCheckBox,JCheckBoxMenuItem,JComponent,JDialog,JFileChooser,JFrame,JLabel,JList,JMenu,JMenuBar,JMenuItem,JPanel,JRadioButton,JScrollPane,JSpinner,JTextArea,JTextField,JTextPane,JViewport,ScrollPaneConstants,SpinnerListModel,SpinnerNumberModel,SwingConstants,SwingUtilities,UIManager,UIDefaults,UnsupportedLookAndFeelException
import javax.swing.event.*;  //CaretEvent,CaretListener,ChangeEvent,ChangeListener,DocumentEvent,DocumentListener,ListSelectionListener

public class PPI {

  //ポート
  //  0x00e9a000-0x00e9bfffは下位3ビットだけデコードされる
  //  偶数アドレスをバイトサイズでリード/ライトするとバスエラーになる
  //  偶数アドレスをワードサイズでリードすると上位バイトは0xffになる
  //
  //  0x00e9a001  PPIポートA  ジョイスティック1
  //         7        6        5        4        3        2        1        0
  //    +--------+--------+--------+--------+--------+--------+--------+--------+
  //    |   PA7  |   PA6  |   PA5  |   PA4  |   PA3  |   PA2  |   PA1  |   PA0  |
  //    |    1   |   JS7  |   JS6  |    1   |   JS4  |   JS3  |   JS2  |   JS1  |
  //    |        |    B   |    A   |        |   →   |   ←   |   ↓   |   ↑   |
  //    +--------+--------+--------+--------+--------+--------+--------+--------+
  //    bit7    PA7  1
  //    bit6    PA6  JS7入力。0=Bが押されている
  //    bit5    PA5  JS6入力。0=Aが押されている
  //    bit4    PA4  1
  //    bit3    PA3  JS4入力。0=→が押されている
  //    bit2    PA2  JS3入力。0=←が押されている
  //    bit1    PA1  JS2入力。0=↓が押されている
  //    bit0    PA0  JS1入力。0=↑が押されている
  //
  //  0x00e9a003  PPIポートB  ジョイスティック2
  //         7        6        5        4        3        2        1        0
  //    +--------+--------+--------+--------+--------+--------+--------+--------+
  //    |   PB7  |   PB6  |   PB5  |   PB4  |   PB3  |   PB2  |   PB1  |   PB0  |
  //    |    1   |   JT7  |   JT6  |    1   |   JT4  |   JT3  |   JT2  |   JT1  |
  //    |        |    B   |    A   |        |   →   |   ←   |   ↓   |   ↑   |
  //    +--------+--------+--------+--------+--------+--------+--------+--------+
  //    bit7    PB7  1
  //    bit6    PB6  JT7入力。0=Bが押されている
  //    bit5    PB5  JT6入力。0=Aが押されている
  //    bit4    PB4  1
  //    bit3    PB3  JT4入力。0=→が押されている
  //    bit2    PB2  JT3入力。0=←が押されている
  //    bit1    PB1  JT2入力。0=↓が押されている
  //    bit0    PB0  JT1入力。0=↑が押されている
  //
  //  0x00e9a005  PPIポートC  ADPCMコントロール。初期値は0x0b
  //         7        6        5        4        3        2        1        0
  //    +--------+--------+--------+--------+--------+--------+--------+--------+
  //    |   PC7  |   PC6  |   PC5  |   PC4  |   PC3  |   PC2  |   PC1  |   PC0  |
  //    |   JS7  |   JS6  |   JT8  |   JS8  |      RATIO      |  LEFT  |  RIGHT |
  //    +--------+--------+--------+--------+--------+--------+--------+--------+
  //    bit7    PC7    JS7出力(負論理)
  //    bit6    PC6    JS6出力(負論理)
  //    bit5    PC5    JT8出力
  //    bit4    PC4    JS8出力
  //    bit3-2  PC3-2  ADPCMの分周比。00=1/1024,01=1/768,10=1/512,11=1/512
  //    bit1    PC1    ADPCMの出力left。0=出力する,1=出力しない
  //    bit0    PC0    ADPCMの出力right。0=出力する,1=出力しない
  //
  //  0x00e9a007  PPIコントロール
  //    bit7=0  ポートCで出力に設定されているビットの操作
  //      bit3-1  ビット番号
  //      bit0    設定値
  //    bit7=1  モードの設定。0x92に固定
  //      bit6-5  グループA(ポートAとポートCの上位)のモード(0=モード0,1=モード1,2/3=モード2)。モード0に固定
  //      bit4    ポートAの方向(0=出力,1=入力)。入力に固定
  //      bit3    ポートCの上位の方向(0=出力,1=入力)。出力に固定
  //      bit2    グループB(ポートBとポートCの下位)のモード(0=モード0,1=モード1)。モード0に固定
  //      bit1    ポートBの方向(0=出力,1=入力)。入力に固定
  //      bit0    ポートCの下位の方向(0=出力,1=入力)。出力に固定
  //
  //  ボタンのマスク
  //    上と下または左と右のキーが同時に押された場合は両方キャンセルする
  //    SELECTボタンは上下同時押し、RUNボタンは左右同時押しに割り当てられる
  //
  //  bit4とbit7は入力できない
  //    8255の足がプルアップされているので実機ではどうすることもできない(外付けの回路でどうにかなるものではない)
  //    エミュレータで入力できるようにするのは簡単だか対応しているソフトは存在しないだろう
  //
  //  参考
  //    電脳倶楽部67号  B/MDPAD/M6PAD_AN.DOC
  //    電脳倶楽部77号  B/6B/T_EXPAD.DOC
  //
  protected static final int PPI_PORT_A  = 0x00e9a001;   //PPIポートA
  protected static final int PPI_PORT_B  = 0x00e9a003;   //PPIポートB
  protected static final int PPI_PORT_C  = 0x00e9a005;   //PPIポートC
  protected static final int PPI_CONTROL = 0x00e9a007;   //PPIコントロール

  //ジョイスティック
  protected static Joystick[] ppiJoysticks;
  protected static Joystick ppiJoystick1;
  protected static Joystick ppiJoystick2;

  //モード
  public static boolean ppiJoyKey;  //true=キーボード入力をジョイスティック入力とみなす
  public static boolean ppiJoyAuto;  //true=ジョイスティックポートが連続的に読み出されている間だけ有効
  public static boolean ppiJoyBlock;  //true=ジョイスティック入力が有効なときはキーボード入力を取り除く

  //自動切り替え
  //  PPIのポートが参照されてから一定時間だけJOYKEYを有効にする
  protected static final long PPI_CONTINUOUS_ACCESS_SPAN = XEiJ.TMR_FREQ / 10;  //自動切り替えの有効期間(TMR_FREQ単位)。1/10秒
  protected static long ppiLastAccessTime;  //最後にPPIのポートがアクセスされた時刻

  //ポートの値
  protected static int ppiPortCData;

  //ウインドウ
  protected static JFrame ppiFrame;

  protected static JScrollPane ppiConfigurationScrollPane;

  //ppiInit ()
  //  PPIを初期化する
  public static void ppiInit () {
    ppiJoyKey = Settings.sgsGetOnOff ("joykey");  //true
    ppiJoyAuto = Settings.sgsGetOnOff ("joyauto");  //true
    ppiJoyBlock = Settings.sgsGetOnOff ("joyblock");  //true

    ppiJoysticks = new Joystick[] {
      new DummyPad (),
      new Normal2ButtonPad (1),
      new Normal2ButtonPad (2),
      new MegaDrive3ButtonPad (1),
      new MegaDrive3ButtonPad (2),
      new MegaDrive6ButtonPad (1),
      new MegaDrive6ButtonPad (2),
      new Shiromadokun (1),
      new Shiromadokun (2),
    };

    String id1 = Settings.sgsGetString ("joystick1");  //ジョイスティックポート1に接続するデバイス
    ppiJoystick1 = ppiJoysticks[0];  //DummyPad
    for (Joystick joystick : ppiJoysticks) {
      if (joystick.getId ().equals (id1)) {
        ppiJoystick1 = joystick;
      }
    }

    String id2 = Settings.sgsGetString ("joystick2");  //ジョイスティックポート2に接続するデバイス
    ppiJoystick2 = ppiJoysticks[0];  //DummyPad
    for (Joystick joystick : ppiJoysticks) {
      if (joystick.getId ().equals (id2)) {
        ppiJoystick2 = joystick;
      }
    }

    ppiReset ();
  }

  public static void ppiTini () {
    for (Joystick joystick : ppiJoysticks) {
      joystick.tini ();
    }
    Settings.sgsPutString ("joystick1", ppiJoystick1.getId ());
    Settings.sgsPutString ("joystick2", ppiJoystick2.getId ());
  }

  //リセット
  public static void ppiReset () {
    ppiPortCData = 0;
    ppiLastAccessTime = 0L;
  }

  //ppiStart ()
  public static void ppiStart () {
    if (RestorableFrame.rfmGetOpened (Settings.SGS_PPI_FRAME_KEY)) {
      ppiOpen ();
    }
  }

  //ppiOpen ()
  //  ジョイスティックの設定ウインドウを開く
  public static void ppiOpen () {
    if (ppiFrame == null) {
      ppiMakeFrame ();
    }
    ppiFrame.setVisible (true);
  }

  //ppiMakeFrame ()
  //  ジョイスティックポートの設定ウインドウを作る
  //  ここでは開かない
  public static void ppiMakeFrame () {

    //アクションリスナー
    ActionListener actionListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        String command = ae.getActionCommand ();
        switch (command) {
        case "Use part of keyboard as a joystick":
          //キーボードの一部をジョイスティックとして使う
          ppiJoyKey = ((JCheckBox) ae.getSource ()).isSelected ();
          break;
        case "Use joysticks only while the port is read continuously":
          //ポートが繰り返し読み出されている間だけジョイスティックを使う
          ppiJoyAuto = ((JCheckBox) ae.getSource ()).isSelected ();
          break;
        case "Remove key input data processed as a joystick operation":
          //ジョイスティック操作として処理されたキー入力データを取り除く
          ppiJoyBlock = ((JCheckBox) ae.getSource ()).isSelected ();
          break;
        }
      }
    };

    //ジョイスティックポートのメニュー
    //       0   1  2
    //   0  ポート  接続
    //   1   1   2
    //   2  -------------------------
    //   3  RB  RB  なし
    //   4  RB  RB  ノーマル2ボタンパッド#1
    //   5  RB  RB  ノーマル2ボタンパッド#2
    //   6  RB  RB  メガドラ3ボタンパッド#1
    //   7  RB  RB  メガドラ3ボタンパッド#2
    //   8  RB  RB  メガドラ6ボタンパッド#1
    //   9  RB  RB  メガドラ6ボタンパッド#2
    //  10  RB  RB  白窓君#1
    //  11  RB  RB  白窓君#2
    ButtonGroup port1Group = new ButtonGroup ();
    ButtonGroup port2Group = new ButtonGroup ();
    ActionListener port1Listener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Joystick joyStick = ppiJoysticks[Integer.parseInt (ae.getActionCommand ())];
        if (ppiJoystick1 != joyStick) {
          ppiJoystick1.reset ();
          ppiJoystick1 = joyStick;
        }
        ppiConfigurationScrollPane.setViewportView (joyStick.getConfigurationPanel ());
      }
    };
    ActionListener port2Listener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Joystick joyStick = ppiJoysticks[Integer.parseInt (ae.getActionCommand ())];
        if (ppiJoystick2 != joyStick) {
          ppiJoystick2.reset ();
          ppiJoystick2 = joyStick;
        }
        ppiConfigurationScrollPane.setViewportView (joyStick.getConfigurationPanel ());
      }
    };
    ArrayList<Object> cellList = new ArrayList<Object> ();
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Port"), "ja", "ポート"));  //(0,0)-(1,0)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Connect to"), "ja", "接続"));  //(2,0)-(2,1)
    cellList.add ("1");  //(0,1)
    cellList.add ("2");  //(1,1)
    cellList.add (ComponentFactory.createHorizontalSeparator ());  //(0,2)-(2,2)
    for (int i = 0; i < ppiJoysticks.length; i++) {
      Joystick joyStick = ppiJoysticks[i];
      cellList.add (ComponentFactory.setText (ComponentFactory.createRadioButton (
        port1Group, joyStick == ppiJoystick1, String.valueOf (i), port1Listener), ""));  //(0,3+i)
      cellList.add (ComponentFactory.setText (ComponentFactory.createRadioButton (
        port2Group, joyStick == ppiJoystick2, String.valueOf (i), port2Listener), ""));  //(1,3+i)
      cellList.add (Multilingual.mlnText (ComponentFactory.createLabel (joyStick.getNameEn ()), "ja", joyStick.getNameJa ()));  //(3,3)
    }
    JScrollPane portMenuPanel = new JScrollPane (
      ComponentFactory.createGridPanel (
        3, 3 + ppiJoysticks.length,
        "paddingLeft=3,paddingRight=3,center",   //gridStyles
        "",  //colStyles
        "italic;italic;colSpan=3,widen",  //rowStyles
        "colSpan=2;rowSpan=2",  //cellStyles
        cellList.toArray (new Object[0])
        )
      );

    //個々のジョイスティックの設定パネル
    ppiConfigurationScrollPane = new JScrollPane (ppiJoystick1.getConfigurationPanel ());

    //ウインドウ
    ppiFrame = Multilingual.mlnTitle (
      ComponentFactory.createRestorableSubFrame (
        Settings.SGS_PPI_FRAME_KEY,
        "Joystick port",
        null,
        ComponentFactory.setEmptyBorder (
          ComponentFactory.createVerticalBox (
            Multilingual.mlnText (
              ComponentFactory.createCheckBox (
                ppiJoyKey,
                "Use part of keyboard as a joystick",
                actionListener),
              "ja", "キーボードの一部をジョイスティックとして使う"),
            Multilingual.mlnText (
              ComponentFactory.createCheckBox (
                ppiJoyAuto,
                "Use joysticks only while the port is read continuously",
                actionListener),
              "ja", "ポートが繰り返し読み出されている間だけジョイスティックを使う"),
            Multilingual.mlnText (
              ComponentFactory.createCheckBox (
                ppiJoyBlock,
                "Remove key input data processed as a joystick operation",
                actionListener),
              "ja", "ジョイスティック操作として処理されたキー入力データを取り除く"),
            Box.createVerticalStrut (5),
            ComponentFactory.createHorizontalSplitPane (
              portMenuPanel,
              ppiConfigurationScrollPane
              )
            ),
          5, 5, 5, 5
          )
        ),
      "ja", "ジョイスティックポート");
  }  //ppiMakeFrame()

  //consume = ppiInput (ke, pressed)
  //  JOYKEYの処理
  //  consume  true=入力をキーボードに渡さない
  public static boolean ppiInput (KeyEvent ke, boolean pressed) {
    boolean consume = false;
    if (ppiJoyKey && (!ppiJoyAuto || XEiJ.mpuClockTime < ppiLastAccessTime + PPI_CONTINUOUS_ACCESS_SPAN)) {
      if (ppiJoystick1.input (ke, pressed) ||
          ppiJoystick2.input (ke, pressed)) {
        //押されたときだけキーボード入力を取り除く
        //  特に自動有効化のときは、押されている間に有効になって離されたデータだけ取り除かれると、
        //  キーボード側は押されたままになっていると判断してリピートが止まらなくなる
        consume = pressed && ppiJoyBlock;
      }
    }
    return consume;
  }

  //  リード
  //  FM音源レジスタのアクセスウエイトのためのPPIの空読みは、ジョイスティックのデータを得ることが目的ではないので、
  //  ジョイスティックポートが連続的に読み出されていると見なさない
  public static int ppiRbz (int a) {
    switch (a & 7) {
    case PPI_PORT_A & 7:
      if (XEiJ.regOC >> 6 != 0b0100_101_000) {  //TST.B以外
        ppiLastAccessTime = XEiJ.mpuClockTime;
      }
      return ppiJoystick1.rbz () & 0xff;
    case PPI_PORT_B & 7:
      if (XEiJ.regOC >> 6 != 0b0100_101_000) {  //TST.B以外
        ppiLastAccessTime = XEiJ.mpuClockTime;
      }
      return ppiJoystick2.rbz () & 0xff;
    case PPI_PORT_C & 7:
      return ppiPortCData;
    case PPI_CONTROL & 7:
      return 0xff;
    }
    return 0xff;
  }

  //  ライト
  public static void ppiWb (int a, int d) {
    switch (a & 7) {
    case PPI_PORT_A & 7:
      ppiJoystick1.wb (d);
      return;
    case PPI_PORT_B & 7:
      ppiJoystick2.wb (d);
      return;
    case PPI_PORT_C & 7:
      ppiPortCData = d &= 0xff;
      ADPCM.pcmSetPan (d);  //パン
      ADPCM.pcmDivider = d >> 2 & 3;  //分周比。0=1/1024,1=1/768,2=1/512,3=1/768
      ADPCM.pcmUpdateRepeatInterval ();
      ppiJoystick1.setPin8 (d >> 4 & 1);
      ppiJoystick2.setPin8 (d >> 5 & 1);
      ppiJoystick1.setPin6 ((d >> 6 & 1) ^ 1);
      ppiJoystick1.setPin7 ((d >> 7 & 1) ^ 1);
      return;
    case PPI_CONTROL & 7:
      if ((d & 0x80) == 0x00) {  //0...nnnx
        int n = (d >> 1) & 7;  //ビット番号
        int x = d & 1;  //データ
        ppiPortCData = ppiPortCData & ~(1 << n) | x << n;
        switch (n) {
        case 0:
        case 1:
          ADPCM.pcmSetPan (ppiPortCData & 3);  //パン
          break;
        case 2:
        case 3:
          ADPCM.pcmDivider = ppiPortCData >> 2 & 3;  //分周比。0=1/1024,1=1/768,2=1/512,3=1/768
          ADPCM.pcmUpdateRepeatInterval ();
          break;
        case 4:
          ppiJoystick1.setPin8 (x);
          break;
        case 5:
          ppiJoystick2.setPin8 (x);
          break;
        case 6:
          ppiJoystick2.setPin6 (x ^ 1);
          break;
        case 7:
          ppiJoystick2.setPin7 (x ^ 1);
          break;
        }
      }
      return;
    }

  }

}  //class PPI



//class Joystick
//  ジョイスティック
abstract class Joystick {
  protected static final int DELAY_MIN = 10;
  protected static final int DELAY_MAX = 2000;
  protected static final int DELAY_STEP = 10;
  protected static final int INTERVAL_MIN = 10;
  protected static final int INTERVAL_MAX = 2000;
  protected static final int INTERVAL_STEP = 10;
  protected String id;
  protected String nameEn;
  protected String nameJa;
  protected JComponent configurationPanel;
  public Joystick () {
    configurationPanel = null;
  }
  public JComponent getConfigurationPanel () {
    if (configurationPanel != null) {
      return configurationPanel;
    }
    return configurationPanel = new JPanel ();
  }
  public void tini () {
  }
  public void reset () {
  }
  public String getId () {
    return id;
  }
  public String getNameEn () {
    return nameEn;
  }
  public String getNameJa () {
    return nameJa;
  }
  public boolean input (KeyEvent ke, boolean pressed) {
    return false;
  }
  public void setPin6 (int pin6) {
  }
  public void setPin7 (int pin7) {
  }
  public void setPin8 (int pin8) {
  }
  public int rbz () {
    return 0xff;
  }
  public void wb (int d) {
  }
}  //class Joystick



//class DummyPad
//  ダミーパッド
//  ジョイスティックポートに何も繋がっていない
class DummyPad extends Joystick {

  public DummyPad () {
    id = "";
    nameEn = "None";
    nameJa = "なし";
  }

}  //class DummyPad



//class Normal2ButtonPad
//  ノーマル2ボタンパッド
class Normal2ButtonPad extends Joystick {

  protected static final int UP      = 0;
  protected static final int DOWN    = 1;
  protected static final int LEFT    = 2;
  protected static final int RIGHT   = 3;
  protected static final int B       = 4;
  protected static final int A       = 5;
  protected static final int SELECT  = 6;
  protected static final int RUN     = 7;
  protected static final int BUTTONS = 8;  //ボタン数

  protected static final String[]       labelOf = { "↑" , "↓" , "←" , "→" , "B" , "A" , "SELECT", "RUN" };  //ラベル
  protected static final boolean[] isRepeatable = { false, false, false, false, true, true, false   , false };  //連射可能

  protected static final int MAP_KEY      = BUTTONS * 0;  //キーコード
  protected static final int MAP_REPEAT   = BUTTONS * 1;  //連射有効
  protected static final int MAP_DELAY    = BUTTONS * 2;  //遅延時間
  protected static final int MAP_INTERVAL = BUTTONS * 3;  //連射間隔
  protected static final int MAP_LENGTH   = BUTTONS * 4;
  protected int[] map;

  protected boolean[] isPressed;  //ボタンの押し下げ状態
  protected long[] startTimeOf;  //連射開始時刻

  protected JTextField[] keyTextFieldOf = new JTextField[BUTTONS];
  protected JCheckBox[] repeatCheckBoxOf = new JCheckBox[BUTTONS];
  protected SpinnerNumberModel[] delayModelOf = new SpinnerNumberModel[BUTTONS];
  protected JSpinner[] delaySpinnerOf = new JSpinner[BUTTONS];
  protected SpinnerNumberModel[] intervalModelOf = new SpinnerNumberModel[BUTTONS];
  protected JSpinner[] intervalSpinnerOf = new JSpinner[BUTTONS];

  public Normal2ButtonPad (int number) {
    id = "normal2button" + number;
    nameEn = "Normal 2 button pad #" + number;
    nameJa = "ノーマル 2 ボタンパッド #" + number;
    if (true) {
      map = Settings.sgsGetIntArray (id, MAP_LENGTH);
    } else {
      map = (number == 1 ?
             new int[] {
               KeyEvent.VK_UP,     //↑
               KeyEvent.VK_DOWN,   //↓
               KeyEvent.VK_LEFT,   //←
               KeyEvent.VK_RIGHT,  //→
               KeyEvent.VK_Z,      //B
               KeyEvent.VK_X,      //A
               KeyEvent.VK_E,      //SELECT
               KeyEvent.VK_R,      //RUN
               //38, 40, 37, 39, 90, 88, 69, 82,  //キーコード
               0, 0, 0, 0, 0, 0, 0, 0,  //連射有効
               0, 0, 0, 0, 50, 50, 0, 0,  //遅延時間
               0, 0, 0, 0, 50, 50, 0, 0,  //連射間隔
             } :
             new int[] {
               KeyEvent.VK_UNDEFINED,  //↑
               KeyEvent.VK_UNDEFINED,  //↓
               KeyEvent.VK_UNDEFINED,  //←
               KeyEvent.VK_UNDEFINED,  //→
               KeyEvent.VK_UNDEFINED,  //B
               KeyEvent.VK_UNDEFINED,  //A
               KeyEvent.VK_UNDEFINED,  //SELECT
               KeyEvent.VK_UNDEFINED,  //RUN
               //0, 0, 0, 0, 0, 0, 0, 0,  //キーコード
               0, 0, 0, 0, 0, 0, 0, 0,  //連射有効
               0, 0, 0, 0, 50, 50, 0, 0,  //遅延時間
               0, 0, 0, 0, 50, 50, 0, 0,  //連射間隔
             });
    }
    isPressed = new boolean[BUTTONS];
    startTimeOf = new long[BUTTONS];
    reset ();
    configurationPanel = null;
  }

  @Override public void tini () {
    Settings.sgsPutIntArray (id, map);
  }

  @Override public void reset () {
    for (int i = 0; i < BUTTONS; i++) {
      isPressed[i] = false;
      startTimeOf[i] = XEiJ.FAR_FUTURE;
    }
  }

  @Override public JComponent getConfigurationPanel () {

    if (configurationPanel != null) {
      return configurationPanel;
    }

    //キーリスナー
    KeyListener keyListener = new KeyAdapter () {
      @Override public void keyPressed (KeyEvent ke) {
        int i = Integer.parseInt (((JTextField) ke.getSource ()).getName ());
        int keyCode = ke.getKeyCode ();
        if (keyCode == KeyEvent.VK_ESCAPE) {  //Escキーは解除とみなす
          keyCode = KeyEvent.VK_UNDEFINED;
        }
        map[MAP_KEY + i] = keyCode;
        keyTextFieldOf[i].setText (
          keyCode == KeyEvent.VK_UNDEFINED ? Multilingual.mlnJapanese ? "なし" : "none" : KeyEvent.getKeyText (keyCode)
          );
        ke.consume ();
      }
      @Override public void keyReleased (KeyEvent ke) {
        ke.consume ();
      }
      @Override public void keyTyped (KeyEvent ke) {
        ke.consume ();
      }
    };

    //アクションリスナー
    ActionListener actionListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        String command = ae.getActionCommand ();
        if (command.startsWith ("Repeat ")) {
          int i = Integer.parseInt (command.substring (7));
          map[MAP_REPEAT + i] = repeatCheckBoxOf[i].isSelected () ? 1 : 0;
        }
      }
    };

    //チェンジリスナー
    ChangeListener changeListener = new ChangeListener () {
      @Override public void stateChanged (ChangeEvent ce) {
        JSpinner spinner = (JSpinner) ce.getSource ();
        String name = spinner.getName ();
        if (name.startsWith ("Delay ")) {
          int i = Integer.parseInt (name.substring (6));
          map[MAP_DELAY + i] = delayModelOf[i].getNumber ().intValue ();
        } else if (name.startsWith ("Interval ")) {
          int i = Integer.parseInt (name.substring (9));
          map[MAP_INTERVAL + i] = intervalModelOf[i].getNumber ().intValue ();
        }
      }
    };

    for (int i = 0; i < BUTTONS; i++) {
      keyTextFieldOf[i] = ComponentFactory.addListener (
        ComponentFactory.setHorizontalAlignment (
          ComponentFactory.setName (
            ComponentFactory.createTextField (
              map[MAP_KEY + i] == KeyEvent.VK_UNDEFINED ? Multilingual.mlnJapanese ? "なし" : "None" : KeyEvent.getKeyText (map[MAP_KEY + i]),
              10),
            String.valueOf (i)
            ),
          JTextField.CENTER
          ),
        keyListener
        );
      if (isRepeatable[i]) {
        repeatCheckBoxOf[i] = ComponentFactory.setText (
          ComponentFactory.createCheckBox (map[MAP_REPEAT + i] != 0, "Repeat " + i, actionListener), "");
        delayModelOf[i] = new SpinnerNumberModel (Math.max (DELAY_MIN, Math.min (DELAY_MAX, map[MAP_DELAY + i])),
                                                  DELAY_MIN, DELAY_MAX, DELAY_STEP);
        delaySpinnerOf[i] = ComponentFactory.setName (
          ComponentFactory.createNumberSpinner (delayModelOf[i], 4, changeListener), "Delay " + i);
        intervalModelOf[i] = new SpinnerNumberModel (Math.max (INTERVAL_MIN, Math.min (INTERVAL_MAX, map[MAP_INTERVAL + i])),
                                                     INTERVAL_MIN, INTERVAL_MAX, INTERVAL_STEP);
        intervalSpinnerOf[i] = ComponentFactory.setName (
          ComponentFactory.createNumberSpinner (intervalModelOf[i], 4, changeListener), "Interval " + i);
      }
    }

    //          0     1     2      3        4
    //    0               Title
    //    1                 -
    //    2  Button  Key  Burst  Delay  Interval
    //    3  -----------------------------------
    //    4    ↑     TF    -      -        -
    //    5    ↓     TF    -      -        -
    //    6    ←     TF    -      -        -
    //    7    →     TF    -      -        -
    //    8     B     TF    CB     S        S
    //    9     A     TF    CB     S        S
    //   10  SELECT   TF    -      -        -
    //   11    RUN    TF    -      -        -
    ArrayList<Object> cellList = new ArrayList<Object> ();
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel (getNameEn ()), "ja", getNameJa ()));  //(0,0)
    cellList.add (null);  //(0,1)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Button"), "ja", "ボタン"));  //(0,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Key"), "ja", "キー"));  //(1,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Burst"), "ja", "連射"));  //(2,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Delay (ms)"), "ja", "開始 (ms)"));  //(3,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Interval (ms)"), "ja", "間隔 (ms)"));  //(4,2)
    cellList.add (ComponentFactory.createHorizontalSeparator ());  //(0,3)-(4,3)
    for (int i = 0; i < BUTTONS; i++) {
      cellList.add (labelOf[i]);  //(0,4+i)
      cellList.add (keyTextFieldOf[i]);  //(1,4+i)
      cellList.add (isRepeatable[i] ? repeatCheckBoxOf[i] : null);  //(2,4+i)
      cellList.add (isRepeatable[i] ? delaySpinnerOf[i] : null);  //(3,4+i)
      cellList.add (isRepeatable[i] ? intervalSpinnerOf[i] : null);  //(4,4+i)
    }
    return configurationPanel = ComponentFactory.createGridPanel (
      5, 4 + BUTTONS,
      "paddingLeft=3,paddingRight=3,center",   //gridStyles
      "",  //colStyles
      "colSpan=5;colSpan=5;italic;colSpan=5,widen",  //rowStyles
      "",  //cellStyles
      cellList.toArray (new Object[0])
      );
  }

  @Override public boolean input (KeyEvent ke, boolean pressed) {
    int keyCode = ke.getKeyCode ();
    for (int i = 0; i < BUTTONS; i++) {
      if (keyCode == map[MAP_KEY + i]) {
        if (isRepeatable[i]) {  //連射可能
          if (map[MAP_REPEAT + i] != 0 && !isPressed[i] && pressed) {  //連射有効で、押し下げられた
            startTimeOf[i] = XEiJ.mpuClockTime + map[MAP_DELAY + i] * (XEiJ.TMR_FREQ / 1000);  //連射開始時刻
          }
        }
        isPressed[i] = pressed;
        return true;
      }
    }
    return false;
  }

  @Override public int rbz () {
    int d = ((isPressed[UP] && !isPressed[DOWN] ? 0b11111110 : 0xff)  //↑が押されていて、↓が押されていない
             &
             (isPressed[DOWN] && !isPressed[UP] ? 0b11111101 : 0xff)  //↓が押されていて、↑が押されていない
             &
             (isPressed[LEFT] && !isPressed[RIGHT] ? 0b11111011 : 0xff)  //←が押されていて、→が押されていない
             &
             (isPressed[RIGHT] && !isPressed[LEFT] ? 0b11110111 : 0xff)  //→が押されていて、←が押されていない
             &
             (isPressed[A] &&  //Aが押されていて
              (map[MAP_REPEAT + A] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[A] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[A]) /
                       (map[MAP_INTERVAL + A] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b11011111
              :
              0xff)
             &
             (isPressed[B] &&  //Bが押されていて
              (map[MAP_REPEAT + B] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[B] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[B]) /
                       (map[MAP_INTERVAL + B] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b10111111
              :
              0xff)
             &
             (isPressed[SELECT] ? 0b11111100 : 0xff)  //SELECTが押されている
             &
             (isPressed[RUN] ? 0b11110011 : 0xff)  //RUNが押されている
             );
    return d;
  }

}  //class Normal2ButtonPad



//class MegaDrive3ButtonPad
//  メガドラ3ボタンパッド
class MegaDrive3ButtonPad extends Joystick {

  protected static final int UP      =  0;
  protected static final int DOWN    =  1;
  protected static final int LEFT    =  2;
  protected static final int RIGHT   =  3;
  protected static final int A       =  4;
  protected static final int B       =  5;
  protected static final int C       =  6;
  protected static final int START   =  7;
  protected static final int BUTTONS =  8;  //ボタン数

  protected static final String[]       labelOf = { "↑" , "↓" , "←" , "→" , "A" , "B" , "C" , "START" };  //ラベル
  protected static final boolean[] isRepeatable = { false, false, false, false, true, true, true, false   };  //連射可能

  protected static final int MAP_KEY      = BUTTONS * 0;  //キーコード
  protected static final int MAP_REPEAT   = BUTTONS * 1;  //連射有効
  protected static final int MAP_DELAY    = BUTTONS * 2;  //遅延時間
  protected static final int MAP_INTERVAL = BUTTONS * 3;  //連射間隔
  protected static final int MAP_LENGTH   = BUTTONS * 4;
  protected int[] map;

  protected boolean[] isPressed;  //ボタンの押し下げ状態
  protected long[] startTimeOf;  //連射開始時刻

  protected int page;  //ページ

  protected JTextField[] keyTextFieldOf = new JTextField[BUTTONS];
  protected JCheckBox[] repeatCheckBoxOf = new JCheckBox[BUTTONS];
  protected SpinnerNumberModel[] delayModelOf = new SpinnerNumberModel[BUTTONS];
  protected JSpinner[] delaySpinnerOf = new JSpinner[BUTTONS];
  protected SpinnerNumberModel[] intervalModelOf = new SpinnerNumberModel[BUTTONS];
  protected JSpinner[] intervalSpinnerOf = new JSpinner[BUTTONS];

  public MegaDrive3ButtonPad (int number) {
    id = "megadrive3button" + number;
    nameEn = "MEGA DRIVE 3 button pad #" + number;
    nameJa = "メガドラ 3 ボタンパッド #" + number;
    if (true) {
      map = Settings.sgsGetIntArray (id, MAP_LENGTH);
    } else {
      map = (number == 1 ?
             new int[] {
               KeyEvent.VK_UP,     //↑
               KeyEvent.VK_DOWN,   //↓
               KeyEvent.VK_LEFT,   //←
               KeyEvent.VK_RIGHT,  //→
               KeyEvent.VK_Z,      //A
               KeyEvent.VK_X,      //B
               KeyEvent.VK_C,      //C
               KeyEvent.VK_R,      //START
               //38, 40, 37, 39, 90, 88, 67, 82,  //キーコード
               0, 0, 0, 0, 0, 0, 0, 0,  //連射有効
               0, 0, 0, 0, 50, 50, 50, 0,  //遅延時間
               0, 0, 0, 0, 50, 50, 50, 0,  //連射間隔
             } :
             new int[] {
               KeyEvent.VK_UNDEFINED,  //↑
               KeyEvent.VK_UNDEFINED,  //↓
               KeyEvent.VK_UNDEFINED,  //←
               KeyEvent.VK_UNDEFINED,  //→
               KeyEvent.VK_UNDEFINED,  //A
               KeyEvent.VK_UNDEFINED,  //B
               KeyEvent.VK_UNDEFINED,  //C
               KeyEvent.VK_UNDEFINED,  //START
               //0, 0, 0, 0, 0, 0, 0, 0,  //キーコード
               0, 0, 0, 0, 0, 0, 0, 0,  //連射有効
               0, 0, 0, 0, 50, 50, 50, 0,  //遅延時間
               0, 0, 0, 0, 50, 50, 50, 0,  //連射間隔
             });
    }
    isPressed = new boolean[BUTTONS];
    startTimeOf = new long[BUTTONS];
    reset ();
    configurationPanel = null;
  }

  @Override public void tini () {
    Settings.sgsPutIntArray (id, map);
  }

  @Override public void reset () {
    for (int i = 0; i < BUTTONS; i++) {
      isPressed[i] = false;
      startTimeOf[i] = XEiJ.FAR_FUTURE;
    }
    page = 0;
  }

  @Override public JComponent getConfigurationPanel () {

    if (configurationPanel != null) {
      return configurationPanel;
    }

    //キーリスナー
    KeyListener keyListener = new KeyAdapter () {
      @Override public void keyPressed (KeyEvent ke) {
        int i = Integer.parseInt (((JTextField) ke.getSource ()).getName ());
        int keyCode = ke.getKeyCode ();
        if (keyCode == KeyEvent.VK_ESCAPE) {  //Escキーは解除とみなす
          keyCode = KeyEvent.VK_UNDEFINED;
        }
        map[MAP_KEY + i] = keyCode;
        keyTextFieldOf[i].setText (
          keyCode == KeyEvent.VK_UNDEFINED ? Multilingual.mlnJapanese ? "なし" : "none" : KeyEvent.getKeyText (keyCode)
          );
        ke.consume ();
      }
      @Override public void keyReleased (KeyEvent ke) {
        ke.consume ();
      }
      @Override public void keyTyped (KeyEvent ke) {
        ke.consume ();
      }
    };

    //アクションリスナー
    ActionListener actionListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        String command = ae.getActionCommand ();
        if (command.startsWith ("Repeat ")) {
          int i = Integer.parseInt (command.substring (7));
          map[MAP_REPEAT + i] = repeatCheckBoxOf[i].isSelected () ? 1 : 0;
        }
      }
    };

    //チェンジリスナー
    ChangeListener changeListener = new ChangeListener () {
      @Override public void stateChanged (ChangeEvent ce) {
        JSpinner spinner = (JSpinner) ce.getSource ();
        String name = spinner.getName ();
        if (name.startsWith ("Delay ")) {
          int i = Integer.parseInt (name.substring (6));
          map[MAP_DELAY + i] = delayModelOf[i].getNumber ().intValue ();
        } else if (name.startsWith ("Interval ")) {
          int i = Integer.parseInt (name.substring (9));
          map[MAP_INTERVAL + i] = intervalModelOf[i].getNumber ().intValue ();
        }
      }
    };

    for (int i = 0; i < BUTTONS; i++) {
      keyTextFieldOf[i] = ComponentFactory.addListener (
        ComponentFactory.setHorizontalAlignment (
          ComponentFactory.setName (
            ComponentFactory.createTextField (
              map[MAP_KEY + i] == KeyEvent.VK_UNDEFINED ? Multilingual.mlnJapanese ? "なし" : "None" : KeyEvent.getKeyText (map[MAP_KEY + i]),
              10),
            String.valueOf (i)
            ),
          JTextField.CENTER
          ),
        keyListener
        );
      if (isRepeatable[i]) {
        repeatCheckBoxOf[i] = ComponentFactory.setText (
          ComponentFactory.createCheckBox (map[MAP_REPEAT + i] != 0, "Repeat " + i, actionListener), "");
        delayModelOf[i] = new SpinnerNumberModel (Math.max (DELAY_MIN, Math.min (DELAY_MAX, map[MAP_DELAY + i])),
                                                  DELAY_MIN, DELAY_MAX, DELAY_STEP);
        delaySpinnerOf[i] = ComponentFactory.setName (
          ComponentFactory.createNumberSpinner (delayModelOf[i], 4, changeListener), "Delay " + i);
        intervalModelOf[i] = new SpinnerNumberModel (Math.max (INTERVAL_MIN, Math.min (INTERVAL_MAX, map[MAP_INTERVAL + i])),
                                                     INTERVAL_MIN, INTERVAL_MAX, INTERVAL_STEP);
        intervalSpinnerOf[i] = ComponentFactory.setName (
          ComponentFactory.createNumberSpinner (intervalModelOf[i], 4, changeListener), "Interval " + i);
      }
    }

    //          0     1     2      3        4
    //    0               Title
    //    1                 -
    //    2  Button  Key  Burst  Delay  Interval
    //    3  -----------------------------------
    //    4    ↑     TF    -      -        -
    //    5    ↓     TF    -      -        -
    //    6    ←     TF    -      -        -
    //    7    →     TF    -      -        -
    //    8     A     TF    CB     S        S
    //    9     B     TF    CB     S        S
    //   10     C     TF    CB     S        S
    //   11   START   TF    -      -        -
    ArrayList<Object> cellList = new ArrayList<Object> ();
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel (getNameEn ()), "ja", getNameJa ()));  //(0,0)
    cellList.add (null);  //(0,1)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Button"), "ja", "ボタン"));  //(0,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Key"), "ja", "キー"));  //(1,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Burst"), "ja", "連射"));  //(2,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Delay (ms)"), "ja", "開始 (ms)"));  //(3,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Interval (ms)"), "ja", "間隔 (ms)"));  //(4,2)
    cellList.add (ComponentFactory.createHorizontalSeparator ());  //(0,1)-(4,3)
    for (int i = 0; i < BUTTONS; i++) {
      cellList.add (labelOf[i]);  //(0,4+i)
      cellList.add (keyTextFieldOf[i]);  //(1,4+i)
      cellList.add (isRepeatable[i] ? repeatCheckBoxOf[i] : null);  //(2,4+i)
      cellList.add (isRepeatable[i] ? delaySpinnerOf[i] : null);  //(3,4+i)
      cellList.add (isRepeatable[i] ? intervalSpinnerOf[i] : null);  //(4,4+i)
    }
    return configurationPanel = ComponentFactory.createGridPanel (
      5, 4 + BUTTONS,
      "paddingLeft=3,paddingRight=3,center",   //gridStyles
      "",  //colStyles
      "colSpan=5;colSpan=5;italic;colSpan=5,widen",  //rowStyles
      "",  //cellStyles
      cellList.toArray (new Object[0])
      );
  }

  @Override public boolean input (KeyEvent ke, boolean pressed) {
    int keyCode = ke.getKeyCode ();
    for (int i = 0; i < BUTTONS; i++) {
      if (keyCode == map[MAP_KEY + i]) {
        if (isRepeatable[i]) {  //連射可能
          if (map[MAP_REPEAT + i] != 0 && !isPressed[i] && pressed) {  //連射有効で、押し下げられた
            startTimeOf[i] = XEiJ.mpuClockTime + map[MAP_DELAY + i] * (XEiJ.TMR_FREQ / 1000);  //連射開始時刻
          }
        }
        isPressed[i] = pressed;
        return true;
      }
    }
    return false;
  }

  @Override public void setPin8 (int pin8) {
    page = pin8;
  }

  @Override public int rbz () {
    int d = (page == 0
             ?
             (isPressed[UP] && !isPressed[DOWN] ? 0b11111110 : 0xff)  //↑が押されていて、↓が押されていない
             &
             (isPressed[DOWN] && !isPressed[UP] ? 0b11111101 : 0xff)  //↓が押されていて、↑が押されていない
             &
             0b11110011 &
             (isPressed[A] &&  //Aが押されていて
              (map[MAP_REPEAT + A] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[A] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[A]) /
                       (map[MAP_INTERVAL + A] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b11011111
              :
              0xff)
             &
             (isPressed[START] ? 0b10111111 : 0xff)  //STARTが押されている
             :
             (isPressed[UP] && !isPressed[DOWN] ? 0b11111110 : 0xff)  //↑が押されていて、↓が押されていない
             &
             (isPressed[DOWN] && !isPressed[UP] ? 0b11111101 : 0xff)  //↓が押されていて、↑が押されていない
             &
             (isPressed[LEFT] && !isPressed[RIGHT] ? 0b11111011 : 0xff)  //←が押されていて、→が押されていない
             &
             (isPressed[RIGHT] && !isPressed[LEFT] ? 0b11110111 : 0xff)  //→が押されていて、←が押されていない
             &
             (isPressed[B] &&  //Bが押されていて
              (map[MAP_REPEAT + B] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[B] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[B]) /
                       (map[MAP_INTERVAL + B] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b11011111
              :
              0xff)
             &
             (isPressed[C] &&  //Cが押されていて
              (map[MAP_REPEAT + C] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[C] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[C]) /
                       (map[MAP_INTERVAL + C] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b10111111
              :
              0xff)
             );
    return d;
  }

}  //class MegaDrive3ButtonPad



//class MegaDrive6ButtonPad
//  メガドラ6ボタンパッド
class MegaDrive6ButtonPad extends Joystick {

  protected static final int UP      =  0;
  protected static final int DOWN    =  1;
  protected static final int LEFT    =  2;
  protected static final int RIGHT   =  3;
  protected static final int A       =  4;
  protected static final int B       =  5;
  protected static final int C       =  6;
  protected static final int X       =  7;
  protected static final int Y       =  8;
  protected static final int Z       =  9;
  protected static final int MODE    = 10;
  protected static final int START   = 11;
  protected static final int BUTTONS = 12;  //ボタン数

  protected static final String[]       labelOf = { "↑" , "↓" , "←" , "→" , "A" , "B" , "C" , "X" , "Y" , "Z" , "MODE", "START" };  //ラベル
  protected static final boolean[] isRepeatable = { false, false, false, false, true, true, true, true, true, true, false , false   };  //連射可能

  protected static final int MAP_KEY      = BUTTONS * 0;  //キーコード
  protected static final int MAP_REPEAT   = BUTTONS * 1;  //連射有効
  protected static final int MAP_DELAY    = BUTTONS * 2;  //遅延時間
  protected static final int MAP_INTERVAL = BUTTONS * 3;  //連射間隔
  protected static final int MAP_LENGTH   = BUTTONS * 4;
  protected int[] map;

  protected boolean[] isPressed;  //ボタンの押し下げ状態
  protected long[] startTimeOf;  //連射開始時刻

  protected int page;  //ページ
  protected long lastPageTime;  //最後のページ切り替えの時刻
  protected final long PAGE_SPAN = XEiJ.TMR_FREQ / 10000;  //前回のページ切り替えから100μs経っていたらページ0,1に戻す

  protected JTextField[] keyTextFieldOf = new JTextField[BUTTONS];
  protected JCheckBox[] repeatCheckBoxOf = new JCheckBox[BUTTONS];
  protected SpinnerNumberModel[] delayModelOf = new SpinnerNumberModel[BUTTONS];
  protected JSpinner[] delaySpinnerOf = new JSpinner[BUTTONS];
  protected SpinnerNumberModel[] intervalModelOf = new SpinnerNumberModel[BUTTONS];
  protected JSpinner[] intervalSpinnerOf = new JSpinner[BUTTONS];

  public MegaDrive6ButtonPad (int number) {
    id = "megadrive6button" + number;
    nameEn = "MEGA DRIVE 6 button pad #" + number;
    nameJa = "メガドラ 6 ボタンパッド #" + number;
    if (true) {
      map = Settings.sgsGetIntArray (id, MAP_LENGTH);
    } else {
      map = (number == 1 ?
             new int[] {
               KeyEvent.VK_UP,     //↑
               KeyEvent.VK_DOWN,   //↓
               KeyEvent.VK_LEFT,   //←
               KeyEvent.VK_RIGHT,  //→
               KeyEvent.VK_Z,      //A
               KeyEvent.VK_X,      //B
               KeyEvent.VK_C,      //C
               KeyEvent.VK_A,      //X
               KeyEvent.VK_S,      //Y
               KeyEvent.VK_D,      //Z
               KeyEvent.VK_E,      //MODE
               KeyEvent.VK_R,      //START
               //38, 40, 37, 39, 90, 88, 67, 65, 83, 68, 69, 82,  //キーコード
               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  //連射有効
               0, 0, 0, 0, 50, 50, 50, 50, 50, 50, 0, 0,  //遅延時間
               0, 0, 0, 0, 50, 50, 50, 50, 50, 50, 0, 0,  //連射間隔
             } :
             new int[] {
               KeyEvent.VK_UNDEFINED,  //↑
               KeyEvent.VK_UNDEFINED,  //↓
               KeyEvent.VK_UNDEFINED,  //←
               KeyEvent.VK_UNDEFINED,  //→
               KeyEvent.VK_UNDEFINED,  //A
               KeyEvent.VK_UNDEFINED,  //B
               KeyEvent.VK_UNDEFINED,  //C
               KeyEvent.VK_UNDEFINED,  //X
               KeyEvent.VK_UNDEFINED,  //Y
               KeyEvent.VK_UNDEFINED,  //Z
               KeyEvent.VK_UNDEFINED,  //MODE
               KeyEvent.VK_UNDEFINED,  //START
               //0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  //キーコード
               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  //連射有効
               0, 0, 0, 0, 50, 50, 50, 50, 50, 50, 0, 0,  //遅延時間
               0, 0, 0, 0, 50, 50, 50, 50, 50, 50, 0, 0,  //連射間隔
             });
    }
    isPressed = new boolean[BUTTONS];
    startTimeOf = new long[BUTTONS];
    reset ();
    configurationPanel = null;
  }

  @Override public void tini () {
    Settings.sgsPutIntArray (id, map);
  }

  @Override public void reset () {
    for (int i = 0; i < BUTTONS; i++) {
      isPressed[i] = false;
      startTimeOf[i] = XEiJ.FAR_FUTURE;
    }
    page = 0;
    lastPageTime = 0L;
  }

  @Override public JComponent getConfigurationPanel () {

    if (configurationPanel != null) {
      return configurationPanel;
    }

    //キーリスナー
    KeyListener keyListener = new KeyAdapter () {
      @Override public void keyPressed (KeyEvent ke) {
        int i = Integer.parseInt (((JTextField) ke.getSource ()).getName ());
        int keyCode = ke.getKeyCode ();
        if (keyCode == KeyEvent.VK_ESCAPE) {  //Escキーは解除とみなす
          keyCode = KeyEvent.VK_UNDEFINED;
        }
        map[MAP_KEY + i] = keyCode;
        keyTextFieldOf[i].setText (
          keyCode == KeyEvent.VK_UNDEFINED ? Multilingual.mlnJapanese ? "なし" : "none" : KeyEvent.getKeyText (keyCode)
          );
        ke.consume ();
      }
      @Override public void keyReleased (KeyEvent ke) {
        ke.consume ();
      }
      @Override public void keyTyped (KeyEvent ke) {
        ke.consume ();
      }
    };

    //アクションリスナー
    ActionListener actionListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        String command = ae.getActionCommand ();
        if (command.startsWith ("Repeat ")) {
          int i = Integer.parseInt (command.substring (7));
          map[MAP_REPEAT + i] = repeatCheckBoxOf[i].isSelected () ? 1 : 0;
        }
      }
    };

    //チェンジリスナー
    ChangeListener changeListener = new ChangeListener () {
      @Override public void stateChanged (ChangeEvent ce) {
        JSpinner spinner = (JSpinner) ce.getSource ();
        String name = spinner.getName ();
        if (name.startsWith ("Delay ")) {
          int i = Integer.parseInt (name.substring (6));
          map[MAP_DELAY + i] = delayModelOf[i].getNumber ().intValue ();
        } else if (name.startsWith ("Interval ")) {
          int i = Integer.parseInt (name.substring (9));
          map[MAP_INTERVAL + i] = intervalModelOf[i].getNumber ().intValue ();
        }
      }
    };

    for (int i = 0; i < BUTTONS; i++) {
      keyTextFieldOf[i] = ComponentFactory.addListener (
        ComponentFactory.setHorizontalAlignment (
          ComponentFactory.setName (
            ComponentFactory.createTextField (
              map[MAP_KEY + i] == KeyEvent.VK_UNDEFINED ? Multilingual.mlnJapanese ? "なし" : "None" : KeyEvent.getKeyText (map[MAP_KEY + i]),
              10),
            String.valueOf (i)
            ),
          JTextField.CENTER
          ),
        keyListener
        );
      if (isRepeatable[i]) {
        repeatCheckBoxOf[i] = ComponentFactory.setText (
          ComponentFactory.createCheckBox (map[MAP_REPEAT + i] != 0, "Repeat " + i, actionListener), "");
        delayModelOf[i] = new SpinnerNumberModel (Math.max (DELAY_MIN, Math.min (DELAY_MAX, map[MAP_DELAY + i])),
                                                  DELAY_MIN, DELAY_MAX, DELAY_STEP);
        delaySpinnerOf[i] = ComponentFactory.setName (
          ComponentFactory.createNumberSpinner (delayModelOf[i], 4, changeListener), "Delay " + i);
        intervalModelOf[i] = new SpinnerNumberModel (Math.max (INTERVAL_MIN, Math.min (INTERVAL_MAX, map[MAP_INTERVAL + i])),
                                                     INTERVAL_MIN, INTERVAL_MAX, INTERVAL_STEP);
        intervalSpinnerOf[i] = ComponentFactory.setName (
          ComponentFactory.createNumberSpinner (intervalModelOf[i], 4, changeListener), "Interval " + i);
      }
    }

    //          0     1     2      3        4
    //    0               Title
    //    1                 -
    //    2  Button  Key  Burst  Delay  Interval
    //    3  -----------------------------------
    //    4    ↑     TF    -      -        -
    //    5    ↓     TF    -      -        -
    //    6    ←     TF    -      -        -
    //    7    →     TF    -      -        -
    //    8     A     TF    CB     S        S
    //    9     B     TF    CB     S        S
    //   10     C     TF    CB     S        S
    //   11     X     TF    CB     S        S
    //   12     Y     TF    CB     S        S
    //   13     Z     TF    CB     S        S
    //   14   MODE    TF    -      -        -
    //   15   START   TF    -      -        -
    ArrayList<Object> cellList = new ArrayList<Object> ();
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel (getNameEn ()), "ja", getNameJa ()));  //(0,0)
    cellList.add (null);  //(0,1)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Button"), "ja", "ボタン"));  //(0,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Key"), "ja", "キー"));  //(1,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Burst"), "ja", "連射"));  //(2,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Delay (ms)"), "ja", "開始 (ms)"));  //(3,2)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Interval (ms)"), "ja", "間隔 (ms)"));  //(4,2)
    cellList.add (ComponentFactory.createHorizontalSeparator ());  //(0,1)-(4,3)
    for (int i = 0; i < BUTTONS; i++) {
      cellList.add (labelOf[i]);  //(0,4+i)
      cellList.add (keyTextFieldOf[i]);  //(1,4+i)
      cellList.add (isRepeatable[i] ? repeatCheckBoxOf[i] : null);  //(2,4+i)
      cellList.add (isRepeatable[i] ? delaySpinnerOf[i] : null);  //(3,4+i)
      cellList.add (isRepeatable[i] ? intervalSpinnerOf[i] : null);  //(4,4+i)
    }
    return configurationPanel = ComponentFactory.createGridPanel (
      5, 4 + BUTTONS,
      "paddingLeft=3,paddingRight=3,center",   //gridStyles
      "",  //colStyles
      "colSpan=5;colSpan=5;italic;colSpan=5,widen",  //rowStyles
      "",  //cellStyles
      cellList.toArray (new Object[0])
      );
  }

  @Override public boolean input (KeyEvent ke, boolean pressed) {
    int keyCode = ke.getKeyCode ();
    for (int i = 0; i < BUTTONS; i++) {
      if (keyCode == map[MAP_KEY + i]) {
        if (isRepeatable[i]) {  //連射可能
          if (map[MAP_REPEAT + i] != 0 && !isPressed[i] && pressed) {  //連射有効で、押し下げられた
            startTimeOf[i] = XEiJ.mpuClockTime + map[MAP_DELAY + i] * (XEiJ.TMR_FREQ / 1000);  //連射開始時刻
          }
        }
        isPressed[i] = pressed;
        return true;
      }
    }
    return false;
  }

  @Override public void setPin8 (int pin8) {
    page += (page ^ pin8) & 1;
    if (page == 6) {
      page = 0;
    }
    long t = XEiJ.mpuClockTime;
    if (lastPageTime + PAGE_SPAN <= t) {
      page &= 1;
    }
    lastPageTime = t;
  }

  @Override public int rbz () {
    //  page=0  1 S A 1 0 0 D U
    //  page=1  1 C B 1 R L D U
    //  page=2  1 S A 1 0 0 D U
    //  page=3  1 C B 1 R L D U
    //  page=4  1 S A 1 0 0 0 0
    //  page=5  1 C B 1 M X Y Z
    int d = (page == 0 || page == 2
             ?
             (isPressed[UP] && !isPressed[DOWN] ? 0b11111110 : 0xff)  //↑が押されていて、↓が押されていない
             &
             (isPressed[DOWN] && !isPressed[UP] ? 0b11111101 : 0xff)  //↓が押されていて、↑が押されていない
             &
             0b11110011
             &
             (isPressed[A] &&  //Aが押されていて
              (map[MAP_REPEAT + A] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[A] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[A]) /
                       (map[MAP_INTERVAL + A] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b11011111
              :
              0xff)
             &
             (isPressed[START] ? 0b10111111 : 0xff)  //STARTが押されている
             :
             page == 4
             ?
             0b11110000
             &
             (isPressed[A] &&  //Aが押されていて
              (map[MAP_REPEAT + A] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[A] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[A]) /
                       (map[MAP_INTERVAL + A] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b11011111
              :
              0xff)
             &
             (isPressed[START] ? 0b10111111 : 0xff)  //STARTが押されている
             :
             page == 1 || page == 3
             ?
             (isPressed[UP] && !isPressed[DOWN] ? 0b11111110 : 0xff)  //↑が押されていて、↓が押されていない
             &
             (isPressed[DOWN] && !isPressed[UP] ? 0b11111101 : 0xff)  //↓が押されていて、↑が押されていない
             &
             (isPressed[LEFT] && !isPressed[RIGHT] ? 0b11111011 : 0xff)  //←が押されていて、→が押されていない
             &
             (isPressed[RIGHT] && !isPressed[LEFT] ? 0b11110111 : 0xff)  //→が押されていて、←が押されていない
             &
             (isPressed[B] &&  //Bが押されていて
              (map[MAP_REPEAT + B] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[B] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[B]) /
                       (map[MAP_INTERVAL + B] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b11011111
              :
              0xff)
             &
             (isPressed[C] &&  //Cが押されていて
              (map[MAP_REPEAT + C] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[C] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[C]) /
                       (map[MAP_INTERVAL + C] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b10111111
              :
              0xff)
             :
             (isPressed[Z] &&  //Zが押されていて
              (map[MAP_REPEAT + Z] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[Z] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[Z]) /
                       (map[MAP_INTERVAL + Z] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b11111110
              :
              0xff)
             &
             (isPressed[Y] &&  //Yが押されていて
              (map[MAP_REPEAT + Y] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[Y] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[Y]) /
                       (map[MAP_INTERVAL + Y] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b11111101
              :
              0xff)
             &
             (isPressed[X] &&  //Xが押されていて
              (map[MAP_REPEAT + X] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[X] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[X]) /
                       (map[MAP_INTERVAL + X] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b11111011
              :
              0xff)
             &
             (isPressed[MODE] ? 0b11110111 : 0xff)  //MODEが押されている
             &
             (isPressed[B] &&  //Bが押されていて
              (map[MAP_REPEAT + B] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[B] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[B]) /
                       (map[MAP_INTERVAL + B] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b11011111
              :
              0xff)
             &
             (isPressed[C] &&  //Cが押されていて
              (map[MAP_REPEAT + C] == 0 ||  //連射が無効か
               XEiJ.mpuClockTime < startTimeOf[C] ||  //連射開始時刻になっていないか
               ((int) ((XEiJ.mpuClockTime - startTimeOf[C]) /
                       (map[MAP_INTERVAL + C] * (XEiJ.TMR_FREQ / 1000))) & 1) != 0)  //連射開始時刻を過ぎた時間を連射間隔で割った商が奇数
              ?
              0b10111111
              :
              0xff)
             );
    return d;
  }

}  //class MegaDrive6ButtonPad



//                                                                                  
//    白窓君 (旧版) の配線図                                                        
//                                  JS                                              
//            ┏━━━━━━━━━━━━━━━━━━━━━┓                        
//            ┃    IOA0    IOA1    IOA2    IOA3    VCC1  ┃                        
//            ┃    1  IOA5 2  IOA6 3  PC4  4  GND  5     ┃                        
//            ┗┓  ○  6   ○  7   ○  8   ○  9   ○  ┏┛                        
//              ┃  │  ○  │  ○  │  ○  │  ○  │  ┃                          
//              ┗━┿━┿━┿━┿━┿━┿━┿━┿━┿━┛                          
//                  黒  青  茶  紫  赤  緑  橙  灰  黄                              
//                  │  │  │  │  │  │  │  桃  │                              
//                  │  │  │  │  │  │  └─┼─┼───────┐              
//                  │  │  │  │  └─┼───┼─┼─────┐  │              
//                  │  │  └─┼───┼───┼─┼───┐  │  │              
//                  └─┼───┼───┼───┼─┼─┐  │  │  │              
//                      └─┐  │  ┌─┘      │  │  │  │  │  │              
//              ┌─────┼─┼─┼─────┘  │  │  │  │  │              
//              灰  ┌───┼─┼─┼───────┘  │  │  │  │              
//              桃  黄      青  紫  緑                  黒  茶  赤  橙              
//          ┏━┿━┿━━━┿━┿━┿━━━━━━━━━┿━┿━┿━┿━━━━┓    
//      ┌─╂─○  ○  ○  ○  ○  ○  ○  ○  ○  ○  ○  ○  ○  ○        ┃    
//      │  ┃  1  2│ 3│  4   5   6   7   8   9   10  11  12  13  14        ┃    
//      ＜  ┃ VSS  │  │  RS  R/W E   DB0 DB1 DB2 DB3 DB4 DB5 DB6 DB7       ┃    
//      ＜←╂───┼─┘                                                    ┃    
//      ＜  ┃      │  VL                                                    ┃    
//      └─╂───┘                     NDM162                             ┃    
//          ┃     VDD                                                        ┃    
//          ┗━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┛    
//                                                                                  
//                                                                                  
//                                                                                  
//    白窓君 (復刻版) の配線図                                                      
//                                                                                  
//                            JS                                                    
//      ┏━━━━━━━━━━━━━━━━━━━━━┓                              
//      ┃    IOA0    IOA1    IOA2    IOA3    VCC1  ┃                              
//      ┃    1  IOA5 2  IOA6 3  PC4  4  GND  5     ┃                              
//      ┗┓  ○  6   ○  7   ○  8   ○  9   ○  ┏┛                              
//        ┃  │  ○  │  ○  │  ○  │  ○  │  ┃                                
//        ┗━┿━┿━┿━┿━┿━┿━┿━┿━┿━┛                                
//            黒  青  茶  紫  赤  緑  橙  灰  黄              SC1602B               
//            │  │  │  │  │  │  │  桃  │    ┏━━━━━━━━━━━━┓    
//            │  │  │  │  │  │  └─┼─┼─橙╂○14 DB7    ○13 DB6    ┃    
//            │  │  │  │  └─┼───┼─┼──╂─────赤┘          ┃    
//            │  │  └─┼───┼───┼─┼─茶╂○12 DB5    ○11 DB4    ┃    
//            └─┼───┼───┼───┼─┼──╂─────黒┘          ┃    
//                │      │      │      │  │    ┃○10 DB3    ○9 DB2     ┃    
//                │      │      │      │  │    ┃                        ┃    
//                │      │      │      │  │    ┃○8 DB1     ○7 DB0     ┃    
//                │      └───┼───┼─┼──╂─────紫┐          ┃    
//                │              └───┼─┼─緑╂○6 E       ○5 R/W     ┃    
//                │                      │  │    ┃                        ┃    
//                └───────────┼─┼─青╂○4 RS  ┌─○3 VO      ┃    
//                                        │  └──╂────┼黄┐          ┃    
//                                        └──灰桃╂○2 VSS │  ○1 VDD     ┃    
//                                                  ┗┿━━━┿━┿━━━━━┛    
//                                                    │      ↓  │                
//                                                    └──∨∨∨┘                
//                                                                                  
//                                                                                  
//  HD44780
//    CLCDコントローラ
//      R/W  RS
//       0    0    コマンドレジスタ書き込み
//       0    1    データレジスタ書き込み
//       1    0    ステータスレジスタ読み出し
//       1    1    データレジスタ読み出し
//    書き込み手順
//      R/W,RS,DB7,DB6,DB5,DB4をセットしてE=1→0
//      R/W,RS,DB3,DB2,DB1,DB0をセットしてE=1→0
//    コマンド
//      0  0  0  0  0  0  0  1    画面クリア,カーソルアドレス0,ホームアドレス0
//      0  0  0  0  0  0  1  *    カーソルアドレス0,ホームアドレス0
//      0  0  0  0  0  1  0  0    文字を書いたらカーソルアドレスをデクリメント(カーソルを左へ移動),左上0→右下64+39,左下64→右上39
//      0  0  0  0  0  1  0  1    文字を書いたらカーソルアドレスをデクリメント,ホームアドレスをデクリメント(画面を右へシフト),0→39
//      0  0  0  0  0  1  1  0    文字を書いたらカーソルアドレスをインクリメント(カーソルを右へ移動),右上39→左下64,右下64+39→左上0
//      0  0  0  0  0  1  1  1    文字を書いたらカーソルアドレスをインクリメント,ホームアドレスをインクリメント(画面を左へシフト),39→0
//      0  0  0  0  1  0  *  *    表示OFF
//      0  0  0  0  1  1  0  0    表示ON,下線カーソルOFF,四角カーソルOFF
//      0  0  0  0  1  1  0  1    表示ON,下線カーソルOFF,四角カーソルON
//      0  0  0  0  1  1  1  0    表示ON,下線カーソルON,四角カーソルOFF
//      0  0  0  0  1  1  1  1    表示ON,下線カーソルON,四角カーソルON
//      0  0  0  1  0  0  *  *    カーソルアドレスデクリメント,左上0→右下64+39,左下64→右上39
//      0  0  0  1  0  1  *  *    カーソルアドレスインクリメント,右上39→左下64,右下64+39→左上0
//      0  0  0  1  1  0  *  *    ホームアドレスインクリメント(画面を左へシフト),39→0
//      0  0  0  1  1  1  *  *    ホームアドレスデクリメント(画面を右へシフト),0→39
//      0  0  1  0  0  0  *  *    4bit,5x8ドット1行      0011 0011 0011 0010で4bitになる。0x33は下位データ待ちを繰り返して0x32になったら抜ける
//      0  0  1  0  0  1  *  *    4bit,5x10ドット1行
//      0  0  1  0  1  *  *  *    4bit,5x8ドット2行
//      0  0  1  1  0  0  *  *    8bit,5x8ドット1行      0011**** 0011**** 0011****で8bitになる
//      0  0  1  1  0  1  *  *    8bit,5x10ドット1行
//      0  0  1  1  1  *  *  *    8bit,5x8ドット2行
//      0  1  c  c  c  y  y  y    フォントアドレス
//      1  y  x  x  x  x  x  x    カーソルアドレス,40..63→64,64+40..64+63→0

class Shiromadokun extends Joystick {

  private static final boolean DEBUG = false;

  //仕様
  public static final int JAPANESE_SPECIFICATION = 0;
  public static final int EUROPEAN_SPECIFICATION = 1;
  protected int specification;

  protected int pinE;  //1→0で書き込む
  protected int busData;  //E=1→0のとき書き込むデータ
  protected int upper;  //上位4bit。-1=上位4bit待ち
  protected int lower;  //下位4bit。-1=下位4bit待ち
  protected int fontOrScreen;  //0=フォント書き込み,1=画面書き込み
  protected int fontAddress;  //フォントアドレス。次にフォントデータを書き込まれるアドレス。0..63
  protected int cursorAddress;  //カーソルアドレス。カーソルが表示されて次に文字を書き込まれるアドレス。0..39,64..64+39
  protected int homeAddress;  //ホームアドレス。画面の左上に表示されている文字のアドレス。0..39
  protected byte[] screenMemory;  //画面メモリ。0..39は1行目,64..64+39は2行目
  protected byte[] fontMemory;  //フォントメモリ。0..7は外字,8..15は0..7と同じ
  protected int fontBaseAddress;  //フォントベースアドレス。8*256*0は日本仕様、8*256*1は欧州仕様

  protected int entryMode;
  protected boolean displayOn;  //表示OFF/表示ON
  protected boolean underlineCursorOn;  //下線カーソルOFF/下線カーソルON

  protected static final long BLINKING_DELAY = 0L;
  protected static final long BLINKING_INTERVAL = 500L;  //実機は300くらいか
  protected boolean blinkingCursorOn;  //点滅カーソルOFF/点滅カーソルON
  protected int blinkingCursorMask;  //点滅カーソルのマスク
  protected TimerTask blinkingCursorTask;  //点滅カーソルのタスク

  //イメージとビットマップ
  //  NDM162のLCDは16文字x2行、1文字が5x7ドット、隙間が1ドット、パディングは左右が6ドット、上下が4ドットくらい
  protected static final int SCREEN_WIDTH = 16;
  protected static final int SCREEN_HEIGHT = 2;
  protected static final int CHARACTER_WIDTH = 5;
  protected static final int CHARACTER_HEIGHT = 8;
  protected static final int SPACE_WIDTH = 1;
  protected static final int SPACE_HEIGHT = 1;
  protected static final int MARGIN_TOP = 2;
  protected static final int MARGIN_BOTTOM = 2;
  protected static final int MARGIN_LEFT = 4;
  protected static final int MARGIN_RIGHT = 4;
  protected static final int DOT_WIDTH = 2;
  protected static final int DOT_HEIGHT = 2;
  protected static final int IMAGE_WIDTH = DOT_WIDTH * (MARGIN_LEFT + (CHARACTER_WIDTH + SPACE_WIDTH) * SCREEN_WIDTH - 1 + MARGIN_RIGHT);
  protected static final int IMAGE_HEIGHT = DOT_HEIGHT * (MARGIN_TOP + (CHARACTER_HEIGHT + SPACE_HEIGHT) * SCREEN_HEIGHT - 1 + MARGIN_BOTTOM);
  protected BufferedImage image;
  protected byte[] bitmap;
  protected JPanel panel;

  public Shiromadokun (int number) {
    id = "shiromadokun" + number;
    nameEn = "Shiromadokun #" + number;
    nameJa = "白窓君 #" + number;
    specification = Settings.sgsGetInt (id);
    if (!(specification == JAPANESE_SPECIFICATION ||
          specification == EUROPEAN_SPECIFICATION)) {
      specification = JAPANESE_SPECIFICATION;
    }
    screenMemory = new byte[64 * 2];
    fontMemory = FontEditor.fntPageSmk6x8.fnpBinaryArray;  //フォントが初期化済みであること
    fontBaseAddress = 8 * 256 * specification;
    image = new BufferedImage (IMAGE_WIDTH,
                               IMAGE_HEIGHT,
                               BufferedImage.TYPE_BYTE_BINARY,
                               new IndexColorModel (1, 2,
                                                    new byte[] { (byte) 0x66,        0x00 },
                                                    new byte[] { (byte) 0xcc, (byte) 0x33 },
                                                    new byte[] { (byte) 0x99, (byte) 0x99 }));
    bitmap = ((DataBufferByte) image.getRaster ().getDataBuffer ()).getData ();
    panel = ComponentFactory.setFixedSize (
      new JPanel () {
        @Override protected void paintComponent (Graphics g) {
          if (displayOn) {
            drawBitmap ();
          }
          g.drawImage (image, 0, 0, null);
        }
      },
      IMAGE_WIDTH, IMAGE_HEIGHT);
    reset ();
    configurationPanel = null;
  }

  @Override public void tini () {
    Settings.sgsPutInt (id, specification);
  }

  @Override public void reset () {
    pinE = 0;
    busData = 0;
    upper = -1;
    lower = -1;
    fontOrScreen = 0;
    fontAddress = 0;
    cursorAddress = 0;
    homeAddress = 0;
    Arrays.fill (screenMemory, (byte) 0);
    //Arrays.fill (fontMemory, (byte) 0);
    entryMode = 2;
    displayOn = true;
    underlineCursorOn = false;
    blinkingCursorOn = false;
    blinkingCursorMask = 0x00;
    blinkingCursorTask = null;
    setSpecification (specification);
  }

  protected void setSpecification (int specification) {
    this.specification = specification;
    fontBaseAddress = 8 * 256 * specification;
  }

  @Override public JComponent getConfigurationPanel () {
    if (configurationPanel != null) {
      return configurationPanel;
    }

    ActionListener actionListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        String command = ae.getActionCommand ();
        switch (command) {
        case "Japanese specification" :  //日本仕様
          setSpecification (JAPANESE_SPECIFICATION);
          panel.repaint ();
          break;
        case "European specification":  //欧州仕様
          setSpecification (EUROPEAN_SPECIFICATION);
          panel.repaint ();
          break;
        }
      }
    };

    ButtonGroup specificationGroup = new ButtonGroup ();
    //           0        1
    //    0        Title
    //    1          -
    //    2        Panel
    //    3          -
    //    4  日本仕様  欧州仕様
    return configurationPanel = ComponentFactory.createGridPanel (
      2, 5,
      "paddingLeft=3,paddingRight=3,center",   //gridStyles
      "",  //colStyles
      "colSpan=2;colSpan=2;colSpan=2;colSpan=2",  //rowStyles
      "",  //cellStyles
      Multilingual.mlnText (ComponentFactory.createLabel (getNameEn ()), "ja", getNameJa ()),  //(0,0)-(1,0)
      null,  //(0,1)-(1,1)
      panel,  //(0,2)-(1,2)
      null,  //(0,3)-(1,3)
      Multilingual.mlnText (
        ComponentFactory.createRadioButton (
          specificationGroup, specification == JAPANESE_SPECIFICATION, "Japanese specification", actionListener),
        "ja", "日本仕様"),  //(0,4)
      Multilingual.mlnText (
        ComponentFactory.createRadioButton (
          specificationGroup, specification == EUROPEAN_SPECIFICATION, "European specification", actionListener),
        "ja", "欧州仕様")  //(1,4)
      );
  }

  @Override public void setPin8 (int pin8) {
    if (pinE == 1 && pin8 == 0) {  //E=1→0
      if ((busData & 0x60) == 0x00) {  //コマンドレジスタ書き込み
        if (upper < 0) {  //上位4bit
          upper = busData & 0x0f;
          lower = -1;
        } else {  //下位4bit
          lower = busData & 0x0f;
          if (upper == 0x3 && lower == 0x3) {  //0x33のときは下位を繰り返す
            lower = -1;
          } else {
            int command = upper << 4 | lower;
            upper = -1;
            lower = -1;
            processCommand (command);  //コマンド処理
          }
        }
      } else if ((busData & 0x60) == 0x20) {  //データレジスタ書き込み
        if (upper < 0) {  //上位4bit
          upper = busData & 0x0f;
          lower = -1;
        } else {  //下位4bit
          lower = busData & 0x0f;
          int data = upper << 4 | lower;
          upper = -1;
          lower = -1;
          processData (data);  //データ処理
        }
      }
    }
    pinE = pin8;
  }
  @Override public int rbz () {
    return 0xff;
  }
  @Override public void wb (int d) {
    busData = d & 0xff;
  }

  protected void processCommand (int command) {
    if (DEBUG) {
      System.out.printf ("command(0x%02x)\n", command);
    }
    if (command <= 0x03) {
      switch (command) {
      case 0x00:
        break;
      case 0x01:  //Clear display
        //画面クリア,カーソルアドレス0,ホームアドレス0
        Arrays.fill (screenMemory, (byte) 0);
        //ここはbreakしない
      case 0x02:  //Return home
      case 0x03:
        //カーソルアドレス0,ホームアドレス0
        homeAddress = 0;
        cursorAddress = 0;
        if (DEBUG) {
          System.out.printf ("homeAddress=%d\n", homeAddress);
          System.out.printf ("cursorAddress=%d\n", cursorAddress);
        }
        break;
      }
    } else if (command <= 0x3f) {
      switch (command >> 2) {
      case 0x04 >> 2:  //Entry mode set
        entryMode = command & 3;
        if (DEBUG) {
          System.out.printf ("entryMode=%d\n", entryMode);
        }
        break;
      case 0x08 >> 2:  //Display on/off control
        displayOn = false;  //表示OFF
        if (DEBUG) {
          System.out.printf ("displayOn=%b\n", displayOn);
        }
        Arrays.fill (bitmap, (byte) 0);
        break;
      case 0x0c >> 2:
        displayOn = true;  //表示ON
        underlineCursorOn = (command & 2) != 0;  //下線カーソルOFF/下線カーソルON
        blinkingCursorOn = (command & 1) != 0;  //点滅カーソルOFF/点滅カーソルON
        if (DEBUG) {
          System.out.printf ("displayOn=%b\n", displayOn);
          System.out.printf ("underlineCursorOn=%b\n", underlineCursorOn);
          System.out.printf ("blinkingCursorOn=%b\n", blinkingCursorOn);
        }
        if (blinkingCursorOn) {
          blinkingStart ();
        } else {
          blinkingEnd ();
        }
        break;
      case 0x10 >> 2:  //Cursor or display shift
        //カーソルアドレスデクリメント,左上0→右下64+39,左下64→右上39
        cursorAddress = cursorAddress == 0 ? 64 + 39 : cursorAddress == 64 ? 39 : cursorAddress - 1;
        if (DEBUG) {
          System.out.printf ("cursorAddress=%d\n", cursorAddress);
        }
        break;
      case 0x14 >> 2:
        //カーソルアドレスインクリメント,右上39→左下64,右下64+39→左上0
        cursorAddress = cursorAddress == 39 ? 64 : cursorAddress == 64 + 39 ? 0 : cursorAddress + 1;
        if (DEBUG) {
          System.out.printf ("cursorAddress=%d\n", cursorAddress);
        }
        break;
      case 0x18 >> 2:
        //ホームアドレスインクリメント(画面を左へシフト),39→0
        homeAddress = homeAddress == 39 ? 0 : homeAddress + 1;
        if (DEBUG) {
          System.out.printf ("homeAddress=%d\n", homeAddress);
        }
        break;
      case 0x1c >> 2:
        //ホームアドレスデクリメント(画面を右へシフト),0→39
        homeAddress = homeAddress == 0 ? 39 : homeAddress - 1;
        if (DEBUG) {
          System.out.printf ("homeAddress=%d\n", homeAddress);
        }
        break;
      case 0x20 >> 2:  //Function set
      case 0x24 >> 2:
      case 0x28 >> 2:
      case 0x2c >> 2:
      case 0x30 >> 2:
      case 0x34 >> 2:
      case 0x38 >> 2:
      case 0x3c >> 2:
        break;
      }
    } else if (command <= 0x7f) {  //フォントアドレス
      fontOrScreen = 0;
      fontAddress = command & 0x3f;
      if (DEBUG) {
        System.out.printf ("fontOrScreen=%d\n", fontOrScreen);
        System.out.printf ("fontAddress=%d\n", fontAddress);
      }
    } else {  //カーソルアドレス,40..63→64,64+40..64+63→0
      fontOrScreen = 1;
      int a = command & 0x7f;
      cursorAddress = a < 40 ? a : a < 64 ? 64 : a < 64 + 40 ? a : 0;
      if (DEBUG) {
        System.out.printf ("fontOrScreen=%d\n", fontOrScreen);
        System.out.printf ("cursorAddress=%d\n", cursorAddress);
      }
    }
    panel.repaint ();
  }

  protected void processData (int data) {
    if (DEBUG) {
      System.out.printf ("data(0x%02x)\n", data);
    }
    if (fontOrScreen == 0) {  //フォント書き込み
      fontMemory[fontBaseAddress + fontAddress] =
        fontMemory[fontBaseAddress + fontAddress + 64] = (byte) (data << 3);  //上位に寄せる
      if (DEBUG) {
        System.out.printf ("fontMemory[%d]=0x%02x\n",
                           fontBaseAddress + fontAddress, (fontMemory[fontBaseAddress + fontAddress] & 0xff) >> 3);  //下位に寄せる
      }
      fontAddress = fontAddress == 63 ? 0 : fontAddress + 1;
      if (DEBUG) {
        System.out.printf ("fontAddress=%d\n", fontAddress);
      }
    } else {  //画面書き込み
      screenMemory[cursorAddress] = (byte) data;
      if (DEBUG) {
        System.out.printf ("screenMemory[%d]=0x%02x\n", cursorAddress, screenMemory[cursorAddress] & 0xff);
      }
      switch (entryMode) {
      case 1:  //文字を書いたらカーソルアドレスをデクリメント,ホームアドレスをデクリメント(画面を右へシフト),0→39
        homeAddress = homeAddress == 0 ? 39 : homeAddress - 1;
        if (DEBUG) {
          System.out.printf ("homeAddress=%d\n", homeAddress);
        }
        //ここはbreakしない
      case 0:  //文字を書いたらカーソルアドレスをデクリメント(カーソルを左へ移動),左上0→右下64+39,左下64→右上39
        cursorAddress = cursorAddress == 0 ? 64 + 39 : cursorAddress == 64 ? 39 : cursorAddress - 1;
        if (DEBUG) {
          System.out.printf ("cursorAddress=%d\n", cursorAddress);
        }
        break;
      case 3:  //文字を書いたらカーソルアドレスをインクリメント,ホームアドレスをインクリメント(画面を左へシフト),39→0
        homeAddress = homeAddress == 39 ? 0 : homeAddress + 1;
        if (DEBUG) {
          System.out.printf ("homeAddress=%d\n", homeAddress);
        }
        //ここはbreakしない
      case 2:  //文字を書いたらカーソルアドレスをインクリメント(カーソルを右へ移動),右上39→左下64,右下64+39→左上0
        cursorAddress = cursorAddress == 39 ? 64 : cursorAddress == 64 + 39 ? 0 : cursorAddress + 1;
        if (DEBUG) {
          System.out.printf ("cursorAddress=%d\n", cursorAddress);
        }
        break;
      }
    }
    panel.repaint ();
  }

  protected void drawBitmap () {
    int offset = (IMAGE_WIDTH + 7) >> 3;
    for (int screenY = 0; screenY < SCREEN_HEIGHT; screenY++) {
      for (int screenX = 0; screenX < SCREEN_WIDTH; screenX++) {
        int screenAddress = (screenX + homeAddress) % 40 + 64 * screenY;  //(screenX,screenY)の位置に表示する文字のアドレス
        int underlineCursorMask = 0x00;  //下線カーソルマスク。最下ラインにorする
        int blinkingMask = 0x00;  //点滅マスク。すべてのラインにorする
        if (screenAddress == cursorAddress) {
          if (underlineCursorOn) {  //下線カーソルON
            underlineCursorMask = 0x1f;
          }
          blinkingMask = blinkingCursorMask;
        }
        int fontStartAddress = (screenMemory[screenAddress] & 0xff) << 3;
        for (int characterY = 0; characterY < CHARACTER_HEIGHT; characterY++) {
          int y0 = DOT_HEIGHT * (MARGIN_TOP + (CHARACTER_HEIGHT + SPACE_HEIGHT) * screenY + characterY);
          int fontRaster = (fontMemory[fontBaseAddress + fontStartAddress + characterY] & 0xff) >> 3;  //下位に寄せる
          if (characterY == CHARACTER_HEIGHT - 1) {  //最下ライン
            fontRaster |= underlineCursorMask;
          }
          fontRaster |= blinkingMask;
          for (int characterX = 0; characterX < CHARACTER_WIDTH; characterX++) {
            int x0 = DOT_WIDTH * (MARGIN_LEFT + (CHARACTER_WIDTH + SPACE_WIDTH) * screenX + characterX);
            if ((fontRaster >> (CHARACTER_WIDTH - 1 - characterX) & 1) == 0) {
              for (int dotY = 0; dotY < DOT_HEIGHT; dotY++) {
                int y = y0 + dotY;
                for (int dotX = 0; dotX < DOT_WIDTH; dotX++) {
                  int x = x0 + dotX;
                  bitmap[offset * y + (x >> 3)] &= ~(0x80 >> (x & 7));
                }
              }
            } else {
              for (int dotY = 0; dotY < DOT_HEIGHT; dotY++) {
                int y = y0 + dotY;
                for (int dotX = 0; dotX < DOT_WIDTH; dotX++) {
                  int x = x0 + dotX;
                  bitmap[offset * y + (x >> 3)] |= 0x80 >> (x & 7);
                }
              }
            }
          }
        }
      }
    }
  }

  //  点滅開始
  protected void blinkingStart () {
    blinkingCursorMask = 0x00;
    if (blinkingCursorTask == null) {
      blinkingCursorTask = new BlinkingCursorTask ();
      XEiJ.tmrTimer.schedule (blinkingCursorTask, BLINKING_DELAY, BLINKING_INTERVAL);
    }
  }

  //  点滅終了
  protected void blinkingEnd () {
    if (blinkingCursorTask != null) {
      blinkingCursorTask.cancel ();
      blinkingCursorTask = null;
    }
    blinkingCursorMask = 0x00;
  }

  //  点滅タスク
  class BlinkingCursorTask extends TimerTask {
    @Override public void run () {
      blinkingCursorMask ^= 0x1f;
      panel.repaint ();
    }
  }

}  //class Shiromadokun



