//========================================================================================
//  ROM.java
//    en:ROM
//    ja:ROM
//  Copyright (C) 2003-2020 Makoto Kamada
//
//  This file is part of the XEiJ (X68000 Emulator in Java).
//  You can use, modify and redistribute the XEiJ if the conditions are met.
//  Read the XEiJ License for more details.
//  https://stdkmd.net/xeij/
//========================================================================================

package xeij;

import java.io.*;  //File
import java.lang.*;
import java.util.*;  //Arrays
import java.util.zip.*;  //CRC32
import javax.swing.*;  //JFileChooser
import javax.swing.filechooser.*;  //FileFilter

public class ROM {

  //========================================================================================
  //$$ROM ROM
  //
  //  ROMマップ
  //    0x00f00000  0x00f40000  CG1  CGROM1
  //    0x00f40000  0x00fc0000  CG2  CGROM2
  //    0x00fc0000  0x00fe0000       ROM30
  //    0x00fe0000  0x01000000       IPLROM
  //
  //  ROMの構築手順
  //    1  ROM(0x00f00000～0x01000000;1MB)
  //       1.1  -romで指定されたROMのイメージファイル(1MB)を読み込む
  //                ROM.DAT(1MB)  実機で作成したもの
  //    2  IPLROM(0x00fe0000～0x01000000;128KB)
  //       2.1  -iplromで指定されたIPLROMのイメージファイル(128KB)を読み込む
  //                IPLROM.DAT(128KB)    IPLROM 1.0
  //                IPLROMXV.DAT(128KB)  IPLROM 1.1
  //                IPLROMCO.DAT(128KB)  IPLROM 1.2
  //                IPLROM30.DAT(128KB)  IPLROM 1.3
  //            -romが指定されていても-iplromが指定されたときは指定されたものを読み込む
  //            -romと-iplromのどちらも指定されなかったかされても読み込めなかったときはリソースからIPLROMXV.DATを読み込む
  //       2.2  IPLROMのバージョンを確認してパッチをあてる
  //    3  ROM30(0x00fc0000～0x00fe0000;128KB)
  //       3.1  -rom30で指定されたX68030のROMの拡張部分のイメージファイル(128KB)を読み込む
  //                ROM30.DAT(128KB)  実機で作成したもの
  //            -romが指定されていても-rom30が指定されたときは指定されたものを読み込む
  //       3.2  -scsiinromで指定されたSCSIINROMのイメージファイル(8KB)を読み込む
  //                SCSIINROM.DAT(8KB)  実機で作成したもの
  //            -romが指定されていても-scsiinromが指定されたときは指定されたものを読み込む
  //    CGROMを構築する
  //

  //パラメータ
  public static String romROMName;  //-rom ROMのイメージファイル(1MB)
  public static String romIPLROMName;  //-iplrom IPLROMのイメージファイル(128KB)
  public static boolean romROMDBOn;  //-romdb IPLでROMデバッガを強制的に起動する(on/off)
  public static String romSCSIINROMName;  //-scsiinrom SCSIINROMのイメージファイル(8KB)
  public static String romROM30Name;  //-rom30 X68030のROMの拡張部分のイメージファイル(128KB)

  public static int romVersionLong;  //IPLROMのバージョン。0x10870318,0x10870507,0x11910111,0x12911024,0x13921127など。-1=不明
  public static int romVersionShort;  //IPLROMのバージョン。romVersionLong>>24&15。0=1.0,1=1.1,2=1.2,3=1.3。-1=不明
  public static int romPreviousVersionLong;  //変更前のIPLROMのバージョン。-1=バージョンを変更していない
  public static int romPreviousVersionShort;  //変更前のIPLROMのバージョン。-1=バージョンを変更していない

  public static int romDebuggerVersion;  //ROMデバッガのバージョン。0x0100=1.0,0x0232=2.32,-1=不明
  public static int romHumanVersion;  //ROM Humanのバージョン。0x0100=1.00,0x0215=2.15,-1=不明

  public static int romSRAMInitialData;        //SRAM初期化データの先頭。0=不明
  public static int romBootDeviceRead;         //起動シーケンスで起動デバイスを読み出すmove.w $00ED0018,d1の先頭
  public static int romBootROMAddressRead;     //起動シーケンスでROM起動ハンドルを読み出すmovea.l $00ED000C,a0の先頭
  public static int romBootRAMAddressRead;     //起動シーケンスでRAM起動アドレスを読み出すmovea.l $00ED0010,a0の先頭
  public static int romBootinfROMAddressRead;  //IOCS _BOOTINFでROM起動ハンドルを読み出すmove.l $00ED000C,d0の先頭
  public static int romBootinfRAMAddressRead;  //IOCS _BOOTINFでRAM起動アドレスを読み出すmove.l $00ED0010,d0の先頭

  public static boolean romROM30Required;  //true=rom30を読み込んでいない
  public static byte[] romROM30Data;  //-romまたは-rom30で読み込まれた0x00fc0000～0x00fe0000のデータ

  public static boolean romSCSIINROMRequired;  //true=scsiinromを読み込んでいない
  public static byte[] romSCSIINROMData;  //-romまたは-scsiinromで読み込まれた0x00fc0000～0x00fc2000のデータ

  //romInit ()
  //  ROMを初期化する
  public static void romInit () {

    romROMName = Settings.sgsGetString ("rom");
    romIPLROMName = Settings.sgsGetString ("iplrom");
    romROMDBOn = Settings.sgsGetOnOff ("romdb");
    romSCSIINROMName = Settings.sgsGetString ("scsiinrom");
    romROM30Name = Settings.sgsGetString ("rom30");

    romVersionLong = -1;
    romVersionShort = -1;
    romPreviousVersionLong = -1;
    romPreviousVersionShort = -1;

    romDebuggerVersion = -1;
    romHumanVersion = -1;

    romSRAMInitialData = 0;
    romBootDeviceRead = 0;
    romBootROMAddressRead = 0;
    romBootRAMAddressRead = 0;
    romBootinfROMAddressRead = 0;
    romBootinfRAMAddressRead = 0;

    romROM30Required = true;
    romROM30Data = null;

    romSCSIINROMRequired = true;
    romSCSIINROMData = null;

    boolean iplromRequired = true;

    //-romで指定されたROMのイメージファイル(1MB)を読み込む
    {
      byte[] array = XEiJ.rscGetFile (romROMName, 1024 * 1024);
      if (array != null) {  //読み込めた
        System.arraycopy (array, 0, MainMemory.mmrM8, 0x00f00000, 1024 * 1024);
        iplromRequired = false;
        //  ROM30
        if (MainMemory.mmrRls (0x00fc021c + 0x003e) == ('2' << 24 | '.' << 16 | '1' << 8 | '5')) {  //Human68k version 2.15があるのでrom30は読み込み済み
          romROM30Data = new byte[1024 * 128];
          System.arraycopy (MainMemory.mmrM8, 0x00fc0000, romROM30Data, 0, 1024 * 128);
          romROM30Required = false;
        }
        //  SCSIINROM
        if (MainMemory.mmrRls (0x00fc0024) == ('S' << 24 | 'C' << 16 | 'S' << 8 | 'I') &&
            MainMemory.mmrRwz (0x00fc0028) == ('I' << 8 | 'N') &&
            (MainMemory.mmrRls (0x00fc0000) & -0x2000) == 0x00fc0000) {  //SCSIINフラグがあってROM起動ハンドルがscsiinromを指しているのでscsiinromは読み込み済み
          romSCSIINROMData = new byte[8192];
          System.arraycopy (MainMemory.mmrM8, 0x00fc0000, romSCSIINROMData, 0, 8192);
          romSCSIINROMRequired = false;
        }
      } else {  //読み込めなかった
        romROMName = "";
      }
    }

    //IPLROMのイメージファイル(128KB)を読み込む
    //  -romが指定されていても-iplromが指定されたときは指定されたものを読み込む
    {
      byte[] array = XEiJ.rscGetFile (romIPLROMName, 1024 * 128);
      if (array != null) {  //読み込めた
        System.arraycopy (array, 0, MainMemory.mmrM8, 0x00fe0000, 1024 * 128);
        iplromRequired = false;
      } else {  //読み込めなかった
        romIPLROMName = "";
      }
    }
    //  -romと-iplromのどちらも指定されなかったかされても読み込めなかったときはリソースからIPLROM*.DATを読み込む
    if (iplromRequired) {
      byte[] array = XEiJ.rscGetResource (XEiJ.mdlIPLROMRequest == 0 ? "IPLROM.DAT" :
                                          XEiJ.mdlIPLROMRequest == 1 ? "IPLROMXV.DAT" :
                                          XEiJ.mdlIPLROMRequest == 2 ? "IPLROMCO.DAT" :
                                          "IPLROM30.DAT", 1024 * 128);
      if (array != null) {  //読み込めた
        System.arraycopy (array, 0, MainMemory.mmrM8, 0x00fe0000, 1024 * 128);
        iplromRequired = false;
      }
    }

    //IPLROMのバージョンを確認する
    romCheck ();
    //IPLROMにパッチをあてる
    romPatch ();

    //-rom30で指定されたX68030のROMの拡張部分のイメージファイル(128KB)を読み込む
    //  -romが指定されていても-rom30が指定されたときは指定されたものを読み込む
    {
      byte[] array = XEiJ.rscGetFile (romROM30Name, 1024 * 128);
      if (array != null) {  //読み込めた
        if (3 <= romVersionShort) {
          System.arraycopy (array, 0, MainMemory.mmrM8, 0x00fc0000, 1024 * 128);
        }
        romROM30Data = array;
        romROM30Required = false;
      } else {  //読み込めなかった
        romROM30Name = "";
      }
    }

    //-scsiinromで指定されたSCSIINROMのイメージファイル(8KB)を読み込む
    //  -romが指定されていても-scsiinromが指定されたときは指定されたものを読み込む
    {
      byte[] array = XEiJ.rscGetFile (romSCSIINROMName, 8192);
      if (array != null) {  //読み込めた
        if (romVersionShort < 3) {
          System.arraycopy (array, 0, MainMemory.mmrM8, 0x00fc0000, 8192);
        }
        romSCSIINROMData = array;
        romSCSIINROMRequired = false;
      } else {  //読み込めなかった
        romSCSIINROMName = "";
      }
    }

    romCheck2 ();  //ROMデバッガのバージョンを確認する
    romPatch2 ();  //ROMデバッガにパッチをあてる
    romCheck3 ();  //ROM Humanのバージョンを確認する
    romPatch3 ();  //ROM Humanにパッチをあてる

  }  //romInit()

  //romCheck ()
  //  IPLROMのバージョンを確認する
  //    IPLROM 1.0 ($10870318)  0x00ff000a  初代
  //    IPLROM 1.0 ($10870507)  0x00ff000a  ACE,EXPERT,PRO,SUPER
  //    IPLROM 1.1 ($11910111)  0x00ff000a  XVI
  //    IPLROM 1.2 ($12911024)  0x00ff000a  Compact
  //    IPLROM 1.3 ($13921127)  0x00ff0032  X68030
  public static void romCheck () {
    int versionLong = MainMemory.mmrRls (0x00ff000a);
    if (versionLong == 0x10870318) {  //IPLROM 1.0 ($10870318)
      romVersionLong = versionLong;
      romSRAMInitialData           = 0x00ff078a;  //SRAM初期化データの先頭
      romBootDeviceRead            = 0x00ff0144;  //起動シーケンスで起動デバイスを読み出すmove.w $00ED0018,d1の先頭
      romBootROMAddressRead        = 0x00ff0236;  //起動シーケンスでROM起動ハンドルを読み出すmovea.l $00ED000C,a0の先頭
      romBootRAMAddressRead        = 0x00ff026e;  //起動シーケンスでRAM起動アドレスを読み出すmovea.l $00ED0010,a0の先頭
      romBootinfROMAddressRead     = 0x00ffaa62;  //IOCS _BOOTINFでROM起動ハンドルを読み出すmove.l $00ED000C,d0の先頭
      romBootinfRAMAddressRead     = 0x00ffaa70;  //IOCS _BOOTINFでRAM起動アドレスを読み出すmove.l $00ED0010,d0の先頭
      FontEditor.fntHan6x12Address = 0x00ffcff6;  //IPLROMにあるANK 6x12フォントの先頭
    } else if (versionLong == 0x10870507) {  //IPLROM 1.0 ($10870507)
      romVersionLong = versionLong;
      romSRAMInitialData           = 0x00ff0792;  //SRAM初期化データの先頭
      romBootDeviceRead            = 0x00ff0144;  //起動シーケンスで起動デバイスを読み出すmove.w $00ED0018,d1の先頭
      romBootROMAddressRead        = 0x00ff0236;  //起動シーケンスでROM起動ハンドルを読み出すmovea.l $00ED000C,a0の先頭
      romBootRAMAddressRead        = 0x00ff026e;  //起動シーケンスでRAM起動アドレスを読み出すmovea.l $00ED0010,a0の先頭
      romBootinfROMAddressRead     = 0x00ffaa88;  //IOCS _BOOTINFでROM起動ハンドルを読み出すmove.l $00ED000C,d0の先頭
      romBootinfRAMAddressRead     = 0x00ffaa96;  //IOCS _BOOTINFでRAM起動アドレスを読み出すmove.l $00ED0010,d0の先頭
      FontEditor.fntHan6x12Address = 0x00ffd018;  //IPLROMにあるANK 6x12フォントの先頭
    } else if (versionLong == 0x11910111) {  //IPLROM 1.1 ($11910111)
      romVersionLong = versionLong;
      romSRAMInitialData           = 0x00ff0842;  //SRAM初期化データの先頭
      romBootDeviceRead            = 0x00ff01e6;  //起動シーケンスで起動デバイスを読み出すmove.w $00ED0018,d1の先頭
      romBootROMAddressRead        = 0x00ff02d8;  //起動シーケンスでROM起動ハンドルを読み出すmovea.l $00ED000C,a0の先頭
      romBootRAMAddressRead        = 0x00ff0310;  //起動シーケンスでRAM起動アドレスを読み出すmovea.l $00ED0010,a0の先頭
      romBootinfROMAddressRead     = 0x00ffadb4;  //IOCS _BOOTINFでROM起動ハンドルを読み出すmove.l $00ED000C,d0の先頭
      romBootinfRAMAddressRead     = 0x00ffadc2;  //IOCS _BOOTINFでRAM起動アドレスを読み出すmove.l $00ED0010,d0の先頭
      FontEditor.fntHan6x12Address = 0x00ffd344;  //IPLROMにあるANK 6x12フォントの先頭
    } else if (versionLong == 0x12911024) {  //IPLROM 1.2 ($12911024)
      romVersionLong = versionLong;
      romSRAMInitialData           = 0x00ff08ee;  //SRAM初期化データの先頭
      romBootDeviceRead            = 0x00ff0236;  //起動シーケンスで起動デバイスを読み出すmove.w $00ED0018,d1の先頭
      romBootROMAddressRead        = 0x00ff032c;  //起動シーケンスでROM起動ハンドルを読み出すmovea.l $00ED000C,a0の先頭
      romBootRAMAddressRead        = 0x00ff0368;  //起動シーケンスでRAM起動アドレスを読み出すmovea.l $00ED0010,a0の先頭
      romBootinfROMAddressRead     = 0x00ffaece;  //IOCS _BOOTINFでROM起動ハンドルを読み出すmove.l $00ED000C,d0の先頭
      romBootinfRAMAddressRead     = 0x00ffaedc;  //IOCS _BOOTINFでRAM起動アドレスを読み出すmove.l $00ED0010,d0の先頭
      FontEditor.fntHan6x12Address = 0x00ffd45e;  //IPLROMにあるANK 6x12フォントの先頭
    } else {
      versionLong = MainMemory.mmrRls (0x00ff0032);
      if (versionLong == 0x13921127 ||  //IPLROM 1.3 ($13921127)
          romPreviousVersionLong == 0x13921127) {
        romVersionLong = versionLong;
        romSRAMInitialData           = 0x00ff09e8;  //SRAM初期化データの先頭
        romBootDeviceRead            = 0x00ff0386;  //起動シーケンスで起動デバイスを読み出すmove.w $00ED0018,d1の先頭
        romBootROMAddressRead        = 0x00ff045e;  //起動シーケンスでROM起動ハンドルを読み出すmovea.l $00ED000C,a0の先頭
        romBootRAMAddressRead        = 0x00ff048e;  //起動シーケンスでRAM起動アドレスを読み出すmovea.l $00ED0010,a0の先頭
        romBootinfROMAddressRead     = 0x00ffa9f0;  //IOCS _BOOTINFでROM起動ハンドルを読み出すmove.l $00ED000C,d0の先頭
        romBootinfRAMAddressRead     = 0x00ffa9fe;  //IOCS _BOOTINFでRAM起動アドレスを読み出すmove.l $00ED0010,d0の先頭
        FontEditor.fntHan6x12Address = 0;  //IPLROM 1.3にはANK 6x12フォントがない
      } else {  //不明
        romVersionLong = -1;
        romVersionShort = -1;
        romPreviousVersionLong = -1;
        romPreviousVersionShort = -1;
        versionLong = -1;
        romSRAMInitialData           = 0;
        romBootDeviceRead            = 0;
        romBootROMAddressRead        = 0;
        romBootRAMAddressRead        = 0;
        romBootinfROMAddressRead     = 0;
        romBootinfRAMAddressRead     = 0;
        FontEditor.fntHan6x12Address = 0;
      }
    }
    if (romVersionLong != -1) {
      romVersionShort = romVersionLong >> 24 & 15;
      System.out.printf (Multilingual.mlnJapanese ?
                         "IPLROM 1.%d ($%08X) があります\n" :
                         "IPLROM 1.%d ($%08X) exists\n",
                         romVersionShort, romVersionLong);
    } else {
      romVersionShort = -1;
      System.out.println (Multilingual.mlnJapanese ?
                          "IPLROM がありません" :
                          "IPLROM does not exist");
    }
  }  //romCheck()

  //romCheck2 ()
  //  ROMデバッガのバージョンを確認する
  public static void romCheck2 () {
    if (MainMemory.mmrRls (0x00fe00d6 + 20) == ('1' << 24 | '.' << 16 | '0' << 8 | 0x0d)) {  //3文字
      romDebuggerVersion = 0x0100;
      System.out.println (Multilingual.mlnJapanese ? "ROM デバッガ 1.0 があります" : "ROM debugger 1.0 exists");
    } else if (MainMemory.mmrRls (0x00fd3a12 + 24) == ('2' << 24 | '.' << 16 | '3' << 8 | '2')) {
      romDebuggerVersion = 0x0232;
      System.out.println (Multilingual.mlnJapanese ? "ROM デバッガ 2.32 があります" : "ROM debugger 2.32 exists");
    } else {
      romDebuggerVersion = -1;
      System.out.println (Multilingual.mlnJapanese ? "ROM デバッガがありません" : "ROM debugger does not exist");
    }
  }  //romCheck2()

  //romCheck3 ()
  //  ROM Humanのバージョンを確認する
  public static void romCheck3 () {
    if (MainMemory.mmrRls (0x00fe5000 + 0x6fc0 - 0x6800 + 37) == ('1' << 24 | '.' << 16 | '0' << 8 | '0')) {  //奇数アドレス
      romHumanVersion = 0x0100;
      System.out.println (Multilingual.mlnJapanese ? "ROM Human 1.00 があります" : "ROM Human 1.00 exists");
    } else if (MainMemory.mmrRls (0x00fc025a) == ('2' << 24 | '.' << 16 | '1' << 8 | '5')) {
      romHumanVersion = 0x0215;
      System.out.println (Multilingual.mlnJapanese ? "ROM Human 2.15 があります" : "ROM Human 2.15 exists");
    } else {
      romHumanVersion = -1;
      System.out.println (Multilingual.mlnJapanese ? "ROM Human がありません" : "ROM Human does not exist");
    }
  }  //romCheck3()

  //  perl misc/rompattobytes.pl ROM_PATCH13 misc/rompat13.x
  public static final int ROM_PATCH13_BASE = 0x00fe4000;
  public static final byte[] ROM_PATCH13_TEXT = "This is an IPLROM modified by XEiJ. Distribution is prohibited.\0C\372\1`!\311\0\20!\311\0,\"OA\372\1\34~\0p\0N{\b\1~\1p\1\300;\2\373g\0\1\bA\372\0004\6\372\0\0\0\32~\2p\bN{\0\2A\372\0\n\362\200\0\0\b\307\0\17`\0\0\346\0\0\0\0\0\376@\230\0\0\0\0\0\0\0\0\360\0\6\317.IA\372\0@Nz\b\2~\3 <\0\0\b\bN{\0\2/<\0\3304@\360\27@\0.\274\0\0\0\0\360\27\b\0\360\27\f\0X\217A\372\0\n\362\200\0\0\b\307\0\17a\0\0\314`\0\0\216A\372\0<Nz\b\5~\4p\0N{\0\2\364\330p\0N{\0\3\365\30p\0N{\0\4N{\0\5N{\0\6N{\0\7A\372\0\n\362\200\0\0\b\307\0\17a\0\0\376`NA\372\0JNz\30\b~\6 <\0\200\0\0N{\0\2\364\330 <\0\0\2\20N{\0\3\365\30p\0N{\0\4N{\0\5N{\0\6N{\0\7p\1N{\b\bA\372\0\n\362\200\0\0\b\307\0\17a\0\0\262`\2~\0\221\310C\371\0\377\7pr\1\340\231p\177 \311\323\301 \311\323\301Q\310\377\366\276<\0\6f\bC\372\1 !\311\0\364C\372\2\f!\311\0`N\371\0\377\0~.IN\320O\357\377\200 O \374\177\377\0\2 \17\350\210T\200\351\210 \200A\370\0\1\"@p\7\"\310\321\374\0 \0\0Q\310\377\366#|\0\360\0\1\377\344\360\27L\0/<\200\3304@\360\0$\0\360\27@\0C\371\0\360\0\0A\371\0 \0\0000<\7\377\261\211V\310\377\374f\4\b\307\0\37\b\227\0\7\360\27@\0\360\0$\0O\357\0\204NuC\370 \0A\351\2\2p\177\"\310Q\310\377\374A\351\2\2p\177\"\310Q\310\377\374p\37\"\374\0\377\4AQ\310\377\370\"|\0\177\300@N{\220\4N{\220\6N{\220\5N{\220\7C\370 \0N{\230\7N{\230\6\"|\0\0\300\0\365\30\364\330N{\220\3A\371\200\377\0\0C\371\200\360\0\0000<\7\377\263\210V\310\377\374f\4\b\307\0\37\"|\0\0\2\0N{\220\3\365\30\364\330\223\311N{\220\4N{\220\6N{\220\5N{\220\7NuHo\0\bH\347\377\376t\5\b/\0\5\0@f\bt\1Nh/H\0< o\0B2\0300\1\300|\3618\260|\1\bg\22L\327\0\7 o\0 O\357\0@N\371\0\377\7pp\7\300A\"w\4 \322\330/H\0B0\1\340HA\367\2\376\322\1eDNz\20\0N{ \0k\22\16\21\0\0\341H\16)\0\0\0\0021@\0\2`\36\16\21\0\0\341\210\16)\0\0\0\2\341\210\16)\0\0\0\4\341\210\16)\0\0\0\6 \200N{\20\0L\327\0\377`FNz\20\1N{ \1k\0240(\0\2\341X\16\21\b\0\341X\16)\b\0\0\2`  \20\341\230\16\21\b\0\341\230\16)\b\0\0\2\341\230\16)\b\0\0\4\341\230\16)\b\0\0\6N{\20\1L\327\0\7L\357\3\0\0 O\357\0@J\27j\4\0|\200\0Nsp\0\209\0\350\340\13F\0\351H\350\b1\300\f\266\f8\0\2\f\274e0 <\0\0!\1\f8\0\4\f\274e\30g\22Nz\b\b\b\300\0\0N{\b\b <\0\200\0\0000<\200\0N{\0\2B9\0\350\340\tA\372\1\ba81\300\f\270O\357\377\342 \17\320|\0\16\300|\377\360X@ @!z\0\350\377\374 \272\0\346!z\0\346\0\4!z\0\344\0\ba\nO\357\0\0361\300\f\272Nut\16@\347\0|\7\0C\371\0\350\200\35\37)\377\354\37)\377\370\37\21\2)\0\317\377\354\2)\0\317\377\370\22\274\0\0J\21f\374\23|\0\0\0\6\23|\0\0\0\b <\0@\0\0\344\250S\200\22\274\0qN\220\22\274\0\0J\21f\374p\0\220)\0\6r\0\222)\0\b\23|\0\310\0\6\23|\0\0\0\b\22\237\23_\377\370\23_\377\354F\337\300\374\0002\260\1c\4\320|\1\0\20\1S@JBg\f\260|\23\210d\6SB`\0\377pC\372\0Br\0\228\f\274\262|\0\6b\4\345I\322\301\"\21\344\251J\200g\24\260\274\0\0\377\377b\f\202\300h\2r\377p\0000\1Nup\0NuNqH@Q\310\377\376H@Q\310\377\366Nu\372\0\0\0\372\0\0\0\372\0\0\0\372\0\0\0\246\252\252\253\0\0\0\0)\252\252\2532<C\20\b8\0\7\b\5f\0302<L\20\b8\0\6\b\4f\f2<L\23\b8\0\5\b\5g\30\23\374\0001\0\350\340\r\23\301\0\355\0\35B9\0\350\340\rp\20NONuH\347\377\376O\357\377\200,Oa2a\0\0\340a\0\1\260a\0\4~a\0\4\342a\0\5La\0\2Nr\377t\377p.NOBAWBp#NOO\357\0\200L\337\177\377Nu NC\372\0va\0\7\6p\217NO\"\0B@\222\200H@\351\210\351\211\350H\350I\341\210\341\211\350H\350I\350\b\350\t\200\2740000\202\2740000\341\230\20\300\20\374\0.\341\230\20\300C\372\0_a\0\6\306\341\230\20\300\341\230\20\300\20\374\0-\341\231\20\301\341\231\20\301\20\374\0-\341\231\20\301\341\231\20\301C\372\0<a\0\6\236\"Np!NONu\33[1m                 BIOS ROM version : \0 (20\0)\33[1m\r\n\0r\0p\254NO*\0 NC\372\0\230a\0\6Nt\377rM\272<\0\6f\22Nz\b\b\340Ht\0\24\0\264<\0\6d\2rX\20\301C\372\0\227a\0\6(\272<\0\3e\36b\b\b\5\0\16f\26`\fJEk\20rL\b\5\0\16f\2rE\20\301\20\374\0C\20\374\0000r0\322\5\20\301\20\374\0000 \2k\n\20\374\0-r\3a\0\5n\20\374\0 \20\374\0( \5B@H@a\0\5\"\20\350\377\377\21|\0.\377\376C\372\0007a\0\5\304\"Np!NONu      Micro-Processing Unit (MPU) : \0C68\0MHz)\r\n\0JEj@ NC\372\0>\272<\0\4d\4C\372\0Ya\0\5tC\372\0v\272<\0\4d\30\362\0\\\1\362\'h\0 \37\200\237\200\237C\372\0gg\4C\372\0ka\0\5N\"Np!NONu        Floating-Point Unit (FPU) : \0Floating-Point Coprocessor (FPCP) : \0on MPU\r\n\0MC68881\r\n\0MC68882\r\n\0\0\272<\0\4e&a\0\0\372k  NC\372\0\202a\0\4\310J\200C\372\0dg\4C\372\0ha\0\4\270\"Np!NOp\0a\0\1\202k  NC\372\0\177a\0\4\240J\200C\372\0<g\4C\372\0@a\0\4\220\"Np!NOp\1a\0\1Zk  NC\372\0|a\0\4xJ\200C\372\0\24g\4C\372\0\30a\0\4h\"Np!NONuMC68881\r\n\0MC68882\r\n\0          Motherboard Coprocessor : \0    Extension Board Coprocessor 1 : \0    Extension Board Coprocessor 2 : \0\0H\347x\340p\1\300;\2\373S\200k\0\0\232A\371\0\2 \0p\377@\302\0|\7\0Nz@\1Nz0\0r\7N{\20\1N{\20\0C\372\0f$x\0\b!\311\0\b\"OBA\16h\30\0\0\6\16h\20\0\0\6\16P\20\0\262|\b\2f\3662<\\\1\16h\30\0\0\n\16P\20\0\262|\b\2f\3662<h\0\16h\30\0\0\n\16P\20\0\262|2\ff\366\16\250\0\0\0\20\16\250\20\0\0\20\200\201\16\250\20\0\0\20\200\201g\2p\1.I!\312\0\bN{0\0N{@\1F\302L\337\7\36J\200NuH\347`\340A\371\0\351\340\0J\200g\4A\350\0\200p\377@\302\0|\7\0C\372\0F$x\0\b!\311\0\b\"OBh\0\6Jh\0\6\fP\b\2f\3721|\\\1\0\n\fP\b\2f\3721|h\0\0\n\fP2\ff\372 (\0\20\"(\0\20\200\201\"(\0\20\200\201g\2p\1.I!\312\0\bF\302L\337\7\6J\200Nu\b\5\0\16g( NC\372\0&a\0\2\252\272<\0\3d\nC\372\0=a\0\2\234`\bC\372\0=a\0\2\222\"Np!NONu     Memory Management Unit (MMU) : \0MC68851\r\n\0on MPU\r\n\0@\347\0|\7\0/8\0\b*O!\374\0\376J\202\0\b\225\312 \nJ\22\325\374\0\20\0\0\265\374\0\300\0\0e\360.M!\337\0\bF\337\"\n NC\372\0\22a\0\2\26a\0\1\0\"Np!NONu               Motherboard Memory : $\0@\347\0|\7\0/8\0\b*O!\374\0\376J\362\0\b\225\312C\370\"\0002<\0\377\225\374\1\0\0\0S\211\22\222\24\201.MQ\311\377\360!\374\0\376K\30\0\b\225\312C\370!\0002<\0\377\20\1R\0\225\374\1\0\0\0\23\0\22\222.MQ\311\377\360!\374\0\376K>\0\b\225\312C\370!\0002<\0\377\225\374\1\0\0\0\223!f\4\24\251\1\0.MQ\311\377\356!\337\0\bF\337\225\312G\370 \1J\23f\b \nf\24$K`\20 \ng\f Na \"Np!NO\225\312R\213\266\374 \377c\334 \ng\n Na\b\"Np!NONu \nH@\341\210\"\13HA\341\211C\372\08a\0\1\24a\0\0\344\20\374\0-\20\374\0$/\0 \1S\200a\0\0\322\20\374\0 \20\374\0( \1\220\237H@\350Ha6C\372\0.a\0\0\344Nu             Daughterboard Memory : $\0MB)\r\n\0H\347\340@J\200f\6\20\374\0000`$j\6\20\374\0-D\200C\372\0D\260\231e\374Y\211`\nR\1\220\202d\372\320\202\20\301r/$\31f\360B\20L\337\2\7NuH\347\340@t\n\224\201d\2t\0\345JC\373 \22g\336\260\251\377\374e\330 )\377\374S\200`\320;\232\312\0\5\365\341\0\0\230\226\200\0\17B@\0\1\206\240\0\0\'\20\0\0\3\350\0\0\0d\0\0\0\n\0\0\0\1\0\0\0\0H\347\340\0t\7\351\230r\17\302\0\20\373\20\16Q\312\377\364B\20L\337\0\7Nu0123456789ABCDEF\20\331f\374S\210NuH\347q\316M\371\0\350\200\0~\0\368\t<HG\262|\377\377g\0\2d>\1B\7\276|C\0g\6\276|L\0f(\23\374\0001\0\350\340\r\276|L\0g\n\b\271\0\0\0\355\0\225`\b\b\371\0\0\0\355\0\225B9\0\350\340\rBGa\0\0024K\372\4\\\332\354\0\4v\7\306,\0\6=|\0\0\246\0JGf8=|\0\0\204\200p\37NO\b\356\0\3\200(=|\0013\200*r\0t\0A\371\0\340\0\0000<?\377 \301 \302Q\310\377\372=|\0003\200*\b\256\0\3\200(4\3\341J\24-\0\21p\37\300.\200)\260|\0\27f\2p\6r\37\302\2\262|\0\27f\2r\6\350X\350Y\262@d.=B\200(J-\0\22f\b\b\256\0\1`\7`\6\b\356\0\1`\7A\356\200\2C\355\0\0020\331 \331 \331 \331=U\200\0`&A\356\200\0C\325 \331 \331 \331 \331=B\200(J-\0\22f\b\b\256\0\1`\7`\6\b\356\0\1`\7p\0\20-\0\23=@\200\20p\0A\356\200\0220\300 \300 \300 \300 \300 \300T\2100\374\0003 \3000\300A\371\0\353\0\0001|\0\20\b\bp\22\300-\0\21\260|\0\22gBr\37\302-\0\21\262<\0\21f\bJ-\0\22g\2S\1p\4\320m\0\0041@\b\fp\na\0\0010p\3\300\1f\0061U\b\n`\0061|\0\377\b\n1m\0\f\b\0161A\b\20`\0322<\0\3771A\b\fp\na\0\1\0041A\b\n1A\b\0161A\b\20!\374\0\340\0\0\tDB\270\tH0\0242,\0\2\346H\350IS@SA1\300\tp1\301\trB\270\tt0\0242,\0\2\266<\0\4d\0204<\2\0\260Bc\0020\2\262Bc\0022\2S@SAB\270\th1\300\tl1\301\tnB\201$8\tlpwNOp\4\300\3X\0\341H!\300\t`Bx\tdp\4\346(W\301\220\1\21\300\t=JGf<a\0\0042A\371\0\355\0.C\356\242\0\"\330\"\330 \0202\0H@2\3002\3002\3002\3000\1\"\300\"\300\"\300\"\300\35y\0\355\0(`\1=|\6\344\245\0p\36NO=|\0 \246\0BGHG \7L\337s\216Nu\302|\0\377\262|\0\34e\6\229\0\355\0\35\21\301\t<\b9\0\0\0\355\0\225g\4\322|\0\34I\372\0H\347I\330\301NuH\347\340\00028\f\270\f8\0\4\f\274e\16g\6\302\374L\315`\n\302\374\0233`\4\302\374\f\315HA4\1`\bH@2\2Q\311\377\376Q\310\377\370H@Q\310\377\360L\337\0\7Nu\2\0\2\0\0\0\4\0\2\0\2\0\0\24\4\0\1\0\1\0\0(\4\0\1\0\1\0\0<\4\0\2\0\2\0\0\0\0\0\2\0\2\0\0\24\0\0\1\0\1\0\0(\0\0\1\0\1\0\0<\0\0\2\0\2\0\0\0\1\0\2\0\2\0\0\24\1\0\1\0\1\0\0(\1\0\1\0\1\0\0<\1\0\2\0\2\0\0\0\3\0\2\0\2\0\0\24\3\0\1\0\1\0\0(\3\0\1\0\1\0\0<\3\0\3\0\2\0\0P\4\0\4\0\1\250\0d\4\0\4\0\3P\0x\4\0\2\200\1\340\0\214\4\0\3\0\2\0\0P\1\0\4\0\1\250\0d\1\0\4\0\3P\0x\1\0\2\200\1\340\0\214\1\0\3\0\2\0\0P\3\0\4\0\1\250\0d\3\0\4\0\3P\0x\3\0\2\200\1\340\0\214\3\0\2\0\2\0\0\240\4\0\2\0\2\0\0\240\4\0\1\0\1\0\0\310\4\0\1\0\1\0\0\334\4\0\2\0\2\0\0\240\0\0\2\0\2\0\0\240\0\0\1\0\1\0\0\310\0\0\1\0\1\0\0\334\0\0\2\0\2\0\0\240\1\0\2\0\2\0\0\240\1\0\1\0\1\0\0\310\1\0\1\0\1\0\0\334\1\0\2\0\2\0\0\240\3\0\2\0\2\0\0\240\3\0\1\0\1\0\0\310\3\0\1\0\1\0\0\334\3\0\3\0\2\0\0\360\4\0\3\0\2X\1\4\4\0\3\0\4\0\1\30\4\0\2\200\1\340\1,\4\0\3\0\2\0\0\360\1\0\3\0\2X\1\4\1\0\3\0\4\0\1\30\1\0\2\200\1\340\1,\1\0\3\0\2\0\0\360\3\0\3\0\2X\1\4\3\0\3\0\4\0\1\30\3\0\2\200\1\340\1,\3\0\0[\0\t\0\21\0Q\0027\0\5\0(\2(\0\25\0\33\0K\0\3\0\5\0E\1\3\0\1\0\20\1\0\0\5\0,\0-\0\4\0\6\0&\0027\0\5\0(\2(\0\20\0\33\0%\0\1\0\0\0 \1\3\0\1\0\20\1\0\0\0\0$\0\211\0\16\0\34\0|\0027\0\5\0(\2(\0\26\0\33\0\257\0\17\0\37\0\237\1\320\0\7\0 \1\310\0\26\0\33\0\257\0\17\0\37\0\237\1\320\0\7\0 \1\310\0\32\0\33\0\211\0\16\0$\0t\0027\0\5\08\2\30\0\26\0\33\0Q\0\5\0\13\0K\2p\0\1\0C\2C\0\25\0\33\0[\0\7\0\21\0Q\2p\0\2\0C\2C\0\26\0\33\0)\0\2\0\3\0#\2p\0\1\0C\2C\0\20\0\33\0-\0\3\0\6\0&\2p\0\2\0C\2C\0\21\1\33\0{\0\b\0\23\0s\2p\0\1\0C\2C\0\26\0\33\0{\0\b\0\23\0s\2p\0\1\0\27\2o\0\26\0\33\0{\0\b\0\23\0s\2p\0\1\0C\2C\0\32\0\33\0c\0\13\0\r\0]\2\f\0\1\0!\2\1\0\27\0\33H\347\360\312M\371\0\350\200\0\228\t<a\0\374\204v\7\306,\0\6\2n\377\340\246\0\35|\0\13\200(A\371\0\300\0\0p\0r\377 \300 \300Q\311\377\372aBp\0170|\4\0r\17\266|\0\1e\22P\301g\16H\201\266|\0\3c\6p\20\321\310r\17!\374\0\300\0\0\t\\!\310\t`1\301\td\35C\200(=C\244\0\201n\246\0L\337S\17NuA\371\0\350 \0\266|\0\1e\bg\24\266|\0\3c\34C\372\0002p\7 \331Q\310\377\374NuC\372\0Dp\177 \331Q\310\377\374Nu\"<\0\1\0\1$<\2\2\2\2p\177 \301\322\202Q\310\377\372Nu\0\0R\224\0 \0>\4\0\7\300\4 \7\376\200\0\370\0\200 \370>\204\0\377\300\255j\377\376\0\0\0\b\0\22\0\32\0$\0,\0006\0>\1\0\1\b\1\22\1\32\1$\1,\0016\1>\2@\2H\2R\2Z\2d\2l\2v\2~\3@\3H\3R\3Z\3d\3l\3v\3~\4\200\4\210\4\222\4\232\4\244\4\254\4\266\4\276\5\200\5\210\5\222\5\232\5\244\5\254\5\266\5\276\6\300\6\310\6\322\6\332\6\344\6\354\6\366\6\376\7\300\7\310\7\322\7\332\7\344\7\354\7\366\7\376P\0P\bP\22P\32P$P,P6P>Q\0Q\bQ\22Q\32Q$Q,Q6Q>R@RHRRRZRdRlRvR~S@SHSRSZSdSlSvS~T\200T\210T\222T\232T\244T\254T\266T\276U\200U\210U\222U\232U\244U\254U\266U\276V\300V\310V\322V\332V\344V\354V\366V\376W\300W\310W\322W\332W\344W\354W\366W\376\250\0\250\b\250\22\250\32\250$\250,\2506\250>\251\0\251\b\251\22\251\32\251$\251,\2516\251>\252@\252H\252R\252Z\252d\252l\252v\252~\253@\253H\253R\253Z\253d\253l\253v\253~\254\200\254\210\254\222\254\232\254\244\254\254\254\266\254\276\255\200\255\210\255\222\255\232\255\244\255\254\255\266\255\276\256\300\256\310\256\322\256\332\256\344\256\354\256\366\256\376\257\300\257\310\257\322\257\332\257\344\257\354\257\366\257\376\370\0\370\b\370\22\370\32\370$\370,\3706\370>\371\0\371\b\371\22\371\32\371$\371,\3716\371>\372@\372H\372R\372Z\372d\372l\372v\372~\373@\373H\373R\373Z\373d\373l\373v\373~\374\200\374\210\374\222\374\232\374\244\374\254\374\266\374\276\375\200\375\210\375\222\375\232\375\244\375\254\375\266\375\276\376\300\376\310\376\322\376\332\376\344\376\354\376\366\376\376\377\300\377\310\377\322\377\332\377\344\377\354\377\366\377\376p\377\f8\0\2\f\274d\4JAf\24\262|\0\4b\16?\1\322A2;\20\nN\273\20\0062\37Nu\0\n\0>\0h\0x\0\246p\f\f8\0\2\f\274e\2p\6\300\370\f\270\320\274\0\0\0002\200\374\0dH@B@J8\f\276V\300\342XJ8\f\275V\300\342X\208\f\274Nup\0\f8\0\2\f\274e\36Nz\0\2\f8\0\4\f\274d\6\342\230\343\30`\4H@\343X\343\230\300\274\0\0\0\3Nu/\2t\0\249\0\355\0\220a2$\37Nu\f8\0\2\f\274e \f8\0\4\f\274d\32/\0Nz\0\2\200|\b\bN{\0\2\300|\367\367N{\0\2 \37Nu\364\370Nua\226\f8\0\2\f\274eP/\1?\2\f8\0\4\f\274d\26r\1\302B\304|\0\2DB\304|!\0\202BN{\20\2`*Nz\20\2\342J\323A\342Y\342J\323\201\342\231N{\20\0024\27FB\304@g\16UBb\6e\6\364x`\4\364x\364\2304\37\"\37Nu/\2t\0a\236$\37Nu/\2t\1a\224$\37Nu".getBytes (XEiJ.ISO_8859_1);
  public static final byte[] ROM_PATCH13_DATA = "\0\377\0000\0\377\0006 <\23\222\0\0\0\6 <\25!\1\21\0\377\0Z\0\377\0~a\0\f\342\0\0\0\6N\371\0\376@@\0\377\1<\0\377\1\234\f8\0\1\0\0\0\bN\271\0\376C\244`X\0\377\1\262\0\377\1\270B9\0\350\0\0\0\6\21|\0\0\300\1\0\377\2R\0\377\2\202r\23C\370\0\0\0\bN\271\0\376E `(\0\377\2\262\0\377\2\272C\372\b\324\0\0\0\2`\6\0\377\0036\0\377\3<p\0N{\0\0\0\6N\271\0\376V\256\0\377\3\200\0\377\3\206p\0N{\0\0\0\6N\271\0\376V\256\0\377\16\210\0\377\16\232a\0\1\202\0\0\0\bN\271\0\376E^`\n\0\377\20\340\0\377\20\356\22<\0\1\0\0\0\16r\1t\2aRr\1t\0aLr\b\0\377;\374\0\377<\0H\347`\0\0\0\0\4H\347p\0\0\377<\26\0\377<\32L\337\0\6\0\0\0\4L\337\0\16\0\377<\34\0\377< L\337\0\6\0\0\0\4L\337\0\16\0\377<Z\0\377<^H\347`\0\0\0\0\4H\347p\0\0\377<v\0\377<zL\337\0\6\0\0\0\4L\337\0\16\0\377DD\0\377D\216?\0?<\0\0\0H?\0/<\0\270\0\20/8\13\374?<\0\2a\0#nBo\0\nRWa\0#dO\357\0\f0\37g\4a\0e\360L\337\177\376NuRA\262|\0\37c\16\262x\trc\b28\trg\2SANu\0\377D\246\0\377D\250RA\351A\0\0\0\2a\314\0\377Z\250\0\377Z\252+\200\n\0\0\0\0\2;\200\0\377j\332\0\377k\30p\377H\347\0\0\0>H\347`@$\1HB x\4XN\220 @\260\374\fFg$H@\260|\0\360e\6\260|\1\0e\0260\1\300\302\320A\320B\20\331Q\310\377\374p\0L\337\2\6Nup\377`\366\0\377k\214\0\377k\222A\371\0\350\0\0\0\6N\371\0\376L\262\0\377\202\204\0\377\202\246/\0\f8\0\0\0\6N\371\0\376V$\0\377\223\242\0\377\223\244v\3C\370\0\0\0\2v\b\0\377\223\362\0\377\223\364v\1NO\0\0\0\2v\5\0\377\225\"\0\377\225$v\3NO\0\0\0\2v\b\0\377\223\262\0\377\223\330\20\21\2\0\0\0\0&pp\300\21\260<\0pf$p\370\320)\0\2e\34\1<\0Gf\252`\24\340M0\5\346H\232@\342MNu\0\377\224:\0\377\224>\340\215\342\215\0\0\0\2a\220\0\377\225t\0\377\225x\340\215\342\215\0\0\0\4a\0\376V\0\377\226\234\0\377\226\256\fE\0\3\0\0\0\20p\377\353\250F\200\324\200\326\200\352\252\352\253Nu\0\377\253\244\0\377\253\250\262|\3\360\0\0\0\4\262|\4\0\0\377\253\264\0\377\253\270\264|\3\360\0\0\0\4\264|\4\0\0\377\254r\0\377\254\206\f8\0\1\0\0\0\nN\271\0\376V\270/\0`\n\0\377\254\350\0\377\254\366\f8\0\1\0\0\0\n$\37N\271\0\376VR`\4\0\377\263&\0\377\2632H\347`@\0\0\0\6N\371\0\376R\246\0\377\267@\0\377\267D\0260\0\0\0\0\0\4\0260\0\2\0\377\307Z\0\377\310\30H\347`\0\0\0\0\6N\371\0\376U\210\0\377\314\236\0\377\314\27009\0\350\0\0\0\30p\22\300y\0\350\0(\260|\0\22f\6p\377X\217Nup\0Nu\0\377\314\270\0\377\316\16H\347Pb\0\0\0\342H\347Pb,z\306~p:\275\374\0\351` f\b\f.\0\377\0\rg\4\262\200e\2\"\0\322A2;\20\fN\273\20\bL\337F\nNu\1&\2\24\1\356\2\342\b\330\b\220\3\302\4\0\4>\4\252\4\276\3~\3>\4z\0v\0v\0v\0v\0v\0v\0v\0v\0v\0v\0v\0v\0v\0v\0v\0v\0v\0v\t*\nZ\n\264\fD\7\b\7\264\13\22\13x\f\30\t\242\t\342\7&\tf\r\20\7D\f\252\f\334\n\"\fx\0v\0v\0v\rD\r\206\r\276\r\366\0vp\377Nu\f8\0\6\f\274g\6\1\316\0\27Nu\341\230\35@\0\27\341\230\35@\0\31\341\230\35@\0\33\341\230\35@\0\35Nu\341\213\272<\0\3d\4\353\253`\2\347\213Nu\0\377\317F\0\377\317J\1\316\0\27\0\0\0\4a\0\376\32\0\377\317\226\0\377\317\232\1\316\0\27\0\0\0\4a\0\375\312\0\377\321\312\0\377\321\316\1\316\0\27\0\0\0\4a\0\373\226\0\377\324`\0\377\324d\341\213\353\253\0\0\0\4a\0\371(\0\377\327x\0\377\327|\341\213\353\253\0\0\0\4a\0\366\20\0\377\327\322\0\377\327\326\341\213\353\253\0\0\0\4a\0\365\266\0\377\330:\0\377\330>\341\213\353\253\0\0\0\4a\0\365N\0\377\330\234\0\377\330\240\341\213\353\253\0\0\0\4a\0\364\354\0\377\3334\0\377\3338\1\316\0\27\0\0\0\4a\0\362,\0\377\333\226\0\377\333\232\1\316\0\27\0\0\0\4a\0\361\312\0\377\320f\0\377\320\252H\347\0\2\0\0\0:H\347p\372,z\302\320A\356\0\tE\356\0\rG\356\0\25I\356\0\13J\203f\0046<\1\0\20\20r \302\0f\0\2\346r\b\302\0f\0\2\322\20\24j\352`\0\1\302\0\377\322`\0\377\323~H\347px\0\0\1\36\300<\0\7\260<\0\1f\0\0\376\35@\0\21\35|\0\0\0\27 \3H@\35@\0\31\341\230\35@\0\33\341\230\35@\0\35\20\220\35|\0\200\0\5p\360\300\22\260<\0\260f\366$\3\346\212g\0\0\246p\1\300;\n\373fx`.p\367\300\20f\0\0\230\b\22\0\0g\n`\0\0\216HBJ\20f\350\b\22\0\1g\366\22\323\22\323\22\323\22\323\22\323\22\323\22\323\22\323Q\312\377\344HBQ\312\377\334\306|\0\7`\\p\367\300\20fZ\b\22\0\0g\6`RJ\20f\356\b\22\0\1g\366\20\23\341H\20\23H@\20\23\341H\20\23\"\300\20\23\341H\20\23H@\20\23\341H\20\23\"\300Q\203d\322\306|\0\7`\32p\367\300\20f\30\b\22\0\0g\f`\20J\20f\356\b\22\0\0f\366\22\323Q\313\377\362p\0\20\20g\374\20\200r\20\302\0g\br\347\302\0f\2p\0L\337_\16Nup\0\20\20\20\200H@\20\24`\356r\0p\365NOp\377`\344\0\377\334\352\0\377\334\356\260|\377\377\0\0\0\4\262|\377\377\0\0\0\0".getBytes (XEiJ.ISO_8859_1);

  //romPatch ()
  //  IPLROMにパッチをあてる
  public static void romPatch () {

    //IPLROMのバグまたはおかしなところを修正する
    int patched = 0;
    int failed = 0;
    switch (romVersionLong) {

    case -1:  //IPLROMがない
      return;

    case 0x10870318:  //IPLROM 1.0 ($10870318)
      //
      //  電卓を使うと実行中のプログラムが誤動作することがある
      //    https://stdkmd.net/bugsx68k/#rom_dentakud3
      if (MainMemory.mmrRls (0x00ff32c4) == 0x48e76000) {
        MainMemory.mmrWw (0x00ff32c4 + 2, 0x7000);  //00FF32C4  48E76000  movem.l d1-d2,-(sp)  →  48E77000  movem.l d1-d3,-(sp)
        MainMemory.mmrWw (0x00ff32de + 2, 0x000e);  //00FF32DE  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        MainMemory.mmrWw (0x00ff32e4 + 2, 0x000e);  //00FF32E4  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        MainMemory.mmrWw (0x00ff332e + 2, 0x7000);  //00FF332E  48E76000  movem.l d1-d2,-(sp)  →  48E77000  movem.l d1-d3,-(sp)
        MainMemory.mmrWw (0x00ff334a + 2, 0x000e);  //00FF334A  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        patched++;
      } else {
        failed++;
      }
      //
      //  カーソルが画面の最下行にあると電卓が画面外に表示される
      //    https://stdkmd.net/bugsx68k/#rom_dentaku64
      //    電卓の表示位置を決めるサブルーチンに電卓のY座標を調整するコードを追加する
      //    直前にある電卓を消すサブルーチンを詰めてできた隙間に電卓のY座標を調整するコードを押し込んでそれを呼び出す
      if (MainMemory.mmrRls (0x00ff3c56) == 0x4fef000a) {
        MainMemory.mmrWwa (
          0x00ff3c56,
          0x21fc, 0x00e6, 0x0000, 0x094c,  //00FF3C56  4FEF000A              lea.l $000A(sp),sp            →  move.l #$00E60000,$094C.w
          //                                 00FF3C5A  23FC00E600000000094C  move.l #$00E60000,$0000094C.l →
          0x426f, 0x0008,                  //00FF3C5E                                                      →  clr.w $0008(sp)
          0x6100, 0x2240,                  //00FF3C62                                                      →  bsr.w $00FF5EA4
          //                                 00FF3C64  4267                  clr.w -(sp)                   →
          0x4fef, 0x000a,                  //00FF3C66  3F3C0010              move.w #$0010,-(sp)           →  lea.l $000A(sp),sp
          0x301f,                          //00FF3C6A  3F3C00B8              move.w #$00B8,-(sp)           →  move.w (sp)+,d0
          0x6704,                          //00FF3C6C                                                      →  beq.s $00FF3C72
          0x6100, 0x6328,                  //00FF3C6E  3F3900000BFE          move.w $00000BFE.l,-(sp)      →  bsr.w $00FF9F98
          0x4cdf, 0x7ffe,                  //00FF3C72                                                      →  movem.l (sp)+,d1-d7/a0-a6
          //                                 00FF3C74  3F3900000BFC          move.w $00000BFC.l,-(sp)      →
          0x4e75,                          //00FF3C76                                                      →  rts
          0x4e71,                          //00FF3C78                                                      →  nop
          0xb27c, 0x001f,                  //00FF3C7A  61002228              bsr.w $00FF5EA4               →  cmp.w #$001F,d1
          0x630e,                          //00FF3C7E  4FEF000A              lea.l $000A(sp),sp            →  bls.s $00FF3C8E
          0xb278, 0x0972,                  //00FF3C                                                        →  cmp.w $0972.w,d1
          //                                 00FF3C82  301F                  move.w (sp)+,d0               →
          0x6308,                          //00FF3C84  6704                  beq.s $00FF3C8A               →  bls.s $00FF3C8E
          0x3238, 0x0972,                  //00FF3C86  61006310              bsr.w $00FF9F98               →  move.w $0972.w,d1
          0x6702,                          //00FF3C8A  4CDF7FFE              movem.l (sp)+,d1-d7/a0-a6     →  beq.s $00FF3C8E
          0x5341);                         //00FF3C                                                        →  subq.w #1,d1
        MainMemory.mmrWwa (
          0x00ff3ca8,
          0x3238, 0x0976,                  //00FF3CA8  323900000976          move.w $00000976.l,d1         →  move.w $0976.w,d1
          0x5241,                          //00FF3C                                                        →  addq.w #1,d1
          0x61ca);                         //00FF3CAE  5241                  addq.w #1,d1                  →  bsr.s $00FF3C7A
        patched++;
      } else {
        failed++;
      }
      //
      //  ソフトキーボードの↑キーの袋文字が閉じていない
      //    https://stdkmd.net/bugsx68k/#rom_softkeyboard
      if (MainMemory.mmrRwz (0x00ff53f8) == 0x2b80) {
        MainMemory.mmrWb (0x00ff53f8, 0x3b);  //00FF53F8  2B80  .dc.b %00101011,%10000000  →  3B80  .dc.b %00111011,%10000000
        patched++;
      } else {
        failed++;
      }
      //
      //  _CRTMODが指定された画面モードと異なる色数でグラフィックパレットを初期化する
      //    https://stdkmd.net/bugsx68k/#rom_crtmod_gpalet
      if (MainMemory.mmrRwz (0x00ff6578) == 0x3039) {
        MainMemory.mmrWwa (
          0x00ff6578,
          0x700c,                          //00FF6578  moveq.l #$0C,d0
          0xc038, 0x093c,                  //00FF657A  and.b   $093C.w,d0
          0xe248,                          //00FF657E  lsr.w   #1,d0
          0x303b, 0x0010,                  //00FF6580  move.w  $00FF6592(pc,d0.w),d0
          0x4ebb, 0x000c,                  //00FF6584  jsr     $00FF6592(pc,d0.w)
          0x08f9, 0x0003, 0x00e8, 0x0028,  //00FF6588  bset.b  #$03,$00E80028
          0x4e75,                          //00FF6590  rts
          0x45be,                          //00FF6592  .dc.w   $00FFAB50-$00FF6592
          0x45be,                          //00FF6594  .dc.w   $00FFAB50-$00FF6592
          0x45d4,                          //00FF6596  .dc.w   $00FFAB66-$00FF6592
          0x45ec,                          //00FF6598  .dc.w   $00FFAB7E-$00FF6592
          0x4e71,                          //00FF659A  nop
          0x4e71,                          //00FF659C  nop
          0x4e71,                          //00FF659E  nop
          0x4e71,                          //00FF65A0  nop
          0x4e71);                         //00FF65A2  nop
        patched++;
      } else {
        failed++;
      }
      //
      //  _GPALETで65536色モードのパレットを正しく取得できない
      //    https://stdkmd.net/bugsx68k/#rom_gpalet
      if (MainMemory.mmrRls (0x00ffaeaa) == 0x16300000) {
        MainMemory.mmrWb (0x00ffaeaa + 3, 0x02);   //00FFAEAA  16300000  move.b $00(a0,d0.w),d3  →  16300002  move.b $02(a0,d0.w),d3
        patched++;
      } else {
        failed++;
      }
      //
      //  _SP_INITでバスエラーが発生する
      //    https://stdkmd.net/bugsx68k/#rom_spinit
      if (MainMemory.mmrRls (0x00ffc338) == 0x303900e8) {
        //00FFC338  303900E80028  move.w  $00E80028,d0  |00FFC338  2048          movea.l a0,a0
        //                                              |00FFC33A  7012          moveq.l #%10010,d0
        //                                              |00FFC33C  C07900E80028  and.w   $00E80028,d0
        //00FFC33E  024000FF      andi.w  #$00FF,d0     |
        //                                              |00FFC342  B07C0012      cmp.w   #%10010,d0
        //00FFC342  0C400016      cmpi.w  #$0016,d0     |
        //00FFC346                                      |00FFC346
        MainMemory.mmrWwa (0x00ffc338, 0x2048, 0x7012, 0xc079, 0x00e8, 0x0028, 0xb07c, 0x0012);
        patched++;
      } else {
        failed++;
      }
      //
      if (patched != 0) {
        System.out.printf (Multilingual.mlnJapanese ?
                           "IPLROM 1.%d ($%08X) にパッチをあてました (%d/%d)\n" :
                           "IPLROM 1.%d ($%08X) was patched (%d/%d)\n",
                           romVersionShort, romVersionLong, patched, patched + failed);
      }
      break;

    case 0x10870507:  //IPLROM 1.0 ($10870507)
      //
      //  電卓を使うと実行中のプログラムが誤動作することがある
      //    https://stdkmd.net/bugsx68k/#rom_dentakud3
      if (MainMemory.mmrRls (0x00ff32d8) == 0x48e76000) {
        MainMemory.mmrWw (0x00ff32d8 + 2, 0x7000);  //00FF32D8  48E76000  movem.l d1-d2,-(sp)  →  48E77000  movem.l d1-d3,-(sp)
        MainMemory.mmrWw (0x00ff32f2 + 2, 0x000e);  //00FF32F2  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        MainMemory.mmrWw (0x00ff32f8 + 2, 0x000e);  //00FF32F8  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        MainMemory.mmrWw (0x00ff3342 + 2, 0x7000);  //00FF3342  48E76000  movem.l d1-d2,-(sp)  →  48E77000  movem.l d1-d3,-(sp)
        MainMemory.mmrWw (0x00ff335e + 2, 0x000e);  //00FF335E  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        patched++;
      } else {
        failed++;
      }
      //
      //  カーソルが画面の最下行にあると電卓が画面外に表示される
      //    https://stdkmd.net/bugsx68k/#rom_dentaku64
      //    電卓の表示位置を決めるサブルーチンに電卓のY座標を調整するコードを追加する
      //    直前にある電卓を消すサブルーチンを詰めてできた隙間に電卓のY座標を調整するコードを押し込んでそれを呼び出す
      if (MainMemory.mmrRls (0x00ff3c6a) == 0x4fef000a) {
        MainMemory.mmrWwa (
          0x00ff3c6a,
          0x21fc, 0x00e6, 0x0000, 0x094c,  //00FF3C6A  4FEF000A              lea.l $000A(sp),sp            →  move.l #$00E60000,$094C.w
          //                                 00FF3C6E  23FC00E600000000094C  move.l #$00E60000,$0000094C.l →
          0x426f, 0x0008,                  //00FF3C72                                                      →  clr.w $0008(sp)
          0x6100, 0x2240,                  //00FF3C76                                                      →  bsr.w $00FF5EB8
          //                                 00FF3C78  4267                  clr.w -(sp)                   →
          0x4fef, 0x000a,                  //00FF3C7A  3F3C0010              move.w #$0010,-(sp)           →  lea.l $000A(sp),sp
          0x301f,                          //00FF3C7E  3F3C00B8              move.w #$00B8,-(sp)           →  move.w (sp)+,d0
          0x6704,                          //00FF3C80                                                      →  beq.s $00FF3C86
          0x6100, 0x633a,                  //00FF3C82  3F3900000BFE          move.w $00000BFE.l,-(sp)      →  bsr.w $00FF9FBE
          0x4cdf, 0x7ffe,                  //00FF3C86                                                      →  movem.l (sp)+,d1-d7/a0-a6
          //                                 00FF3C88  3F3900000BFC          move.w $00000BFC.l,-(sp)      →
          0x4e75,                          //00FF3C8A                                                      →  rts
          0x4e71,                          //00FF3C8C                                                      →  nop
          0xb27c, 0x001f,                  //00FF3C8E  61002228              bsr.w $00FF5EB8               →  cmp.w #$001F,d1
          0x630e,                          //00FF3C92  4FEF000A              lea.l $000A(sp),sp            →  bls.s $00FF3CA2
          0xb278, 0x0972,                  //00FF3C94                                                      →  cmp.w $0972.w,d1
          //                                 00FF3C96  301F                  move.w (sp)+,d0               →
          0x6308,                          //00FF3C98  6704                  beq.s $00FF3C9E               →  bls.s $00FF3CA2
          0x3238, 0x0972,                  //00FF3C9A  61006322              bsr.w $00FF9FBE               →  move.w $0972.w,d1
          0x6702,                          //00FF3C9E  4CDF7FFE              movem.l (sp)+,d1-d7/a0-a6     →  beq.s $00FF3CA2
          0x5341);                         //00FF3CA0                                                      →  subq.w #1,d1
        MainMemory.mmrWwa (
          0x00ff3cbc,
          0x3238, 0x0976,                  //00FF3CBC  323900000976          move.w $00000976.l,d1         →  move.w $0976.w,d1
          0x5241,                          //00FF3CC0                                                      →  addq.w #1,d1
          0x61ca);                         //00FF3CC2  5241                  addq.w #1,d1                  →  bsr.s $00FF3C8E
        patched++;
      } else {
        failed++;
      }
      //
      //  ソフトキーボードの↑キーの袋文字が閉じていない
      //    https://stdkmd.net/bugsx68k/#rom_softkeyboard
      if (MainMemory.mmrRwz (0x00ff540c) == 0x2b80) {
        MainMemory.mmrWb (0x00ff540c, 0x3b);  //00FF540C  2B80  .dc.b %00101011,%10000000  →  3B80  .dc.b %00111011,%10000000
        patched++;
      } else {
        failed++;
      }
      //
      //  _CRTMODが指定された画面モードと異なる色数でグラフィックパレットを初期化する
      //    https://stdkmd.net/bugsx68k/#rom_crtmod_gpalet
      if (MainMemory.mmrRwz (0x00ff658c) == 0x3039) {
        MainMemory.mmrWwa (
          0x00ff658c,
          0x700c,                          //00FF658C  moveq.l #$0C,d0
          0xc038, 0x093c,                  //00FF658E  and.b   $093C.w,d0
          0xe248,                          //00FF6592  lsr.w   #1,d0
          0x303b, 0x0010,                  //00FF6594  move.w  $00FF65A6(pc,d0.w),d0
          0x4ebb, 0x000c,                  //00FF6598  jsr     $00FF65A6(pc,d0.w)
          0x08f9, 0x0003, 0x00e8, 0x0028,  //00FF659C  bset.b  #$03,$00E80028
          0x4e75,                          //00FF65A4  rts
          0x45d0,                          //00FF65A6  .dc.w   $00FFAB76-$00FF65A6
          0x45d0,                          //00FF65A8  .dc.w   $00FFAB76-$00FF65A6
          0x45e6,                          //00FF65AA  .dc.w   $00FFAB8C-$00FF65A6
          0x45fe,                          //00FF65AC  .dc.w   $00FFABA4-$00FF65A6
          0x4e71,                          //00FF65AE  nop
          0x4e71,                          //00FF65B0  nop
          0x4e71,                          //00FF65B2  nop
          0x4e71,                          //00FF65B4  nop
          0x4e71);                         //00FF65B6  nop
        patched++;
      } else {
        failed++;
      }
      //
      //  _GPALETで65536色モードのパレットを正しく取得できない
      //    https://stdkmd.net/bugsx68k/#rom_gpalet
      if (MainMemory.mmrRls (0x00ffaecc) == 0x16300000) {
        MainMemory.mmrWb (0x00ffaecc + 3, 0x02);   //00FFAECC  16300000  move.b $00(a0,d0.w),d3  →  16300002  move.b $02(a0,d0.w),d3
        patched++;
      } else {
        failed++;
      }
      //
      //  _SP_INITでバスエラーが発生する
      //    https://stdkmd.net/bugsx68k/#rom_spinit
      if (MainMemory.mmrRls (0x00ffc35a) == 0x303900e8) {
        //00FFC35A  303900E80028  move.w  $00E80028,d0  |00FFC35A  2048          movea.l a0,a0
        //                                              |00FFC35C  7012          moveq.l #%10010,d0
        //                                              |00FFC35E  C07900E80028  and.w   $00E80028,d0
        //00FFC360  024000FF      andi.w  #$00FF,d0     |
        //                                              |00FFC364  B07C0012      cmp.w   #%10010,d0
        //00FFC364  0C400016      cmpi.w  #$0016,d0     |
        //00FFC368                                      |00FFC368
        MainMemory.mmrWwa (0x00ffc35a, 0x2048, 0x7012, 0xc079, 0x00e8, 0x0028, 0xb07c, 0x0012);
        patched++;
      } else {
        failed++;
      }
      //
      if (patched != 0) {
        System.out.printf (Multilingual.mlnJapanese ?
                           "IPLROM 1.%d ($%08X) にパッチをあてました (%d/%d)\n" :
                           "IPLROM 1.%d ($%08X) was patched (%d/%d)\n",
                           romVersionShort, romVersionLong, patched, patched + failed);
      }
      break;

    case 0x11910111:  //IPLROM 1.1 ($11910111)
      //
      //  電卓を使うと実行中のプログラムが誤動作することがある
      //    https://stdkmd.net/bugsx68k/#rom_dentakud3
      if (MainMemory.mmrRls (0x00ff360e) == 0x48e76000) {
        MainMemory.mmrWw (0x00ff360e + 2, 0x7000);  //00FF360E  48E76000  movem.l d1-d2,-(sp)  →  48E77000  movem.l d1-d3,-(sp)
        MainMemory.mmrWw (0x00ff3628 + 2, 0x000e);  //00FF3628  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        MainMemory.mmrWw (0x00ff362e + 2, 0x000e);  //00FF362E  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        MainMemory.mmrWw (0x00ff3678 + 2, 0x7000);  //00FF3678  48E76000  movem.l d1-d2,-(sp)  →  48E77000  movem.l d1-d3,-(sp)
        MainMemory.mmrWw (0x00ff3694 + 2, 0x000e);  //00FF3694  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        patched++;
      } else {
        failed++;
      }
      //
      //  カーソルが画面の最下行にあると電卓が画面外に表示される
      //    https://stdkmd.net/bugsx68k/#rom_dentaku64
      //    電卓の表示位置を決めるサブルーチンに電卓のY座標を調整するコードを追加する
      //    直前にある電卓を消すサブルーチンを詰めてできた隙間に電卓のY座標を調整するコードを押し込んでそれを呼び出す
      if (MainMemory.mmrRls (0x00ff3fa0) == 0x4fef000a) {
        MainMemory.mmrWwa (
          0x00ff3fa0,
          0x21fc, 0x00e6, 0x0000, 0x094c,  //00FF3FA0  4FEF000A              lea.l $000A(sp),sp            →  move.l #$00E60000,$094C.w
          //                                 00FF3FA4  23FC00E600000000094C  move.l #$00E60000,$0000094C.l →
          0x426f, 0x0008,                  //00FF3FA8                                                      →  clr.w $0008(sp)
          0x6100, 0x2240,                  //00FF3FAC                                                      →  bsr.w $00FF61EE
          //                                 00FF3FAE  4267                  clr.w -(sp)                   →
          0x4fef, 0x000a,                  //00FF3FB0  3F3C0010              move.w #$0010,-(sp)           →  lea.l $000A(sp),sp
          0x301f,                          //00FF3FB4  3F3C00B8              move.w #$00B8,-(sp)           →  move.w (sp)+,d0
          0x6704,                          //00FF3FB6                                                      →  beq.s $00FF3FBC
          0x6100, 0x6384,                  //00FF3FB8  3F3900000BFE          move.w $00000BFE.l,-(sp)      →  bsr.w $00FFA33E
          0x4cdf, 0x7ffe,                  //00FF3FBC                                                      →  movem.l (sp)+,d1-d7/a0-a6
          //                                 00FF3FBE  3F3900000BFC          move.w $00000BFC.l,-(sp)      →
          0x4e75,                          //00FF3FC0                                                      →  rts
          0x4e71,                          //00FF3FC2                                                      →  nop
          0xb27c, 0x001f,                  //00FF3FC4  61002228              bsr.w $00FF61EE               →  cmp.w #$001F,d1
          0x630e,                          //00FF3FC8  4FEF000A              lea.l $000A(sp),sp            →  bls.s $00FF3FD8
          0xb278, 0x0972,                  //00FF3FCA                                                      →  cmp.w $0972.w,d1
          //                                 00FF3FCC  301F                  move.w (sp)+,d0               →
          0x6308,                          //00FF3FCE  6704                  beq.s $00FF3FD4               →  bls.s $00FF3FD8
          0x3238, 0x0972,                  //00FF3FD0  6100636C              bsr.w $00FFA33E               →  move.w $0972.w,d1
          0x6702,                          //00FF3FD4  4CDF7FFE              movem.l (sp)+,d1-d7/a0-a6     →  beq.s $00FF3FD8
          0x5341);                         //00FF3FD6                                                      →  subq.w #1,d1
        MainMemory.mmrWwa (
          0x00ff3ff2,
          0x3238, 0x0976,                  //00FF3FF2  323900000976          move.w $00000976.l,d1         →  move.w $0976.w,d1
          0x5241,                          //00FF3FF6                                                      →  addq.w #1,d1
          0x61ca);                         //00FF3FF8  5241                  addq.w #1,d1                  →  bsr.s $00FF3FC4
        patched++;
      } else {
        failed++;
      }
      //
      //  ソフトキーボードの↑キーの袋文字が閉じていない
      //    https://stdkmd.net/bugsx68k/#rom_softkeyboard
      if (MainMemory.mmrRwz (0x00ff5742) == 0x2b80) {
        MainMemory.mmrWb (0x00ff5742, 0x3b);  //00FF5742  2B80  .dc.b %00101011,%10000000  →  3B80  .dc.b %00111011,%10000000
        patched++;
      } else {
        failed++;
      }
      //
      //  _CRTMODが指定された画面モードと異なる色数でグラフィックパレットを初期化する
      //    https://stdkmd.net/bugsx68k/#rom_crtmod_gpalet
      if (MainMemory.mmrRwz (0x00ff68d4) == 0x3039) {
        MainMemory.mmrWwa (0x00ff68d4,
                0x700c,                          //00FF68D4  moveq.l #$0C,d0
                0xc038, 0x093c,                  //00FF68D6  and.b   $093C.w,d0
                0xe248,                          //00FF68DA  lsr.w   #1,d0
                0x303b, 0x0010,                  //00FF68DC  move.w  $00FF68EE(pc,d0.w),d0
                0x4ebb, 0x000c,                  //00FF68E0  jsr     $00FF68EE(pc,d0.w)
                0x08f9, 0x0003, 0x00e8, 0x0028,  //00FF68E4  bset.b  #$03,$00E80028
                0x4e75,                          //00FF68EC  rts
                0x45b4,                          //00FF68EE  .dc.w   $00FFAEA2-$00FF68EE
                0x45b4,                          //00FF68F0  .dc.w   $00FFAEA2-$00FF68EE
                0x45ca,                          //00FF68F2  .dc.w   $00FFAEB8-$00FF68EE
                0x45e2,                          //00FF68F4  .dc.w   $00FFAED0-$00FF68EE
                0x4e71,                          //00FF68F6  nop
                0x4e71,                          //00FF68F8  nop
                0x4e71,                          //00FF68FA  nop
                0x4e71,                          //00FF68FC  nop
                0x4e71);                         //00FF68FE  nop
        patched++;
      } else {
        failed++;
      }
      //
      //  IOCS _MS_PATSTが機能しない
      //    https://stdkmd.net/bugsx68k/#rom_mspatst
      //    ここではパターンアドレステーブルのオフセットを計算するlsl.w #2,d1の手前にd1.wを復元するmove.w $0AE6-$0A7A(a5),d1を押し込む
      //    設定後の11ワードは直後にある_MS_SELと同じなのでbra.sで飛ばせば10ワード確保できるのだが設定までの7ワードに入り切ってしまった
      if (MainMemory.mmrRwz (0x00ffa606) == 0xe549) {
        MainMemory.mmrWwa (0x00ffa606,
                0x7240,  //00FFA606  E549      lsl.w #2,d1                  →  7240          moveq.l #$40,d1
                0xd26d,  //00FFA608  41ED00F6  lea.l $0B70-$0A7A(a5),a0     →  D26D006C      add.w $0AE6-$0A7A(a5),d1
                0x006c,
                0xe549,  //00FFA60C  41F01000  lea.l $00(a0,d1.w),a0        →  E549          lsl.w #2,d1
                0x2bad,  //00FFA60E                                         →  2BAD007210F6  move.l $0AEC-$0A7A(a5),$0B70-$0A7A-$40*4(a5,d1.w)
                0x0072,  //00FFA610  20AD0072  move.l $0AEC-$0A7A(a5),(a0)  →
                0x10f6);
        patched++;
      } else {
        failed++;
      }
      //
      //  _GPALETで65536色モードのパレットを正しく取得できない
      //    https://stdkmd.net/bugsx68k/#rom_gpalet
      if (MainMemory.mmrRls (0x00ffb1f8) == 0x16300000) {
        MainMemory.mmrWb (0x00ffb1f8 + 3, 0x02);   //00FFB1F8  16300000  move.b $00(a0,d0.w),d3  →  16300002  move.b $02(a0,d0.w),d3
        patched++;
      } else {
        failed++;
      }
      //
      //  _SP_INITでバスエラーが発生する
      //    https://stdkmd.net/bugsx68k/#rom_spinit
      if (MainMemory.mmrRls (0x00ffc686) == 0x303900e8) {
        //00FFC686  303900E80028  move.w  $00E80028,d0  |00FFC686  2048          movea.l a0,a0
        //                                              |00FFC688  7012          moveq.l #%10010,d0
        //                                              |00FFC68A  C07900E80028  and.w   $00E80028,d0
        //00FFC68C  024000FF      andi.w  #$00FF,d0     |
        //                                              |00FFC690  B07C0012      cmp.w   #%10010,d0
        //00FFC690  0C400016      cmpi.w  #$0016,d0     |
        //00FFC694                                      |00FFC694
        MainMemory.mmrWwa (0x00ffc686, 0x2048, 0x7012, 0xc079, 0x00e8, 0x0028, 0xb07c, 0x0012);
        patched++;
      } else {
        failed++;
      }
      //
      if (patched != 0) {
        System.out.printf (Multilingual.mlnJapanese ?
                           "IPLROM 1.%d ($%08X) にパッチをあてました (%d/%d)\n" :
                           "IPLROM 1.%d ($%08X) was patched (%d/%d)\n",
                           romVersionShort, romVersionLong, patched, patched + failed);
      }
      break;

    case 0x12911024:  //IPLROM 1.2 ($12911024)
      //
      //  電卓を使うと実行中のプログラムが誤動作することがある
      //    https://stdkmd.net/bugsx68k/#rom_dentakud3
      if (MainMemory.mmrRls (0x00ff36ca) == 0x48e76000) {
        MainMemory.mmrWw (0x00ff36ca + 2, 0x7000);  //00FF36CA  48E76000  movem.l d1-d2,-(sp)  →  48E77000  movem.l d1-d3,-(sp)
        MainMemory.mmrWw (0x00ff36e4 + 2, 0x000e);  //00FF36E4  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        MainMemory.mmrWw (0x00ff36ea + 2, 0x000e);  //00FF36EA  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        MainMemory.mmrWw (0x00ff3734 + 2, 0x7000);  //00FF3734  48E76000  movem.l d1-d2,-(sp)  →  48E77000  movem.l d1-d3,-(sp)
        MainMemory.mmrWw (0x00ff3750 + 2, 0x000e);  //00FF3750  4CDF0006  movem.l (sp)+,d1-d2  →  4CDF000E  movem.l (sp)+,d1-d3
        patched++;
      } else {
        failed++;
      }
      //
      //  カーソルが画面の最下行にあると電卓が画面外に表示される
      //    https://stdkmd.net/bugsx68k/#rom_dentaku64
      //    電卓の表示位置を決めるサブルーチンに電卓のY座標を調整するコードを追加する
      //    直前にある電卓を消すサブルーチンを詰めてできた隙間に電卓のY座標を調整するコードを押し込んでそれを呼び出す
      if (MainMemory.mmrRls (0x00ff405c) == 0x4fef000a) {
        MainMemory.mmrWwa (
          0x00ff405c,
          0x21fc, 0x00e6, 0x0000, 0x094c,  //00FF405C  4FEF000A              lea.l $000A(sp),sp            →  move.l #$00E60000,$094C.w
          //                                 00FF4060  23FC00E600000000094C  move.l #$00E60000,$0000094C.l →
          0x426f, 0x0008,                  //00FF4064                                                      →  clr.w $0008(sp)
          0x6100, 0x2240,                  //00FF4068                                                      →  bsr.w $00FF62AA
          //                                 00FF406A  4267                  clr.w -(sp)                   →
          0x4fef, 0x000a,                  //00FF406C  3F3C0010              move.w #$0010,-(sp)           →  lea.l $000A(sp),sp
          0x301f,                          //00FF4070  3F3C00B8              move.w #$00B8,-(sp)           →  move.w (sp)+,d0
          0x6704,                          //00FF4072                                                      →  beq.s $00FF4078
          0x6100, 0x63e2,                  //00FF4074  3F3900000BFE          move.w $00000BFE.l,-(sp)      →  bsr.w $00FFA458
          0x4cdf, 0x7ffe,                  //00FF4078                                                      →  movem.l (sp)+,d1-d7/a0-a6
          //                                 00FF407A  3F3900000BFC          move.w $00000BFC.l,-(sp)      →
          0x4e75,                          //00FF407C                                                      →  rts
          0x4e71,                          //00FF407E                                                      →  nop
          0xb27c, 0x001f,                  //00FF4080  61002228              bsr.w $00FF62AA               →  cmp.w #$001F,d1
          0x630e,                          //00FF4084  4FEF000A              lea.l $000A(sp),sp            →  bls.s $00FF4094
          0xb278, 0x0972,                  //00FF4086                                                      →  cmp.w $0972.w,d1
          //                                 00FF4088  301F                  move.w (sp)+,d0               →
          0x6308,                          //00FF408A  6704                  beq.s $00FF4090               →  bls.s $00FF4094
          0x3238, 0x0972,                  //00FF408C  610063CA              bsr.w $00FFA458               →  move.w $0972.w,d1
          0x6702,                          //00FF4090  4CDF7FFE              movem.l (sp)+,d1-d7/a0-a6     →  beq.s $00FF4094
          0x5341);                         //00FF4092                                                      →  subq.w #1,d1
        MainMemory.mmrWwa (
          0x00ff40ae,
          0x3238, 0x0976,                  //00FF40AE  323900000976          move.w $00000976.l,d1         →  move.w $0976.w,d1
          0x5241,                          //00FF40B2                                                      →  addq.w #1,d1
          0x61ca);                         //00FF40B4  5241                  addq.w #1,d1                  →  bsr.s $00FF4080
        patched++;
      } else {
        failed++;
      }
      //
      //  ソフトキーボードの↑キーの袋文字が閉じていない
      //    https://stdkmd.net/bugsx68k/#rom_softkeyboard
      if (MainMemory.mmrRwz (0x00ff57fe) == 0x2b80) {
        MainMemory.mmrWb (0x00ff57fe, 0x3b);  //00FF57FE  2B80  .dc.b %00101011,%10000000  →  3B80  .dc.b %00111011,%10000000
        patched++;
      } else {
        failed++;
      }
      //
      //  _CRTMODが指定された画面モードと異なる色数でグラフィックパレットを初期化する
      //    https://stdkmd.net/bugsx68k/#rom_crtmod_gpalet
      if (MainMemory.mmrRwz (0x00ff69ac) == 0x3039) {
        MainMemory.mmrWwa (0x00ff69ac,
                0x700c,                          //00FF69AC  moveq.l #$0C,d0
                0xc038, 0x093c,                  //00FF69AE  and.b   $093C.w,d0
                0xe248,                          //00FF69B2  lsr.w   #1,d0
                0x303b, 0x0010,                  //00FF69B4  move.w  $00FF69C6(pc,d0.w),d0
                0x4ebb, 0x000c,                  //00FF69B8  jsr     $00FF69C6(pc,d0.w)
                0x08f9, 0x0003, 0x00e8, 0x0028,  //00FF69BC  bset.b  #$03,$00E80028
                0x4e75,                          //00FF69C4  rts
                0x45f6,                          //00FF69C6  .dc.w   $00FFAFBC-$00FF69C6
                0x45f6,                          //00FF69C8  .dc.w   $00FFAFBC-$00FF69C6
                0x460c,                          //00FF69CA  .dc.w   $00FFAFD2-$00FF69C6
                0x4624,                          //00FF69CC  .dc.w   $00FFAFEA-$00FF69C6
                0x4e71,                          //00FF69CE  nop
                0x4e71,                          //00FF69D0  nop
                0x4e71,                          //00FF69D2  nop
                0x4e71,                          //00FF69D4  nop
                0x4e71);                         //00FF69D6  nop
        patched++;
      } else {
        failed++;
      }
      //
      //  _GPALETで65536色モードのパレットを正しく取得できない
      //    https://stdkmd.net/bugsx68k/#rom_gpalet
      if (MainMemory.mmrRls (0x00ffb312) == 0x16300000) {
        MainMemory.mmrWb (0x00ffb312 + 3, 0x02);   //00FFB312  16300000  move.b $00(a0,d0.w),d3  →  16300002  move.b $02(a0,d0.w),d3
        patched++;
      } else {
        failed++;
      }
      //
      //  _SP_INITでバスエラーが発生する
      //    https://stdkmd.net/bugsx68k/#rom_spinit
      if (MainMemory.mmrRls (0x00ffc7a0) == 0x303900e8) {
        //00FFC7A0  303900E80028  move.w  $00E80028,d0  |00FFC7A0  2048          movea.l a0,a0
        //                                              |00FFC7A2  7012          moveq.l #%10010,d0
        //                                              |00FFC7A4  C07900E80028  and.w   $00E80028,d0
        //00FFC7A6  024000FF      andi.w  #$00FF,d0     |
        //                                              |00FFC7AA  B07C0012      cmp.w   #%10010,d0
        //00FFC7AA  0C400016      cmpi.w  #$0016,d0     |
        //00FFC7AE                                      |00FFC7AE
        MainMemory.mmrWwa (0x00ffc7a0, 0x2048, 0x7012, 0xc079, 0x00e8, 0x0028, 0xb07c, 0x0012);
        patched++;
      } else {
        failed++;
      }
      //
      if (patched != 0) {
        System.out.printf (Multilingual.mlnJapanese ?
                           "IPLROM 1.%d ($%08X) にパッチをあてました (%d/%d)\n" :
                           "IPLROM 1.%d ($%08X) was patched (%d/%d)\n",
                           romVersionShort, romVersionLong, patched, patched + failed);
      }
      break;

    case 0x13921127:  //IPLROM 1.3 ($13921127)
      //パッチをあてる
      //  挿入コード
      for (int i = 0; i < ROM_PATCH13_DATA.length; ) {
        int a = ByteArray.byaRls (ROM_PATCH13_DATA, i);  //挿入開始アドレス(0=終了)
        if (a == 0) {
          break;
        }
        int e = ByteArray.byaRls (ROM_PATCH13_DATA, i + 4);  //挿入終了アドレス
        int d = ByteArray.byaRls (ROM_PATCH13_DATA, i + 8);  //元のデータ
        int l = ByteArray.byaRls (ROM_PATCH13_DATA, i + 12);  //挿入コードの長さ(偶数)
        if (MainMemory.mmrRls (a) == d) {
          System.arraycopy (ROM_PATCH13_DATA, i + 16, MainMemory.mmrM8, a, l);
          for (a += l; a < e; a += 2) {
            MainMemory.mmrWw (a, 0x2048);  //残りをMOVEA.L A0,A0で充填する。必須
          }
          patched++;
        } else {
          failed++;
        }
        i += 16 + l;
      }
      //  追加コード
      System.arraycopy (ROM_PATCH13_TEXT, 0, MainMemory.mmrM8, ROM_PATCH13_BASE, ROM_PATCH13_TEXT.length);
      //  IPLROM 1.3に6x12 ANKフォントを入れる
      //    X68000のCGROMとX68030のIPLROMを組み合わせると6x12 ANKフォントを表示できない
      //    エミュレータならばCGROM.DATをCGROM30.DATに差し替えるだけだが、実機でROMが一体になっていない機種はCGROMを簡単に差し替えられない
      //    手順
      //      IPLROM 1.2の$00FFD45E～$00FFE045(3048バイト)をIPLROM 1.3の$00FEF400～$00FEFFE7へコピーする
      //      IPLROM 1.3の$00FEFFE8～$00FEFFFF(24バイト)を$00で充填する
      //      IPLROM 1.3の$00FF73F8にある$00FBF400を$00FEF400に変更する
      {
        byte[] iplrom12 = XEiJ.rscGetResource ("IPLROMCO.DAT");
        if (iplrom12 != null &&
            MainMemory.mmrRls (0x00fef400) == 0xffffffff &&
            MainMemory.mmrRls (0x00ff73f8) == 0x00fbf400) {
          //254文字コピーする
          System.arraycopy (iplrom12, 0x00ffd45e - 0x00fe0000,  //コピー元
                            MainMemory.mmrM8, 0x00fef400 - 0x00000000,  //コピー先
                            12 * 254);  //長さ
          //2文字クリアする
          Arrays.fill (MainMemory.mmrM8, 0x00fef400 + 12 * 254 - 0x00000000,  //開始位置
                       0x00fef400 + 12 * 256 - 0x00000000,  //終了位置
                       (byte) 0x00);  //値
          //_FNTADRを変更する
          MainMemory.mmrWl (0x00ff73f8, 0x00fef400);
          patched++;
        } else {
          failed++;
        }
      }
      //  バージョンが変わる
      romPreviousVersionLong = romVersionLong;
      romPreviousVersionShort = romVersionShort;
      romVersionLong = MainMemory.mmrRls (0x00ff0032);
      romVersionShort = romVersionLong >> 24 & 15;
      if (patched != 0) {
        System.out.printf (Multilingual.mlnJapanese ?
                           "IPLROM 1.%d ($%08X) にパッチをあてました (%d/%d)\n" :
                           "IPLROM 1.%d ($%08X) was patched (%d/%d)\n",
                           romPreviousVersionShort, romPreviousVersionLong, patched, patched + failed);
        System.out.printf (Multilingual.mlnJapanese ?
                           "IPLROM 1.%d ($%08X) が IPLROM 1.%d ($%08X) になりました\n" :
                           "IPLROM 1.%d ($%08X) was changed to IPLROM 1.%d ($%08X)\n",
                           romPreviousVersionShort, romPreviousVersionLong, romVersionShort, romVersionLong);
      }
      break;

    }  //switch romVersionLong

  }  //romPatch()

  //romPatch2 ()
  //  ROMデバッガにパッチをあてる
  public static void romPatch2 () {

    //ROMデバッガのバグまたはおかしなところを修正する
    int patched = 0;
    int failed = 0;
    switch (romDebuggerVersion) {

    case -1:  //ROMデバッガがない
      //00000000  48E7C080     movem.l d0-d1/a0,-(sp)
      //00000004  41FA0012     lea.l   3f(pc),a0
      //00000008  6004         bra.s   2f
      //0000000A  70354E4F  1: IOCS    _OUT232C
      //0000000E  1218      2: move.b  (a0)+,d1
      //00000010  66F8         bne.s   1b
      //00000012  4CDF0103     movem.l (sp)+,d0-d1/a0
      //00000016  4E75         rts
      //00000018            3: .dc.b 'ROM debugger does not exist',$0D,$0A,$00
      int entry = romVersionShort < 3 ? 0x00fe0000 : 0x00fd3800;  //エントリポイント
      MainMemory.mmrWwa (entry, 0x48e7, 0xc080, 0x41fa, 0x0012, 0x6004, 0x7035, 0x4e4f, 0x1218, 0x66f8, 0x4cdf, 0x0103, 0x4e75);
      MainMemory.mmrWstr (entry + 0x0018, "ROM debugger does not exist\r\n\0");
      return;

    case 0x0100:  //ROMデバッガ1.0
      //
      //  実効アドレスの計算で絶対ショートアドレスが符号拡張されない
      //    https://stdkmd.net/bugsx68k/#db_absoluteshort
      if (MainMemory.mmrRwz (0x00fe0a26) == 0x4240) {
        MainMemory.mmrWw (0x00fe0a26, 0x3015);  //00FE0A26  4240  clr.w d0        →  3015  move.w (a5),d0
        MainMemory.mmrWw (0x00fe0a28, 0x48c0);  //00FE0A26  3015  move.w (a5),d0  →  48C0  ext.l d0
        MainMemory.mmrWl (0x00fe0a40, 0x322dfffe);  //00FE0A40  4281      clr.l d1              →  322DFFFE  move.w -$0002(a5),d1
        //                                00FE0A42  222DFFFE  move.l -$0002(a5),d1
        MainMemory.mmrWw (0x00fe0a44, 0x48c1);      //00FE0A44                                  →  48C1      ext.l d1
        patched++;
      } else {
        failed++;
      }
      //
      //  CCRの未定義ビットを操作している
      //    https://stdkmd.net/bugsx68k/#db_ccrundefbit
      if (MainMemory.mmrRls (0x00fe2d10) == 0x023c007b) {
        MainMemory.mmrWb (0x00fe2d10 + 3, 0xfb);  //00FE2D10  023C007B  andi.b #$7B,ccr  →  023C00FB  andi.b #$FB,ccr
        patched++;
      } else {
        failed++;
      }
      //
      //  リモートターミナルからの入力が1文字置きになる
      //    https://stdkmd.net/bugsx68k/#db_ctrls
      //    ^Sのときだけ空読みする方法もあるがターミナルの仕様に合わないのでここでは^Sで一時停止する機能そのものを削除する
      //    ターミナルはスクロールすれば過去に出力されたメッセージを読むことができるので一時停止する機能は通常は必要ない
      //    1.0と2.32でオフセットが違うことに注意
      if (MainMemory.mmrRwz (0x00fe4ba6) == 0x6100) {
        MainMemory.mmrWw (0x00fe4ba6, 0x600e);  //00FE4BA6  610000EE  bsr.w ~FE4C96  →  600E  bra.s ~FE4BB6
        patched++;
      } else {
        failed++;
      }
      //
      if (patched != 0) {
        System.out.printf (Multilingual.mlnJapanese ?
                           "ROM デバッガ 1.0 にパッチをあてました (%d/%d)\n" :
                           "ROM debugger 1.0 was patched (%d/%d)\n",
                           patched, patched + failed);
      }
      break;

    case 0x0232:  //ROMデバッガ2.32
      //
      //  X68030でROMデバッガを有効にすると起動できない
      //    https://stdkmd.net/bugsx68k/#db_dirtya6
      //    ここではA6相対のディスプレースメントから$1000を引いて絶対ショートにする
      //    ..........101110 (d16,A6)
      //    ..........111000 (xxx).W
      //    ....110101...... (d16,A6)
      //    ....000111...... (xxx).W
      if (MainMemory.mmrRwz (0x00fd3834) == 0x2d4f) {
        MainMemory.mmrWl (0x00fd3834, 0x21cf1172);  //00FD3834  2D4F0172  move.l sp,$0172(a6)  →  21CF1172  move.l sp,$1172.w
        MainMemory.mmrWw (0x00fd3868, 0x21e8);  //00FD3868  2D680080059C  move.l $0080(a0),$059C(a6)  →  21E80080159C  move.l $0080(a0),$159C.w
        MainMemory.mmrWw (0x00fd3868 + 4, 0x159c);
        MainMemory.mmrWl (0x00fd3876, 0x40f81042);  //00FD3876  40EE0042  move.w sr,$0042(a6)  →  40F81042  move.w sr,$1042.w
        MainMemory.mmrWl (0x00fd387a, 0x21cf116a);  //00FD387A  2D4F016A  move.l sp,$016A(a6)  →  21CF116A  move.l sp,$116A.w
        MainMemory.mmrWw (0x00fd387e, 0x21fc);  //00FD387E  21FC00FDE4660046  move.l #$00FDE466,$0046(a6)  →  21FC00FDE4661046  move.l #$00FDE466,$1046.w
        MainMemory.mmrWw (0x00fd387e + 6, 0x1046);
        MainMemory.mmrWl (0x00fd3886, 0x4a781690);  //00FD3886  4A6E0690  tst.w $0690(a6)  →  4A781690  tst.w $1690.w
        MainMemory.mmrWw (0x00fd388c, 0x21fc);  //00FD388C  2D7C00FD46FA0176  move.l #$00FD46FA,$0176(a6)  →  21FC00FD46FA1176  move.l #$00FD46FA,$1176.w
        MainMemory.mmrWw (0x00fd388c + 6, 0x1176);
        MainMemory.mmrWw (0x00fd3896, 0x21fc);  //00FD3896  2D7C00FE0CE80176  move.l #$00FE0CE8,$0176(a6)  →  21FC00FE0CE81176  move.l #$00FE0CE8,$1176.w
        MainMemory.mmrWw (0x00fd3896 + 6, 0x1176);
        patched++;
      } else {
        failed++;
      }
      //
      //  FSINCOS.X FPm,FPc:FPs を正しくアセンブルできない
      //    https://stdkmd.net/bugsx68k/#db_fsincos
      if (MainMemory.mmrRwz (0x00fd6faa) == 0x8128) {
        MainMemory.mmrWw (0x00fd6faa, 0x812c);  //00FD6FAA  81280001  or.b d0,$0001(a0)  →  812C0001  or.b d0,$0001(a4)
        patched++;
      } else {
        failed++;
      }
      //
      //  リモートターミナルからの入力が1文字置きになる
      //    https://stdkmd.net/bugsx68k/#db_ctrls
      //    ^Sのときだけ空読みする方法もあるがターミナルの仕様に合わないのでここでは^Sで一時停止する機能そのものを削除する
      //    ターミナルはスクロールすれば過去に出力されたメッセージを読むことができるので一時停止する機能は通常は必要ない
      //    1.0と2.32でオフセットが違うことに注意
      if (MainMemory.mmrRwz (0x00fd8eb8) == 0x6100) {
        MainMemory.mmrWw (0x00fd8eb8, 0x6012);  //00FD8EB8  61000104  bsr.w $00FD8FBE  →  6012  bra.s $00FD8ECC
        patched++;
      } else {
        failed++;
      }
      //
      //  実効アドレスの計算でオペレーションサイズが.Q/.S/.D/.X/.Pのとき-(Ar)のオフセットが違う
      //    https://stdkmd.net/bugsx68k/#db_predecrement
      if (MainMemory.mmrRwz (0x00fda99c) == 0x0244) {
        MainMemory.mmrWwa (0x00fda99c,
                           0x183b, 0x400a,   //00FDA99C  02440003      andi.w #$0003,d4       →  183B400A      move.b @f(pc,d4.w),d4
                           0x90c4,           //00FDA9A0  D844          add.w d4,d4            →  90C4          suba.w d4,a0
                           0x2d48, 0x0328,   //00FDA9A2  D0FB4008      adda.w @f(pc,d4.w),a0  →  2D480328      move.l a0,$0328(a6)
                           0x4e75,           //00FDA9A6  2D480328      move.l a0,$0328(a6)    →  4E75          rts
                           0x0201,           //00FDA9A8                                       →  0201      @@: .dc.b 2,1
                           0x0204,           //00FDA9AA  4E75          rts                    →  0204          .dc.b 2,4
                           0x0804,           //00FDA9AC  FFFE      @@: .dc.w -2               →  0804          .dc.b 8,4
                           0x080c,           //00FDA9AE  FFFF      @@: .dc.w -1               →  080C          .dc.b 8,12
                           0x0c00,           //00FDA9B0  FFFE          .dc.w -2               →  0C00          .dc.b 12,0
                           0x0000);          //00FDA9B2  FFFC          .dc.w -4               →  0000          .dc.b 0,0
        patched++;
      } else {
        failed++;
      }
      //
      //  実効アドレスの計算で(d8,Ar,Rn.wl)/(d8,PC,Rn.wl)のd8が負数のときAr/PCがサプレスされる
      //    https://stdkmd.net/bugsx68k/#db_briefsuppress
      if (MainMemory.mmrRwz (0x00fda9c6) == 0x91c8) {
        MainMemory.mmrWwa (0x00fda9c6,
                           0x4e71,   //00FDA9C6  91C8      suba.l a0,a0     →  4E71  nop
                           0x4e71,   //00FDA9C8  4A2D0001  tst.b $0001(a5)  →  4E71  nop
                           0x4e71,   //00FDA9CA                             →  4E71  nop
                           0x4e71);  //00FDA9CC  6B04      bmi.s $00FDA9D2  →  4E71  nop
        MainMemory.mmrWwa (0x00fdab18,
                           0x4e71,   //00FDAB18  91C8      suba.l a0,a0     →  4E71  nop
                           0x4e71,   //00FDAB1A  4A2D0001  tst.b $0001(a5)  →  4E71  nop
                           0x4e71,   //00FDAB1C                             →  4E71  nop
                           0x4e71);  //00FDAB1E  6B02      bmi.s $00FDAB22  →  4E71  nop
        patched++;
      } else {
        failed++;
      }
      //
      //  実効アドレスの計算でインデックスサプレスが反映されない
      //    https://stdkmd.net/bugsx68k/#db_indexsuppress
      if (MainMemory.mmrRls (0x00fda9f8) == 0x61000238) {
        MainMemory.mmrWw (0x00fda9f8 + 2, 0x0256);  //00FDA9F8  61000238  bsr.w $00FDAC32  →  61000256  bsr.w $00FDAC50
        MainMemory.mmrWw (0x00fdaa5a + 2, 0x01f4);  //00FDAA5A  610001D6  bsr.w $00FDAC32  →  610001F4  bsr.w $00FDAC50
        MainMemory.mmrWw (0x00fdaaa0 + 2, 0x01ae);  //00FDAAA0  61000190  bsr.w $00FDAC32  →  610001AE  bsr.w $00FDAC50
        MainMemory.mmrWw (0x00fdab46 + 2, 0x0108);  //00FDAB46  610000EA  bsr.w $00FDAC32  →  61000108  bsr.w $00FDAC50
        MainMemory.mmrWw (0x00fdaba6 + 2, 0x00a8);  //00FDABA6  6100008A  bsr.w $00FDAC32  →  610000A8  bsr.w $00FDAC50
        MainMemory.mmrWw (0x00fdabfa, 0x6154);      //00FDABFA  6136      bsr.s $00FDAC32  →  6154      bsr.s $00FDAC50
        MainMemory.mmrWwa (0x00fdac48,
                           0x303b, 0x0012,          //00FDAC48  303B0012      move.w $00FDAC5C(pc,d0.w),d0
                           0x4efb, 0x000e,          //00FDAC4C  4EFB000E      jmp $00FDAC5C(pc,d0.w)
                           0x7000,                  //00FDAC50  7000          moveq.l #0,d0
                           0x082d, 0x0006, 0x0001,  //00FDAC52  082D00060001  btst.b #6,1(a5)
                           0x67d8,                  //00FDAC58  67D8          beq.s $00FDAC32
                           0x4e75,                  //00FDAC5A  4E75          rts
                           0x0040, 0x0056,          //00FDAC5C  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC60  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC64  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC68  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC6C  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC70  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC74  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC78  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC7C  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC80  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC84  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC88  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC8C  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC90  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0040, 0x0056,          //00FDAC94  00400056      .dc.w $00FDAC9C-$00FDAC5C,$00FDACB2-$00FDAC5C
                           0x0106, 0x013c,          //00FDAC98  0106013C      .dc.w $00FDAD62-$00FDAC5C,$00FDAD98-$00FDAC5C
                           0x3015,                  //00FDAC9C  3015          move.w (a5),d0
                           0xc07c, 0xf000,          //00FDAC9E  C07CF000      and.w #$F000,d0
                           0xed58,                  //00FDACA2  ED58          rol.w #6,d0
                           0x2f08,                  //00FDACA4  2F08          move.l a0,-(sp)
                           0x3040,                  //00FDACA6  3040          move.w d0,a0
                           0x3028, 0x1058,          //00FDACA8  30281058      move.w $1058(a0),d0
                           0x205f,                  //00FDACAC  205F          movea.l (sp)+,a0
                           0x48c0,                  //00FDACAE  48C0          ext.l d0
                           0x4e75,                  //00FDACB0  4E75          rts
                           0x3015,                  //00FDACB2  3015          move.w (a5),d0
                           0xc07c, 0xf000,          //00FDACB4  C07CF000      and.w #$F000,d0
                           0xed58,                  //00FDACB8  ED58          rol.w #6,d0
                           0x2f08,                  //00FDACBA  2F08          move.l a0,-(sp)
                           0x3040,                  //00FDACBC  3040          move.w d0,a0
                           0x2028, 0x1056,          //00FDACBE  20281056      move.l $1056(a0),d0
                           0x205f,                  //00FDACC2  205F          movea.l (sp)+,a0
                           0x4e75);                 //00FDACC4  4E75          rts
        patched++;
      } else {
        failed++;
      }
      //
      //  実効アドレスの計算でスケールファクタを掛ける代わりにスケールファクタで割る
      //    https://stdkmd.net/bugsx68k/#db_scalefactor
      if (MainMemory.mmrRwz (0x00fdac3c) == 0xe2a8) {
        MainMemory.mmrWw (0x00fdac3c, 0xe3a8);  //00FDAC3C  E2A8  lsr.l d1,d0  →  E3A8  lsl.l d1,d0
        patched++;
      } else {
        failed++;
      }
      //
      //  実効アドレスの計算でインデックスレジスタの選択を間違える
      //    https://stdkmd.net/bugsx68k/#db_indexregister
      if (MainMemory.mmrRls (0x00fdac42) == 0x0240f000) {
        MainMemory.mmrWw (0x00fdac42 + 2, 0xf800);  //00FDAC42  0240F000  andi.w #$F000,d0  →  0240F800  andi.w #$F800,d0
        MainMemory.mmrWw (0x00fdac46, 0xed58);      //00FDAC46  EB58      rol.w #5,d0       →  ED58      rol.w #6,d0
        patched++;
      } else {
        failed++;
      }
      //
      //  CCRの未定義ビットを操作している
      //    https://stdkmd.net/bugsx68k/#db_ccrundefbit
      if (MainMemory.mmrRls (0x00fde432) == 0x023c007b) {
        MainMemory.mmrWb (0x00fde432 + 3, 0xfb);  //00FDE432  023C007B  andi.b #$7B,ccr  →  023C00FB  andi.b #$FB,ccr
        MainMemory.mmrWb (0x00fe00d4 + 3, 0xfe);  //00FE00D4  023C007E  andi.b #$7E,ccr  →  023C00FE  andi.b #$FE,ccr
        MainMemory.mmrWb (0x00fe0520 + 3, 0xfe);  //00FE0520  023C007E  andi.b #$7E,ccr  →  023C00FE  andi.b #$FE,ccr
        patched++;
      } else {
        failed++;
      }
      //
      //  例外発生時に浮動小数点レジスタを保存するときFSAVE/FRESTOREしていない
      //    https://stdkmd.net/bugsx68k/#db_fsavefrestore
      if (MainMemory.mmrRwz (0x00fe17cc) == 0x0000) {
        MainMemory.mmrWw (0x00fe17cc, 0x4e71);  //00FE17CC  0000  →  4E71  nop
        MainMemory.mmrWw (0x00fe17e0, 0x4e71);  //00FE17E0  0000  →  4E71  nop
        MainMemory.mmrWw (0x00fe17e2, 0xf35f);  //00FE17E2  F379  →  F35F  frestore (sp)+
        patched++;
      } else {
        failed++;
      }
      //
      //  浮動小数点数の10の累乗のテーブルの誤差が大きい
      //    https://stdkmd.net/bugsx68k/#db_base10
      if (MainMemory.mmrRls (0x00fe2dee + 8) == 0xa0000005) {
        MainMemory.mmrWla (0x00fe2dee,
                           0x40300000, 0xe35fa931, 0xa0000000,
                           0x402d0000, 0xb5e620f4, 0x80000000,
                           0x402a0000, 0x9184e72a, 0x00000000,
                           0x40260000, 0xe8d4a510, 0x00000000,
                           0x40230000, 0xba43b740, 0x00000000,
                           0x40200000, 0x9502f900, 0x00000000,
                           0x401c0000, 0xee6b2800, 0x00000000,
                           0x40190000, 0xbebc2000, 0x00000000,
                           0x40160000, 0x98968000, 0x00000000,
                           0x40120000, 0xf4240000, 0x00000000,
                           0x400f0000, 0xc3500000, 0x00000000,
                           0x400c0000, 0x9c400000, 0x00000000,
                           0x40080000, 0xfa000000, 0x00000000,
                           0x40050000, 0xc8000000, 0x00000000,
                           0x40020000, 0xa0000000, 0x00000000,
                           0x3fff0000, 0x80000000, 0x00000000,
                           0x3ffb0000, 0xcccccccc, 0xcccccccd,
                           0x3ff80000, 0xa3d70a3d, 0x70a3d70a,
                           0x3ff50000, 0x83126e97, 0x8d4fdf3b,
                           0x3ff10000, 0xd1b71758, 0xe219652c,
                           0x3fee0000, 0xa7c5ac47, 0x1b478423,
                           0x3feb0000, 0x8637bd05, 0xaf6c69b6,
                           0x3fe70000, 0xd6bf94d5, 0xe57a42bc,
                           0x3fe40000, 0xabcc7711, 0x8461cefd,
                           0x3fe10000, 0x89705f41, 0x36b4a597,
                           0x3fdd0000, 0xdbe6fece, 0xbdedd5bf,
                           0x3fda0000, 0xafebff0b, 0xcb24aaff,
                           0x3fd70000, 0x8cbccc09, 0x6f5088cc,
                           0x3fd30000, 0xe12e1342, 0x4bb40e13,
                           0x3fd00000, 0xb424dc35, 0x095cd80f,
                           0x3fcd0000, 0x901d7cf7, 0x3ab0acd9,
                           0x3fc90000, 0xe69594be, 0xc44de15b);
        patched++;
      } else {
        failed++;
      }
      //
      //  software emulationのスペルが間違っている
      //    https://stdkmd.net/bugsx68k/#db_software
      if (MainMemory.mmrRwz (0x00fe34d6) == 0x7765) {  //we
        MainMemory.mmrWw (0x00fe34d6, 0x7761);  //Softwere emulation.  →  Software emulation.
        patched++;
      } else {
        failed++;
      }
      if (patched != 0) {
        System.out.printf (Multilingual.mlnJapanese ?
                           "ROM デバッガ 2.32 にパッチをあてました (%d/%d)\n" :
                           "ROM debugger 2.32 was patched (%d/%d)\n",
                           patched, patched + failed);
      }
      break;

    }  //switch romDebuggerVersion

  }  //romPatch2()

  //romPatch3 ()
  //  ROM Humanにパッチをあてる
  public static void romPatch3 () {

    //ROM Humanのバグまたはおかしなところを修正する
    int patched = 0;
    int failed = 0;
    switch (romHumanVersion) {

    case 0x0100:  //ROM Human 1.00
      break;

    case 0x0215:  //ROM Human 2.15
      //RAMまたはROMから起動してDISK2HDを初期化するときリクエストヘッダの初期化コマンドを設定していない(human215,human301,human302)
      //
      //x形式実行ファイルのメモリアロケーションモードが必要最小ブロックかどうかを確認するビット番号が間違っている(human215,human301,human302)
      //  00FC3476  08010001      btst.l  #$01,d1       →  00FC3476  08010000      btst.l  #$00,d1
      if (MainMemory.mmrRls (0x00fc3476) == 0x08010001) {
        MainMemory.mmrWb (0x00fc3476 + 3, 0x00);
        patched++;
      } else {
        failed++;
      }
      //
      //仮想ディレクトリを展開して実体のドライブに移るときドライブ管理テーブルのアドレスを変更する命令のオペレーションサイズが間違っている(human302)
      //
      //ディレクトリを延長するときルートディレクトリかどうかを判断するためにセクタ番号をデータ部の先頭セクタ番号と比較するとき上位ワードを無視している(human215,human301,human302)
      //                                                →  xxxxxxxx  7000          moveq.l #$00,d0
      //  00FC531C  30280014      move.w  $0014(a0),d0
      //  00FC5320  B240          cmp.w   d0,d1         →  00FC5320  B280          cmp.l   d0,d1
      //  00FC5322  6406          bcc.s   $00FC532A
      //  00FC5324  5241          addq.w  #$01,d1       →  00FC5324  5281          addq.l  #$01,d1
      //  00FC5326  B240          cmp.w   d0,d1         →  00FC5326  B280          cmp.l   d0,d1
      //  00FC5328  4E75          rts
      //
      //FILESのバッファのアドレスのbit31がセットされているとき拡張部分をコピーするループのループカウンタのレジスタが間違っている(human215)
      //  00FC565A  7255          moveq.l #$55,d1
      //  00FC565C  12D8          move.b  (a0)+,(a1)+
      //  00FC565E  51C8FFFC      dbra.w  d0,$00FC565C  →  00FC565E  51C9FFFC      dbra.w  d1,$00FC565C
      if (MainMemory.mmrRls (0x00fc565e) == 0x51c8fffc) {
        MainMemory.mmrWb (0x00fc565e + 1, 0xc9);
        patched++;
      } else {
        failed++;
      }
      //
      //リモートデバイスに対するchmodコマンドのコマンド番号が間違っている(human215)
      //  00FC7264  7057          moveq.l #$57,d0       →  00FC7264  7046          moveq.l #$46,d0
      if (MainMemory.mmrRwz (0x00fc7264) == 0x7057) {
        MainMemory.mmrWb (0x00fc7264 + 1, 0x46);
        patched++;
      } else {
        failed++;
      }
      //
      //サブのメモリ空間を削除するときサブの管理下で常駐したブロックをメインのメモリ空間からサブのメモリ空間に入る方向に繋いでいない(human215,human301,human302)
      //
      //スレッドを切り替えるためのTimer-D割り込みルーチンがMC68030のコプロセッサ命令途中割り込みに対応していない(human215,human301,human302)
      //
      //IOCTRL(19,1)でBPBテーブルをコピーする長さとPDAとイジェクトフラグを書き込む位置が間違っている(human215,human301,human302)
      //  00FCA504  700B          moveq.l #$0B,d0       →  00FCA504  700F          moveq.l #$0F,d0
      //  00FCA506  10DE          move.b  (a6)+,(a0)+
      //  00FCA508  51C8FFFC      dbra.w  d0,$00FCA506
      if (MainMemory.mmrRwz (0x00fca504) == 0x700b) {
        MainMemory.mmrWb (0x00fca504 + 1, 0x0f);
        patched++;
      } else {
        failed++;
      }
      //
      //IOCTRL(19,0)でBPBテーブルのハンドルをBPBテーブルのアドレスとして参照しようとしている(human215)
      //  00FCA520  61000084      bsr.w   $00FCA5A6
      //  00FCA524  206D000E      movea.l $000E(a5),a0
      //  ;BPBテーブルのハンドルを求めるときにd0.w=(d0.w&3)*4を計算しているのでd0.wの上位バイトは既に0になっている
      //  00FCA528  4240          clr.w   d0            →  00FCA528  2C56          movea.l (a6),a6
      //  00FCA52A  102E000A      move.b  $000A(a6),d0
      //  00FCA52E  3080          move.w  d0,(a0)
      if (MainMemory.mmrRwz (0x00fca528) == 0x4240) {
        MainMemory.mmrWw (0x00fca528 + 0, 0x2c56);
        patched++;
      } else {
        failed++;
      }
      //
      if (patched != 0) {
        System.out.printf (Multilingual.mlnJapanese ?
                           "ROM Human 2.15 にパッチをあてました (%d/%d)\n" :
                           "ROM Human 2.15 was patched (%d/%d)\n",
                           patched, patched + failed);
      }
      break;

    }  //switch romHumanVersion

  }  //romPatch3()

  //romSetROMDBOn (on)
  //  ROMデバッガをON/OFFする
  //  SRAMの初期化データに細工をする方法
  //    SRAMを初期化する必要がある
  //    MainMemory.mmrWb (sramInitialData + 0x58, romROMDBOn ? 0xff : 0x00)  //ROMデバッガ起動フラグの初期値
  //  SRAMのROMデバッガ起動フラグを書き換える方法
  //    SRAMの内容が変化する
  //    MainMemory.mmrWb (0x00ed0058, romROMDBOn ? 0xff : 0x00);  //ROMデバッガ起動フラグ
  //  IPLのコードを書き換える方法
  //    SRAMの内容が変化しない
  //    無条件にROMデバッガを起動することができる
  //    IPLのバージョンによって書き換える場所が異なる
  //    IPLROM 1.0 $10870318,$10870507共通
  //        00FF00CE    clr.b   $000009DE.l         ;ROMデバッガ起動済みフラグ($00=未起動,$01=起動済み)をクリア
  //        00FF00D4    move.b  $0000080E.l,d0      ;NUM|OPT.2|OPT.1|CTRL|SHIFTキー押し下げフラグ
  //        00FF00DA    move.b  $00ED0058,d1        ;ROMデバッガ起動フラグ($00=起動しない,$FF=起動する)
  //        00FF00E0    eor.b   d1,d0
  //        00FF00E2    btst.l  #$03,d0             ;OPT.2キー押し下げフラグとROMデバッガ起動フラグの状態が同じときは
  //        00FF00E6    beq.s   $00FF00F6           ;ROMデバッガを起動しない
  //        00FF00E8    move.b  #$01,$000009DE.l    ;ROMデバッガ起動済みフラグ($00=未起動,$01=起動済み)をセット
  //        00FF00F0    jsr     $00FE0000           ;ROMデバッガを起動する
  //        00FF00F6
  //      0x00ff00e6のbeq.sをnopに変更する
  //    IPLROM 1.1
  //      アドレスがずれるだけで該当部分のコードはIPLROM 1.0と同じ
  //      0x00ff0188のbeq.sをnopに変更する
  //    IPLROM 1.2
  //      アドレスがずれるだけで該当部分のコードはIPLROM 1.0と同じ
  //      0x00ff01d4のbeq.sをnopに変更する
  //    IPLROM 1.3
  //        00FF0286    clr.b   $09DE.w             ;ROMデバッガ起動済みフラグ($00=未起動,$01=起動済み)をクリア
  //        00FF028A    move.b  $080E.w,d0          ;NUM|OPT.2|OPT.1|CTRL|SHIFTキー押し下げフラグ
  //        00FF028E    move.b  $00ED0058,d1        ;ROMデバッガ起動フラグ($00=起動しない,$FF=起動する)
  //        00FF0294    eor.b   d1,d0
  //        00FF0296    btst.l  #$03,d0             ;OPT.2キー押し下げフラグとROMデバッガ起動フラグの状態が同じときは
  //        00FF029A    beq.s   $00FF02A8           ;ROMデバッガを起動しない
  //        00FF029C    move.b  #$01,$09DE.w        ;ROMデバッガ起動済みフラグ($00=未起動,$01=起動済み)をセット
  //        00FF02A2    movea.l $00FF0008(pc),a0    ;ROMデバッガエントリ
  //        00FF02A6    jsr     (a0)                ;ROMデバッガを起動する
  //        00FF02A8
  //      IPLROM 1.0との違いは絶対ロングが絶対ショートになったこととROMデバッガエントリが間接参照になったこと
  //      0x00ff029aのbeq.sをnopに変更する
  public static void romSetROMDBOn (boolean on) {
    if (romVersionLong == 0x10870318) {         //IPLROM 1.0 ($10870318)
      MainMemory.mmrWw (0x00ff00e6, on ? 0x4e71 : 0x670e);  //nop/beq.s
    } else if (romVersionLong == 0x10870507) {  //IPLROM 1.0 ($10870507)
      MainMemory.mmrWw (0x00ff00e6, on ? 0x4e71 : 0x670e);  //nop/beq.s
    } else if (romVersionLong == 0x11910111) {  //IPLROM 1.1 ($11910111)
      MainMemory.mmrWw (0x00ff0188, on ? 0x4e71 : 0x670e);  //nop/beq.s
    } else if (romVersionLong == 0x12911024) {  //IPLROM 1.2 ($12911024)
      MainMemory.mmrWw (0x00ff01d4, on ? 0x4e71 : 0x670e);  //nop/beq.s
    } else if (romVersionLong == 0x13921127 ||  //IPLROM 1.3 ($13921127)
               romPreviousVersionLong == 0x13921127) {
      MainMemory.mmrWw (0x00ff029a, on ? 0x4e71 : 0x670c);  //nop/beq.s
    } else {
      return;
    }
    romROMDBOn = on;
    if (on) {
      System.out.println (Multilingual.mlnJapanese ?
                          "ROM デバッガを起動するために IPLROM にパッチをあてました" :
                          "IPLROM was patched to start ROM debugger");
    }
  }  //romSetROMDBOn(boolean)



  public static File romLastXEIJROMFile = new File ("XEIJROM.DAT");
  public static File romLastEVEN030File = new File ("EVEN030.BIN");
  public static File romLastODD030File = new File ("ODD030.BIN");
  public static File romLastEVENXVIFile = new File ("EVENXVI.BIN");
  public static File romLastODDXVIFile = new File ("ODDXVI.BIN");

  //romOpenXEIJROMDialog ()
  //  XEIJROM.DATを保存するダイアログを開く
  public static void romOpenXEIJROMDialog () {
    romLastXEIJROMFile = romOpenDATDialog ("Create XEIJROM.DAT : XEiJ ROM data file", romLastXEIJROMFile, romMakeXEIJROMArray ());
  }  //romOpenXEIJROMDialog

  //romOpenEVEN030Dialog ()
  //  EVEN030.BINを保存するダイアログを開く
  public static void romOpenEVEN030Dialog () {
    romLastEVEN030File = romOpenBINDialog ("Create EVEN030.BIN : EVEN ROM binary file for X68030", romLastEVEN030File, romMakeEVEN030Array ());
  }  //romOpenEVEN030Dialog

  //romOpenODD030Dialog ()
  //  ODD030.BINを保存するダイアログを開く
  public static void romOpenODD030Dialog () {
    romLastODD030File = romOpenBINDialog ("Create ODD030.BIN : ODD ROM binary file for X68030", romLastODD030File, romMakeODD030Array ());
  }  //romOpenODD030Dialog

  //romOpenEVENXVIDialog ()
  //  EVENXVI.BINを保存するダイアログを開く
  public static void romOpenEVENXVIDialog () {
    romLastEVENXVIFile = romOpenBINDialog ("Create EVENXVI.BIN : EVEN ROM binary file for X68000 XVI", romLastEVENXVIFile, romMakeEVENXVIArray ());
  }  //romOpenEVENXVIDialog

  //romOpenODDXVIDialog ()
  //  ODDXVI.BINを保存するダイアログを開く
  public static void romOpenODDXVIDialog () {
    romLastODDXVIFile = romOpenBINDialog ("Create ODDXVI.BIN : ODD ROM binary file for X68000 XVI", romLastODDXVIFile, romMakeODDXVIArray ());
  }  //romOpenODDXVIDialog

  //romOpenDATDialog (datFile, datArray)
  //  データファイルを保存するダイアログを開く
  public static File romOpenDATDialog (String title, File datFile, byte[] datArray) {
    if (datFile == null || datArray == null) {
      return null;
    }
    JFileChooser2 fileChooser = new JFileChooser2 ();
    fileChooser.setDialogTitle (title);
    fileChooser.setFileFilter (new javax.swing.filechooser.FileFilter () {  //java.io.FileFilterと紛らわしい
      @Override public boolean accept (File file) {
        String name = file.getName ();
        String upperName = name.toUpperCase ();
        return (file.isDirectory () ||
                (file.isFile () &&
                 upperName.endsWith (".DAT")));
      }
      @Override public String getDescription () {
        return (Multilingual.mlnJapanese ?
                "データファイル (*.DAT)" :
                "data file (*.DAT)");
      }
    });
    fileChooser.setSelectedFile (datFile);
    if (fileChooser.showSaveDialog (null) == JFileChooser.APPROVE_OPTION) {
      datFile = fileChooser.getSelectedFile ();
      XEiJ.rscPutFile (datFile.getPath (), datArray);
    }
    return datFile;
  }  //romOpenDATDialog

  //romOpenBINDialog (binFile, binArray)
  //  バイナリファイルを保存するダイアログを開く
  public static File romOpenBINDialog (String title, File binFile, byte[] binArray) {
    if (binFile == null || binArray == null) {
      return null;
    }
    JFileChooser2 fileChooser = new JFileChooser2 ();
    fileChooser.setDialogTitle (title);
    fileChooser.setFileFilter (new javax.swing.filechooser.FileFilter () {  //java.io.FileFilterと紛らわしい
      @Override public boolean accept (File file) {
        String name = file.getName ();
        String upperName = name.toUpperCase ();
        return (file.isDirectory () ||
                (file.isFile () &&
                 upperName.endsWith (".BIN")));
      }
      @Override public String getDescription () {
        return (Multilingual.mlnJapanese ?
                "バイナリファイル (*.BIN)" :
                "binary file (*.BIN)");
      }
    });
    fileChooser.setSelectedFile (binFile);
    if (fileChooser.showSaveDialog (null) == JFileChooser.APPROVE_OPTION) {
      binFile = fileChooser.getSelectedFile ();
      XEiJ.rscPutFile (binFile.getPath (), binArray);
    }
    return binFile;
  }  //romOpenBINDialog

  //datArray = romMakeXEIJROMArray ()
  //  XEIJROM.DATの配列を作る
  //  1MB
  public static byte[] romMakeXEIJROMArray () {
    //バージョンを確認する
    if (romVersionShort <= 3) {
      return null;
    }
    //現在のROMをコピーする
    byte[] datArray = new byte[1024 * 1024];
    System.arraycopy (MainMemory.mmrM8, 0x00f00000 - 0x00000000,  //コピー元
                      datArray, 0x00f00000 - 0x00f00000,  //コピー先
                      1024 * 1024);  //長さ
    //SRAM初期化データを初期値に戻す
    final int address = 0x00ff09e8;
    final int length = 91;
    byte[] iplrom13 = XEiJ.rscGetResource ("IPLROM30.DAT");
    System.arraycopy (iplrom13, address - 0x00fe0000,  //コピー元
                      datArray, address - 0x00f00000,  //コピー先
                      length);  //長さ
    return datArray;
  }  //romMakeXEIJROMArray

  //binArray = romMakeEVEN030Array ()
  //  EVEN030.BINの配列を作る
  //  X68030用EVEN ROM。256Kbit*16=4Mbit(512KB)。リトルエンディアン
  public static byte[] romMakeEVEN030Array () {
    byte[] datArray = romMakeXEIJROMArray ();
    if (datArray == null) {
      return null;
    }
    byte[] binArray = new byte[1024 * 512];
    for (int i = 0; i < 1024 * 1024 / 4; i++) {
      binArray[2 * i + 1] = datArray[4 * i    ];
      binArray[2 * i    ] = datArray[4 * i + 1];
    }
    return binArray;
  }  //romMakeEVEN030Array

  //binArray = romMakeODD030Array ()
  //  ODD030.BINの配列を作る
  //  X68030用ODD ROM。256Kbit*16=4Mbit(512KB)。リトルエンディアン
  public static byte[] romMakeODD030Array () {
    byte[] datArray = romMakeXEIJROMArray ();
    if (datArray == null) {
      return null;
    }
    byte[] binArray = new byte[1024 * 512];
    for (int i = 0; i < 1024 * 1024 / 4; i++) {
      binArray[2 * i + 1] = datArray[4 * i + 2];
      binArray[2 * i    ] = datArray[4 * i + 3];
    }
    return binArray;
  }  //romMakeODD030Array

  //binArray = romMakeEVENXVIArray ()
  //  EVENXVI.BINの配列を作る
  //  XVI用(増設RAMボードCZ-6BE2A用)EVEN ROM。128Kbit*8=1024Mbit(128KB)
  public static byte[] romMakeEVENXVIArray () {
    byte[] datArray = romMakeXEIJROMArray ();
    if (datArray == null) {
      return null;
    }
    byte[] binArray = new byte[1024 * 128];
    for (int i = 0; i < 1024 * 256 / 2; i++) {
      binArray[i] = datArray[1024 * 768 + 2 * i    ];
    }
    return binArray;
  }  //romMakeEVENXVIArray

  //binArray = romMakeODDXVIArray ()
  //  ODDXVI.BINの配列を作る
  //  XVI用(増設RAMボードCZ-6BE2A用)ODD ROM。128Kbit*8=1024Mbit(128KB)
  public static byte[] romMakeODDXVIArray () {
    byte[] datArray = romMakeXEIJROMArray ();
    if (datArray == null) {
      return null;
    }
    byte[] binArray = new byte[1024 * 128];
    for (int i = 0; i < 1024 * 256 / 2; i++) {
      binArray[i] = datArray[1024 * 768 + 2 * i + 1];
    }
    return binArray;
  }  //romMakeODDXVIArray



}  //class ROM



