//========================================================================================
//  FontPage.java
//    en:Font editor
//    ja:フォントエディタ
//  Copyright (C) 2003-2019 Makoto Kamada
//
//  This file is part of the XEiJ (X68000 Emulator in Java).
//  You can use, modify and redistribute the XEiJ if the conditions are met.
//  Read the XEiJ License for more details.
//  https://stdkmd.net/xeij/
//========================================================================================

package xeij;

import java.awt.*;  //BasicStroke,BorderLayout,BoxLayout,Color,Component,Container,Cursor,Desktop,Dimension,Font,FlowLayout,Frame,Graphics,Graphics2D,GraphicsDevice,GraphicsEnvironment,GridLayout,Image,Insets,Paint,Point,Rectangle,RenderingHints,Robot,Shape,Stroke,TexturePaint,Toolkit
import java.awt.event.*;  //ActionEvent,ActionListener,ComponentAdapter,ComponentEvent,ComponentListener,FocusAdapter,FocusEvent,FocusListener,InputEvent,KeyAdapter,KeyEvent,KeyListener,MouseAdapter,MouseEvent,MouseListener,MouseMotionAdapter,MouseWheelEvent,WindowAdapter,WindowEvent,WindowListener,WindowStateListener
import java.awt.font.*;  //FontRenderContext,LineMetrics,TextLayout
import java.awt.geom.*;  //AffineTransform,GeneralPath,Point2D,Rectangle2D
import java.awt.image.*;  //BufferedImage,DataBuffer,DataBufferByte,DataBufferInt,IndexColorModel
import java.io.*;  //BufferedInputStream,BufferedOutputStream,BufferedReader,BufferedWriter,ByteArrayOutputStream,File,FileInputStream,FileNotFoundException,FileReader,InputStream,InputStreamReader,IOException,OutputStreamWriter,RandomAccessFile,UnsupportedEncodingException
import java.lang.*;  //Boolean,Character,Class,Comparable,Double,Exception,Float,IllegalArgumentException,Integer,InterruptedException,Long,Math,Number,Object,Runnable,SecurityException,String,StringBuilder,System
import javax.swing.*;  //AbstractButton,AbstractSpinnerModel,Box,ButtonGroup,DefaultListModel,ImageIcon,JButton,JCheckBox,JCheckBoxMenuItem,JComponent,JDialog,JFileChooser,JFrame,JLabel,JList,JMenu,JMenuBar,JMenuItem,JPanel,JRadioButton,JScrollPane,JSpinner,JTextArea,JTextField,JTextPane,JViewport,ScrollPaneConstants,SpinnerListModel,SpinnerNumberModel,SwingConstants,SwingUtilities,UIManager,UIDefaults,UnsupportedLookAndFeelException
import java.util.*;  //ArrayList,Arrays,Calendar,GregorianCalendar,HashMap,Map,Map.Entry,TimeZone,Timer,TimerTask,TreeMap

//========================================================================================
//
//  フォントマップ
//       開始        終了        ID       種類    サイズ  文字数  備考
//    0x00f00000  0x00f388c0  Zen16x16  全角漢字   16x16    7238  94点×77区(1区～8区,16区～84区)
//    0x00f388c0  0x00f3a000                                      空き(0x00で充填)
//    0x00f3a000  0x00f3a800  Yon8x8    1/4角ANK    8x8      256  16桁×16行
//    0x00f3a800  0x00f3b800  Han8x8    半角ANK     8x16     256  16桁×16行
//    0x00f3b800  0x00f3d000  Yon12x12  1/4角ANK   12x12     256  16桁×16行
//    0x00f3d000  0x00f40000  Han12x24  半角ANK    12x24     256  16桁×16行
//    0x00f40000  0x00fbf3b0  Zen24x24  全角漢字   24x24    7238  94点×77区(1区～8区,16区～84区)
//    0x00fbf3b0  0x00fbf400                                      空き(0x00で充填)
//    0x00fbf400  0x00fc0000  Han6x12   半角ANK     6x12     256  16桁×16行(0x00～0xff)。IPLROM1.3
//    0x00ffd018  0x00ffdc00  Han6x12   半角ANK     6x12     254  16桁×16行(0x00～0xfd)。IPLROM1.0
//    0x00ffd344  0x00ffdf2c  Han6x12   半角ANK     6x12     254  16桁×16行(0x00～0xfd)。IPLROM1.1
//    0x00ffd45e  0x00ffe046  Han6x12   半角ANK     6x12     254  16桁×16行(0x00～0xfd)。IPLROM1.2
//
//  CGROMの構築手順
//    起動時
//      -rom=～で指定されたROMファイル(1MB)を読み込む
//      -model～で選択されたIPLROMファイル(128KB)を読み込む
//      -iplrom=～で指定されたIPLROMファイル(128KB)を読み込む
//      -cgrom=～で指定されたCGROM.DATファイル(768KB)を読み込む
//      -cgrom=～が指定されなかったときリソースファイルCGROM_XEiJ.DAT(768KB)を読み込む
//      -fon=～などで指定されたフォントデータファイルを読み込む
//    CGROM1(0x00f00000～0x00f40000)が最初にアクセスされたとき
//      Zen16x16がなければ描画して作る
//      Yon8x8がなければ描画して作る
//      Han8x16がなければ描画して作る
//      Yon12x12がなければ描画して作る
//      Han12x24がなければ描画して作る
//    CGROM2(0x00f40000～0x00fc0000)が最初にアクセスされたとき
//      Zen24x24がなければ描画して作る
//      Han6x12がなければ描画して作る
//
//    CGROMを書き換えても@IOCSのバッファやローカルメモリにコピーされたROMは更新されないことに注意
//    リソースファイルのCGROM_XEiJ.DATはMISAKI_G.F8,GOX80.F12,GOL80.FON,MIN.F24を組み合わせたもの

public class FontEditor {

  //スイッチ
  public static final boolean FNT_CREATE_CGROM_LATER = true;  //足りないフォントは後で必要になったときに作る

  //ページの名前
  //  英語
  public static final String FNT_EN_Yon4x4   = "Quarter ANK 4x4";
  public static final String FNT_EN_Yon6x6   = "Quarter ANK 6x6";
  public static final String FNT_EN_Yon8x8   = "Quarter ANK 8x8 (CGROM)";
  public static final String FNT_EN_Yon12x12 = "Quarter ANK 12x12 (CGROM)";
  public static final String FNT_EN_Yon16x16 = "Quarter ANK 16x16";
  public static final String FNT_EN_Yon24x24 = "Quarter ANK 24x24";
  public static final String FNT_EN_Han4x8   = "Half ANK 4x8";
  public static final String FNT_EN_Han6x12  = "Half ANK 6x12 (CGROM)";
  public static final String FNT_EN_Han8x16  = "Half ANK 8x16 (CGROM)";
  public static final String FNT_EN_Han12x24 = "Half ANK 12x24 (CGROM)";
  public static final String FNT_EN_Han16x32 = "Half ANK 16x32";
  public static final String FNT_EN_Han24x48 = "Half ANK 24x48";
  public static final String FNT_EN_Zen8x8   = "Full Kanji 8x8";
  public static final String FNT_EN_Zen12x12 = "Full Kanji 12x12";
  public static final String FNT_EN_Zen16x16 = "Full Kanji 16x16 (CGROM)";
  public static final String FNT_EN_Zen24x24 = "Full Kanji 24x24 (CGROM)";
  public static final String FNT_EN_Zen32x32 = "Full Kanji 32x32";
  public static final String FNT_EN_Zen48x48 = "Full Kanji 48x48";
  public static final String FNT_EN_Pic36x46 = "CZ-8PC4 Pica 36x46";
  public static final String FNT_EN_Eli30x46 = "CZ-8PC4 Elite 30x46";
  public static final String FNT_EN_Sma18x46 = "CZ-8PC4 Small 18x46";
  public static final String FNT_EN_Scr28x32 = "CZ-8PC4 Script 28x32";
  public static final String FNT_EN_Pan24x48 = "CZ-8PC4 Half 24x48";
  public static final String FNT_EN_Lcd4x6   = "LCD 4x6";
  public static final String FNT_EN_Lcd6x8   = "LCD 6x8";
  public static final String FNT_EN_Smk6x8   = "Shiromadokun 6x8";
  //  日本語
  public static final String FNT_JA_Yon4x4   = "1/4角 ANK 4x4";
  public static final String FNT_JA_Yon6x6   = "1/4角 ANK 6x6";
  public static final String FNT_JA_Yon8x8   = "1/4角 ANK 8x8 (CGROM)";
  public static final String FNT_JA_Yon12x12 = "1/4角 ANK 12x12 (CGROM)";
  public static final String FNT_JA_Yon16x16 = "1/4角 ANK 16x16";
  public static final String FNT_JA_Yon24x24 = "1/4角 ANK 24x24";
  public static final String FNT_JA_Han4x8   = "半角 ANK 4x8";
  public static final String FNT_JA_Han6x12  = "半角 ANK 6x12 (CGROM)";
  public static final String FNT_JA_Han8x16  = "半角 ANK 8x16 (CGROM)";
  public static final String FNT_JA_Han12x24 = "半角 ANK 12x24 (CGROM)";
  public static final String FNT_JA_Han16x32 = "半角 ANK 16x32";
  public static final String FNT_JA_Han24x48 = "半角 ANK 24x48";
  public static final String FNT_JA_Zen8x8   = "全角 漢字 8x8";
  public static final String FNT_JA_Zen12x12 = "全角 漢字 12x12";
  public static final String FNT_JA_Zen16x16 = "全角 漢字 16x16 (CGROM)";
  public static final String FNT_JA_Zen24x24 = "全角 漢字 24x24 (CGROM)";
  public static final String FNT_JA_Zen32x32 = "全角 漢字 32x32";
  public static final String FNT_JA_Zen48x48 = "全角 漢字 48x48";
  public static final String FNT_JA_Pic36x46 = "CZ-8PC4 パイカ 36x46";
  public static final String FNT_JA_Eli30x46 = "CZ-8PC4 エリート 30x46";
  public static final String FNT_JA_Sma18x46 = "CZ-8PC4 縮小 18x46";
  public static final String FNT_JA_Scr28x32 = "CZ-8PC4 スクリプト 28x32";
  public static final String FNT_JA_Pan24x48 = "CZ-8PC4 半角 24x48";
  public static final String FNT_JA_Lcd4x6   = "LCD 4x6";
  public static final String FNT_JA_Lcd6x8   = "LCD 6x8";
  public static final String FNT_JA_Smk6x8   = "白窓君 6x8";

  //ページ
  public static FontPage.Yon fntPageYon4x4;
  public static FontPage.Yon fntPageYon6x6;
  public static FontPage.Yon fntPageYon8x8;
  public static FontPage.Yon fntPageYon12x12;
  public static FontPage.Yon fntPageYon16x16;
  public static FontPage.Yon fntPageYon24x24;
  public static FontPage.Han fntPageHan4x8;
  public static FontPage.Han fntPageHan6x12;
  public static FontPage.Han fntPageHan8x16;
  public static FontPage.Han fntPageHan12x24;
  public static FontPage.Han fntPageHan16x32;
  public static FontPage.Han fntPageHan24x48;
  public static FontPage.Zen fntPageZen8x8;
  public static FontPage.Zen fntPageZen12x12;
  public static FontPage.Zen fntPageZen16x16;
  public static FontPage.Zen fntPageZen24x24;
  public static FontPage.Zen fntPageZen32x32;
  public static FontPage.Zen fntPageZen48x48;
  public static FontPage.Prn fntPagePic36x46;
  public static FontPage.Prn fntPageEli30x46;
  public static FontPage.Prn fntPageSma18x46;
  public static FontPage.Prn fntPageScr28x32;
  public static FontPage.Prn fntPagePan24x48;
  public static FontPage.Lcd fntPageLcd4x6;
  public static FontPage.Lcd fntPageLcd6x8;
  public static FontPage.Smk fntPageSmk6x8;
  public static FontPage[] fntPageArray;
  public static FontPage[] fntCGROMPageArray;
  public static FontPage fntSelectedPage;

  //パラメータ
  //  CGROM.DATファイル名
  public static String fntCGROMName;  //-cgrom
  //  フォントファイル名
  public static String fntFq4Name;   //-fq4   Yon4x4
  public static String fntFq6Name;   //-fq6   Yon6x6
  public static String fntFq8Name;   //-fq8   Yon8x8
  public static String fntFq12Name;  //-fq12  Yon12x12
  public static String fntFq16Name;  //-fq16  Yon16x16
  public static String fntFq24Name;  //-fq24  Yon24x24
  public static String fntF8Name;    //-f8    Han4x8+Zen8x8
  public static String fntF12Name;   //-f12   Han6x12+Zen12x12
  public static String fntFonName;   //-fon   Han8x16+Zen16x16
  public static String fntF24Name;   //-f24   Han12x24+Zen24x24
  public static String fntF32Name;   //-f32   Han16x32+Zen32x32
  public static String fntF48Name;   //-f48   Han24x48+Zen48x48
  public static String fntFpicName;  //-fpic  Pic36x46
  public static String fntFeliName;  //-feli  Eli30x46
  public static String fntFsmaName;  //-fsma  Sma18x46
  public static String fntFscrName;  //-fscr  Scr28x32
  public static String fntFpanName;  //-fhan  Pan24x48
  public static String fntF4x6Name;  //-f4x6  Lcd4x6
  public static String fntF6x8Name;  //-f6x8  Lcd6x8
  public static String fntFsmkName;  //-fsmk  Smk6x8
  //  おむすびフォント
  public static boolean fntOmusubiOn;  //-omusubi

  //アドレス
  public static final int FNT_ADDRESS_Zen16x16 = 0x00f00000;  //全角 漢字 16x16
  public static final int FNT_ADDRESS_Yon8x8   = 0x00f3a000;  //1/4角 ANK 8x8
  public static final int FNT_ADDRESS_Han8x16  = 0x00f3a800;  //半角 ANK 8x16
  public static final int FNT_ADDRESS_Yon12x12 = 0x00f3b800;  //1/4角 ANK 12x12
  public static final int FNT_ADDRESS_Han12x24 = 0x00f3d000;  //半角 ANK 12x24
  public static final int FNT_ADDRESS_Zen24x24 = 0x00f40000;  //全角 漢字 24x24
  public static final int FNT_ADDRESS_Han6x12  = 0x00fbf400;  //半角 ANK 6x12
  public static int fntHan6x12Address;  //IPLROMにあるHan6x12の先頭アドレス。IPLROMのバージョンを確認したときに設定される。0=なし

  //等幅フォントファミリ
  public static final String[] FNT_GOTHIC_FAMILIES = {  //ゴシック
    //"游ゴシック",  //Windows,Mac
    "ＭＳ ゴシック", "MS Gothic",  //Windows
    "ヒラギノ角ゴ ProN W3", "Hiragino Kaku Gothic ProN",  //Mac
    "ヒラギノ角ゴ Pro W3", "Hiragino Kaku Gothic Pro",
    "Osaka－等幅", "Osaka-Mono",
    "VL ゴシック", "VL Gothic",  //Linux
    "Takaoゴシック", "TakaoGothic",
    "IPAゴシック", "IPAGothic",
  };
  public static final String[] FNT_MINCHO_FAMILIES = {  //明朝
    //"游明朝",  //Windows,Mac
    "ＭＳ 明朝", "MS Mincho",  //Windows
    "ヒラギノ明朝 ProN W3", "Hiragino Mincho ProN",  //Mac
    "ヒラギノ明朝 Pro W3", "Hiragino Mincho Pro",
    "さざなみ明朝", "Sazanami Mincho",  //Linux
    "Takao明朝", "TakaoMincho",
    "IPA明朝", "IPAMincho",
  };
  public static String[] fntAvailableFamilies;  //使えるフォントの一覧
  public static String fntGothicFamily;  //ゴシック
  public static String fntMinchoFamily;  //明朝

  //fntInit ()
  //  初期化
  //  -romと-iplromを処理してから呼び出すこと
  //  -cgromはここで処理する
  public static void fntInit () {

    //設定
    //  CGROM.DATファイル名
    fntCGROMName = Settings.sgsGetString ("cgrom");
    //  フォントファイル名
    fntFq4Name  = Settings.sgsGetString ("fq4");
    fntFq6Name  = Settings.sgsGetString ("fq6");
    fntFq8Name  = Settings.sgsGetString ("fq8");
    fntFq12Name = Settings.sgsGetString ("fq12");
    fntFq16Name = Settings.sgsGetString ("fq16");
    fntFq24Name = Settings.sgsGetString ("fq24");
    fntF8Name   = Settings.sgsGetString ("f8");
    fntF12Name  = Settings.sgsGetString ("f12");
    fntFonName  = Settings.sgsGetString ("fon");
    fntF24Name  = Settings.sgsGetString ("f24");
    fntF32Name  = Settings.sgsGetString ("f32");
    fntF48Name  = Settings.sgsGetString ("f48");
    fntFpicName = Settings.sgsGetString ("fpic");
    fntFeliName = Settings.sgsGetString ("feli");
    fntFsmaName = Settings.sgsGetString ("fsma");
    fntFscrName = Settings.sgsGetString ("fscr");
    fntFpanName = Settings.sgsGetString ("fpan");
    fntF4x6Name = Settings.sgsGetString ("f4x6");
    fntF6x8Name = Settings.sgsGetString ("f6x8");
    fntFsmkName = Settings.sgsGetString ("fsmk");
    //  おむすびフォント
    fntOmusubiOn = Settings.sgsGetOnOff ("omusubi");

    //ページ
    fntPageArray = new FontPage[] {
      fntPageYon4x4   = new FontPage.Yon ( 4,  4, FNT_EN_Yon4x4,   FNT_JA_Yon4x4,   "./yon4x4.dat",   "./yon4x4.png"),
      fntPageYon6x6   = new FontPage.Yon ( 6,  6, FNT_EN_Yon6x6,   FNT_JA_Yon6x6,   "./yon6x6.dat",   "./yon6x6.png"),
      fntPageYon8x8   = new FontPage.Yon ( 8,  8, FNT_EN_Yon8x8,   FNT_JA_Yon8x8,   "./yon8x8.dat",   "./yon8x8.png",
                                          MainMemory.mmrM8, FNT_ADDRESS_Yon8x8),
      fntPageYon12x12 = new FontPage.Yon (12, 12, FNT_EN_Yon12x12, FNT_JA_Yon12x12, "./yon12x12.dat", "./yon12x12.png",
                                          MainMemory.mmrM8, FNT_ADDRESS_Yon12x12),
      fntPageYon16x16 = new FontPage.Yon (16, 16, FNT_EN_Yon16x16, FNT_JA_Yon16x16, "./yon16x16.dat", "./yon16x16.png"),
      fntPageYon24x24 = new FontPage.Yon (24, 24, FNT_EN_Yon24x24, FNT_JA_Yon24x24, "./yon24x24.dat", "./yon24x24.png"),
      fntPageHan4x8   = new FontPage.Han ( 4,  8, FNT_EN_Han4x8,   FNT_JA_Han4x8,   "./han4x8.f8",    "./han4x8.png"),
      fntPageHan6x12  = new FontPage.Han ( 6, 12, FNT_EN_Han6x12,  FNT_JA_Han6x12,  "./han6x12.f12",  "./han6x12.png",
                                          MainMemory.mmrM8, FNT_ADDRESS_Han6x12),
      fntPageHan8x16  = new FontPage.Han ( 8, 16, FNT_EN_Han8x16,  FNT_JA_Han8x16,  "./han8x16.fon",  "./han8x16.png",
                                          MainMemory.mmrM8, FNT_ADDRESS_Han8x16),
      fntPageHan12x24 = new FontPage.Han (12, 24, FNT_EN_Han12x24, FNT_JA_Han12x24, "./han12x24.f24", "./han12x24.png",
                                          MainMemory.mmrM8, FNT_ADDRESS_Han12x24),
      fntPageHan16x32 = new FontPage.Han (16, 32, FNT_EN_Han16x32, FNT_JA_Han16x32, "./han16x32.f32", "./han16x32.png"),
      fntPageHan24x48 = new FontPage.Han (24, 48, FNT_EN_Han24x48, FNT_JA_Han24x48, "./han24x48.f48", "./han24x48.png"),
      fntPageZen8x8   = new FontPage.Zen ( 8,  8, FNT_EN_Zen8x8,   FNT_JA_Zen8x8,   "./zen8x8.f8",    "./zen8x8.png"),
      fntPageZen12x12 = new FontPage.Zen (12, 12, FNT_EN_Zen12x12, FNT_JA_Zen12x12, "./zen12x12.f12", "./zen12x12.png"),
      fntPageZen16x16 = new FontPage.Zen (16, 16, FNT_EN_Zen16x16, FNT_JA_Zen16x16, "./zen16x16.fon", "./zen16x16.png",
                                          MainMemory.mmrM8, FNT_ADDRESS_Zen16x16),
      fntPageZen24x24 = new FontPage.Zen (24, 24, FNT_EN_Zen24x24, FNT_JA_Zen24x24, "./zen24x24.f24", "./zen24x24.png",
                                          MainMemory.mmrM8, FNT_ADDRESS_Zen24x24),
      fntPageZen32x32 = new FontPage.Zen (32, 32, FNT_EN_Zen32x32, FNT_JA_Zen32x32, "./zen32x32.f32", "./zen32x32.png"),
      fntPageZen48x48 = new FontPage.Zen (48, 48, FNT_EN_Zen48x48, FNT_JA_Zen48x48, "./zen48x48.f48", "./zen48x48.png"),
      fntPagePic36x46 = new FontPage.Prn (36, 46, FNT_EN_Pic36x46, FNT_JA_Pic36x46, "./pic36x46.dat", "./pic36x46.png"),
      fntPageEli30x46 = new FontPage.Prn (30, 46, FNT_EN_Eli30x46, FNT_JA_Eli30x46, "./eli30x46.dat", "./eli30x46.png"),
      fntPageSma18x46 = new FontPage.Prn (18, 46, FNT_EN_Sma18x46, FNT_JA_Sma18x46, "./sma18x46.dat", "./sma18x46.png"),
      fntPageScr28x32 = new FontPage.Prn (28, 32, FNT_EN_Scr28x32, FNT_JA_Scr28x32, "./scr28x32.dat", "./scr28x32.png"),
      fntPagePan24x48 = new FontPage.Prn (24, 48, FNT_EN_Pan24x48, FNT_JA_Pan24x48, "./pan24x48.dat", "./pan24x48.png"),
      fntPageLcd4x6   = new FontPage.Lcd ( 4,  6, FNT_EN_Lcd4x6,   FNT_JA_Lcd4x6,   "./lcd4x6.dat",   "./lcd4x6.png"),
      fntPageLcd6x8   = new FontPage.Lcd ( 6,  8, FNT_EN_Lcd6x8,   FNT_JA_Lcd6x8,   "./lcd6x8.dat",   "./lcd6x8.png"),
      fntPageSmk6x8   = new FontPage.Smk ( 6,  8, FNT_EN_Smk6x8,   FNT_JA_Smk6x8,   "./smk6x8.dat",   "./smk6x8.png"),
    };
    fntCGROMPageArray = new FontPage[] {
      fntPageZen16x16,
      fntPageYon8x8,
      fntPageHan8x16,
      fntPageYon12x12,
      fntPageHan12x24,
      fntPageZen24x24,
      fntPageHan6x12,
    };

    //半角と全角のペア
    //  Han4x8+Zen8x8
    fntPageHan4x8.setZenPage (fntPageZen8x8);
    fntPageZen8x8.setHanPage (fntPageHan4x8);
    //  Han6x12+Zen12x12
    fntPageHan6x12.setZenPage (fntPageZen12x12);
    fntPageZen12x12.setHanPage (fntPageHan6x12);
    //  Han8x16+Zen16x16
    fntPageHan8x16.setZenPage (fntPageZen16x16);
    fntPageZen16x16.setHanPage (fntPageHan8x16);
    //  Han12x24+Zen24x24
    fntPageHan12x24.setZenPage (fntPageZen24x24);
    fntPageZen24x24.setHanPage (fntPageHan12x24);
    //  Han16x32+Zen32x32
    fntPageHan16x32.setZenPage (fntPageZen32x32);
    fntPageZen32x32.setHanPage (fntPageHan16x32);
    //  Han24x48+Zen48x48
    fntPageHan24x48.setZenPage (fntPageZen48x48);
    fntPageZen48x48.setHanPage (fntPageHan24x48);

    //現在選択されているページ
    fntSelectedPage = fntPageZen16x16;

    //-rom
    //  ロード済みのCGROMでフォントを構築する
    fntPageZen16x16.fnpInputMemory ();
    fntPageYon8x8.fnpInputMemory ();
    fntPageHan8x16.fnpInputMemory ();
    fntPageYon12x12.fnpInputMemory ();
    fntPageHan12x24.fnpInputMemory ();
    fntPageZen24x24.fnpInputMemory ();
    fntPageHan6x12.fnpInputMemory ();

    //-fz16dataなど
    //  設定ファイルに保存されていたフォントデータを読み込む
    fntPageYon4x4.fnpLoadFontDataArray (Settings.sgsGetData ("fq4data"));  //-fq4data
    fntPageYon6x6.fnpLoadFontDataArray (Settings.sgsGetData ("fq6data"));  //-fq6data
    if (!fntPageYon8x8.fnpReady) {
      fntPageYon8x8.fnpLoadFontDataArray (Settings.sgsGetData ("fq8data"));  //-fq8data
    }
    if (!fntPageYon12x12.fnpReady) {
      fntPageYon12x12.fnpLoadFontDataArray (Settings.sgsGetData ("fq12data"));  //-fq12data
    }
    fntPageYon16x16.fnpLoadFontDataArray (Settings.sgsGetData ("fq16data"));  //-fq16data
    fntPageYon24x24.fnpLoadFontDataArray (Settings.sgsGetData ("fq24data"));  //-fq24data
    fntPageHan4x8.fnpLoadFontDataArray (Settings.sgsGetData ("fh8data"));  //-fh8data
    if (!fntPageHan6x12.fnpReady) {
      fntPageHan6x12.fnpLoadFontDataArray (Settings.sgsGetData ("fh12data"));  //-fh12data
    }
    if (!fntPageHan8x16.fnpReady) {
      fntPageHan8x16.fnpLoadFontDataArray (Settings.sgsGetData ("fh16data"));  //-fh16data
    }
    if (!fntPageHan12x24.fnpReady) {
      fntPageHan12x24.fnpLoadFontDataArray (Settings.sgsGetData ("fh24data"));  //-fh24data
    }
    fntPageHan16x32.fnpLoadFontDataArray (Settings.sgsGetData ("fh32data"));  //-fh32data
    fntPageHan24x48.fnpLoadFontDataArray (Settings.sgsGetData ("fh48data"));  //-fh48data
    fntPageZen8x8.fnpLoadFontDataArray (Settings.sgsGetData ("fz8data"));  //-fz8data
    fntPageZen12x12.fnpLoadFontDataArray (Settings.sgsGetData ("fz12data"));  //-fz12data
    if (!fntPageZen16x16.fnpReady) {
      fntPageZen16x16.fnpLoadFontDataArray (Settings.sgsGetData ("fz16data"));  //-fz16data
    }
    if (!fntPageZen24x24.fnpReady) {
      fntPageZen24x24.fnpLoadFontDataArray (Settings.sgsGetData ("fz24data"));  //-fz24data
    }
    fntPageZen32x32.fnpLoadFontDataArray (Settings.sgsGetData ("fz32data"));  //-fz32data
    fntPageZen48x48.fnpLoadFontDataArray (Settings.sgsGetData ("fz48data"));  //-fz48data
    fntPagePic36x46.fnpLoadFontDataArray (Settings.sgsGetData ("fpicdata"));  //-fpicdata
    fntPageEli30x46.fnpLoadFontDataArray (Settings.sgsGetData ("felidata"));  //-felidata
    fntPageSma18x46.fnpLoadFontDataArray (Settings.sgsGetData ("fsmadata"));  //-fsmadata
    fntPageScr28x32.fnpLoadFontDataArray (Settings.sgsGetData ("fscrdata"));  //-fscrdata
    fntPagePan24x48.fnpLoadFontDataArray (Settings.sgsGetData ("fpandata"));  //-fpandata
    fntPageLcd4x6.fnpLoadFontDataArray (Settings.sgsGetData ("f4x6data"));  //-f4x6data
    fntPageLcd6x8.fnpLoadFontDataArray (Settings.sgsGetData ("f6x8data"));  //-f6x8data
    fntPageSmk6x8.fnpLoadFontDataArray (Settings.sgsGetData ("fsmkdata"));  //-fsmkdata

    //-cgrom
    //  Zen16x16,Yon8x8,Han8x16,Yon12x12,Han12x24,Zen24x24,Han6x12
    {
      byte[] array = XEiJ.rscGetFile (fntCGROMName);  //-cgromで指定されたファイルを読み込む
      if (array != null) {  //-cgromが指定されていて、読み込めた
        if (array.length == 1024 * 768 ||
            array.length == 1024 * 1024) {  //768KBまたは1MB
          for (FontPage page : fntCGROMPageArray) {
            page.fnpLoadFontDataArray (array, page.fnpMemoryAddress - 0x00f00000, page.fnpMemoryBytes);
          }
        } else {  //768KBまたは1MBではない
          System.out.println (Multilingual.mlnJapanese ?
                              XEiJ.rscLastFileName + " のサイズが違います" :
                              XEiJ.rscLastFileName + " size is wrong");
        }
      } else {  //読み込めなかった
        fntCGROMName = "";  //保存しない
      }
    }

    //-fonなど
    //  指定されたフォントを上書きする
    fntPageYon4x4.fnpLoadFontDataFiles (fntFq4Name);  //-fq4
    fntPageYon6x6.fnpLoadFontDataFiles (fntFq6Name);  //-fq6
    fntPageYon8x8.fnpLoadFontDataFiles (fntFq8Name);  //-fq8
    fntPageYon12x12.fnpLoadFontDataFiles (fntFq12Name);  //-fq12
    fntPageYon16x16.fnpLoadFontDataFiles (fntFq16Name);  //-fq16
    fntPageYon24x24.fnpLoadFontDataFiles (fntFq24Name);  //-fq24
    fntPageHan4x8.fnpLoadFontDataFiles (fntF8Name);  //-f8
    fntPageHan6x12.fnpLoadFontDataFiles (fntF12Name);  //-f12
    fntPageHan8x16.fnpLoadFontDataFiles (fntFonName);  //-fon
    fntPageHan12x24.fnpLoadFontDataFiles (fntF24Name);  //-f24
    fntPageHan16x32.fnpLoadFontDataFiles (fntF32Name);  //-f32
    fntPageHan24x48.fnpLoadFontDataFiles (fntF48Name);  //-f48
    fntPageZen8x8.fnpLoadFontDataFiles (fntF8Name);  //-f8
    fntPageZen12x12.fnpLoadFontDataFiles (fntF12Name);  //-f12
    fntPageZen16x16.fnpLoadFontDataFiles (fntFonName);  //-fon
    fntPageZen24x24.fnpLoadFontDataFiles (fntF24Name);  //-f24
    fntPageZen32x32.fnpLoadFontDataFiles (fntF32Name);  //-f32
    fntPageZen48x48.fnpLoadFontDataFiles (fntF48Name);  //-f48
    fntPagePic36x46.fnpLoadFontDataFiles (fntFpicName);  //-fpic
    fntPageEli30x46.fnpLoadFontDataFiles (fntFeliName);  //-feli
    fntPageSma18x46.fnpLoadFontDataFiles (fntFsmaName);  //-fsma
    fntPageScr28x32.fnpLoadFontDataFiles (fntFscrName);  //-fscr
    fntPagePan24x48.fnpLoadFontDataFiles (fntFpanName);  //-fpan
    fntPageLcd4x6.fnpLoadFontDataFiles (fntF4x6Name);  //-f4x6
    fntPageLcd6x8.fnpLoadFontDataFiles (fntF6x8Name);  //-f6x8
    fntPageSmk6x8.fnpLoadFontDataFiles (fntFsmkName);  //-fsmk

    //おむすびフォント
    //  -omusubiは-fonよりも強い
    if (fntOmusubiOn) {
      fntPageHan8x16.fnpLoadFontDataArray (FNT_OMUSUBI_DATA, 0, 1 * 16 * 256);
    }

    //IPLROM.DAT
    //  Han6x12
    if (!fntPageHan6x12.fnpReady) {
      byte[] array = XEiJ.rscGetResource ("IPLROM.DAT");  //リソースファイルを読み込む
      System.arraycopy (array, 0x00ffd018 - 0x00fe0000, MainMemory.mmrM8, 0x00fbf400, 1 * 12 * 254);  //IPLROMからCGROMに254文字コピーする
      Arrays.fill (MainMemory.mmrM8, 0x00fbf400 + 1 * 12 * 254, 0x00fbf400 + 1 * 12 * 256, (byte) 0);  //255文字目と256文字目をクリアする
      fntPageHan6x12.fnpInputMemory ();
    }

    //CGROM_XEiJ.DAT
    //  CGROMに準備ができていないフォントがあるときCGROM_XEiJ.DATを読み込む
    {
      boolean ready = true;
      for (FontPage page : fntCGROMPageArray) {
        if (!page.fnpReady) {
          ready = false;
          break;
        }
      }
      if (!ready) {  //CGROMに準備ができていないフォントがある
        final String name = "CGROM_XEiJ.DAT";  //リソースファイル名
        byte[] array = XEiJ.rscGetResource (name);  //リソースファイルを読み込む
        if (array != null) {  //リソースファイルを読み込めた
          for (FontPage page : fntCGROMPageArray) {
            if (!page.fnpReady) {
              page.fnpLoadFontDataArray (array, page.fnpMemoryAddress - 0x00f00000, page.fnpMemoryBytes);
            }
          }
        }
      }
    }

    //足りないフォントを作る
    fntAvailableFamilies = null;
    fntGothicFamily = null;
    fntMinchoFamily = null;
    if (!(fntPageZen16x16.fnpReady &&
          fntPageYon8x8.fnpReady &&
          fntPageHan8x16.fnpReady &&
          fntPageYon12x12.fnpReady &&
          fntPageHan12x24.fnpReady)) {  //CGROM1が揃っていない
      if (FNT_CREATE_CGROM_LATER) {  //後で作る
        XEiJ.busSuper (MemoryMappedDevice.MMD_CG1, 0x00f00000, 0x00f40000);
      } else {  //今作る
        fntCreateCGROM1 ();
      }
    }
    if (!(fntPageZen24x24.fnpReady &&
          fntPageHan6x12.fnpReady)) {  //CGROM2が揃っていない
      if (FNT_CREATE_CGROM_LATER) {  //後で作る
        XEiJ.busSuper (MemoryMappedDevice.MMD_CG2, 0x00f40000, 0x00fc0000);
      } else {  //今作る
        fntCreateCGROM2 ();
      }
    }

    //ウインドウ
    fntCanvas = null;

  }  //fntInit()

  //fntTini ()
  //  後始末
  public static void fntTini () {

    //設定
    //  CGROM.DATファイル名
    Settings.sgsPutString ("cgrom", fntCGROMName);
    //  フォントファイル名
    Settings.sgsPutString ("fq4", fntFq4Name);
    Settings.sgsPutString ("fq6", fntFq6Name);
    Settings.sgsPutString ("fq8", fntFq8Name);
    Settings.sgsPutString ("fq12", fntFq12Name);
    Settings.sgsPutString ("fq16", fntFq16Name);
    Settings.sgsPutString ("fq24", fntFq24Name);
    Settings.sgsPutString ("f8", fntF8Name);
    Settings.sgsPutString ("f12", fntF12Name);
    Settings.sgsPutString ("fon", fntFonName);
    Settings.sgsPutString ("f24", fntF24Name);
    Settings.sgsPutString ("f32", fntF32Name);
    Settings.sgsPutString ("f48", fntF48Name);
    Settings.sgsPutString ("fpic", fntFpicName);
    Settings.sgsPutString ("feli", fntFeliName);
    Settings.sgsPutString ("fsma", fntFsmaName);
    Settings.sgsPutString ("fscr", fntFscrName);
    Settings.sgsPutString ("fpan", fntFpanName);
    Settings.sgsPutString ("f4x6", fntF4x6Name);
    Settings.sgsPutString ("f6x8", fntF6x8Name);
    Settings.sgsPutString ("fsmk", fntFsmkName);
    //  フォントデータ
    Settings.sgsPutData ("fq4data", fntPageYon4x4.fnpBinaryArray);
    Settings.sgsPutData ("fq6data", fntPageYon6x6.fnpBinaryArray);
    Settings.sgsPutData ("fq8data", fntPageYon8x8.fnpBinaryArray);
    Settings.sgsPutData ("fq12data", fntPageYon12x12.fnpBinaryArray);
    Settings.sgsPutData ("fq16data", fntPageYon16x16.fnpBinaryArray);
    Settings.sgsPutData ("fq24data", fntPageYon24x24.fnpBinaryArray);
    Settings.sgsPutData ("fh8data", fntPageHan4x8.fnpBinaryArray);
    Settings.sgsPutData ("fh12data", fntPageHan6x12.fnpBinaryArray);
    Settings.sgsPutData ("fh16data", fntPageHan8x16.fnpBinaryArray);
    Settings.sgsPutData ("fh24data", fntPageHan12x24.fnpBinaryArray);
    Settings.sgsPutData ("fh32data", fntPageHan16x32.fnpBinaryArray);
    Settings.sgsPutData ("fh48data", fntPageHan24x48.fnpBinaryArray);
    Settings.sgsPutData ("fz8data", fntPageZen8x8.fnpBinaryArray);
    Settings.sgsPutData ("fz12data", fntPageZen12x12.fnpBinaryArray);
    Settings.sgsPutData ("fz16data", fntPageZen16x16.fnpBinaryArray);
    Settings.sgsPutData ("fz24data", fntPageZen24x24.fnpBinaryArray);
    Settings.sgsPutData ("fz32data", fntPageZen32x32.fnpBinaryArray);
    Settings.sgsPutData ("fz48data", fntPageZen48x48.fnpBinaryArray);
    Settings.sgsPutData ("fpicdata", fntPagePic36x46.fnpBinaryArray);
    Settings.sgsPutData ("felidata", fntPageEli30x46.fnpBinaryArray);
    Settings.sgsPutData ("fsmadata", fntPageSma18x46.fnpBinaryArray);
    Settings.sgsPutData ("fscrdata", fntPageScr28x32.fnpBinaryArray);
    Settings.sgsPutData ("fpandata", fntPagePan24x48.fnpBinaryArray);
    Settings.sgsPutData ("f4x6data", fntPageLcd4x6.fnpBinaryArray);
    Settings.sgsPutData ("f6x8data", fntPageLcd6x8.fnpBinaryArray);
    Settings.sgsPutData ("fsmkdata", fntPageSmk6x8.fnpBinaryArray);
    //  おむすびフォント
    Settings.sgsPutOnOff ("omusubi", fntOmusubiOn);

  }



  //fntGetAvailableFamilies ()
  public static String[] fntGetAvailableFamilies () {
    if (fntAvailableFamilies == null) {
      //フォントファミリの一覧を用意する
      fntAvailableFamilies = GraphicsEnvironment.getLocalGraphicsEnvironment ().getAvailableFontFamilyNames ();
      Arrays.sort (fntAvailableFamilies, null);
    }
    return fntAvailableFamilies;
  }  //fntGetAvailableFamilies()

  //fntGetGothicFamily ()
  public static String fntGetGothicFamily () {
    if (fntGothicFamily == null) {
      fntGothicFamily = "Monospaced";
      //使えそうなフォントを探す
      String[] availableFamilies = fntGetAvailableFamilies ();
      for (String family : FNT_GOTHIC_FAMILIES) {
        if (0 <= Arrays.binarySearch (availableFamilies, family, null)) {
          fntGothicFamily = family;
          break;
        }
      }
      System.out.println ((Multilingual.mlnJapanese ? "ゴシック体のフォント: " : "Gothic font: ") + fntGothicFamily);
    }
    return fntGothicFamily;
  }  //fntGetGothicFamily()

  //fntGetMinchoFamily ()
  public static String fntGetMinchoFamily () {
    if (fntMinchoFamily == null) {
      fntMinchoFamily = "Monospaced";
      //使えそうなフォントを探す
      String[] availableFamilies = fntGetAvailableFamilies ();
      for (String family : FNT_MINCHO_FAMILIES) {
        if (0 <= Arrays.binarySearch (availableFamilies, family, null)) {
          fntMinchoFamily = family;
          break;
        }
      }
      System.out.println ((Multilingual.mlnJapanese ? "明朝体のフォント: " : "Mincho font: ") + fntMinchoFamily);
    }
    return fntMinchoFamily;
  }  //fntGetMinchoFamily()



  //fntCreateCGROM1 ()
  //  CGROM1が最初にアクセスされたときフォントを作る
  public static void fntCreateCGROM1 () {
    if (!fntPageZen16x16.fnpReady) {
      fntPageZen16x16.fnpCreateImage (fntGetGothicFamily ());
    }
    if (!fntPageYon8x8.fnpReady) {
      fntPageYon8x8.fnpCreateImage (fntGetGothicFamily ());
    }
    if (!fntPageHan8x16.fnpReady) {
      fntPageHan8x16.fnpCreateImage (fntGetGothicFamily ());
    }
    if (!fntPageYon12x12.fnpReady) {
      fntPageYon12x12.fnpCreateImage (fntGetGothicFamily ());
    }
    if (!fntPageHan12x24.fnpReady) {
      fntPageHan12x24.fnpCreateImage (fntGetMinchoFamily ());
    }
    XEiJ.busSuper (MemoryMappedDevice.MMD_ROM, 0x00f00000, 0x00f40000);
  }  //fntCreateCGROM1

  //fntCreateCGROM2 ()
  //  CGROM2が最初にアクセスされたときフォントを作る
  public static void fntCreateCGROM2 () {
    if (!fntPageZen24x24.fnpReady) {
      fntPageZen24x24.fnpCreateImage (fntGetMinchoFamily ());
    }
    if (!fntPageHan6x12.fnpReady) {
      fntPageHan6x12.fnpCreateImage (fntGetGothicFamily ());
    }
    XEiJ.busSuper (MemoryMappedDevice.MMD_ROM, 0x00f40000, 0x00fc0000);
  }  //fntCreateCGROM2



  //編集モード
  public static final int FNT_MODE_BROWSE    = 0;
  public static final int FNT_MODE_PENCIL    = 1;
  public static final int FNT_MODE_ERASER    = 2;
  public static final int FNT_MODE_INVERTER  = 3;

  //ウインドウ
  public static DrawingCanvas fntCanvas;
  public static JTextField fntStatusTextField;
  public static JFrame fntFrame;
  public static boolean fntUseMouseWheel;
  public static boolean fntReverseMouseWheel;
  public static int fntScaleShift;
  public static JRadioButtonMenuItem[] fntScaleMenuItem;
  public static int fntMode;  //編集モード
  public static int fntLastX;  //最後に編集した座標
  public static int fntLastY;



  //fntStart ()
  public static void fntStart () {
    if (RestorableFrame.rfmGetOpened (Settings.SGS_FNT_FRAME_KEY)) {
      fntOpen ();
    }
  }  //fntStart()



  //fntOpen ()
  //  ウインドウを開く
  public static void fntOpen () {
    if (fntFrame == null) {
      fntMakeFrame ();
    }
    fntFrame.setVisible (true);
  }  //fntOpen()

  //fntMakeFrame ()
  //  ウインドウを作る
  //  ここでは開かない
  public static void fntMakeFrame () {

    fntUseMouseWheel = true;
    fntReverseMouseWheel = false;
    fntScaleShift = 0;
    fntMode = FNT_MODE_BROWSE;
    fntLastX = -1;
    fntLastY = -1;

    //ステータステキストフィールド
    fntStatusTextField = ComponentFactory.setEditable (
      ComponentFactory.createTextField ("", 20),
      false);

    //キャンバス
    fntCanvas = new DrawingCanvas (fntSelectedPage.fnpImageObject);
    fntCanvas.setMatColor (new Color (LnF.LNF_RGB[10]));
    fntCanvas.setCenterPoint (new Point (0, 0));
    ComponentFactory.addListener (
      fntCanvas,
      new MouseAdapter () {
        @Override public void mousePressed (MouseEvent me) {
          int x = me.getX ();
          int y = me.getY ();
          if (fntMode == FNT_MODE_PENCIL ||
              fntMode == FNT_MODE_ERASER ||
              fntMode == FNT_MODE_INVERTER) {
            fntLastX = x;
            fntLastY = y;
            fntSelectedPage.fnpEditPixel (x, y, fntMode);
            fntCanvas.repaint ();
            me.consume ();
          }
        }
        @Override public void mouseReleased (MouseEvent me) {
          if (fntMode == FNT_MODE_PENCIL ||
              fntMode == FNT_MODE_ERASER ||
              fntMode == FNT_MODE_INVERTER) {
            fntLastX = -1;
            fntLastY = -1;
            me.consume ();
          }
        }
      });
    ComponentFactory.addListener (
      fntCanvas,
      new MouseMotionAdapter () {
        @Override public void mouseDragged (MouseEvent me) {
          int x = me.getX ();
          int y = me.getY ();
          if (fntMode == FNT_MODE_PENCIL ||
              fntMode == FNT_MODE_ERASER ||
              fntMode == FNT_MODE_INVERTER) {
            if (!(fntLastX == x && fntLastY == y)) {
              fntLastX = x;
              fntLastY = y;
              fntSelectedPage.fnpEditPixel (x, y, fntMode);
              fntCanvas.repaint ();
            }
            me.consume ();
          }
        }
        @Override public void mouseMoved (MouseEvent me) {
          int x = me.getX ();
          int y = me.getY ();
          if (!me.isShiftDown ()) {  //Shiftキーが押されていない
            //ステータステキストを表示する
            fntStatusTextField.setText (fntSelectedPage.fnpGetStatusText (x, y));
            fntStatusTextField.repaint ();
          }
        }
      });
    ComponentFactory.addListener (
      fntCanvas,
      new MouseWheelListener () {
        @Override public void mouseWheelMoved (MouseWheelEvent mwe) {
          if (fntUseMouseWheel) {
            MouseWheelEvent2D mwe2D = (MouseWheelEvent2D) mwe;
            int n = mwe2D.getWheelRotation ();
            if (fntReverseMouseWheel) {
              n = -n;
            }
            n = n < 0 ? Math.min (4, fntScaleShift + 1) : Math.max (-4, fntScaleShift - 1);
            if (fntScaleShift != n) {
              fntScaleShift = n;
              fntCanvas.setScaleShift (n, mwe2D.getPoint2D ());
              fntScaleMenuItem[4 + n].setSelected (true);
            }
          }
          mwe.consume ();
        }
      });

    //アクションリスナー
    ActionListener listener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        Object source = ae.getSource ();
        String command = ae.getActionCommand ();
        switch (command) {
          //ファイル
        case "Load font data file":
          fntSelectedPage.fnpOpenLoadFontDataFileChooser ();
          break;
        case "Save font data file":
          fntSelectedPage.fnpOpenSaveFontDataFileChooser ();
          break;
        case "Read font image file":
          fntSelectedPage.fnpOpenReadFontImageFileChooser ();
          break;
        case "Write font image file":
          fntSelectedPage.fnpOpenWriteFontImageFileChooser ();
          break;
        case "Close":
          fntFrame.setVisible (false);
          break;
          //ページ
        case FNT_EN_Yon4x4:
          fntSelectPage (fntPageYon4x4);
          break;
        case FNT_EN_Yon6x6:
          fntSelectPage (fntPageYon6x6);
          break;
        case FNT_EN_Yon8x8:
          fntSelectPage (fntPageYon8x8);
          break;
        case FNT_EN_Yon12x12:
          fntSelectPage (fntPageYon12x12);
          break;
        case FNT_EN_Yon16x16:
          fntSelectPage (fntPageYon16x16);
          break;
        case FNT_EN_Yon24x24:
          fntSelectPage (fntPageYon24x24);
          break;
        case FNT_EN_Han4x8:
          fntSelectPage (fntPageHan4x8);
          break;
        case FNT_EN_Han6x12:
          fntSelectPage (fntPageHan6x12);
          break;
        case FNT_EN_Han8x16:
          fntSelectPage (fntPageHan8x16);
          break;
        case FNT_EN_Han12x24:
          fntSelectPage (fntPageHan12x24);
          break;
        case FNT_EN_Han16x32:
          fntSelectPage (fntPageHan16x32);
          break;
        case FNT_EN_Han24x48:
          fntSelectPage (fntPageHan24x48);
          break;
        case FNT_EN_Zen8x8:
          fntSelectPage (fntPageZen8x8);
          break;
        case FNT_EN_Zen12x12:
          fntSelectPage (fntPageZen12x12);
          break;
        case FNT_EN_Zen16x16:
          fntSelectPage (fntPageZen16x16);
          break;
        case FNT_EN_Zen24x24:
          fntSelectPage (fntPageZen24x24);
          break;
        case FNT_EN_Zen32x32:
          fntSelectPage (fntPageZen32x32);
          break;
        case FNT_EN_Zen48x48:
          fntSelectPage (fntPageZen48x48);
          break;
        case FNT_EN_Pic36x46:
          fntSelectPage (fntPagePic36x46);
          break;
        case FNT_EN_Eli30x46:
          fntSelectPage (fntPageEli30x46);
          break;
        case FNT_EN_Sma18x46:
          fntSelectPage (fntPageSma18x46);
          break;
        case FNT_EN_Scr28x32:
          fntSelectPage (fntPageScr28x32);
          break;
        case FNT_EN_Pan24x48:
          fntSelectPage (fntPagePan24x48);
          break;
        case FNT_EN_Lcd4x6:
          fntSelectPage (fntPageLcd4x6);
          break;
        case FNT_EN_Lcd6x8:
          fntSelectPage (fntPageLcd6x8);
          break;
        case FNT_EN_Smk6x8:
          fntSelectPage (fntPageSmk6x8);
          break;
          //自動生成
/*
          //変換
          //  → ANK 1/4角 8x8
        case "from " + FNT_EN_Half8x16 + " to " + FNT_EN_Yon8x8:
          fntConvertHalf8x16toYon8x8 ();
          break;
        case "from " + FNT_EN_Full8x8 + " to " + FNT_EN_Yon8x8:
          fntConvertFull8x8toYon8x8 ();
          break;
          //  → ANK 1/4角 12x12
        case "from " + FNT_EN_Yon8x8 + " to " + FNT_EN_Yon12x12:
          fntConvertYon8x8toYon12x12 ();
          break;
        case "from " + FNT_EN_Half6x12 + " to " + FNT_EN_Yon12x12:
          fntConvertHalf6x12toYon12x12 ();
          break;
        case "from " + FNT_EN_Full12x12 + " to " + FNT_EN_Yon12x12:
          fntConvertFull12x12toYon12x12 ();
          break;
          //  → ANK 1/4角 16x16
        case "from " + FNT_EN_Full16x16 + " to " + FNT_EN_Yon16x16:
          fntConvertFull16x16toYon16x16 ();
          break;
          //  → ANK 1/4角 24x24
        case "from " + FNT_EN_Full24x24 + " to " + FNT_EN_Yon24x24:
          fntConvertFull24x24toYon24x24 ();
          break;
          //  → ANK 半角 6x12
        case "from " + FNT_EN_Half8x16 + " to " + FNT_EN_Half6x12 + " (SX-Window)":
          fntConvertHalf8x16toHalf6x12FSX ();
          break;
          //  → ANK 半角 8x16
        case "from " + FNT_EN_Yon8x8 + " to " + FNT_EN_Half8x16:
          fntConvertYon8x8toHalf8x16 ();
          break;
        case "from " + FNT_EN_Half6x12 + " to " + FNT_EN_Half8x16:
          fntConvertHalf6x12toHalf8x16 ();
          break;
        case "from " + FNT_EN_Full16x16 + " to " + FNT_EN_Half8x16:
          fntConvertFull16x16toHalf8x16 ();
          break;
          //  → ANK 半角 12x24
        case "from " + FNT_EN_Full24x24 + " to " + FNT_EN_Half12x24:
          fntConvertFull24x24toHalf12x24 ();
          break;
          //  → ANK 半角 16x32
          //  → ANK 半角 24x48
          //  → 漢字 全角 8x8
          //  → 漢字 全角 12x12
        case "from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (SX-Window)":
          fntConvertFull16x16toFull12x12FSX ();
          break;
        case "from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (Test1)":
          fntConvertFull16x16toFull12x12Test1 ();
          break;
        case "from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (Test2)":
          fntConvertFull16x16toFull12x12Test2 ();
          break;
        case "from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (Test3)":
          fntConvertFull16x16toFull12x12Test3 ();
          break;
        case "from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (Test4)":
          fntConvertFull16x16toFull12x12Test4 ();
          break;
        case "from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (Test5)":
          fntConvertFull16x16toFull12x12Test5 ();
          break;
        case "from " + FNT_EN_Full24x24 + " to " + FNT_EN_Full12x12 + " (IOCS)":
          fntConvertFull24x24toFull12x12IOCS ();
          break;
          //  → 漢字 全角 16x16
          //  → 漢字 全角 24x24
          //  → 漢字 全角 32x32
          //  → 漢字 全角 48x48
          //
*/
          //編集
        case "Browse":
          fntMode = FNT_MODE_BROWSE;
          break;
        case "Pencil":
          fntMode = FNT_MODE_PENCIL;
          break;
        case "Eraser":
          fntMode = FNT_MODE_ERASER;
          break;
        case "Inverter":
          fntMode = FNT_MODE_INVERTER;
          break;
        case "Zero clear":
          fntSelectedPage.fnpZeroClear ();
          break;
          //拡大縮小
        case "6.25%":
          fntCanvas.setScaleShift (-4);
          break;
        case "12.5%":
          fntCanvas.setScaleShift (-3);
          break;
        case "25%":
          fntCanvas.setScaleShift (-2);
          break;
        case "50%":
          fntCanvas.setScaleShift (-1);
          break;
        case "100%":
          fntCanvas.setScaleShift (0);
          break;
        case "200%":
          fntCanvas.setScaleShift (1);
          break;
        case "400%":
          fntCanvas.setScaleShift (2);
          break;
        case "800%":
          fntCanvas.setScaleShift (3);
          break;
        case "1600%":
          fntCanvas.setScaleShift (4);
          break;
        case "Use mouse wheel":
          fntUseMouseWheel = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        case "Reverse mouse wheel":
          fntReverseMouseWheel = ((JCheckBoxMenuItem) source).isSelected ();
          break;
        }
      }  //actionPerformed
    };  //new ActionListener

    //ホストマシンのフォント
    Font[] hostFonts = GraphicsEnvironment.getLocalGraphicsEnvironment ().getAllFonts ();
    String[] hostFontNames = new String[hostFonts.length];
    for (int i = 0; i < hostFonts.length; i++) {
      hostFontNames[i] = hostFonts[i].getName ();
    }
    Arrays.sort (hostFontNames, new Comparator<String> () {
      @Override public int compare (String o1, String o2) {
        return o1.toUpperCase ().compareTo (o2.toUpperCase ());
      }
    });
    ActionListener fontListener = new ActionListener () {
      @Override public void actionPerformed (ActionEvent ae) {
        fntSelectedPage.fnpCreateImage (ae.getActionCommand ());
        fntCanvas.repaint ();
      }
    };
    JMenu hostFontMenu = Multilingual.mlnText (ComponentFactory.createMenu ("Fonts on host machine", 'H'), "ja", "ホストマシンのフォント");
    //フォント名の先頭の文字が同じフォントが多数あるのでメニューを先頭1文字と先頭2文字の2段階にする
    //!!! 2段階でも足りなくなる可能性がある。先頭の文字の偏り方によって深さを可変にしたい
    String h1 = "";  //前回の先頭1文字
    String h2 = "";  //前回の先頭2文字
    JMenu hostFontSubMenu1 = null;
    JMenu hostFontSubMenu2 = null;
    for (int i = 0; i < hostFontNames.length; i++) {
      String name = hostFontNames[i];
      if (name == null || name.length () == 0) {  //念のため
        continue;
      }
      String c1 = name.substring (0, 1).toUpperCase ();  //今回の先頭1文字
      if (!h1.equalsIgnoreCase (c1)) {  //先頭1文字が前回と異なる
        h1 = c1;
        h2 = "";
        hostFontSubMenu1 = ComponentFactory.createMenu (c1);
        hostFontSubMenu2 = null;
        hostFontMenu.add (hostFontSubMenu1);
      }
      if (name.length () == 1) {  //1文字だけ
        hostFontSubMenu1.add (ComponentFactory.createMenuItem (name, fontListener));
      } else {  //2文字以上ある
        String c2 = name.substring (0, 2).toUpperCase ();  //今回の先頭2文字
        if (!h2.equalsIgnoreCase (c2)) {  //先頭2文字が前回と異なる
          h2 = c2;
          hostFontSubMenu2 = ComponentFactory.createMenu (c2);
          hostFontSubMenu1.add (hostFontSubMenu2);
        }
        hostFontSubMenu2.add (ComponentFactory.createMenuItem (name, fontListener));
      }
    }

    //メニューバー
    ButtonGroup setGroup = new ButtonGroup ();
    ButtonGroup editGroup = new ButtonGroup ();
    ButtonGroup zoomGroup = new ButtonGroup ();
    fntScaleMenuItem = new JRadioButtonMenuItem[9];
    JMenuBar fntMenuBar = ComponentFactory.createMenuBar (
      Multilingual.mlnText (
        ComponentFactory.createMenu (
          "File", 'F',
          Multilingual.mlnText (ComponentFactory.createMenuItem ("Load font data file" , 'L', listener),
                                "ja", "フォントデータファイルを読み込む"),
          Multilingual.mlnText (ComponentFactory.createMenuItem ("Save font data file" , 'S', listener),
                                "ja", "フォントデータファイルに書き出す"),
          ComponentFactory.createHorizontalSeparator (),
          Multilingual.mlnText (ComponentFactory.createMenuItem ("Read font image file" , 'R', listener),
                                "ja", "フォント画像ファイルを読み込む"),
          Multilingual.mlnText (ComponentFactory.createMenuItem ("Write font image file" , 'W', listener),
                                "ja", "フォント画像ファイルに書き出す"),
          ComponentFactory.createHorizontalSeparator (),
          Multilingual.mlnText (ComponentFactory.createMenuItem ("Zero clear" , 'Z', listener), "ja", "ゼロクリア"),
          ComponentFactory.createHorizontalSeparator (),
          Multilingual.mlnText (ComponentFactory.createMenuItem ("Close", 'C', listener), "ja", "閉じる")
          ),
        "ja", "ファイル"),
      Multilingual.mlnText (
        ComponentFactory.createMenu (
          "Page", 'P',
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageYon4x4,   FNT_EN_Yon4x4,        listener), "ja", FNT_JA_Yon4x4),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageYon6x6,   FNT_EN_Yon6x6,        listener), "ja", FNT_JA_Yon6x6),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageYon8x8,   FNT_EN_Yon8x8,   '1', listener), "ja", FNT_JA_Yon8x8),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageYon12x12, FNT_EN_Yon12x12, '2', listener), "ja", FNT_JA_Yon12x12),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageYon16x16, FNT_EN_Yon16x16,      listener), "ja", FNT_JA_Yon16x16),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageYon24x24, FNT_EN_Yon24x24,      listener), "ja", FNT_JA_Yon24x24),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageHan4x8,   FNT_EN_Han4x8,        listener), "ja", FNT_JA_Han4x8),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageHan6x12,  FNT_EN_Han6x12,  '3', listener), "ja", FNT_JA_Han6x12),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageHan8x16,  FNT_EN_Han8x16,  '4', listener), "ja", FNT_JA_Han8x16),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageHan12x24, FNT_EN_Han12x24, '5', listener), "ja", FNT_JA_Han12x24),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageHan16x32, FNT_EN_Han16x32,      listener), "ja", FNT_JA_Han16x32),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageHan24x48, FNT_EN_Han24x48,      listener), "ja", FNT_JA_Han24x48),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageZen8x8,   FNT_EN_Zen8x8,        listener), "ja", FNT_JA_Zen8x8),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageZen12x12, FNT_EN_Zen12x12,      listener), "ja", FNT_JA_Zen12x12),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageZen16x16, FNT_EN_Zen16x16, '6', listener), "ja", FNT_JA_Zen16x16),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageZen24x24, FNT_EN_Zen24x24, '7', listener), "ja", FNT_JA_Zen24x24),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageZen32x32, FNT_EN_Zen32x32,      listener), "ja", FNT_JA_Zen32x32),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageZen48x48, FNT_EN_Zen48x48,      listener), "ja", FNT_JA_Zen48x48),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPagePic36x46, FNT_EN_Pic36x46,      listener), "ja", FNT_JA_Pic36x46),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageEli30x46, FNT_EN_Eli30x46,      listener), "ja", FNT_JA_Eli30x46),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageSma18x46, FNT_EN_Sma18x46,      listener), "ja", FNT_JA_Sma18x46),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageScr28x32, FNT_EN_Scr28x32,      listener), "ja", FNT_JA_Scr28x32),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPagePan24x48, FNT_EN_Pan24x48,      listener), "ja", FNT_JA_Pan24x48),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageLcd4x6,   FNT_EN_Lcd4x6,        listener), "ja", FNT_JA_Lcd4x6),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageLcd6x8,   FNT_EN_Lcd6x8,        listener), "ja", FNT_JA_Lcd6x8),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            setGroup, fntSelectedPage == fntPageSmk6x8,   FNT_EN_Smk6x8,        listener), "ja", FNT_JA_Smk6x8)
          ),
        "ja", "ページ"),
      Multilingual.mlnText (
        ComponentFactory.createMenu (
          "Automatic creation", 'A',
          hostFontMenu//,
/*
          Multilingual.mlnText (
            ComponentFactory.createMenu (
              "Convert from another page", 'C',
              //  → ANK 1/4角 8x8
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Half8x16 + " to " + FNT_EN_Yon8x8, listener),
                                    "ja", FNT_JA_Half8x16 + " から " + FNT_JA_Yon8x8 + " へ"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full8x8 + " to " + FNT_EN_Yon8x8, listener),
                                    "ja", FNT_JA_Full8x8 + " から " + FNT_JA_Yon8x8 + " へ"),
              //  → ANK 1/4角 12x12
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Yon8x8 + " to " + FNT_EN_Yon12x12, listener),
                                    "ja", FNT_JA_Yon8x8 + " から " + FNT_JA_Yon12x12 + " へ"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Half6x12 + " to " + FNT_EN_Yon12x12, listener),
                                    "ja", FNT_JA_Half6x12 + " から " + FNT_JA_Yon12x12 + " へ"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full12x12 + " to " + FNT_EN_Yon12x12, listener),
                                    "ja", FNT_JA_Full12x12 + " から " + FNT_JA_Yon12x12 + " へ"),
              //  → ANK 1/4角 16x16
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full16x16 + " to " + FNT_EN_Yon16x16, listener),
                                    "ja", FNT_JA_Full16x16 + " から " + FNT_JA_Yon16x16 + " へ"),
              //  → ANK 1/4角 24x24
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full24x24 + " to " + FNT_EN_Yon24x24, listener),
                                    "ja", FNT_JA_Full24x24 + " から " + FNT_JA_Yon24x24 + " へ"),
              //  → ANK 半角 6x12
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Half8x16 + " to " + FNT_EN_Half6x12 + " (SX-Window)", listener),
                                    "ja", FNT_JA_Half8x16 + " から " + FNT_JA_Half6x12 + " へ (SX-Window)"),
              //  → ANK 半角 8x16
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Yon8x8 + " to " + FNT_EN_Half8x16, listener),
                                    "ja", FNT_JA_Yon8x8 + " から " + FNT_JA_Half8x16 + " へ"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Half6x12 + " to " + FNT_EN_Half8x16, listener),
                                    "ja", FNT_JA_Half6x12 + " から " + FNT_JA_Half8x16 + " へ"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full16x16 + " to " + FNT_EN_Half8x16, listener),
                                    "ja", FNT_JA_Full16x16 + " から " + FNT_JA_Half8x16 + " へ"),
              //  → ANK 半角 12x24
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full24x24 + " to " + FNT_EN_Half12x24, listener),
                                    "ja", FNT_JA_Full24x24 + " から " + FNT_JA_Half12x24 + " へ"),
              //  → ANK 半角 16x32
              //  → ANK 半角 24x48
              //  → 漢字 全角 8x8
              //  → 漢字 全角 12x12
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (SX-Window)", listener),
                                    "ja", FNT_JA_Full16x16 + " から " + FNT_JA_Full12x12 + " へ (SX-Window)"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (Test1)", listener),
                                    "ja", FNT_JA_Full16x16 + " から " + FNT_JA_Full12x12 + " へ (Test1)"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (Test2)", listener),
                                    "ja", FNT_JA_Full16x16 + " から " + FNT_JA_Full12x12 + " へ (Test2)"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (Test3)", listener),
                                    "ja", FNT_JA_Full16x16 + " から " + FNT_JA_Full12x12 + " へ (Test3)"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (Test4)", listener),
                                    "ja", FNT_JA_Full16x16 + " から " + FNT_JA_Full12x12 + " へ (Test4)"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full16x16 + " to " + FNT_EN_Full12x12 + " (Test5)", listener),
                                    "ja", FNT_JA_Full16x16 + " から " + FNT_JA_Full12x12 + " へ (Test5)"),
              Multilingual.mlnText (ComponentFactory.createMenuItem ("from " + FNT_EN_Full24x24 + " to " + FNT_EN_Full12x12 + " (IOCS)", listener),
                                    "ja", FNT_JA_Full24x24 + " から " + FNT_JA_Full12x12 + " へ (IOCS)")
              ),
            "ja", "他のページのフォントから変換する")
*/
          ),
        "ja", "自動生成"),
      Multilingual.mlnText (
        ComponentFactory.createMenu (
          "Edit", 'E',
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            editGroup, fntMode == FNT_MODE_BROWSE,    "Browse",  'B', listener), "ja", "閲覧"),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            editGroup, fntMode == FNT_MODE_PENCIL,    "Pencil",  'P', listener), "ja", "鉛筆"),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            editGroup, fntMode == FNT_MODE_ERASER,    "Eraser",  'E', listener), "ja", "消しゴム"),
          Multilingual.mlnText (ComponentFactory.createRadioButtonMenuItem (
            editGroup, fntMode == FNT_MODE_INVERTER, "Inverter", 'I', listener), "ja", "白黒反転")
          ),
        "ja", "編集"),
      Multilingual.mlnText (
        ComponentFactory.createMenu (
          "Zoom in and out", 'Z',
          fntScaleMenuItem[0] = ComponentFactory.createRadioButtonMenuItem (zoomGroup, fntScaleShift == -4, "6.25%", '1', listener),
          fntScaleMenuItem[1] = ComponentFactory.createRadioButtonMenuItem (zoomGroup, fntScaleShift == -3, "12.5%", '2', listener),
          fntScaleMenuItem[2] = ComponentFactory.createRadioButtonMenuItem (zoomGroup, fntScaleShift == -2, "25%"  , '3', listener),
          fntScaleMenuItem[3] = ComponentFactory.createRadioButtonMenuItem (zoomGroup, fntScaleShift == -1, "50%"  , '4', listener),
          fntScaleMenuItem[4] = ComponentFactory.createRadioButtonMenuItem (zoomGroup, fntScaleShift ==  0, "100%" , '5', listener),
          fntScaleMenuItem[5] = ComponentFactory.createRadioButtonMenuItem (zoomGroup, fntScaleShift ==  1, "200%" , '6', listener),
          fntScaleMenuItem[6] = ComponentFactory.createRadioButtonMenuItem (zoomGroup, fntScaleShift ==  2, "400%" , '7', listener),
          fntScaleMenuItem[7] = ComponentFactory.createRadioButtonMenuItem (zoomGroup, fntScaleShift ==  3, "800%" , '8', listener),
          fntScaleMenuItem[8] = ComponentFactory.createRadioButtonMenuItem (zoomGroup, fntScaleShift ==  4, "1600%", '9', listener),
          ComponentFactory.createHorizontalSeparator (),
          Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (
            fntUseMouseWheel    , "Use mouse wheel"    , listener), "ja", "マウスホイールを使う"),
          Multilingual.mlnText (ComponentFactory.createCheckBoxMenuItem (
            fntReverseMouseWheel, "Reverse mouse wheel", listener), "ja", "マウスホイールの向きを逆にする")
          ),
        "ja", "拡大縮小")
      );

    //ウインドウ
    fntFrame = Multilingual.mlnTitle (
      ComponentFactory.createRestorableSubFrame (
        Settings.SGS_FNT_FRAME_KEY,
        "Font Editor",
        fntMenuBar,
        ComponentFactory.createBorderPanel (
          ComponentFactory.setPreferredSize (fntCanvas, 800, 600),
          null,
          null,
          fntStatusTextField
          )
        ),
      "ja", "フォントエディタ");

  }  //fntMakeFrame

  //fntSelectPage (page)
  //  ページを選択する
  public static void fntSelectPage (FontPage page) {
    fntSelectedPage = page;
    if (fntCanvas != null) {
      fntCanvas.setImage (page.fnpImageObject);
    }
  }  //fntSelectPage(FontPage)



/*



  //loaded = fntLoadFontDataFileSub (names, requested)
  //  フォントファイルを読み込む
  //  loaded     読み込んだフォントの種類のマスク。一度に複数の種類のフォントを読み込む場合がある
  //  name       フォントファイル名。','で区切って複数のファイルを指定できる
  //             1つのファイルに複数の種類のフォントが含まれている場合がある
  //  requested  読み込むフォントの種類のマスク。一度に複数の種類を指定できる
  //             指定されなかったフォントはファイルに含まれていても読み込まない
  //             指定されたフォントを読み込めなくてもエラーにならない
  public static int fntLoadFontDataFileSub (String names, int requested) {
    int loaded = 0;
    byte[] bb = new byte[Math.max (786432, 3 * 48 * 256 + 6 * 48 * 94 * 94)];
    for (String name : names.split (",")) {
      name = name.trim ();
      if (name.length () == 0) {
        continue;
      }
      String upperName = name.toUpperCase ();
      int guess = (upperName.startsWith ("CGROM") ? FNT_MASK_CGROM :
                   upperName.endsWith (".F8") || upperName.endsWith (".F8.GZ") ? FNT_MASK_Half4x8 | FNT_MASK_Full8x8 :
                   upperName.endsWith (".F12") || upperName.endsWith (".F12.GZ") ? FNT_MASK_Half6x12 | FNT_MASK_Full12x12 :
                   upperName.endsWith (".FON") || upperName.endsWith (".FON.GZ") ? FNT_MASK_Half8x16 | FNT_MASK_Full16x16 :
                   upperName.endsWith (".F24") || upperName.endsWith (".F24.GZ") ? FNT_MASK_Half12x24 | FNT_MASK_Full24x24 :
                   upperName.endsWith (".F32") || upperName.endsWith (".F32.GZ") ? FNT_MASK_Half16x32 | FNT_MASK_Full32x32 :
                   upperName.endsWith (".F48") || upperName.endsWith (".F48.GZ") ? FNT_MASK_Half24x48 | FNT_MASK_Full48x48 :
                   FNT_MASK_ALL);  //このファイルに含まれている可能性のあるフォントの種類
      if ((requested & guess) == 0) {  //このファイルに含まれている可能性のある種類のフォントは不要
        continue;
      }
      //読み込む
      InputStream in = XEiJ.ismOpen (name);
      if (in == null) {  //読み込めない
        continue;
      }
      int length = XEiJ.ismRead (
        in, bb, 0,
        guess == (FNT_MASK_Half4x8 | FNT_MASK_Full8x8) ? 1 * 8 * 256 + 1 * 8 * 94 * 94 :  //2048+70688=72736  Half4x8+Full8x8
        guess == (FNT_MASK_Half6x12 | FNT_MASK_Full12x12) ? 1 * 12 * 256 + 2 * 12 * 94 * 94 :  //3072+212064=215136  Half6x12+Full12x12
        guess == (FNT_MASK_Half8x16 | FNT_MASK_Full16x16) ? 1 * 16 * 256 + 2 * 16 * 94 * 94 :  //4096+282752=286848  Half8x16+Full16x16
        guess == (FNT_MASK_Half12x24 | FNT_MASK_Full24x24) ? 2 * 24 * 256 + 3 * 24 * 94 * 94 :  //12288+636192=648480  Half12x24+Full24x24
        guess == (FNT_MASK_Half16x32 | FNT_MASK_Full32x32) ? 2 * 32 * 256 + 4 * 32 * 94 * 94 :  //16384+1131008=1147392  Half16x32+Full32x32
        guess == (FNT_MASK_Half24x48 | FNT_MASK_Full48x48) ? 3 * 48 * 256 + 6 * 48 * 94 * 94 :  //36864+2544768=2581632  Half24x48+Full48x48
        bb.length);  //実際に読み込んだ長さ
      XEiJ.ismClose (in);
      //読み込んだ長さで改めてフォントの種類を判断する
      if (length == 786432) {  //CGROM.DAT
        if ((requested & FNT_MASK_CGROM) != 0) {
          System.arraycopy (bb, 0, MainMemory.mmrM8, 0x00f00000, 786432);
          requested &= ~(FNT_MASK_CGROM & ~FNT_MASK_Half6x12);
          loaded |= FNT_MASK_CGROM & ~FNT_MASK_Half6x12;
          fntPageFull16x16.fnpMemoryToImage ();
          fntPageYon8x8.fnpMemoryToImage ();
          fntPageHalf8x16.fnpMemoryToImage ();
          fntPageYon12x12.fnpMemoryToImage ();
          fntPageHalf12x24.fnpMemoryToImage ();
          fntPageFull24x24.fnpMemoryToImage ();
          if (MainMemory.mmrM8[FNT_ADDRESS_Half6x12 + 12 * 'A' + 6] != 0) {  //Half6x12が含まれている
            //ROM1.0～ROM1.2のときはCGROMに読み込んだHalf6x12をIPLROMにコピーする
            if (fntHalf6x12Address != 0) {
              System.arraycopy (MainMemory.mmrM8, FNT_ADDRESS_Half6x12, MainMemory.mmrM8, fntHalf6x12Address, 1 * 12 * 254);
            }
            requested &= ~FNT_MASK_Half6x12;
            loaded |= FNT_MASK_Half6x12;
            fntPageHalf6x12.fnpMemoryToImage ();
          }
        }
      } else if (length == 1 * 8 * 256 + 1 * 8 * 94 * 94) {  //Half4x8+Full8x8
        if ((requested & FNT_MASK_Half4x8) != 0) {
          System.arraycopy (bb, 0, fntPageHalf4x8.fnpBitmapArray, 0, 1 * 8 * 256);
          requested &= ~FNT_MASK_Half4x8;
          loaded |= FNT_MASK_Half4x8;
          fntPageHalf4x8.fnpMemoryToImage ();
        }
        if ((requested & FNT_MASK_Full8x8) != 0) {
          System.arraycopy (bb, 1 * 8 * 256, fntPageFull8x8.fnpBitmapArray, 0, 1 * 8 * 94 * 8);  //1区～8区
          System.arraycopy (bb, 1 * 8 * 256 + 1 * 8 * 94 * 15, fntPageFull8x8.fnpBitmapArray, 1 * 8 * 94 * 8, 1 * 8 * 94 * 69);  //16区～84区
          requested &= ~FNT_MASK_Full8x8;
          loaded |= FNT_MASK_Full8x8;
          fntPageFull8x8.fnpMemoryToImage ();
        }
      } else if (length == 1 * 12 * 256 + 2 * 12 * 94 * 94) {  //Half6x12+Full12x12
        if ((requested & FNT_MASK_Half6x12) != 0) {
          System.arraycopy (bb, 0, MainMemory.mmrM8, FNT_ADDRESS_Half6x12, 1 * 12 * 256);
          requested &= ~FNT_MASK_Half6x12;
          loaded |= FNT_MASK_Half6x12;
          fntPageHalf6x12.fnpMemoryToImage ();
        }
        if ((requested & FNT_MASK_Full12x12) != 0) {
          System.arraycopy (bb, 1 * 12 * 256, fntPageFull12x12.fnpBitmapArray, 0, 2 * 12 * 94 * 8);  //1区～8区
          System.arraycopy (bb, 1 * 12 * 256 + 2 * 12 * 94 * 15, fntPageFull12x12.fnpBitmapArray, 2 * 12 * 94 * 8, 2 * 12 * 94 * 69);  //16区～84区
          requested &= ~FNT_MASK_Full12x12;
          loaded |= FNT_MASK_Full12x12;
          fntPageFull12x12.fnpMemoryToImage ();
        }
      } else if (length == 1 * 16 * 256 + 2 * 16 * 94 * 94) {  //Half8x16+Full16x16
        if ((requested & FNT_MASK_Half8x16) != 0) {
          System.arraycopy (bb, 0, MainMemory.mmrM8, FNT_ADDRESS_Half8x16, 1 * 16 * 256);
          requested &= ~FNT_MASK_Half8x16;
          loaded |= FNT_MASK_Half8x16;
          fntPageHalf8x16.fnpMemoryToImage ();
        }
        if ((requested & FNT_MASK_Full16x16) != 0) {
          System.arraycopy (bb, 1 * 16 * 256, MainMemory.mmrM8, FNT_ADDRESS_Full16x16, 2 * 16 * 94 * 8);  //1区～8区
          System.arraycopy (bb, 1 * 16 * 256 + 2 * 16 * 94 * 15, MainMemory.mmrM8, FNT_ADDRESS_Full16x16 + 2 * 16 * 94 * 8, 2 * 16 * 94 * 69);  //16区～84区
          requested &= ~FNT_MASK_Full16x16;
          loaded |= FNT_MASK_Full16x16;
          fntPageFull16x16.fnpMemoryToImage ();
        }
      } else if (length == 2 * 24 * 256 + 3 * 24 * 94 * 94) {  //Half12x24+Full24x24
        if ((requested & FNT_MASK_Half12x24) != 0) {
          System.arraycopy (bb, 0, MainMemory.mmrM8, FNT_ADDRESS_Half12x24, 2 * 24 * 256);
          requested &= ~FNT_MASK_Half12x24;
          loaded |= FNT_MASK_Half12x24;
          fntPageHalf12x24.fnpMemoryToImage ();
        }
        if ((requested & FNT_MASK_Full24x24) != 0) {
          System.arraycopy (bb, 2 * 24 * 256, MainMemory.mmrM8, FNT_ADDRESS_Full24x24, 3 * 24 * 94 * 8);  //1区～8区
          System.arraycopy (bb, 2 * 24 * 256 + 3 * 24 * 94 * 15, MainMemory.mmrM8, FNT_ADDRESS_Full24x24 + 3 * 24 * 94 * 8, 3 * 24 * 94 * 69);  //16区～84区
          requested &= ~FNT_MASK_Full24x24;
          loaded |= FNT_MASK_Full24x24;
          fntPageFull24x24.fnpMemoryToImage ();
        }
      } else if (length == 2 * 32 * 256 + 4 * 32 * 94 * 94) {  //Half16x32+Full32x32
        if ((requested & FNT_MASK_Half16x32) != 0) {
          System.arraycopy (bb, 0, fntPageHalf16x32.fnpBitmapArray, 0, 2 * 32 * 256);
          requested &= ~FNT_MASK_Half16x32;
          loaded |= FNT_MASK_Half16x32;
          fntPageHalf16x32.fnpMemoryToImage ();
        }
        if ((requested & FNT_MASK_Full32x32) != 0) {
          System.arraycopy (bb, 2 * 32 * 256, fntPageFull32x32.fnpBitmapArray, 0, 4 * 32 * 94 * 8);  //1区～8区
          System.arraycopy (bb, 2 * 32 * 256 + 4 * 32 * 94 * 15, fntPageFull32x32.fnpBitmapArray, 4 * 32 * 94 * 8, 4 * 32 * 94 * 69);  //16区～84区
          requested &= ~FNT_MASK_Full32x32;
          loaded |= FNT_MASK_Full32x32;
          fntPageFull32x32.fnpMemoryToImage ();
        }
      } else if (length == 3 * 48 * 256 + 6 * 48 * 94 * 94) {  //Half24x48+Full48x48
        if ((requested & FNT_MASK_Half24x48) != 0) {
          System.arraycopy (bb, 0, fntPageHalf24x48.fnpBitmapArray, 0, 3 * 48 * 256);
          requested &= ~FNT_MASK_Half24x48;
          loaded |= FNT_MASK_Half24x48;
          fntPageHalf24x48.fnpMemoryToImage ();
        }
        if ((requested & FNT_MASK_Full48x48) != 0) {
          System.arraycopy (bb, 3 * 48 * 256, fntPageFull48x48.fnpBitmapArray, 0, 6 * 48 * 94 * 8);  //1区～8区
          System.arraycopy (bb, 3 * 48 * 256 + 6 * 48 * 94 * 15, fntPageFull48x48.fnpBitmapArray, 6 * 48 * 94 * 8, 6 * 48 * 94 * 69);  //16区～84区
          requested &= ~FNT_MASK_Full48x48;
          loaded |= FNT_MASK_Full48x48;
          fntPageFull48x48.fnpMemoryToImage ();
        }
      } else {
        for (int k = 0; k < FNT_COUNT; k++) {
          FontPage page = fntPageArray[(fntSelectedPage.fnpId + k) % FNT_COUNT];  //選択されているページから探し始める
          if ((requested & page.fnpMask) != 0 && length == page.fnpDataSize) {
            System.arraycopy (bb, 0, page.fnpBitmapArray, 0, page.fnpDataSize);
            requested &= ~page.fnpMask;
            loaded |= page.fnpMask;
            page.fnpMemoryToImage ();
            break;
          }
        }
      }
      if (requested == 0) {
        break;
      }
    }  //for name
    return loaded;
  }  //fntLoadFontDataFileSub(String,int)

  //fntSaveFontDataFile ()
  //  フォントデータファイルを名前を指定して書き出す
  //  コマンドラインのみ
  public static void fntSaveFontDataFile () {
    String dataName = fntSelectedPage.fnpDataName;
    if (dataName == null) {
      return;
    }
    JFileChooser2 fileChooser = new JFileChooser2 (new File (dataName));
    fileChooser.setFileFilter (new javax.swing.filechooser.FileFilter () {  //java.io.FileFilterと紛らわしい
      @Override public boolean accept (File file) {
        String name = file.getName ();
        String upperName = name.toUpperCase ();
        return (file.isDirectory () ||
                (file.isFile () &&
                 (upperName.startsWith ("CGROM") ||
                  upperName.endsWith (".F8") ||
                  upperName.endsWith (".F12") ||
                  upperName.endsWith (".FON") ||
                  upperName.endsWith (".F24") ||
                  upperName.endsWith (".F32") ||
                  upperName.endsWith (".F48"))));
      }
      @Override public String getDescription () {
        return (Multilingual.mlnJapanese ?
                "フォントデータファイル (CGROM*.*,*.F8,*.F12,*.FON,*.F24,*.F32,*.F48)" :
                "Font data files (CGROM*.*,*.F8,*.F12,*.FON,*.F24,*.F32,*.F48)");
      }
    });
    if (fileChooser.showSaveDialog (null) == JFileChooser.APPROVE_OPTION) {
      File file = fileChooser.getSelectedFile ();
      String name = file.getPath ();
      //出力するページを決める
      //  *.F12のとき現在表示されているページがHalf6x12ならばFull12x12を出力しない
      //  *.FONのとき現在表示されているページがHalf8x16ならばFull16x16を出力しない
      //  *.F24のとき現在表示されているページがHalf12x24ならばFull24x24を出力しない
      //  *.F32のとき現在表示されているページがHalf16x32ならばFull32x32を出力しない
      //  *.F48のとき現在表示されているページがHalf24x48ならばFull48x48を出力しない
      String upperName = name.toUpperCase ();
      int mask = (
        upperName.startsWith ("CGROM") ? FNT_MASK_CGROM :
        upperName.endsWith (".F8") ? fntSelectedPage == fntPageHalf4x8 ? FNT_MASK_Half4x8 : FNT_MASK_Half4x8 | FNT_MASK_Full8x8 :
        upperName.endsWith (".F12") ? fntSelectedPage == fntPageHalf6x12 ? FNT_MASK_Half6x12 : FNT_MASK_Half6x12 | FNT_MASK_Full12x12 :
        upperName.endsWith (".FON") ? fntSelectedPage == fntPageHalf8x16 ? FNT_MASK_Half8x16 : FNT_MASK_Half8x16 | FNT_MASK_Full16x16 :
        upperName.endsWith (".F24") ? fntSelectedPage == fntPageHalf12x24 ? FNT_MASK_Half12x24 : FNT_MASK_Half12x24 | FNT_MASK_Full24x24 :
        upperName.endsWith (".F32") ? fntSelectedPage == fntPageHalf16x32 ? FNT_MASK_Half16x32 : FNT_MASK_Half16x32 | FNT_MASK_Full32x32 :
        upperName.endsWith (".F48") ? fntSelectedPage == fntPageHalf24x48 ? FNT_MASK_Half24x48 : FNT_MASK_Half24x48 | FNT_MASK_Full48x48 :
        FNT_MASK_CGROM);  //出力するページ
      File fileTmp = new File (name + ".tmp");
      File fileBak = new File (name + ".bak");
      try (OutputStream out = new BufferedOutputStream (new FileOutputStream (fileTmp))) {  //try-with-resourcesは1.7から
        if (mask == FNT_MASK_CGROM) {
          out.write (MainMemory.mmrM8, 0x00f00000, 0x00fc0000 - 0x00f00000);
        } else if (mask == FNT_MASK_Yon4x4) {  //ANK 1/4角 4x4
          out.write (fntPageYon4x4.fnpBitmapArray, 0, 1 * 4 * 16 * 16);
        } else if (mask == FNT_MASK_Yon6x6) {  //ANK 1/4角 6x6
          out.write (fntPageYon6x6.fnpBitmapArray, 0, 1 * 6 * 16 * 16);
        } else if (mask == FNT_MASK_Yon8x8) {  //ANK 1/4角 8x8
          out.write (MainMemory.mmrM8, FNT_ADDRESS_Yon8x8, 1 * 8 * 16 * 16);
        } else if (mask == FNT_MASK_Yon12x12) {  //ANK 1/4角 12x12
          out.write (MainMemory.mmrM8, FNT_ADDRESS_Yon12x12, 2 * 12 * 16 * 16);
        } else if (mask == FNT_MASK_Yon16x16) {  //ANK 1/4角 16x16
          out.write (fntPageYon16x16.fnpBitmapArray, 0, 2 * 16 * 16 * 16);
        } else if (mask == FNT_MASK_Yon24x24) {  //ANK 1/4角 24x24
          out.write (fntPageYon24x24.fnpBitmapArray, 0, 3 * 24 * 16 * 16);
        } else if (mask == FNT_MASK_Full8x8) {
          byte[] w = new byte[Math.max (1 * 8 * 256, 1 * 8 * 94 * 10)];
          out.write (w, 0, 1 * 8 * 256);  //Half4x8
          out.write (fntPageFull8x8.fnpBitmapArray, 0, 1 * 8 * 94 * 8);  //Full8x8 1区～8区
          out.write (w, 0, 1 * 8 * 94 * 7);  //Full8x8 9区～15区
          out.write (fntPageFull8x8.fnpBitmapArray, 1 * 8 * 94 * 8, 1 * 8 * 94 * 69);  //Full8x8 16区～84区
          out.write (w, 0, 1 * 8 * 94 * 10);  //Full8x8 85区～94区
        } else if ((mask & FNT_MASK_Half4x8) != 0) {
          out.write (fntPageHalf4x8.fnpBitmapArray, 0, 1 * 8 * 256);  //Half4x8
          if ((mask & FNT_MASK_Full8x8) != 0) {
            byte[] w = new byte[1 * 8 * 94 * 10];
            out.write (fntPageFull8x8.fnpBitmapArray, 0, 1 * 8 * 94 * 8);  //Full8x8 1区～8区
            out.write (w, 0, 1 * 8 * 94 * 7);  //Full8x8 9区～15区
            out.write (fntPageFull8x8.fnpBitmapArray, 1 * 8 * 94 * 8, 1 * 8 * 94 * 69);  //Full8x8 16区～84区
            out.write (w, 0, 1 * 8 * 94 * 10);  //Full8x8 85区～94区
          }
        } else if ((mask & FNT_MASK_Half6x12) != 0) {
          out.write (MainMemory.mmrM8, FNT_ADDRESS_Half6x12, 1 * 12 * 256);  //Half6x12
          if ((mask & FNT_MASK_Full12x12) != 0) {
            byte[] w = new byte[2 * 12 * 94 * 10];
            out.write (fntPageFull12x12.fnpBitmapArray, 0, 2 * 12 * 94 * 8);  //Full12x12 1区～8区
            out.write (w, 0, 2 * 12 * 94 * 7);  //Full12x12 9区～15区
            out.write (fntPageFull12x12.fnpBitmapArray, 2 * 12 * 94 * 8, 2 * 12 * 94 * 69);  //Full12x12 16区～84区
            out.write (w, 0, 2 * 12 * 94 * 10);  //Full12x12 85区～94区
          }
        } else if ((mask & FNT_MASK_Half8x16) != 0) {
          out.write (MainMemory.mmrM8, FNT_ADDRESS_Half8x16, 1 * 16 * 256);  //Half8x16
          if ((mask & FNT_MASK_Full16x16) != 0) {
            byte[] w = new byte[2 * 16 * 94 * 10];
            out.write (MainMemory.mmrM8, FNT_ADDRESS_Full16x16, 2 * 16 * 94 * 8);  //Full16x16 1区～8区
            //!!! 外字を出力したい
            out.write (w, 0, 2 * 16 * 94 * 7);  //Full16x16 9区～15区
            out.write (MainMemory.mmrM8, FNT_ADDRESS_Full16x16 + 2 * 16 * 94 * 8, 2 * 16 * 94 * 69);  //Full16x16 16区～84区
            out.write (w, 0, 2 * 16 * 94 * 10);  //Full16x16 85区～94区
          }
        } else if ((mask & FNT_MASK_Half12x24) != 0) {
          out.write (MainMemory.mmrM8, FNT_ADDRESS_Half12x24, 2 * 24 * 256);  //Half12x24
          if ((mask & FNT_MASK_Full24x24) != 0) {
            byte[] w = new byte[3 * 24 * 94 * 10];
            out.write (MainMemory.mmrM8, FNT_ADDRESS_Full24x24, 3 * 24 * 94 * 8);  //Full24x24 1区～8区
            //!!! 外字を出力したい
            out.write (w, 0, 3 * 24 * 94 * 7);  //Full24x24 9区～15区
            out.write (MainMemory.mmrM8, FNT_ADDRESS_Full24x24 + 3 * 24 * 94 * 8, 3 * 24 * 94 * 69);  //Full24x24 16区～84区
            out.write (w, 0, 3 * 24 * 94 * 10);  //Full24x24 85区～94区
          }
        } else if ((mask & FNT_MASK_Half16x32) != 0) {
          out.write (fntPageHalf16x32.fnpBitmapArray, 0, 2 * 32 * 256);  //Half16x32
          if ((mask & FNT_MASK_Full32x32) != 0) {
            byte[] w = new byte[4 * 32 * 94 * 10];
            out.write (fntPageFull32x32.fnpBitmapArray, 0, 4 * 32 * 94 * 8);  //Full32x32 1区～8区
            out.write (w, 0, 4 * 32 * 94 * 7);  //Full32x32 9区～15区
            out.write (fntPageFull32x32.fnpBitmapArray, 4 * 32 * 94 * 8, 4 * 32 * 94 * 69);  //Full32x32 16区～84区
            out.write (w, 0, 4 * 32 * 94 * 10);  //Full32x32 85区～94区
          }
        } else if ((mask & FNT_MASK_Half24x48) != 0) {
          out.write (fntPageHalf24x48.fnpBitmapArray, 0, 3 * 48 * 256);  //Half24x48
          if ((mask & FNT_MASK_Full48x48) != 0) {
            byte[] w = new byte[6 * 48 * 94 * 10];
            out.write (fntPageFull48x48.fnpBitmapArray, 0, 6 * 48 * 94 * 8);  //Full48x48 1区～8区
            out.write (w, 0, 6 * 48 * 94 * 7);  //Full48x48 9区～15区
            out.write (fntPageFull48x48.fnpBitmapArray, 6 * 48 * 94 * 8, 6 * 48 * 94 * 69);  //Full48x48 16区～84区
            out.write (w, 0, 6 * 48 * 94 * 10);  //Full48x48 85区～94区
          }
        } else if (mask == FNT_MASK_PrnPica36x46) {
          out.write (fntPagePrnPica36x46.fnpBitmapArray, 0, 5 * 46 * 16 * 32);
        } else if (mask == FNT_MASK_PrnElite30x46) {
          out.write (fntPagePrnElite30x46.fnpBitmapArray, 0, 4 * 46 * 16 * 32);
        } else if (mask == FNT_MASK_PrnSmall18x46) {
          out.write (fntPagePrnSmall18x46.fnpBitmapArray, 0, 3 * 46 * 16 * 32);
        } else if (mask == FNT_MASK_PrnScript28x32) {
          out.write (fntPagePrnScript28x32.fnpBitmapArray, 0, 4 * 32 * 16 * 32);
        } else if (mask == FNT_MASK_PrnHankaku24x48) {
          out.write (fntPagePrnHankaku24x48.fnpBitmapArray, 0, 3 * 48 * 16 * 32);
        } else if (mask == FNT_MASK_Lcd4x6) {
          out.write (fntPageLcd4x6.fnpBitmapArray, 0, 1 * 6 * 16 * 16);
        } else if (mask == FNT_MASK_Lcd6x8) {
          out.write (fntPageLcd6x8.fnpBitmapArray, 0, 1 * 8 * 16 * 16);
        } else if (mask == FNT_MASK_Shiromadokun6x8) {
          out.write (fntPageShiromadokun6x8.fnpBitmapArray, 0, 1 * 8 * 16 * 32);
        }
      } catch (IOException ioe) {
        JOptionPane.showMessageDialog (null, Multilingual.mlnJapanese ? "失敗しました" : "Failed");
        return;
      }
      if (file.exists ()) {
        if (fileBak.exists ()) {
          fileBak.delete ();  //javaのFileのrenameToはPerlと違って上書きしてくれないので明示的に削除する必要がある
        }
        file.renameTo (fileBak);
      }
      fileTmp.renameTo (file);
      fntCGROM = name;
      //成功したのでヒストリに加える
      fileChooser.addHistory (file);
    }
  }  //fntSaveFontDataFile()

  //fntReadFontImageFile ()
  //  フォント画像ファイルを選択して読み込む
  //  コマンドラインのみ
  public static void fntReadFontImageFile () {
    JFileChooser2 fileChooser = new JFileChooser2 (new File (fntSelectedPage.fnpImageName));
    fileChooser.setFileFilter (new javax.swing.filechooser.FileFilter () {  //java.io.FileFilterと紛らわしい
      @Override public boolean accept (File file) {
        String name = file.getName ();
        String upperName = name.toUpperCase ();
        return (file.isDirectory () ||
                (file.isFile () &&
                 (upperName.endsWith ("GIF") ||
                  upperName.endsWith ("PNG"))));
      }
      @Override public String getDescription () {
        return (Multilingual.mlnJapanese ?
                "フォント画像ファイル (*.GIF,*.PNG)" :
                "Font image files (*.GIF,*.PNG)");
      }
    });
    if (fileChooser.showOpenDialog (null) == JFileChooser.APPROVE_OPTION) {
      File file = fileChooser.getSelectedFile ();
      if (file.isFile ()) {
        String name = file.getPath ();
        BufferedImage image = XEiJ.loadImage (name);
        if (image != null) {
          int imageWidth = image.getWidth ();
          int imageHeight = image.getHeight ();
          for (int k = 0; k < FNT_COUNT; k++) {
            FontPage page = fntPageArray[(fntSelectedPage.fnpId + k) % FNT_COUNT];  //選択されているページから探し始める
            if (imageWidth == page.fnpImageWidth && imageHeight == page.fnpImageHeight) {
              page.fnpFromBufferedImage (image);
              page.fnpImageName = name;
              if (k != 0) {
                fntSelectPage (page);
              }
              //成功したのでヒストリに加える
              fileChooser.addHistory (file);
              return;
            }
          }
        }
      }
    }
    JOptionPane.showMessageDialog (null, Multilingual.mlnJapanese ? "失敗しました" : "Failed");
  }  //fntReadFontImageFile()

  //fntWriteFontImageFile ()
  //  フォント画像ファイルを名前を指定して書き出す
  //  コマンドラインのみ
  public static void fntWriteFontImageFile () {
    JFileChooser2 fileChooser = new JFileChooser2 (new File (fntSelectedPage.fnpImageName));
    fileChooser.setFileFilter (new javax.swing.filechooser.FileFilter () {  //java.io.FileFilterと紛らわしい
      @Override public boolean accept (File file) {
        String name = file.getName ();
        String upperName = name.toUpperCase ();
        return (file.isDirectory () ||
                (file.isFile () &&
                 (upperName.startsWith ("GIF") ||
                  upperName.endsWith (".PNG"))));
      }
      @Override public String getDescription () {
        return (Multilingual.mlnJapanese ?
                "フォント画像ファイル (*.GIF,*.PNG)" :
                "Font image files (*.GIF,*.PNG)");
      }
    });
    if (fileChooser.showSaveDialog (null) == JFileChooser.APPROVE_OPTION) {
      File file = fileChooser.getSelectedFile ();
      String name = file.getPath ();
      if (file.isFile ()) {  //既にある
        File fileBak = new File (name + ".bak");
        if (fileBak.exists ()) {
          fileBak.delete ();  //javaのFileのrenameToはPerlと違って上書きしてくれないので明示的に削除する必要がある
        }
        file.renameTo (fileBak);
      }
      if (XEiJ.saveImage (fntSelectedPage.fnpImageObject, name)) {
        fntSelectedPage.fnpImageName = name;  //成功したときだけ更新する
        //成功したのでヒストリに加える
        fileChooser.addHistory (file);
      } else {
        JOptionPane.showMessageDialog (null, Multilingual.mlnJapanese ? "失敗しました" : "Failed");
      }
    }
  }  //fntWriteFontImageFile()


  //  → ANK 1/4角 8x8

  //fntConvertHalf8x16toYon8x8 ()
  //  Half8x16を縮小してYon8x8を作る
  //  □■□■□■□■    □■□■□■□■
  //  □■□■□■□■    ■□■□■□■□
  //  ■□■□■□■□    □■□■□■□■
  //  ■□■□■□■□ → ■□■□■□■□
  //  □■□■□■□■    □■□■□■□■
  //  □■□■□■□■    ■□■□■□■□
  //  ■□■□■□■□    □■□■□■□■
  //  ■□■□■□■□    ■□■□■□■□
  //  □■□■□■□■
  //  □■□■□■□■
  //  ■□■□■□■□
  //  ■□■□■□■□
  //  □■□■□■□■
  //  □■□■□■□■
  //  ■□■□■□■□
  //  ■□■□■□■□
  public static void fntConvertHalf8x16toYon8x8 () {
    FontPage src = fntPageHalf8x16;
    FontPage dst = fntPageYon8x8;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    for (int c = 0; c < 8 * 256; c++) {
      mdst[adst] = (byte) (msrc[asrc] | msrc[asrc + 1]);
      asrc += 1 * (16 / 8);
      adst += 1 * (8 / 8);
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertHalf8x16toYon8x8()

  //fntConvertFull8x8toYon8x8 ()
  //  Full8x8からYon8x8へ変換する
  public static void fntConvertFull8x8toYon8x8 () {
    FontPage src = fntPageFull8x8;
    FontPage dst = fntPageYon8x8;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    for (int c = 0; c < 256; c++) {
      if (c == 0x82) {  //BROKEN BARを作る
        mdst[adst    ] = 0b00010000;
        mdst[adst + 1] = 0b00010000;
        mdst[adst + 2] = 0b00010000;
        mdst[adst + 3] = 0b00000000;
        mdst[adst + 4] = 0b00010000;
        mdst[adst + 5] = 0b00010000;
        mdst[adst + 6] = 0b00010000;
        mdst[adst + 7] = 0b00000000;
      } else {
        int s = CharacterCode.chrCharToSJIS[FontPage.Quater.FULL_BASE[c]];
        int sh = s >> 8;
        int sl = s & 255;
        int t = asrc + ((sh - (sh < 0xe0 ? 0x81 : 0xe0 - (0xa0 - 0x81))) * 188 +
                        (sl - (sl < 0x80 ? 0x40 : 0x80 - (0x7f - 0x40))) << 3);
        mdst[adst    ] = msrc[t    ];
        mdst[adst + 1] = msrc[t + 1];
        mdst[adst + 2] = msrc[t + 2];
        mdst[adst + 3] = msrc[t + 3];
        mdst[adst + 4] = msrc[t + 4];
        mdst[adst + 5] = msrc[t + 5];
        mdst[adst + 6] = msrc[t + 6];
        mdst[adst + 7] = msrc[t + 7];
      }
      adst += 1 * 8;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull8x8toYon8x8()


  //  → ANK 1/4角 12x12

  //fntConvertYon8x8toYon12x12 ()
  //  Yon8x8を拡大してYon12x12を作る
  //  □■□■□■□■    □□■□□■□□ ■□□■
  //  ■□■□■□■□    □□■□□■□□ ■□□■
  //  □■□■□■□■    ■■□■■□■■ □■■□
  //  ■□■□■□■□ → □□■□□■□□ ■□□■
  //  □■□■□■□■    □□■□□■□□ ■□□■
  //  ■□■□■□■□    ■■□■■□■■ □■■□
  //  □■□■□■□■    □□■□□■□□ ■□□■
  //  ■□■□■□■□    □□■□□■□□ ■□□■
  //                      ■■□■■□■■ □■■□
  //                      □□■□□■□□ ■□□■
  //                      □□■□□■□□ ■□□■
  //                      ■■□■■□■■ □■■□
  public static void fntConvertYon8x8toYon12x12 () {
    FontPage src = fntPageYon8x8;
    FontPage dst = fntPageYon12x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w = new int[256];
    if (true) {
      w[0] = 0b000;
      w[1] = 0b001;
      w[2] = 0b110;
      w[3] = 0b111;
      for (int i = 4; i < 16; i++) {
        w[i] = w[i >> 2] << 3 | w[i & 3];
      }
      for (int i = 16; i < 256; i++) {
        w[i] = w[i >> 4] << 6 | w[i & 15];
      }
    } else {
      for (int i = 0; i < 256; i++) {
        w[i] = ((i & 0b00000001) << 4 - 0 |
                (i & 0b00000010) * (0b00000000_01100000 >> 1) |
                (i & 0b00000100) << 7 - 2 |
                (i & 0b00001000) * (0b00000011_00000000 >> 3) |
                (i & 0b00010000) << 10 - 4 |
                (i & 0b00100000) * (0b00011000_00000000 >> 5) |
                (i & 0b01000000) << 13 - 6 |
                (i & 0b10000000) * (0b11000000_00000000 >> 7));
      }
    }
    for (int c = 0; c < 4 * 256; c++) {
      int t;
      t = w[msrc[asrc    ] & 255];
      mdst[adst + 2] = mdst[adst    ] = (byte) (t >> 4);
      mdst[adst + 3] = mdst[adst + 1] = (byte) (t << 4);
      t = w[msrc[asrc + 1] & 255];
      mdst[adst + 4]                  = (byte) (t >> 4);
      mdst[adst + 5]                  = (byte) (t << 4);
      asrc += 2;
      adst += 6;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertYon8x8toYon12x12()

  //fntConvertHalf6x12toYon12x12 ()
  //  Half6x12を拡大してYon12x12を作る
  //  □■□■□■　　    □□■■□□■■ □□■■　　　　
  //  ■□■□■□　　    ■■□□■■□□ ■■□□　　　　
  //  □■□■□■　　    □□■■□□■■ □□■■　　　　
  //  ■□■□■□　　    ■■□□■■□□ ■■□□　　　　
  //  □■□■□■　　    □□■■□□■■ □□■■　　　　
  //  ■□■□■□　　 → ■■□□■■□□ ■■□□　　　　
  //  □■□■□■　　    □□■■□□■■ □□■■　　　　
  //  ■□■□■□　　    ■■□□■■□□ ■■□□　　　　
  //  □■□■□■　　    □□■■□□■■ □□■■　　　　
  //  ■□■□■□　　    ■■□□■■□□ ■■□□　　　　
  //  □■□■□■　　    □□■■□□■■ □□■■　　　　
  //  ■□■□■□　　    ■■□□■■□□ ■■□□　　　　
  public static void fntConvertHalf6x12toYon12x12 () {
    FontPage src = fntPageHalf6x12;
    FontPage dst = fntPageYon12x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w = new int[256];
    w[0] = 0b0000;
    w[1] = 0b0011;
    w[2] = 0b1100;
    w[3] = 0b1111;
    for (int i = 4; i < 16; i++) {
      w[i] = w[i >> 2] << 4 | w[i & 3];
    }
    for (int i = 16; i < 256; i++) {
      w[i] = w[i >> 4] << 8 | w[i & 15];
    }
    for (int c = 0; c < 12 * 256; c++) {
      int t = w[(msrc[asrc] & 255) >> 2];
      mdst[adst    ] = (byte) (t >> 4);
      mdst[adst + 1] = (byte) (t << 4);
      asrc += 1;
      adst += 2;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertHalf6x12toYon12x12()

  //fntConvertFull12x12toYon12x12 ()
  //  Full12x12からYon12x12へ変換する
  public static void fntConvertFull12x12toYon12x12 () {
    FontPage src = fntPageFull12x12;
    FontPage dst = fntPageYon12x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    for (int c = 0; c < 256; c++) {
      if (c == 0x82) {  //BROKEN BARを作る
        Arrays.fill (mdst, adst, adst + 2 * 12, (byte) 0);
        for (int y = 0; y < 12 / 2 - 1; y++) {
          mdst[adst + 2 * y + 1] = 0b00000100;
        }
        for (int y = 12 / 2 + 0; y < 12 - 1; y++) {
          mdst[adst + 2 * y + 1] = 0b00000100;
        }
      } else {
        int s = CharacterCode.chrCharToSJIS[FontPage.QUA.FULLWIDTH[c]];
        int sh = s >> 8;
        int sl = s & 255;
        int t = asrc + 2 * 12 * ((sh - (sh < 0xe0 ? 0x81 : 0xe0 - (0xa0 - 0x81))) * 188 +
                                 (sl - (sl < 0x80 ? 0x40 : 0x80 - (0x7f - 0x40))));
        System.arraycopy (msrc, t, mdst, adst, 2 * 12);
      }
      adst += 2 * 12;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull12x12toYon12x12()


  //  → ANK 1/4角 16x16

  //fntConvertFull16x16toYon16x16 ()
  //  Full16x16からYon16x16へ変換する
  public static void fntConvertFull16x16toYon16x16 () {
    FontPage src = fntPageFull16x16;
    FontPage dst = fntPageYon16x16;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    for (int c = 0; c < 256; c++) {
      if (c == 0x82) {  //BROKEN BARを作る
        Arrays.fill (mdst, adst, adst + 2 * 16, (byte) 0);
        for (int y = 1; y < 16 / 2 - 1; y++) {
          mdst[adst + 2 * y] = 0b00000001;
        }
        for (int y = 16 / 2 + 1; y < 16 - 1; y++) {
          mdst[adst + 2 * y] = 0b00000001;
        }
      } else {
        int s = CharacterCode.chrCharToSJIS[FontPage.QUA.FULLWIDTH[c]];
        int sh = s >> 8;
        int sl = s & 255;
        int t = asrc + 2 * 16 * ((sh - (sh < 0xe0 ? 0x81 : 0xe0 - (0xa0 - 0x81))) * 188 +
                                 (sl - (sl < 0x80 ? 0x40 : 0x80 - (0x7f - 0x40))));
        System.arraycopy (msrc, t, mdst, adst, 2 * 16);
      }
      adst += 2 * 16;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull16x16toYon16x16()


  //  → ANK 1/4角 24x24

  //fntConvertFull24x24toYon24x24 ()
  //  Full24x24からYon24x24へ変換する
  public static void fntConvertFull24x24toYon24x24 () {
    FontPage src = fntPageFull24x24;
    FontPage dst = fntPageYon24x24;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    for (int c = 0; c < 256; c++) {
      if (c == 0x82) {  //BROKEN BARを作る
        Arrays.fill (mdst, adst, adst + 3 * 24, (byte) 0);
        for (int y = 1; y < 24 / 2 - 1; y++) {
          mdst[adst + 3 * y + 1] = 0b00010000;
        }
        for (int y = 24 / 2 + 1; y < 24 - 1; y++) {
          mdst[adst + 3 * y + 1] = 0b00010000;
        }
      } else {
        int s = CharacterCode.chrCharToSJIS[FontPage.QUA.FULLWIDTH[c]];
        int sh = s >> 8;
        int sl = s & 255;
        int t = asrc + 3 * 24 * ((sh - (sh < 0xe0 ? 0x81 : 0xe0 - (0xa0 - 0x81))) * 188 +
                                 (sl - (sl < 0x80 ? 0x40 : 0x80 - (0x7f - 0x40))));
        System.arraycopy (msrc, t, mdst, adst, 3 * 24);
      }
      adst += 3 * 24;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull24x24toYon24x24()


  //  → ANK 半角 6x12

  //fntConvertHalf8x16toHalf6x12FSX ()
  //  Half8x16を縮小してHalf6x12を作る(SX-Window)
  //  FSX310.X 0x02b03e
  //  □□■□■■□■    □■□■□■　　
  //  □□■□■■□■    ■□■□■□　　
  //  ■■□■□□■□    □■□■□■　　
  //  □□■□■■□■    ■□■□■□　　
  //  ■■□■□□■□    □■□■□■　　
  //  ■■□■□□■□ → ■□■□■□　　
  //  □□■□■■□■    □■□■□■　　
  //  ■■□■□□■□    ■□■□■□　　
  //  □□■□■■□■    □■□■□■　　
  //  □□■□■■□■    ■□■□■□　　
  //  ■■□■□□■□    □■□■□■　　
  //  □□■□■■□■    ■□■□■□　　
  //  ■■□■□□■□
  //  ■■□■□□■□
  //  □□■□■■□■
  //  ■■□■□□■□
  public static void fntConvertHalf8x16toHalf6x12FSX () {
    FontPage src = fntPageHalf8x16;
    FontPage dst = fntPageHalf6x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w = new int[256];
    for (int i = 0; i < 256; i++) {
      w[i] = (i & 0b10000000) | (i & 0b01111000) << 1 | (i & 0b00000111) << 2;
    }
    for (int c = 0; c < 4 * 256; c++) {
      mdst[adst    ] = (byte) w[(msrc[asrc    ] | msrc[asrc + 1]) & 255];
      mdst[adst + 1] = (byte) w[ msrc[asrc + 2]                   & 255];
      mdst[adst + 2] = (byte) w[ msrc[asrc + 3]                   & 255];
      asrc += 1 * (16 / 4);
      adst += 1 * (12 / 4);
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertHalf8x16toHalf6x12FSX()


  //  → ANK 半角 8x16

  //fntConvertYon8x8toHalf8x16 ()
  //  Yon8x8を拡大してHalf8x16を作る
  //  □■□■□■□■    □■□■□■□■
  //  ■□■□■□■□    □■□■□■□■
  //  □■□■□■□■    ■□■□■□■□
  //  ■□■□■□■□ → ■□■□■□■□
  //  □■□■□■□■    □■□■□■□■
  //  ■□■□■□■□    □■□■□■□■
  //  □■□■□■□■    ■□■□■□■□
  //  ■□■□■□■□    ■□■□■□■□
  //                      □■□■□■□■
  //                      □■□■□■□■
  //                      ■□■□■□■□
  //                      ■□■□■□■□
  //                      □■□■□■□■
  //                      □■□■□■□■
  //                      ■□■□■□■□
  //                      ■□■□■□■□
  public static void fntConvertYon8x8toHalf8x16 () {
    FontPage src = fntPageYon8x8;
    FontPage dst = fntPageHalf8x16;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    for (int c = 0; c < 8 * 256; c++) {
      mdst[adst    ] = msrc[asrc];
      mdst[adst + 1] = msrc[asrc];
      asrc += 1;
      adst += 2;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertYon8x8toHalf8x16()

  //fntConvertHalf6x12toHalf8x16 ()
  //  Half6x12を拡大してHalf8x16を作る
  //  □■□■□■　　    □□■□■■□■
  //  ■□■□■□　　    □□■□■■□■
  //  □■□■□■　　    ■■□■□□■□
  //  ■□■□■□　　    □□■□■■□■
  //  □■□■□■　　    ■■□■□□■□
  //  ■□■□■□　　 → ■■□■□□■□
  //  □■□■□■　　    □□■□■■□■
  //  ■□■□■□　　    ■■□■□□■□
  //  □■□■□■　　    □□■□■■□■
  //  ■□■□■□　　    □□■□■■□■
  //  □■□■□■　　    ■■□■□□■□
  //  ■□■□■□　　    □□■□■■□■
  //                      ■■□■□□■□
  //                      ■■□■□□■□
  //                      □□■□■■□■
  //                      ■■□■□□■□
  public static void fntConvertHalf6x12toHalf8x16 () {
    FontPage src = fntPageHalf6x12;
    FontPage dst = fntPageHalf8x16;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w = new int[64];
    w[0] = 0b0000;
    w[1] = 0b0001;
    w[2] = 0b0010;
    w[3] = 0b0011;
    w[4] = 0b1100;
    w[5] = 0b1101;
    w[6] = 0b1110;
    w[7] = 0b1111;
    for (int i = 8; i < 64; i++) {
      w[i] = w[i >> 3] << 4 | w[i & 7];
    }
    for (int c = 0; c < (16 / 4) * 256; c++) {
      mdst[adst + 1] = mdst[adst] = (byte) w[(msrc[asrc    ] & 255) >> 2];
      mdst[adst + 2]              = (byte) w[(msrc[asrc + 1] & 255) >> 2];
      mdst[adst + 3]              = (byte) w[(msrc[asrc + 2] & 255) >> 2];
      asrc += 3;
      adst += 4;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertHalf6x12toHalf8x16()

  //fntConvertFull16x16toHalf8x16 ()
  //  Full16x16を縮小してHalf8x16を作る
  //  □□■■□□■■ □□■■□□■■    □■□■□■□■
  //  ■■□□■■□□ ■■□□■■□□    ■□■□■□■□
  //  □□■■□□■■ □□■■□□■■    □■□■□■□■
  //  ■■□□■■□□ ■■□□■■□□    ■□■□■□■□
  //  □□■■□□■■ □□■■□□■■    □■□■□■□■
  //  ■■□□■■□□ ■■□□■■□□    ■□■□■□■□
  //  □□■■□□■■ □□■■□□■■    □■□■□■□■
  //  ■■□□■■□□ ■■□□■■□□ → ■□■□■□■□
  //  □□■■□□■■ □□■■□□■■    □■□■□■□■
  //  ■■□□■■□□ ■■□□■■□□    ■□■□■□■□
  //  □□■■□□■■ □□■■□□■■    □■□■□■□■
  //  ■■□□■■□□ ■■□□■■□□    ■□■□■□■□
  //  □□■■□□■■ □□■■□□■■    □■□■□■□■
  //  ■■□□■■□□ ■■□□■■□□    ■□■□■□■□
  //  □□■■□□■■ □□■■□□■■    □■□■□■□■
  //  ■■□□■■□□ ■■□□■■□□    ■□■□■□■□
  public static void fntConvertFull16x16toHalf8x16 () {
    FontPage src = fntPageFull16x16;
    FontPage dst = fntPageHalf8x16;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w = new int[256];
    for (int i = 0; i < 256; i++) {
      w[i] = (i & 0b10000000) | (i & 0b01100000) << 1 | (i & 0b00011000) << 2 | (i & 0b00000110) << 3 | (i & 0b00000001) << 4;
    }
    for (int c = 0; c < 256; c++) {
      if (c == 0x82) {  //BROKEN BARを作る
        mdst[adst     ] = 0b00010000;
        mdst[adst +  1] = 0b00010000;
        mdst[adst +  2] = 0b00010000;
        mdst[adst +  3] = 0b00010000;
        mdst[adst +  4] = 0b00010000;
        mdst[adst +  5] = 0b00010000;
        mdst[adst +  6] = 0b00010000;
        mdst[adst +  7] = 0b00000000;
        mdst[adst +  8] = 0b00010000;
        mdst[adst +  9] = 0b00010000;
        mdst[adst + 10] = 0b00010000;
        mdst[adst + 11] = 0b00010000;
        mdst[adst + 12] = 0b00010000;
        mdst[adst + 13] = 0b00010000;
        mdst[adst + 14] = 0b00010000;
        mdst[adst + 15] = 0b00000000;
      } else {
        int s = CharacterCode.chrCharToSJIS[FontPage.QUA.FULLWIDTH[c]];
        int sh = s >> 8;
        int sl = s & 255;
        int t = asrc + ((sh - (sh < 0xe0 ? 0x81 : 0xe0 - (0xa0 - 0x81))) * 188 +
                        (sl - (sl < 0x80 ? 0x40 : 0x80 - (0x7f - 0x40))) << 5);
        mdst[adst     ] = (byte) (w[msrc[t     ] & 255] | w[msrc[t +  1] & 255] >> 4);
        mdst[adst +  1] = (byte) (w[msrc[t +  2] & 255] | w[msrc[t +  3] & 255] >> 4);
        mdst[adst +  2] = (byte) (w[msrc[t +  4] & 255] | w[msrc[t +  5] & 255] >> 4);
        mdst[adst +  3] = (byte) (w[msrc[t +  6] & 255] | w[msrc[t +  7] & 255] >> 4);
        mdst[adst +  4] = (byte) (w[msrc[t +  8] & 255] | w[msrc[t +  9] & 255] >> 4);
        mdst[adst +  5] = (byte) (w[msrc[t + 10] & 255] | w[msrc[t + 11] & 255] >> 4);
        mdst[adst +  6] = (byte) (w[msrc[t + 12] & 255] | w[msrc[t + 13] & 255] >> 4);
        mdst[adst +  7] = (byte) (w[msrc[t + 14] & 255] | w[msrc[t + 15] & 255] >> 4);
        mdst[adst +  8] = (byte) (w[msrc[t + 16] & 255] | w[msrc[t + 17] & 255] >> 4);
        mdst[adst +  9] = (byte) (w[msrc[t + 18] & 255] | w[msrc[t + 19] & 255] >> 4);
        mdst[adst + 10] = (byte) (w[msrc[t + 20] & 255] | w[msrc[t + 21] & 255] >> 4);
        mdst[adst + 11] = (byte) (w[msrc[t + 22] & 255] | w[msrc[t + 23] & 255] >> 4);
        mdst[adst + 12] = (byte) (w[msrc[t + 24] & 255] | w[msrc[t + 25] & 255] >> 4);
        mdst[adst + 13] = (byte) (w[msrc[t + 26] & 255] | w[msrc[t + 27] & 255] >> 4);
        mdst[adst + 14] = (byte) (w[msrc[t + 28] & 255] | w[msrc[t + 29] & 255] >> 4);
        mdst[adst + 15] = (byte) (w[msrc[t + 30] & 255] | w[msrc[t + 31] & 255] >> 4);
      }
      adst += 16;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull16x16toHalf8x16()


  //  → ANK 半角 12x24

  //fntConvertFull24x24toHalf12x24 ()
  //  Full24x24を縮小してHalf12x24を作る
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□ → ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  public static void fntConvertFull24x24toHalf12x24 () {
    FontPage src = fntPageFull24x24;
    FontPage dst = fntPageHalf12x24;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w = new int[256];
    for (int i = 0; i < 256; i++) {
      w[i] = (i & 0b10000000) | (i & 0b01100000) << 1 | (i & 0b00011000) << 2 | (i & 0b00000110) << 3 | (i & 0b00000001) << 4;
    }
    for (int c = 0; c < 256; c++) {
      if (c == 0x82) {  //BROKEN BARを作る
        for (int i = 0; i < 24; i++) {
          mdst[adst    ] = (byte) (i == 11 || i == 12 ? 0b00000000 : 0b00000100);
          mdst[adst + 1] = 0b00000000;
          adst += 2;
        }
      } else {
        int s = CharacterCode.chrCharToSJIS[FontPage.QUA.FULLWIDTH[c]];
        int sh = s >> 8;
        int sl = s & 255;
        int t = asrc + 3 * 24 * ((sh - (sh < 0xe0 ? 0x81 : 0xe0 - (0xa0 - 0x81))) * 188 +
                                 (sl - (sl < 0x80 ? 0x40 : 0x80 - (0x7f - 0x40))));
        for (int i = 0; i < 24; i++) {
          mdst[adst    ] = (byte) (w[msrc[t    ] & 255] | w[msrc[t + 1] & 255] >> 4);
          mdst[adst + 1] = (byte)  w[msrc[t + 2] & 255];
          t += 3;
          adst += 2;
        }
      }
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull24x24toHalf12x24()


  //  → ANK 半角 16x32


  //  → ANK 半角 24x48


  //  → 漢字 全角 8x8


  //  → 漢字 全角 12x12

  //fntConvertFull16x16toFull12x12FSX ()
  //  Full16x16を縮小してFull12x12を作る(SX-Window)
  //  FSX310.X 0x02b11e
  //  □□■□■■□■ □□■□■■□■    □■□■□■□■ □■□■
  //  □□■□■■□■ □□■□■■□■    ■□■□■□■□ ■□■□
  //  ■■□■□□■□ ■■□■□□■□    □■□■□■□■ □■□■
  //  □□■□■■□■ □□■□■■□■    ■□■□■□■□ ■□■□
  //  ■■□■□□■□ ■■□■□□■□    □■□■□■□■ □■□■
  //  ■■□■□□■□ ■■□■□□■□ → ■□■□■□■□ ■□■□
  //  □□■□■■□■ □□■□■■□■    □■□■□■□■ □■□■
  //  ■■□■□□■□ ■■□■□□■□    ■□■□■□■□ ■□■□
  //  □□■□■■□■ □□■□■■□■    □■□■□■□■ □■□■
  //  □□■□■■□■ □□■□■■□■    ■□■□■□■□ ■□■□
  //  ■■□■□□■□ ■■□■□□■□    □■□■□■□■ □■□■
  //  □□■□■■□■ □□■□■■□■    ■□■□■□■□ ■□■□
  //  ■■□■□□■□ ■■□■□□■□
  //  ■■□■□□■□ ■■□■□□■□
  //  □□■□■■□■ □□■□■■□■
  //  ■■□■□□■□ ■■□■□□■□
  public static void fntConvertFull16x16toFull12x12FSX () {
    FontPage src = fntPageFull16x16;
    FontPage dst = fntPageFull12x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w = new int[256];
    for (int i = 0; i < 256; i++) {
      w[i] = (i & 0b10000000) | (i & 0b01111000) << 1 | (i & 0b00000111) << 2;
    }
    for (int i = 0; i < 4 * 94 * 77; i++) {
      int t =                  w[(msrc[asrc + 1] | msrc[asrc + 3]) & 255];
      mdst[adst    ] = (byte) (w[(msrc[asrc    ] | msrc[asrc + 2]) & 255] | t >> 6);
      mdst[adst + 1] = (byte) (t << 2);
      t =                      w[ msrc[asrc + 5]                   & 255];
      mdst[adst + 2] = (byte) (w[ msrc[asrc + 4]                   & 255] | t >> 6);
      mdst[adst + 3] = (byte) (t << 2);
      t =                      w[ msrc[asrc + 7]                   & 255];
      mdst[adst + 4] = (byte) (w[ msrc[asrc + 6]                   & 255] | t >> 6);
      mdst[adst + 5] = (byte) (t << 2);
      asrc += 2 * (16 / 4);
      adst += 2 * (12 / 4);
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull16x16toFull12x12FSX()

  //fntConvertFull16x16toFull12x12Test1 ()
  //  Full16x16を縮小してFull12x12を作る(テスト1)
  //  □□■■□■□■ □■□■□□■■    □■□■□■□■ □■□■
  //  □□■■□■□■ □■□■□□■■    ■□■□■□■□ ■□■□
  //  ■■□□■□■□ ■□■□■■□□    □■□■□■□■ □■□■
  //  ■■□□■□■□ ■□■□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□■□■ □■□■□□■■    □■□■□■□■ □■□■
  //  ■■□□■□■□ ■□■□■■□□ → ■□■□■□■□ ■□■□
  //  □□■■□■□■ □■□■□□■■    □■□■□■□■ □■□■
  //  ■■□□■□■□ ■□■□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□■□■ □■□■□□■■    □■□■□■□■ □■□■
  //  ■■□□■□■□ ■□■□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□■□■ □■□■□□■■    □■□■□■□■ □■□■
  //  ■■□□■□■□ ■□■□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□■□■ □■□■□□■■
  //  □□■■□■□■ □■□■□□■■
  //  ■■□□■□■□ ■□■□■■□□
  //  ■■□□■□■□ ■□■□■■□□
  public static void fntConvertFull16x16toFull12x12Test1 () {
    FontPage src = fntPageFull16x16;
    FontPage dst = fntPageFull12x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w0 = new int[256];
    int[] w1 = new int[256];
    for (int i = 0; i < 256; i++) {
      w0[i] = (i & 0b10000000) | (i & 0b01100000) << 1 | (i & 0b00011111) << 2;
      w1[i] = (i & 0b11111000) | (i & 0b00000110) << 1 | (i & 0b00000001) << 2;
    }
    for (int i = 0; i < 94 * 77; i++) {
      int t =                   w1[(msrc[asrc +  1] | msrc[asrc +  3]) & 255];
      mdst[adst     ] = (byte) (w0[(msrc[asrc     ] | msrc[asrc +  2]) & 255] | t >> 6);
      mdst[adst +  1] = (byte) (t << 2);
      t =                       w1[(msrc[asrc +  5] | msrc[asrc +  7]) & 255];
      mdst[adst +  2] = (byte) (w0[(msrc[asrc +  4] | msrc[asrc +  6]) & 255] | t >> 6);
      mdst[adst +  3] = (byte) (t << 2);
      t =                       w1[ msrc[asrc +  9]                    & 255];
      mdst[adst +  4] = (byte) (w0[ msrc[asrc +  8]                    & 255] | t >> 6);
      mdst[adst +  5] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 11]                    & 255];
      mdst[adst +  6] = (byte) (w0[ msrc[asrc + 10]                    & 255] | t >> 6);
      mdst[adst +  7] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 13]                    & 255];
      mdst[adst +  8] = (byte) (w0[ msrc[asrc + 12]                    & 255] | t >> 6);
      mdst[adst +  9] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 15]                    & 255];
      mdst[adst + 10] = (byte) (w0[ msrc[asrc + 14]                    & 255] | t >> 6);
      mdst[adst + 11] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 17]                    & 255];
      mdst[adst + 12] = (byte) (w0[ msrc[asrc + 16]                    & 255] | t >> 6);
      mdst[adst + 13] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 19]                    & 255];
      mdst[adst + 14] = (byte) (w0[ msrc[asrc + 18]                    & 255] | t >> 6);
      mdst[adst + 15] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 21]                    & 255];
      mdst[adst + 16] = (byte) (w0[ msrc[asrc + 20]                    & 255] | t >> 6);
      mdst[adst + 17] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 23]                    & 255];
      mdst[adst + 18] = (byte) (w0[ msrc[asrc + 22]                    & 255] | t >> 6);
      mdst[adst + 19] = (byte) (t << 2);
      t =                       w1[(msrc[asrc + 25] | msrc[asrc + 27]) & 255];
      mdst[adst + 20] = (byte) (w0[(msrc[asrc + 24] | msrc[asrc + 26]) & 255] | t >> 6);
      mdst[adst + 21] = (byte) (t << 2);
      t =                       w1[(msrc[asrc + 29] | msrc[asrc + 31]) & 255];
      mdst[adst + 22] = (byte) (w0[(msrc[asrc + 28] | msrc[asrc + 30]) & 255] | t >> 6);
      mdst[adst + 23] = (byte) (t << 2);
      asrc += 2 * 16;
      adst += 2 * 12;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull16x16toFull12x12Test1()

  //fntConvertFull16x16toFull12x12Test2 ()
  //  Full16x16を縮小してFull12x12を作る(テスト2)
  //  □□■■□■□■ □■□■■□□■    □■□■□■□■ □■□■
  //  □□■■□■□■ □■□■■□□■    ■□■□■□■□ ■□■□
  //  ■■□□■□■□ ■□■□□■■□    □■□■□■□■ □■□■
  //  ■■□□■□■□ ■□■□□■■□    ■□■□■□■□ ■□■□
  //  □□■■□■□■ □■□■■□□■    □■□■□■□■ □■□■
  //  ■■□□■□■□ ■□■□□■■□ → ■□■□■□■□ ■□■□
  //  □□■■□■□■ □■□■■□□■    □■□■□■□■ □■□■
  //  ■■□□■□■□ ■□■□□■■□    ■□■□■□■□ ■□■□
  //  □□■■□■□■ □■□■■□□■    □■□■□■□■ □■□■
  //  ■■□□■□■□ ■□■□□■■□    ■□■□■□■□ ■□■□
  //  □□■■□■□■ □■□■■□□■    □■□■□■□■ □■□■
  //  ■■□□■□■□ ■□■□□■■□    ■□■□■□■□ ■□■□
  //  ■■□□■□■□ ■□■□□■■□
  //  □□■■□■□■ □■□■■□□■
  //  □□■■□■□■ □■□■■□□■
  //  ■■□□■□■□ ■□■□□■■□
  public static void fntConvertFull16x16toFull12x12Test2 () {
    FontPage src = fntPageFull16x16;
    FontPage dst = fntPageFull12x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w0 = new int[256];
    int[] w1 = new int[256];
    for (int i = 0; i < 256; i++) {
      w0[i] = (i & 0b10000000) | (i & 0b01100000) << 1 | (i & 0b00011111) << 2;
      w1[i] = (i & 0b11110000) | (i & 0b00001100) << 1 | (i & 0b00000011) << 2;
    }
    for (int i = 0; i < 94 * 77; i++) {
      int t =                   w1[(msrc[asrc +  1] | msrc[asrc +  3]) & 255];
      mdst[adst     ] = (byte) (w0[(msrc[asrc     ] | msrc[asrc +  2]) & 255] | t >> 6);
      mdst[adst +  1] = (byte) (t << 2);
      t =                       w1[(msrc[asrc +  5] | msrc[asrc +  7]) & 255];
      mdst[adst +  2] = (byte) (w0[(msrc[asrc +  4] | msrc[asrc +  6]) & 255] | t >> 6);
      mdst[adst +  3] = (byte) (t << 2);
      t =                       w1[ msrc[asrc +  9]                    & 255];
      mdst[adst +  4] = (byte) (w0[ msrc[asrc +  8]                    & 255] | t >> 6);
      mdst[adst +  5] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 11]                    & 255];
      mdst[adst +  6] = (byte) (w0[ msrc[asrc + 10]                    & 255] | t >> 6);
      mdst[adst +  7] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 13]                    & 255];
      mdst[adst +  8] = (byte) (w0[ msrc[asrc + 12]                    & 255] | t >> 6);
      mdst[adst +  9] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 15]                    & 255];
      mdst[adst + 10] = (byte) (w0[ msrc[asrc + 14]                    & 255] | t >> 6);
      mdst[adst + 11] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 17]                    & 255];
      mdst[adst + 12] = (byte) (w0[ msrc[asrc + 16]                    & 255] | t >> 6);
      mdst[adst + 13] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 19]                    & 255];
      mdst[adst + 14] = (byte) (w0[ msrc[asrc + 18]                    & 255] | t >> 6);
      mdst[adst + 15] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 21]                    & 255];
      mdst[adst + 16] = (byte) (w0[ msrc[asrc + 20]                    & 255] | t >> 6);
      mdst[adst + 17] = (byte) (t << 2);
      t =                       w1[(msrc[asrc + 23] | msrc[asrc + 25]) & 255];
      mdst[adst + 18] = (byte) (w0[(msrc[asrc + 22] | msrc[asrc + 24]) & 255] | t >> 6);
      mdst[adst + 19] = (byte) (t << 2);
      t =                       w1[(msrc[asrc + 27] | msrc[asrc + 29]) & 255];
      mdst[adst + 20] = (byte) (w0[(msrc[asrc + 26] | msrc[asrc + 28]) & 255] | t >> 6);
      mdst[adst + 21] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 31]                    & 255];
      mdst[adst + 22] = (byte) (w0[ msrc[asrc + 30]                    & 255] | t >> 6);
      mdst[adst + 23] = (byte) (t << 2);
      asrc += 2 * 16;
      adst += 2 * 12;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull16x16toFull12x12Test2()

  //fntConvertFull16x16toFull12x12Test3 ()
  //  Full16x16を縮小してFull12x12を作る(テスト3)
  //  □□■□□■□■ □■□□■□□■    □■□■□■□■ □■□■
  //  □□■□□■□■ □■□□■□□■    ■□■□■□■□ ■□■□
  //  ■■□■■□■□ ■□■■□■■□    □■□■□■□■ □■□■
  //  □□■□□■□■ □■□□■□□■    ■□■□■□■□ ■□■□
  //  □□■□□■□■ □■□□■□□■    □■□■□■□■ □■□■
  //  ■■□■■□■□ ■□■■□■■□ → ■□■□■□■□ ■□■□
  //  □□■□□■□■ □■□□■□□■    □■□■□■□■ □■□■
  //  ■■□■■□■□ ■□■■□■■□    ■□■□■□■□ ■□■□
  //  □□■□□■□■ □■□□■□□■    □■□■□■□■ □■□■
  //  ■■□■■□■□ ■□■■□■■□    ■□■□■□■□ ■□■□
  //  □□■□□■□■ □■□□■□□■    □■□■□■□■ □■□■
  //  □□■□□■□■ □■□□■□□■    ■□■□■□■□ ■□■□
  //  ■■□■■□■□ ■□■■□■■□
  //  □□■□□■□■ □■□□■□□■
  //  □□■□□■□■ □■□□■□□■
  //  ■■□■■□■□ ■□■■□■■□
  public static void fntConvertFull16x16toFull12x12Test3 () {
    FontPage src = fntPageFull16x16;
    FontPage dst = fntPageFull12x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w0 = new int[256];
    int[] w1 = new int[256];
    for (int i = 0; i < 256; i++) {
      w0[i] = (i & 0b10000000) | (i & 0b01110000) << 1 | (i & 0b00001111) << 2;
      w1[i] = (i & 0b11100000) | (i & 0b00011100) << 1 | (i & 0b00000011) << 2;
    }
    for (int i = 0; i < 94 * 77; i++) {
      int t =                   w1[(msrc[asrc +  1] | msrc[asrc +  3]) & 255];
      mdst[adst     ] = (byte) (w0[(msrc[asrc     ] | msrc[asrc +  2]) & 255] | t >> 6);
      mdst[adst +  1] = (byte) (t << 2);
      t =                       w1[ msrc[asrc +  5]                    & 255];
      mdst[adst +  2] = (byte) (w0[ msrc[asrc +  4]                    & 255] | t >> 6);
      mdst[adst +  3] = (byte) (t << 2);
      t =                       w1[(msrc[asrc +  7] | msrc[asrc +  9]) & 255];
      mdst[adst +  4] = (byte) (w0[(msrc[asrc +  6] | msrc[asrc +  8]) & 255] | t >> 6);
      mdst[adst +  5] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 11]                    & 255];
      mdst[adst +  6] = (byte) (w0[ msrc[asrc + 10]                    & 255] | t >> 6);
      mdst[adst +  7] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 13]                    & 255];
      mdst[adst +  8] = (byte) (w0[ msrc[asrc + 12]                    & 255] | t >> 6);
      mdst[adst +  9] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 15]                    & 255];
      mdst[adst + 10] = (byte) (w0[ msrc[asrc + 14]                    & 255] | t >> 6);
      mdst[adst + 11] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 17]                    & 255];
      mdst[adst + 12] = (byte) (w0[ msrc[asrc + 16]                    & 255] | t >> 6);
      mdst[adst + 13] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 19]                    & 255];
      mdst[adst + 14] = (byte) (w0[ msrc[asrc + 18]                    & 255] | t >> 6);
      mdst[adst + 15] = (byte) (t << 2);
      t =                       w1[(msrc[asrc + 21] | msrc[asrc + 23]) & 255];
      mdst[adst + 16] = (byte) (w0[(msrc[asrc + 20] | msrc[asrc + 22]) & 255] | t >> 6);
      mdst[adst + 17] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 25]                    & 255];
      mdst[adst + 18] = (byte) (w0[ msrc[asrc + 24]                    & 255] | t >> 6);
      mdst[adst + 19] = (byte) (t << 2);
      t =                       w1[(msrc[asrc + 27] | msrc[asrc + 29]) & 255];
      mdst[adst + 20] = (byte) (w0[(msrc[asrc + 26] | msrc[asrc + 28]) & 255] | t >> 6);
      mdst[adst + 21] = (byte) (t << 2);
      t =                       w1[ msrc[asrc + 31]                    & 255];
      mdst[adst + 22] = (byte) (w0[ msrc[asrc + 30]                    & 255] | t >> 6);
      mdst[adst + 23] = (byte) (t << 2);
      asrc += 2 * 16;
      adst += 2 * 12;
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull16x16toFull12x12Test3()

  //fntConvertFull16x16toFull12x12Test4 ()
  //  Full16x16を縮小してFull12x12を作る(テスト4)
  //  □□□□■■■■ □□□□■■■■    □□□■■■□□ □■■■
  //  □□□□■■■■ □□□□■■■■    □□□■■■□□ □■■■
  //  □□□□■■■■ □□□□■■■■    □□□■■■□□ □■■■
  //  □□□□■■■■ □□□□■■■■    ■■■□□□■■ ■□□□
  //  ■■■■□□□□ ■■■■□□□□    ■■■□□□■■ ■□□□
  //  ■■■■□□□□ ■■■■□□□□ → ■■■□□□■■ ■□□□
  //  ■■■■□□□□ ■■■■□□□□    □□□■■■□□ □■■■
  //  ■■■■□□□□ ■■■■□□□□    □□□■■■□□ □■■■
  //  □□□□■■■■ □□□□■■■■    □□□■■■□□ □■■■
  //  □□□□■■■■ □□□□■■■■    ■■■□□□■■ ■□□□
  //  □□□□■■■■ □□□□■■■■    ■■■□□□■■ ■□□□
  //  □□□□■■■■ □□□□■■■■    ■■■□□□■■ ■□□□
  //  ■■■■□□□□ ■■■■□□□□
  //  ■■■■□□□□ ■■■■□□□□
  //  ■■■■□□□□ ■■■■□□□□
  //  ■■■■□□□□ ■■■■□□□□
  //   a b c d     e f g
  //  □□□□ → □□□    e = a | b & ~c & d
  //  □□□■ → □□■    f = b & c | ~a & (b | c) & ~d
  //  □□■□ → □■□    g = a & ~b & c | d
  //  □□■■ → □□■
  //  □■□□ → □■□
  //  □■□■ → ■□■
  //  □■■□ → □■□
  //  □■■■ → □■■
  //  ■□□□ → ■□□
  //  ■□□■ → ■□■
  //  ■□■□ → ■□■
  //  ■□■■ → ■□■
  //  ■■□□ → ■□□
  //  ■■□■ → ■□■
  //  ■■■□ → ■■□
  //  ■■■■ → ■■■
  //  for (int a = 0; a <= 1; a++) {
  //    for (int b = 0; b <= 1; b++) {
  //      for (int c = 0; c <= 1; c++) {
  //        for (int d = 0; d <= 1; d++) {
  //          int e = a | b & ~c & d;
  //          int f = b & c | ~a & (b | c) & ~d;
  //          int g = a & ~b & c | d;
  //          System.out.printf ("%d%d%d%d %d%d%d\n", a, b, c, d, e, f, g);
  //        }
  //      }
  //    }
  //  }
  //private static final long FNT_TEST4_BASE = 0b1110_1100_1010_1000_1010_1010_1010_1000_0110_0100_1010_0100_0010_0100_0010_0000L;
  public static final long FNT_TEST4_BASE = (
    0b000_0L       |  //0000
    0b001_0L <<  4 |  //0001
    0b010_0L <<  8 |  //0010
    0b011_0L << 12 |  //0011 *
    0b010_0L << 16 |  //0100
    0b101_0L << 20 |  //0101
    0b010_0L << 24 |  //0110
    0b011_0L << 28 |  //0111
    0b100_0L << 32 |  //1000
    0b101_0L << 36 |  //1001
    0b101_0L << 40 |  //1010
    0b101_0L << 44 |  //1011
    0b110_0L << 48 |  //1100 *
    0b101_0L << 52 |  //1101
    0b110_0L << 56 |  //1110
    0b111_0L << 60);  //1111
  public static int[] fntReverseMap;
  public static int[] fntTest4Map;
  public static void fntConvertFull16x16toFull12x12Test4 () {
    int[] r = fntReverseMap;
    if (r == null) {
      r = fntReverseMap = new int[65536];
      for (int i = 0; i < 65536; i++) {
        //  0123    048c
        //  4567 → 159d
        //  89ab    26ae
        //  cdef    37bf
        r[i] = VideoController.VCN_TXP3[i >> 12] | VideoController.VCN_TXP2[i >> 8 & 15] | VideoController.VCN_TXP1[i >> 4 & 15] | VideoController.VCN_TXP0[i & 15];
        //System.out.printf ("%04x %04x\n", i, r[i]);
      }
    }
    int[] m = fntTest4Map;
    if (m == null) {
      int[] n = new int[65536];
      for (int i = 0; i < 65536; i++) {
        n[i] = ((int) (FNT_TEST4_BASE >> (i >> 12 - 2 & 15 << 2)) << 12 & 0xf000 |
                (int) (FNT_TEST4_BASE >> (i >>  8 - 2 & 15 << 2)) <<  8 & 0x0f00 |
                (int) (FNT_TEST4_BASE >> (i >>  4 - 2 & 15 << 2)) <<  4 & 0x00f0 |
                (int) (FNT_TEST4_BASE >> (i <<      2 & 15 << 2))       & 0x000f);
        //System.out.printf ("%04x %04x\n", i, n[i]);
      }
      m = fntTest4Map = new int[65536];
      for (int i = 0; i < 65536; i++) {
        m[i] = r[n[r[n[i]]]] | n[r[n[r[i]]]];
        //System.out.printf ("%04x %04x\n", i, m[i]);
      }
    }
    FontPage src = fntPageFull16x16;
    FontPage dst = fntPageFull12x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    for (int i = 0; i < 4 * 94 * 77; i++) {
      int s0 = msrc[asrc    ];
      int s1 = msrc[asrc + 1];
      int s2 = msrc[asrc + 2];
      int s3 = msrc[asrc + 3];
      int s4 = msrc[asrc + 4];
      int s5 = msrc[asrc + 5];
      int s6 = msrc[asrc + 6];
      int s7 = msrc[asrc + 7];
      int t0 = m[s0 <<  8 & 0xf000 | s2 << 4 & 0x0f00 | s4      & 0x00f0 | s6 >> 4 & 0x000f];
      int t1 = m[s0 << 12 & 0xf000 | s2 << 8 & 0x0f00 | s4 << 4 & 0x00f0 | s6      & 0x000f];
      int t2 = m[s1 <<  8 & 0xf000 | s3 << 4 & 0x0f00 | s5      & 0x00f0 | s7 >> 4 & 0x000f];
      int t3 = m[s1 << 12 & 0xf000 | s3 << 8 & 0x0f00 | s5 << 4 & 0x00f0 | s7      & 0x000f];
      mdst[adst    ] = (byte) (t0 >> 13 - 5 & 0b11100000 | t1 >> 13 - 2 & 0b00011100 | t2 >> 13 + 1 & 0b00000011);
      mdst[adst + 1] = (byte) (t2 >> 13 - 7 & 0b10000000 | t3 >> 13 - 4 & 0b01110000);
      mdst[adst + 2] = (byte) (t0 >>  9 - 5 & 0b11100000 | t1 >>  9 - 2 & 0b00011100 | t2 >>  9 + 1 & 0b00000011);
      mdst[adst + 3] = (byte) (t2 >>  9 - 7 & 0b10000000 | t3 >>  9 - 4 & 0b01110000);
      mdst[adst + 4] = (byte) (t0           & 0b11100000 | t1 >>  5 - 2 & 0b00011100 | t2 >>  5 + 1 & 0b00000011);
      mdst[adst + 5] = (byte) (t2 <<  7 - 5 & 0b10000000 | t3 >>  5 - 4 & 0b01110000);
      asrc += 2 * (16 / 4);
      adst += 2 * (12 / 4);
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull16x16toFull12x12Test4()

  //fntConvertFull16x16toFull12x12Test5 ()
  //  Full16x16を縮小してFull12x12を作る(テスト5)
  public static int[] fntTest5Map;
  public static void fntConvertFull16x16toFull12x12Test5 () {
    int[] map = fntTest5Map;
    if (map == null) {
      map = fntTest5Map = new int[65536];
      int[] o = new int[33];
      //  perl -e "use integer;for$i(0..32){printf'%d,',-($i>>1^-($i&1)),10}"
      //  0,1,-1,2,-2,3,-3,4,-4,5,-5,6,-6,7,-7,8,-8,9,-9,10,-10,11,-11,12,-12,13,-13,14,-14,15,-15,16,-16,
      for (int i = 0; i <= 32; i++) {
        o[i] = -(i >> 1 ^ - (i & 1));
      }
      for (int i = 0; i < 65536; i++) {
        //  0→1,1→0の変化点を列挙する。最初の0→1から最後の1→0まで最大で16箇所
        int x = i << 1 ^ i;  //変化点。マスクは17bit。変化点は偶数個なので必ず隙間がある
        //  変化点が14箇所以上あるとき中央に近い連続している2個で1→0,0→1であるものを省くことを繰り返して12箇所以内にする
        for (int j = 0, k = Integer.bitCount (x); 12 < k; j++) {
          int p = 8 + o[j];
          if ((~x & 3 << p) == 0 &&  //2個連続している
              (Integer.bitCount (x & (p - 1)) & 1) != 0) {  //これより右側に変化点が奇数個あるすなわちこの2個は1→0,0→1である
            x -= 3 << p;  //2個省く
            k -= 2;
          }
        }
        //  中央に近い変化点から3/4の位置に動かす。入らないときは最も近い隙間に入れる。順序が変わるが連続している変化点がずれるだけ
        int y = 0;
        for (int j = 0; j <= 16; j++) {
          int p = 8 + o[j];  //元の位置
          if ((x & 1 << p) != 0) {  //変化点がある
            int q = p * 3 + 2 >> 2;  //3/4に最も近い位置
            int s = q << 2 <= p * 3 ? 1 : -1;  //3/4に最も近い位置が3/4と等しいか小さいとき次の候補は左側、3/4よりも大きいとき次の候補は右側
            for (int k = 0; k <= 32; k++) {
              int r = q + s * o[k];  //3/4に近い位置
              if (0 <= r && r <= 12 && (y & 1 << r) == 0) {  //範囲内で隙間がある
                y |= 1 << r;  //変化点を入れる
                break;
              }
            }
          }
        }
        //  変化点を0→1,1→0に戻す
        x = 0;
        for (int q = 12; 0 < q; q--) {
          if ((y & 1 << q) != 0) {
            x ^= (1 << 4 << q) - 1;  //左に寄せる
          }
        }
        map[i] = x;
      }
    }
    int[] rev = fntReverseMap;
    if (rev == null) {
      rev = fntReverseMap = new int[65536];
      for (int i = 0; i < 65536; i++) {
        //  0123    048c
        //  4567 → 159d
        //  89ab    26ae
        //  cdef    37bf
        rev[i] = VideoController.VCN_TXP3[i >> 12] | VideoController.VCN_TXP2[i >> 8 & 15] | VideoController.VCN_TXP1[i >> 4 & 15] | VideoController.VCN_TXP0[i & 15];
        //System.out.printf ("%04x %04x\n", i, r[i]);
      }
    }
    FontPage src = fntPageFull16x16;
    FontPage dst = fntPageFull12x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w1 = new int[16];
    int[] w2 = new int[16];
    for (int i = 0; i < 94 * 77; i++) {
      if (false) {
        //とりあえず横方向だけ圧縮してみる
        for (int v = 0; v < 12; v++) {
          int t = map[(char) (msrc[asrc] << 8 | msrc[asrc + 1] & 255)];
          mdst[adst    ] = (byte) (t >> 8);
          mdst[adst + 1] = (byte) t;
          asrc += 2;
          adst += 2;
        }
        asrc += 2 * (16 - 12);
      } else if (false) {
        //横方向に圧縮してから縦方向に圧縮する
        //入力
        for (int v = 0; v < 16; v++) {
          w2[v] = (char) (msrc[asrc] << 8 | msrc[asrc + 1] & 255);
          asrc += 2;
        }
        //圧縮
        for (int v = 0; v < 16; v++) {
          w1[v] = map[w2[v]];
          w2[v] = 0;
        }
        //反転
        for (int v = 0; v < 16; v += 4) {
          for (int u = 0; u < 16; u += 4) {
            int t = rev[w1[u    ] << v       & 0xf000 |
                        w1[u + 1] << v >>  4 & 0x0f00 |
                        w1[u + 2] << v >>  8 & 0x00f0 |
                        w1[u + 3] << v >> 12 & 0x000f];
            w2[v    ] |= (t & 0xf000)       >> u;
            w2[v + 1] |= (t & 0x0f00) <<  4 >> u;
            w2[v + 2] |= (t & 0x00f0) <<  8 >> u;
            w2[v + 3] |= (t & 0x000f) << 12 >> u;
          }
        }
        //圧縮
        for (int v = 0; v < 16; v++) {
          w1[v] = map[w2[v]];
          w2[v] = 0;
        }
        //反転
        for (int v = 0; v < 16; v += 4) {
          for (int u = 0; u < 16; u += 4) {
            int t = rev[w1[u    ] << v       & 0xf000 |
                        w1[u + 1] << v >>  4 & 0x0f00 |
                        w1[u + 2] << v >>  8 & 0x00f0 |
                        w1[u + 3] << v >> 12 & 0x000f];
            w2[v    ] |= (t & 0xf000)       >> u;
            w2[v + 1] |= (t & 0x0f00) <<  4 >> u;
            w2[v + 2] |= (t & 0x00f0) <<  8 >> u;
            w2[v + 3] |= (t & 0x000f) << 12 >> u;
          }
        }
        //出力
        for (int v = 0; v < 12; v++) {
          int t = w2[v];
          mdst[adst    ] = (byte) (t >> 8);
          mdst[adst + 1] = (byte) t;
          adst += 2;
        }
      } else {
        //縦方向に圧縮してから横方向に圧縮する
        //入力
        for (int v = 0; v < 16; v++) {
          w1[v] = (char) (msrc[asrc] << 8 | msrc[asrc + 1] & 255);
          w2[v] = 0;
          asrc += 2;
        }
        //反転
        for (int v = 0; v < 16; v += 4) {
          for (int u = 0; u < 16; u += 4) {
            int t = rev[w1[u    ] << v       & 0xf000 |
                        w1[u + 1] << v >>  4 & 0x0f00 |
                        w1[u + 2] << v >>  8 & 0x00f0 |
                        w1[u + 3] << v >> 12 & 0x000f];
            w2[v    ] |= (t & 0xf000)       >> u;
            w2[v + 1] |= (t & 0x0f00) <<  4 >> u;
            w2[v + 2] |= (t & 0x00f0) <<  8 >> u;
            w2[v + 3] |= (t & 0x000f) << 12 >> u;
          }
        }
        //圧縮
        for (int v = 0; v < 16; v++) {
          w1[v] = map[w2[v]];
          w2[v] = 0;
        }
        //反転
        for (int v = 0; v < 16; v += 4) {
          for (int u = 0; u < 16; u += 4) {
            int t = rev[w1[u    ] << v       & 0xf000 |
                        w1[u + 1] << v >>  4 & 0x0f00 |
                        w1[u + 2] << v >>  8 & 0x00f0 |
                        w1[u + 3] << v >> 12 & 0x000f];
            w2[v    ] |= (t & 0xf000)       >> u;
            w2[v + 1] |= (t & 0x0f00) <<  4 >> u;
            w2[v + 2] |= (t & 0x00f0) <<  8 >> u;
            w2[v + 3] |= (t & 0x000f) << 12 >> u;
          }
        }
        //圧縮
        for (int v = 0; v < 16; v++) {
          w1[v] = map[w2[v]];
        }
        //出力
        for (int v = 0; v < 12; v++) {
          int t = w1[v];
          mdst[adst    ] = (byte) (t >> 8);
          mdst[adst + 1] = (byte) t;
          adst += 2;
        }
      }
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull16x16toFull12x12Test5()

  //fntConvertFull24x24toFull12x12IOCS ()
  //  Full24x24を縮小してFull12x12を作る(IOCS)
  //  ROM1.0  0x00ff6c90
  //  ROM1.1  0x00ff6fde
  //  ROM1.2  0x00ff70d6
  //  ROM1.3  0x00ff74b0
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    ■□■□■□■□ ■□■□
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■ → ■□■□■□■□ ■□■□
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    □■□■□■□■ □■□■
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■    ■□■□■□■□ ■□■□
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    □■□■□■□■ □■□■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□    ■□■□■□■□ ■□■□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■
  //  □□■■□□■■ □□■■□□■■ □□■■□□■■
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□
  //  ■■□□■■□□ ■■□□■■□□ ■■□□■■□□
  public static void fntConvertFull24x24toFull12x12IOCS () {
    FontPage src = fntPageFull24x24;
    FontPage dst = fntPageFull12x12;
    byte[] msrc = src.fnpBitmapArray;
    byte[] mdst = dst.fnpBitmapArray;
    int asrc = 0;
    int adst = 0;
    int[] w = new int[256];
    for (int i = 0; i < 256; i++) {
      w[i] = (i & 0b10000000) | (i & 0b01100000) << 1 | (i & 0b00011000) << 2 | (i & 0b00000110) << 3 | (i & 0b00000001) << 4;
    }
    for (int i = 0; i < 12 * 94 * 77; i++) {
      mdst[adst    ] = (byte) (w[(msrc[asrc    ] | msrc[asrc + 3]) & 255] | w[(msrc[asrc + 1] | msrc[asrc + 4]) & 255] >> 4);
      mdst[adst + 1] = (byte)  w[(msrc[asrc + 2] | msrc[asrc + 5]) & 255];
      asrc += 3 * (24 / 12);
      adst += 2 * (12 / 12);
    }
    dst.fnpMemoryToImage ();
    if (dst == fntSelectedPage) {
      fntCanvas.repaint ();
    }
  }  //fntConvertFull24x24toFull12x12IOCS()

  //! μEmacsの方法とKo-Windowの方法も確認する

*/


  //おむすびフォント
  //  XEiJの作者がXEiJのスクリーンショットに特徴を付けるために描いたANK8x16フォント
  //  三角のおむすびを意識して、上を狭く丸く、下を広く平らにした
  //  OmusubiのOに海苔を付けて同じ形の0と区別できるようにした(が、読み難い)
/*
  public static final byte[] FNT_OMUSUBI_DATA = {
    //  perl misc/sjdump.pl misc/omusubi.fon -code
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000000  ................
    0x00,0x00,0x38,0x40,0x38,0x04,0x78,0x00,0x22,0x22,0x3e,0x22,0x22,0x00,0x00,0x00,  //00000010  ..8@8.x."">""...
    0x00,0x00,0x38,0x40,0x38,0x04,0x78,0x00,0x22,0x14,0x08,0x14,0x22,0x00,0x00,0x00,  //00000020  ..8@8.x."..."...
    0x00,0x00,0x78,0x40,0x78,0x40,0x7c,0x00,0x22,0x14,0x08,0x14,0x22,0x00,0x00,0x00,  //00000030  ..x@x@|."..."...
    0x00,0x00,0x78,0x40,0x78,0x40,0x7c,0x00,0x3e,0x08,0x08,0x08,0x08,0x00,0x00,0x00,  //00000040  ..x@x@|.>.......
    0x00,0x00,0x78,0x40,0x78,0x40,0x7c,0x00,0x1c,0x22,0x2a,0x24,0x1a,0x00,0x00,0x00,  //00000050  ..x@x@|.."*$....
    0x00,0x00,0x38,0x44,0x7c,0x44,0x44,0x00,0x22,0x24,0x38,0x24,0x22,0x00,0x00,0x00,  //00000060  ..8D|DD."$8$"...
    0x00,0x00,0x78,0x44,0x78,0x44,0x78,0x00,0x20,0x20,0x20,0x20,0x3e,0x00,0x00,0x00,  //00000070  ..xDxDx.    >...
    0x00,0x00,0x78,0x44,0x78,0x44,0x78,0x00,0x1c,0x20,0x1c,0x02,0x3c,0x00,0x00,0x00,  //00000080  ..xDxDx.. ..<...
    0x00,0x00,0x44,0x44,0x7c,0x44,0x44,0x00,0x3e,0x08,0x08,0x08,0x08,0x00,0x00,0x00,  //00000090  ..DD|DD.>.......
    0x00,0x00,0x40,0x40,0x40,0x40,0x7c,0x00,0x3c,0x20,0x3c,0x20,0x20,0x00,0x00,0x00,  //000000a0  ..@@@@|.< <  ...
    0x00,0x00,0x44,0x44,0x44,0x28,0x10,0x00,0x3e,0x08,0x08,0x08,0x08,0x00,0x00,0x00,  //000000b0  ..DDD(..>.......
    0x00,0x00,0x78,0x40,0x78,0x40,0x40,0x00,0x3c,0x20,0x3c,0x20,0x20,0x00,0x00,0x00,  //000000c0  ..x@x@@.< <  ...
    0x00,0x00,0x38,0x40,0x40,0x40,0x3c,0x00,0x3c,0x22,0x3c,0x24,0x22,0x00,0x00,0x00,  //000000d0  ..8@@@<.<"<$"...
    0x00,0x00,0x38,0x40,0x38,0x04,0x78,0x00,0x1c,0x22,0x22,0x22,0x1c,0x00,0x00,0x00,  //000000e0  ..8@8.x.."""....
    0x00,0x00,0x38,0x40,0x38,0x04,0x78,0x00,0x1c,0x08,0x08,0x08,0x1c,0x00,0x00,0x00,  //000000f0  ..8@8.x.........
    0x00,0x00,0x78,0x44,0x44,0x44,0x78,0x00,0x3c,0x20,0x3c,0x20,0x3e,0x00,0x00,0x00,  //00000100  ..xDDDx.< < >...
    0x00,0x00,0x78,0x44,0x44,0x44,0x78,0x00,0x18,0x08,0x08,0x08,0x08,0x00,0x00,0x00,  //00000110  ..xDDDx.........
    0x00,0x00,0x78,0x44,0x44,0x44,0x78,0x00,0x1c,0x02,0x1c,0x20,0x3e,0x00,0x00,0x00,  //00000120  ..xDDDx.... >...
    0x00,0x00,0x78,0x44,0x44,0x44,0x78,0x00,0x1c,0x02,0x1c,0x02,0x3c,0x00,0x00,0x00,  //00000130  ..xDDDx.....<...
    0x00,0x00,0x78,0x44,0x44,0x44,0x78,0x00,0x04,0x0c,0x14,0x3e,0x04,0x00,0x00,0x00,  //00000140  ..xDDDx....>....
    0x00,0x00,0x44,0x64,0x54,0x4c,0x44,0x00,0x22,0x24,0x38,0x24,0x22,0x00,0x00,0x00,  //00000150  ..DdTLD."$8$"...
    0x00,0x00,0x38,0x40,0x38,0x04,0x78,0x00,0x22,0x32,0x2a,0x26,0x22,0x00,0x00,0x00,  //00000160  ..8@8.x."2*&"...
    0x00,0x00,0x78,0x40,0x78,0x40,0x7c,0x00,0x3c,0x22,0x3c,0x22,0x3c,0x00,0x00,0x00,  //00000170  ..x@x@|.<"<"<...
    0x00,0x00,0x38,0x40,0x40,0x40,0x3c,0x00,0x22,0x32,0x2a,0x26,0x22,0x00,0x00,0x00,  //00000180  ..8@@@<."2*&"...
    0x00,0x00,0x78,0x40,0x78,0x40,0x7c,0x00,0x22,0x36,0x2a,0x22,0x22,0x00,0x00,0x00,  //00000190  ..x@x@|."6*""...
    0x00,0x00,0x38,0x40,0x38,0x04,0x78,0x00,0x3c,0x22,0x3c,0x22,0x3c,0x00,0x00,0x00,  //000001a0  ..8@8.x.<"<"<...
    0x00,0x00,0x78,0x40,0x78,0x40,0x7c,0x00,0x1c,0x20,0x20,0x20,0x1e,0x00,0x00,0x00,  //000001b0  ..x@x@|..   ....
    0x00,0x00,0x00,0x00,0x10,0x08,0x04,0xfe,0x04,0x08,0x10,0x00,0x00,0x00,0x00,0x00,  //000001c0  ................
    0x00,0x00,0x00,0x00,0x10,0x20,0x40,0xfe,0x40,0x20,0x10,0x00,0x00,0x00,0x00,0x00,  //000001d0  ..... @.@ ......
    0x00,0x00,0x00,0x00,0x10,0x38,0x54,0x92,0x10,0x10,0x10,0x00,0x00,0x00,0x00,0x00,  //000001e0  .....8T.........
    0x00,0x00,0x00,0x00,0x10,0x10,0x10,0x92,0x54,0x38,0x10,0x00,0x00,0x00,0x00,0x00,  //000001f0  ........T8......
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000200  ................
    0x00,0x00,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x00,0x10,0x10,0x00,0x00,0x00,  //00000210  ................
    0x00,0x24,0x6c,0xd8,0x90,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000220  .$l.............
    0x00,0x00,0x44,0x44,0x44,0xfe,0x44,0x44,0x44,0xfe,0x44,0x44,0x44,0x00,0x00,0x00,  //00000230  ..DDD.DDD.DDD...
    0x00,0x00,0x7c,0x90,0x90,0x90,0x78,0x14,0x12,0x12,0x12,0x12,0xfc,0x00,0x00,0x00,  //00000240  ..|...x.........
    0x00,0x00,0x62,0x92,0x94,0x68,0x08,0x10,0x20,0x2c,0x52,0x92,0x8c,0x00,0x00,0x00,  //00000250  ..b..h.. ,R.....
    0x00,0x00,0x70,0x88,0x88,0x50,0x20,0x52,0x8a,0x84,0x84,0x8a,0x72,0x00,0x00,0x00,  //00000260  ..p..P R....r...
    0x00,0x08,0x18,0x30,0x20,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000270  ...0 ...........
    0x00,0x08,0x10,0x10,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x10,0x10,0x08,0x00,0x00,  //00000280  ....       .....
    0x00,0x20,0x10,0x10,0x08,0x08,0x08,0x08,0x08,0x08,0x08,0x10,0x10,0x20,0x00,0x00,  //00000290  . ........... ..
    0x00,0x00,0x00,0x00,0x10,0x92,0x54,0x38,0x54,0x92,0x10,0x00,0x00,0x00,0x00,0x00,  //000002a0  ......T8T.......
    0x00,0x00,0x00,0x00,0x10,0x10,0x10,0xfe,0x10,0x10,0x10,0x00,0x00,0x00,0x00,0x00,  //000002b0  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x08,0x18,0x30,0x20,0x00,0x00,  //000002c0  ............0 ..
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xfe,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //000002d0  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x10,0x38,0x10,0x00,0x00,0x00,  //000002e0  ...........8....
    0x00,0x00,0x02,0x02,0x04,0x08,0x08,0x10,0x20,0x20,0x40,0x80,0x80,0x00,0x00,0x00,  //000002f0  ........  @.....
    0x00,0x00,0x38,0x44,0x44,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x7c,0x00,0x00,0x00,  //00000300  ..8DD.......|...
    0x00,0x00,0x70,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x00,0x00,0x00,  //00000310  ..p.............
    0x00,0x00,0x78,0x04,0x04,0x04,0x38,0x40,0x80,0x80,0x80,0x80,0xfe,0x00,0x00,0x00,  //00000320  ..x...8@........
    0x00,0x00,0x78,0x04,0x04,0x04,0x78,0x04,0x02,0x02,0x02,0x02,0xfc,0x00,0x00,0x00,  //00000330  ..x...x.........
    0x00,0x00,0x0c,0x14,0x24,0x44,0x44,0x84,0x84,0x84,0xfe,0x04,0x04,0x00,0x00,0x00,  //00000340  ....$DD.........
    0x00,0x00,0x7c,0x40,0x40,0x40,0x78,0x04,0x02,0x02,0x02,0x02,0xfc,0x00,0x00,0x00,  //00000350  ..|@@@x.........
    0x00,0x00,0x38,0x40,0x40,0x80,0xb8,0xc4,0x82,0x82,0x82,0x82,0x7c,0x00,0x00,0x00,  //00000360  ..8@@.......|...
    0x00,0x00,0x7e,0x02,0x04,0x04,0x08,0x08,0x08,0x10,0x10,0x10,0x10,0x00,0x00,0x00,  //00000370  ..~.............
    0x00,0x00,0x38,0x44,0x44,0x44,0x38,0x44,0x82,0x82,0x82,0x82,0x7c,0x00,0x00,0x00,  //00000380  ..8DDD8D....|...
    0x00,0x00,0x38,0x44,0x82,0x82,0x82,0x82,0x7e,0x02,0x04,0x04,0x38,0x00,0x00,0x00,  //00000390  ..8D....~...8...
    0x00,0x00,0x00,0x10,0x38,0x10,0x00,0x00,0x00,0x10,0x38,0x10,0x00,0x00,0x00,0x00,  //000003a0  ....8.....8.....
    0x00,0x00,0x00,0x10,0x38,0x10,0x00,0x00,0x00,0x00,0x08,0x18,0x30,0x20,0x00,0x00,  //000003b0  ....8.......0 ..
    0x00,0x00,0x02,0x04,0x08,0x10,0x20,0x40,0x20,0x10,0x08,0x04,0x02,0x00,0x00,0x00,  //000003c0  ...... @ .......
    0x00,0x00,0x00,0x00,0x00,0xfe,0x00,0x00,0x00,0xfe,0x00,0x00,0x00,0x00,0x00,0x00,  //000003d0  ................
    0x00,0x00,0x80,0x40,0x20,0x10,0x08,0x04,0x08,0x10,0x20,0x40,0x80,0x00,0x00,0x00,  //000003e0  ...@ ..... @....
    0x00,0x00,0xfc,0x02,0x02,0x02,0x04,0x08,0x10,0x10,0x00,0x10,0x10,0x00,0x00,0x00,  //000003f0  ................
    0x00,0x00,0x38,0x44,0x44,0x82,0x9e,0xa2,0xa2,0xa2,0x9e,0x80,0x7e,0x00,0x00,0x00,  //00000400  ..8DD.......~...
    0x00,0x00,0x38,0x44,0x44,0x82,0xfe,0x82,0x82,0x82,0x82,0x82,0x82,0x00,0x00,0x00,  //00000410  ..8DD...........
    0x00,0x00,0xf8,0x84,0x84,0x84,0xf8,0x84,0x82,0x82,0x82,0x82,0xfc,0x00,0x00,0x00,  //00000420  ................
    0x00,0x00,0x3c,0x40,0x40,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x7e,0x00,0x00,0x00,  //00000430  ..<@@.......~...
    0x00,0x00,0xf8,0x84,0x84,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0xfc,0x00,0x00,0x00,  //00000440  ................
    0x00,0x00,0xfc,0x80,0x80,0x80,0xfc,0x80,0x80,0x80,0x80,0x80,0xfe,0x00,0x00,0x00,  //00000450  ................
    0x00,0x00,0xfc,0x80,0x80,0x80,0xfc,0x80,0x80,0x80,0x80,0x80,0x80,0x00,0x00,0x00,  //00000460  ................
    0x00,0x00,0x3c,0x40,0x40,0x80,0x8e,0x82,0x82,0x82,0x82,0x82,0x7c,0x00,0x00,0x00,  //00000470  ..<@@.......|...
    0x00,0x00,0x82,0x82,0x82,0x82,0xfe,0x82,0x82,0x82,0x82,0x82,0x82,0x00,0x00,0x00,  //00000480  ................
    0x00,0x00,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x00,0x00,0x00,  //00000490  ................
    0x00,0x00,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x82,0x44,0x38,0x00,0x00,0x00,  //000004a0  ...........D8...
    0x00,0x00,0x84,0x88,0x90,0xa0,0xe0,0x90,0x88,0x84,0x84,0x82,0x82,0x00,0x00,0x00,  //000004b0  ................
    0x00,0x00,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0x80,0xfe,0x00,0x00,0x00,  //000004c0  ................
    0x00,0x00,0x82,0x82,0xc6,0xc6,0xaa,0xaa,0xaa,0x92,0x92,0x92,0x82,0x00,0x00,0x00,  //000004d0  ................
    0x00,0x00,0x82,0xc2,0xa2,0x92,0x8a,0x8a,0x86,0x86,0x82,0x82,0x82,0x00,0x00,0x00,  //000004e0  ................
    0x00,0x00,0x38,0x44,0x44,0x82,0x82,0x82,0xaa,0x82,0xaa,0x82,0x7c,0x00,0x00,0x00,  //000004f0  ..8DD.......|...
    0x00,0x00,0xfc,0x82,0x82,0x82,0xfc,0x80,0x80,0x80,0x80,0x80,0x80,0x00,0x00,0x00,  //00000500  ................
    0x00,0x00,0x38,0x44,0x44,0x82,0x82,0x82,0x82,0xb2,0xca,0x84,0x7a,0x00,0x00,0x00,  //00000510  ..8DD.......z...
    0x00,0x00,0xf8,0x84,0x84,0x84,0xf8,0x90,0x88,0x84,0x84,0x82,0x82,0x00,0x00,0x00,  //00000520  ................
    0x00,0x00,0x7c,0x80,0x80,0x80,0x78,0x04,0x02,0x02,0x02,0x02,0xfc,0x00,0x00,0x00,  //00000530  ..|...x.........
    0x00,0x00,0xfe,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x00,0x00,0x00,  //00000540  ................
    0x00,0x00,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x7c,0x00,0x00,0x00,  //00000550  ............|...
    0x00,0x00,0x82,0x82,0x82,0x82,0x44,0x44,0x44,0x28,0x28,0x10,0x10,0x00,0x00,0x00,  //00000560  ......DDD((.....
    0x00,0x00,0x82,0x92,0x92,0x92,0xaa,0xaa,0xaa,0x44,0x44,0x44,0x44,0x00,0x00,0x00,  //00000570  .........DDDD...
    0x00,0x00,0x82,0x44,0x28,0x10,0x28,0x28,0x44,0x44,0x82,0x82,0x82,0x00,0x00,0x00,  //00000580  ...D(.((DD......
    0x00,0x00,0x82,0x44,0x28,0x28,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x00,0x00,0x00,  //00000590  ...D((..........
    0x00,0x00,0x7c,0x04,0x08,0x08,0x10,0x20,0x20,0x40,0x40,0x80,0xfe,0x00,0x00,0x00,  //000005a0  ..|....  @@.....
    0x00,0x00,0x3c,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x20,0x3c,0x00,0x00,0x00,  //000005b0  ..<         <...
    0x00,0x00,0x44,0x44,0x28,0x28,0x10,0xfe,0x10,0xfe,0x10,0x10,0x10,0x00,0x00,0x00,  //000005c0  ..DD((..........
    0x00,0x00,0x78,0x08,0x08,0x08,0x08,0x08,0x08,0x08,0x08,0x08,0x78,0x00,0x00,0x00,  //000005d0  ..x.........x...
    0x00,0x00,0x10,0x28,0x44,0x82,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //000005e0  ...(D...........
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xfe,0x00,0x00,0x00,  //000005f0  ................
    0x00,0x20,0x30,0x18,0x08,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000600  . 0.............
    0x00,0x00,0x00,0x00,0x00,0x00,0x7c,0x02,0x02,0x7e,0x82,0x82,0x7e,0x00,0x00,0x00,  //00000610  ......|..~..~...
    0x00,0x00,0x80,0x80,0x80,0x80,0xb8,0xc4,0x82,0x82,0x82,0x82,0xfc,0x00,0x00,0x00,  //00000620  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x3c,0x40,0x80,0x80,0x80,0x80,0x7e,0x00,0x00,0x00,  //00000630  ......<@....~...
    0x00,0x00,0x02,0x02,0x02,0x02,0x3a,0x46,0x82,0x82,0x82,0x82,0x7e,0x00,0x00,0x00,  //00000640  ......:F....~...
    0x00,0x00,0x00,0x00,0x00,0x00,0x7c,0x82,0x82,0xfe,0x80,0x80,0x7e,0x00,0x00,0x00,  //00000650  ......|.....~...
    0x00,0x00,0x1c,0x20,0x20,0x20,0xfc,0x20,0x20,0x20,0x20,0x20,0x20,0x00,0x00,0x00,  //00000660  ...   .      ...
    0x00,0x00,0x00,0x00,0x00,0x00,0x3a,0x46,0x82,0x82,0x86,0x7a,0x02,0x04,0x78,0x00,  //00000670  ......:F...z..x.
    0x00,0x00,0x80,0x80,0x80,0x80,0xb8,0xc4,0x82,0x82,0x82,0x82,0x82,0x00,0x00,0x00,  //00000680  ................
    0x00,0x00,0x10,0x10,0x00,0x00,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x00,0x00,0x00,  //00000690  ................
    0x00,0x00,0x08,0x08,0x00,0x00,0x08,0x08,0x08,0x08,0x08,0x08,0x08,0x10,0xe0,0x00,  //000006a0  ................
    0x00,0x00,0x40,0x40,0x40,0x40,0x46,0x48,0x50,0x68,0x44,0x42,0x42,0x00,0x00,0x00,  //000006b0  ..@@@@FHPhDBB...
    0x00,0x00,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x1c,0x00,0x00,0x00,  //000006c0  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0xec,0x92,0x92,0x92,0x92,0x92,0x92,0x00,0x00,0x00,  //000006d0  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0xbc,0xc2,0x82,0x82,0x82,0x82,0x82,0x00,0x00,0x00,  //000006e0  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x38,0x44,0x82,0x82,0x82,0x82,0x7c,0x00,0x00,0x00,  //000006f0  ......8D....|...
    0x00,0x00,0x00,0x00,0x00,0x00,0xb8,0xc4,0x82,0x82,0x82,0x82,0xfc,0x80,0x80,0x80,  //00000700  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x3a,0x46,0x82,0x82,0x82,0x82,0x7e,0x02,0x02,0x02,  //00000710  ......:F....~...
    0x00,0x00,0x00,0x00,0x00,0x00,0x5e,0x60,0x40,0x40,0x40,0x40,0x40,0x00,0x00,0x00,  //00000720  ......^`@@@@@...
    0x00,0x00,0x00,0x00,0x00,0x00,0x7c,0x80,0x80,0x7c,0x02,0x02,0xfc,0x00,0x00,0x00,  //00000730  ......|..|......
    0x00,0x00,0x20,0x20,0x20,0x20,0xfc,0x20,0x20,0x20,0x20,0x20,0x1c,0x00,0x00,0x00,  //00000740  ..    .     ....
    0x00,0x00,0x00,0x00,0x00,0x00,0x82,0x82,0x82,0x82,0x82,0x86,0x7a,0x00,0x00,0x00,  //00000750  ............z...
    0x00,0x00,0x00,0x00,0x00,0x00,0x82,0x82,0x82,0x44,0x44,0x28,0x10,0x00,0x00,0x00,  //00000760  .........DD(....
    0x00,0x00,0x00,0x00,0x00,0x00,0x82,0x92,0x92,0xaa,0xaa,0x44,0x44,0x00,0x00,0x00,  //00000770  ...........DD...
    0x00,0x00,0x00,0x00,0x00,0x00,0xc6,0x28,0x10,0x28,0x44,0x82,0x82,0x00,0x00,0x00,  //00000780  .......(.(D.....
    0x00,0x00,0x00,0x00,0x00,0x00,0x82,0x82,0x82,0x82,0x86,0x7a,0x02,0x04,0x78,0x00,  //00000790  ...........z..x.
    0x00,0x00,0x00,0x00,0x00,0x00,0x7c,0x04,0x08,0x10,0x20,0x40,0xfe,0x00,0x00,0x00,  //000007a0  ......|... @....
    0x00,0x18,0x20,0x20,0x20,0x20,0x20,0xc0,0x20,0x20,0x20,0x20,0x20,0x18,0x00,0x00,  //000007b0  ..     .     ...
    0x00,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x00,0x00,  //000007c0  ................
    0x00,0x30,0x08,0x08,0x08,0x08,0x08,0x06,0x08,0x08,0x08,0x08,0x08,0x30,0x00,0x00,  //000007d0  .0...........0..
    0x00,0x00,0xfe,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //000007e0  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //000007f0  ................
    0x00,0x00,0x80,0x80,0x40,0x20,0x20,0x10,0x08,0x08,0x04,0x02,0x02,0x00,0x00,0x00,  //00000800  ....@  .........
    0x00,0x00,0x60,0x92,0x0c,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000810  ..`.............
    0x00,0x10,0x10,0x10,0x10,0x10,0x10,0x00,0x10,0x10,0x10,0x10,0x10,0x10,0x00,0x00,  //00000820  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000830  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000840  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000850  ................
    0x00,0x00,0x10,0x10,0x78,0x20,0x4c,0x30,0x48,0x84,0x80,0x80,0x7e,0x00,0x00,0x00,  //00000860  ....x L0H...~...
    0x00,0x00,0x00,0x00,0x20,0x20,0x78,0x20,0x7c,0xaa,0x92,0xa2,0x44,0x00,0x00,0x00,  //00000870  ....  x |...D...
    0x00,0x00,0x00,0x00,0x00,0x00,0x44,0x82,0x82,0x82,0x82,0xa2,0x40,0x00,0x00,0x00,  //00000880  ......D.....@...
    0x00,0x00,0x00,0x00,0x30,0x00,0x78,0x04,0x02,0x02,0x02,0x04,0x38,0x00,0x00,0x00,  //00000890  ....0.x.....8...
    0x00,0x00,0x00,0x00,0x70,0x00,0xf8,0x10,0x20,0x50,0x92,0x92,0x8c,0x00,0x00,0x00,  //000008a0  ....p... P......
    0x00,0x00,0x00,0x00,0x20,0x24,0x72,0x20,0x7c,0xa2,0xa2,0xa2,0x64,0x00,0x00,0x00,  //000008b0  .... $r |...d...
    0x00,0x00,0x00,0x00,0x48,0x48,0x5c,0xe2,0x42,0x4c,0x20,0x20,0x20,0x00,0x00,0x00,  //000008c0  ....HH\.BL   ...
    0x00,0x00,0x00,0x00,0x10,0x88,0xbc,0xca,0x8a,0x8a,0xbc,0x08,0x10,0x00,0x00,0x00,  //000008d0  ................
    0x00,0x00,0x00,0x00,0x10,0x10,0x1c,0x10,0x10,0x78,0x94,0x92,0x60,0x00,0x00,0x00,  //000008e0  .........x..`...
    0x00,0x00,0x00,0x00,0x00,0xf8,0x04,0x02,0x02,0x02,0x04,0x18,0x60,0x00,0x00,0x00,  //000008f0  ............`...
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000900  ................
    0x00,0x00,0x20,0x20,0x78,0x20,0x7c,0xaa,0xaa,0x92,0x92,0xa2,0x44,0x00,0x00,0x00,  //00000910  ..  x |.....D...
    0x00,0x00,0x00,0x44,0x44,0x82,0x82,0x82,0x82,0x82,0x82,0xa2,0x40,0x00,0x00,0x00,  //00000920  ...DD.......@...
    0x00,0x00,0x70,0x00,0x00,0xf8,0x04,0x02,0x02,0x02,0x04,0x18,0x60,0x00,0x00,0x00,  //00000930  ..p.........`...
    0x00,0x00,0x70,0x00,0xf8,0x10,0x20,0x40,0x60,0x90,0x92,0x92,0x8c,0x00,0x00,0x00,  //00000940  ..p... @`.......
    0x00,0x00,0x20,0x24,0x72,0x20,0x20,0x7c,0xa2,0xa2,0xa2,0xa2,0x64,0x00,0x00,0x00,  //00000950  .. $r  |....d...
    0x00,0x00,0x24,0x24,0x22,0xf2,0x28,0x24,0x44,0x44,0x44,0x84,0x98,0x00,0x00,0x00,  //00000960  ..$$".($DDD.....
    0x00,0x00,0x20,0x10,0x7c,0x08,0x3e,0x04,0x3e,0x40,0x80,0x80,0x7e,0x00,0x00,0x00,  //00000970  .. .|.>.>@..~...
    0x00,0x00,0x04,0x04,0x08,0x10,0x20,0x40,0x20,0x10,0x08,0x04,0x04,0x00,0x00,0x00,  //00000980  ...... @ .......
    0x00,0x00,0x84,0x84,0xbe,0x84,0x84,0x84,0x84,0x84,0x88,0x88,0x90,0x00,0x00,0x00,  //00000990  ................
    0x00,0x00,0x00,0x7c,0x04,0x00,0x00,0x00,0x00,0x00,0x80,0x80,0x7e,0x00,0x00,0x00,  //000009a0  ...|........~...
    0x00,0x00,0x20,0x10,0x7c,0x08,0x04,0x3e,0x40,0x80,0x80,0x80,0x7e,0x00,0x00,0x00,  //000009b0  .. .|..>@...~...
    0x00,0x00,0x00,0x80,0x80,0x80,0x80,0x82,0x82,0x82,0x82,0x44,0x38,0x00,0x00,0x00,  //000009c0  ...........D8...
    0x00,0x00,0x04,0x04,0xfe,0x04,0x34,0x4c,0x84,0x8c,0x74,0x04,0x38,0x00,0x00,0x00,  //000009d0  ......4L..t.8...
    0x00,0x00,0x44,0x44,0xfe,0x44,0x44,0x44,0x44,0x40,0x40,0x40,0x3c,0x00,0x00,0x00,  //000009e0  ..DD.DDDD@@@<...
    0x00,0x00,0x7c,0x08,0x10,0x20,0xfe,0x18,0x20,0x40,0x40,0x40,0x3c,0x00,0x00,0x00,  //000009f0  ..|.. .. @@@<...
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000a00  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x30,0x48,0x48,0x30,0x00,0x00,  //00000a10  ..........0HH0..
    0x00,0x1e,0x10,0x10,0x10,0x10,0x10,0x10,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000a20  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x08,0x08,0x08,0x08,0x08,0x08,0x78,0x00,0x00,  //00000a30  .............x..
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x20,0x30,0x18,0x08,0x00,0x00,  //00000a40  .......... 0....
    0x00,0x00,0x00,0x00,0x00,0x00,0x10,0x38,0x10,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000a50  .......8........
    0x00,0x00,0x00,0xfe,0x02,0x02,0xfe,0x02,0x02,0x02,0x02,0x04,0x38,0x00,0x00,0x00,  //00000a60  ............8...
    0x00,0x00,0x00,0x00,0x00,0x00,0xfe,0x02,0x22,0x24,0x38,0x20,0xc0,0x00,0x00,0x00,  //00000a70  ........"$8 ....
    0x00,0x00,0x00,0x00,0x00,0x00,0x02,0x04,0x08,0x38,0xc8,0x08,0x08,0x00,0x00,0x00,  //00000a80  .........8......
    0x00,0x00,0x00,0x00,0x10,0x10,0xfe,0x82,0x82,0x02,0x02,0x04,0x38,0x00,0x00,0x00,  //00000a90  ............8...
    0x00,0x00,0x00,0x00,0x00,0x00,0x7c,0x10,0x10,0x10,0x10,0x10,0xfe,0x00,0x00,0x00,  //00000aa0  ......|.........
    0x00,0x00,0x00,0x00,0x04,0x04,0xfe,0x04,0x0c,0x14,0x24,0x44,0x84,0x00,0x00,0x00,  //00000ab0  ..........$D....
    0x00,0x00,0x00,0x00,0x20,0x20,0xfe,0x22,0x24,0x28,0x20,0x20,0x20,0x00,0x00,0x00,  //00000ac0  ....  ."$(   ...
    0x00,0x00,0x00,0x00,0x00,0x00,0x78,0x08,0x08,0x08,0x08,0x08,0xfe,0x00,0x00,0x00,  //00000ad0  ......x.........
    0x00,0x00,0x00,0x00,0x00,0x00,0x7e,0x02,0x02,0x7e,0x02,0x02,0xfe,0x00,0x00,0x00,  //00000ae0  ......~..~......
    0x00,0x00,0x00,0x00,0x00,0x00,0x92,0x92,0x92,0x02,0x02,0x04,0x18,0x00,0x00,0x00,  //00000af0  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x7c,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000b00  .......|........
    0x00,0x00,0x00,0xfe,0x02,0x02,0x22,0x24,0x38,0x20,0x20,0x20,0xc0,0x00,0x00,0x00,  //00000b10  ......"$8   ....
    0x00,0x00,0x00,0x02,0x04,0x08,0x38,0xc8,0x08,0x08,0x08,0x08,0x08,0x00,0x00,0x00,  //00000b20  ......8.........
    0x00,0x00,0x10,0x10,0xfe,0x82,0x82,0x82,0x02,0x02,0x02,0x04,0x18,0x00,0x00,0x00,  //00000b30  ................
    0x00,0x00,0x00,0x7c,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0x10,0xfe,0x00,0x00,0x00,  //00000b40  ...|............
    0x00,0x00,0x04,0x04,0xfe,0x04,0x04,0x0c,0x14,0x24,0x44,0x84,0x04,0x00,0x00,0x00,  //00000b50  .........$D.....
    0x00,0x00,0x20,0x20,0xfe,0x22,0x22,0x22,0x42,0x42,0x42,0x82,0x9e,0x00,0x00,0x00,  //00000b60  ..  ."""BBB.....
    0x00,0x00,0x10,0x10,0xfe,0x10,0x10,0xfe,0x10,0x10,0x10,0x10,0x10,0x00,0x00,0x00,  //00000b70  ................
    0x00,0x00,0x00,0x7e,0x42,0x42,0x42,0x42,0x82,0x02,0x02,0x04,0x18,0x00,0x00,0x00,  //00000b80  ...~BBBB........
    0x00,0x00,0x40,0x40,0x7e,0x88,0x88,0x88,0x08,0x08,0x08,0x10,0x60,0x00,0x00,0x00,  //00000b90  ..@@~.......`...
    0x00,0x00,0x00,0x7e,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0x02,0xfe,0x00,0x00,0x00,  //00000ba0  ...~............
    0x00,0x00,0x44,0x44,0xfe,0x44,0x44,0x04,0x04,0x04,0x04,0x08,0x70,0x00,0x00,0x00,  //00000bb0  ..DD.DD.....p...
    0x00,0x00,0x00,0xf0,0x02,0x02,0xf2,0x02,0x02,0x02,0x02,0x04,0xf8,0x00,0x00,0x00,  //00000bc0  ................
    0x00,0x00,0x00,0x7e,0x02,0x02,0x02,0x04,0x08,0x14,0x22,0x42,0x82,0x00,0x00,0x00,  //00000bd0  ...~......"B....
    0x00,0x00,0x40,0x40,0xfe,0x42,0x44,0x48,0x40,0x40,0x40,0x40,0x3e,0x00,0x00,0x00,  //00000be0  ..@@.BDH@@@@>...
    0x00,0x00,0x00,0x82,0x42,0x22,0x22,0x02,0x04,0x04,0x08,0x10,0x60,0x00,0x00,0x00,  //00000bf0  ....B"".....`...
    0x00,0x00,0x00,0x7e,0x42,0x42,0x7e,0x42,0x82,0x02,0x02,0x04,0x18,0x00,0x00,0x00,  //00000c00  ...~BB~B........
    0x00,0x00,0x00,0x7c,0x08,0x08,0xfe,0x08,0x08,0x08,0x08,0x10,0x60,0x00,0x00,0x00,  //00000c10  ...|........`...
    0x00,0x00,0x00,0x92,0x92,0x92,0x92,0x92,0x02,0x02,0x02,0x04,0x18,0x00,0x00,0x00,  //00000c20  ................
    0x00,0x00,0x00,0x7c,0x00,0x00,0xfe,0x08,0x08,0x08,0x08,0x10,0x60,0x00,0x00,0x00,  //00000c30  ...|........`...
    0x00,0x00,0x00,0x40,0x40,0x40,0x70,0x4c,0x42,0x40,0x40,0x40,0x40,0x00,0x00,0x00,  //00000c40  ...@@@pLB@@@@...
    0x00,0x00,0x10,0x10,0xfe,0x10,0x10,0x10,0x10,0x10,0x10,0x20,0xc0,0x00,0x00,0x00,  //00000c50  ........... ....
    0x00,0x00,0x00,0x00,0x7c,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xfe,0x00,0x00,0x00,  //00000c60  ....|...........
    0x00,0x00,0x00,0x7e,0x02,0x02,0x02,0x24,0x14,0x08,0x14,0x22,0xc0,0x00,0x00,0x00,  //00000c70  ...~...$..."....
    0x00,0x00,0x10,0x10,0xfe,0x04,0x08,0x10,0x38,0x54,0x92,0x10,0x10,0x00,0x00,0x00,  //00000c80  ........8T......
    0x00,0x00,0x00,0x02,0x02,0x02,0x02,0x04,0x04,0x08,0x10,0x20,0xc0,0x00,0x00,0x00,  //00000c90  ........... ....
    0x00,0x00,0x00,0x44,0x42,0x42,0x42,0x42,0x42,0x42,0x42,0x82,0x82,0x00,0x00,0x00,  //00000ca0  ...DBBBBBBB.....
    0x00,0x00,0x80,0x80,0x84,0x98,0xe0,0x80,0x80,0x80,0x80,0x80,0x7e,0x00,0x00,0x00,  //00000cb0  ............~...
    0x00,0x00,0x00,0xfe,0x02,0x02,0x02,0x02,0x02,0x04,0x04,0x08,0x30,0x00,0x00,0x00,  //00000cc0  ............0...
    0x00,0x00,0x00,0x20,0x20,0x50,0x50,0x88,0x88,0x04,0x04,0x02,0x02,0x00,0x00,0x00,  //00000cd0  ...  PP.........
    0x00,0x00,0x10,0x10,0xfe,0x10,0x10,0x54,0x54,0x92,0x92,0x92,0x92,0x00,0x00,0x00,  //00000ce0  .......TT.......
    0x00,0x00,0x00,0xfe,0x02,0x02,0x02,0x02,0x44,0x28,0x10,0x08,0x04,0x00,0x00,0x00,  //00000cf0  ........D(......
    0x00,0x00,0x00,0x7c,0x00,0x00,0x7c,0x00,0x00,0x00,0x00,0x00,0xfe,0x00,0x00,0x00,  //00000d00  ...|..|.........
    0x00,0x00,0x00,0x38,0x40,0x80,0x80,0x82,0x82,0x82,0x82,0x82,0xfe,0x00,0x00,0x00,  //00000d10  ...8@...........
    0x00,0x00,0x02,0x02,0x02,0x02,0x22,0x14,0x08,0x14,0x22,0x40,0x80,0x00,0x00,0x00,  //00000d20  ......"..."@....
    0x00,0x00,0x00,0x7c,0x20,0x20,0xfe,0x20,0x20,0x20,0x20,0x20,0x1e,0x00,0x00,0x00,  //00000d30  ...|  .     ....
    0x00,0x00,0x40,0x40,0xfe,0x42,0x42,0x42,0x44,0x48,0x40,0x40,0x40,0x00,0x00,0x00,  //00000d40  ..@@.BBBDH@@@...
    0x00,0x00,0x00,0x78,0x08,0x08,0x08,0x08,0x08,0x08,0x08,0x08,0xfe,0x00,0x00,0x00,  //00000d50  ...x............
    0x00,0x00,0x00,0x7e,0x02,0x02,0x7e,0x02,0x02,0x02,0x02,0x02,0xfe,0x00,0x00,0x00,  //00000d60  ...~..~.........
    0x00,0x00,0x00,0x7c,0x00,0x00,0xfe,0x02,0x02,0x02,0x02,0x04,0x38,0x00,0x00,0x00,  //00000d70  ...|........8...
    0x00,0x00,0x00,0x82,0x82,0x82,0x82,0x82,0x02,0x02,0x02,0x04,0x38,0x00,0x00,0x00,  //00000d80  ............8...
    0x00,0x00,0x00,0x50,0x50,0x50,0x50,0x50,0x50,0x52,0x52,0x94,0x98,0x00,0x00,0x00,  //00000d90  ...PPPPPPRR.....
    0x00,0x00,0x00,0x80,0x80,0x80,0x80,0x80,0x80,0x82,0x84,0x98,0xe0,0x00,0x00,0x00,  //00000da0  ................
    0x00,0x00,0x00,0xfe,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0x82,0xfe,0x00,0x00,0x00,  //00000db0  ................
    0x00,0x00,0x00,0xfe,0x82,0x82,0x82,0x02,0x02,0x02,0x02,0x04,0x38,0x00,0x00,0x00,  //00000dc0  ............8...
    0x00,0x00,0x00,0xf0,0x00,0x02,0x02,0x02,0x02,0x02,0x02,0x04,0xf8,0x00,0x00,0x00,  //00000dd0  ................
    0x00,0x00,0x90,0x90,0x48,0x48,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000de0  ....HH..........
    0x00,0x00,0x30,0x48,0x48,0x30,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000df0  ..0HH0..........
    0x00,0x00,0x20,0x20,0xfc,0x20,0x40,0x5c,0x40,0x80,0xa0,0xa0,0x9e,0x00,0x00,0x00,  //00000e00  ..  . @\@.......
    0x00,0x00,0x20,0x20,0xf8,0x20,0x40,0x78,0x84,0x02,0x02,0x02,0x7c,0x00,0x00,0x00,  //00000e10  ..  . @x....|...
    0x00,0x00,0x00,0xf8,0x04,0x02,0x02,0x02,0x02,0x02,0x04,0x18,0x60,0x00,0x00,0x00,  //00000e20  ............`...
    0x00,0x00,0x00,0xfe,0x0c,0x10,0x20,0x20,0x20,0x20,0x20,0x10,0x0c,0x00,0x00,0x00,  //00000e30  ......     .....
    0x00,0x00,0x40,0x40,0x26,0x18,0x20,0x40,0x80,0x80,0x80,0x40,0x3e,0x00,0x00,0x00,  //00000e40  ..@@&. @...@>...
    0x00,0x00,0x28,0x24,0xfa,0x20,0x48,0x88,0x3c,0x4a,0x88,0x88,0x70,0x00,0x00,0x00,  //00000e50  ..($. H.<J..p...
    0x00,0x00,0x00,0x9c,0x84,0x80,0x80,0x80,0x80,0x80,0xa0,0xa0,0x9e,0x00,0x00,0x00,  //00000e60  ................
    0x00,0x00,0x44,0x44,0x7c,0xca,0xaa,0xaa,0x92,0x96,0xaa,0xaa,0x4c,0x00,0x00,0x00,  //00000e70  ..DD|.......L...
    0x00,0x00,0x20,0x2c,0xf2,0x22,0x62,0x66,0xaa,0xaa,0xaa,0xaa,0xac,0x00,0x00,0x00,  //00000e80  .. ,."bf........
    0x00,0x00,0x00,0x38,0x54,0x92,0x92,0x92,0x92,0x92,0x92,0x92,0x64,0x00,0x00,0x00,  //00000e90  ...8T.......d...
    0x00,0x00,0x84,0x84,0xbe,0x84,0x84,0x84,0x9e,0xa4,0xa4,0xa4,0x98,0x00,0x00,0x00,  //00000ea0  ................
    0x00,0x00,0x00,0xee,0x28,0x44,0x44,0x82,0x82,0x82,0x82,0x44,0x38,0x00,0x00,0x00,  //00000eb0  ....(DD....D8...
    0x00,0x00,0x38,0x08,0x10,0x54,0x54,0x92,0x8a,0x8a,0x8a,0x8a,0x30,0x00,0x00,0x00,  //00000ec0  ..8..TT.....0...
    0x00,0x00,0x00,0x00,0x00,0x20,0x50,0x88,0x04,0x02,0x00,0x00,0x00,0x00,0x00,0x00,  //00000ed0  ..... P.........
    0x00,0x00,0x00,0xbe,0x84,0x84,0xbe,0x84,0x84,0x9e,0xa4,0xa4,0x98,0x00,0x00,0x00,  //00000ee0  ................
    0x00,0x00,0x10,0x7c,0x10,0x10,0x7c,0x10,0x10,0x78,0x94,0x92,0x60,0x00,0x00,0x00,  //00000ef0  ...|..|..x..`...
    0x00,0x00,0x70,0x10,0x12,0x12,0x3e,0x52,0x92,0x92,0x92,0x92,0x64,0x00,0x00,0x00,  //00000f00  ..p...>R....d...
    0x00,0x00,0x10,0x14,0x7a,0x10,0x70,0x92,0x92,0xe2,0x42,0x42,0x3c,0x00,0x00,0x00,  //00000f10  ....z.p...BB<...
    0x00,0x00,0x44,0x44,0x7c,0xca,0xaa,0xaa,0x92,0x92,0xa2,0xa2,0x44,0x00,0x00,0x00,  //00000f20  ..DD|.......D...
    0x00,0x00,0x10,0x7c,0x10,0x20,0xfc,0x20,0x40,0x42,0x42,0x42,0x3c,0x00,0x00,0x00,  //00000f30  ...|. . @BBB<...
    0x00,0x00,0x08,0x48,0x48,0x5c,0xe2,0x42,0x42,0x4c,0x20,0x20,0x20,0x00,0x00,0x00,  //00000f40  ...HH\.BBL   ...
    0x00,0x00,0x10,0x08,0xbc,0xca,0x8a,0x8a,0x8a,0x8a,0xbc,0x08,0x10,0x00,0x00,0x00,  //00000f50  ................
    0x00,0x00,0x10,0x10,0x1c,0x10,0x10,0x10,0x78,0x94,0x92,0x90,0x60,0x00,0x00,0x00,  //00000f60  ........x...`...
    0x00,0x00,0x30,0x08,0x40,0x40,0x80,0xbc,0xc2,0x82,0x02,0x04,0x38,0x00,0x00,0x00,  //00000f70  ..0.@@......8...
    0x00,0x00,0x00,0x9c,0xa2,0xc2,0x82,0x82,0x82,0x02,0x02,0x04,0x38,0x00,0x00,0x00,  //00000f80  ............8...
    0x00,0x00,0x78,0x08,0x10,0x20,0x78,0x84,0x02,0x62,0x92,0x92,0x7c,0x00,0x00,0x00,  //00000f90  ..x.. x..b..|...
    0x00,0x00,0x20,0x2c,0xf2,0x22,0x62,0x64,0xa4,0xa8,0xa8,0xa8,0xa6,0x00,0x00,0x00,  //00000fa0  .. ,."bd........
    0x00,0x00,0x78,0x08,0x10,0x20,0x78,0x84,0x02,0x02,0x02,0x02,0x7c,0x00,0x00,0x00,  //00000fb0  ..x.. x.....|...
    0x00,0x00,0x20,0x20,0x2c,0xf2,0x22,0x22,0x62,0x62,0xa2,0xa2,0x2c,0x00,0x00,0x00,  //00000fc0  ..  ,.""bb..,...
    0x00,0x00,0x10,0x10,0x20,0x20,0x40,0x40,0x60,0x92,0x92,0x92,0x8c,0x00,0x00,0x00,  //00000fd0  ....  @@`.......
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000fe0  ................
    0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,  //00000ff0  ................
  };
*/
  //  perl misc/itob.pl xeij/FontEditor.java FNT_OMUSUBI_DATA
  public static final byte[] FNT_OMUSUBI_DATA = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\08@8\4x\0\"\">\"\"\0\0\0\0\08@8\4x\0\"\24\b\24\"\0\0\0\0\0x@x@|\0\"\24\b\24\"\0\0\0\0\0x@x@|\0>\b\b\b\b\0\0\0\0\0x@x@|\0\34\"*$\32\0\0\0\0\08D|DD\0\"$8$\"\0\0\0\0\0xDxDx\0    >\0\0\0\0\0xDxDx\0\34 \34\2<\0\0\0\0\0DD|DD\0>\b\b\b\b\0\0\0\0\0@@@@|\0< <  \0\0\0\0\0DDD(\20\0>\b\b\b\b\0\0\0\0\0x@x@@\0< <  \0\0\0\0\08@@@<\0<\"<$\"\0\0\0\0\08@8\4x\0\34\"\"\"\34\0\0\0\0\08@8\4x\0\34\b\b\b\34\0\0\0\0\0xDDDx\0< < >\0\0\0\0\0xDDDx\0\30\b\b\b\b\0\0\0\0\0xDDDx\0\34\2\34 >\0\0\0\0\0xDDDx\0\34\2\34\2<\0\0\0\0\0xDDDx\0\4\f\24>\4\0\0\0\0\0DdTLD\0\"$8$\"\0\0\0\0\08@8\4x\0\"2*&\"\0\0\0\0\0x@x@|\0<\"<\"<\0\0\0\0\08@@@<\0\"2*&\"\0\0\0\0\0x@x@|\0\"6*\"\"\0\0\0\0\08@8\4x\0<\"<\"<\0\0\0\0\0x@x@|\0\34   \36\0\0\0\0\0\0\0\20\b\4\376\4\b\20\0\0\0\0\0\0\0\0\0\20 @\376@ \20\0\0\0\0\0\0\0\0\0\208T\222\20\20\20\0\0\0\0\0\0\0\0\0\20\20\20\222T8\20\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\20\20\20\20\20\20\20\20\0\20\20\0\0\0\0$l\330\220\0\0\0\0\0\0\0\0\0\0\0\0\0DDD\376DDD\376DDD\0\0\0\0\0|\220\220\220x\24\22\22\22\22\374\0\0\0\0\0b\222\224h\b\20 ,R\222\214\0\0\0\0\0p\210\210P R\212\204\204\212r\0\0\0\0\b\0300 \0\0\0\0\0\0\0\0\0\0\0\0\b\20\20       \20\20\b\0\0\0 \20\20\b\b\b\b\b\b\b\20\20 \0\0\0\0\0\0\20\222T8T\222\20\0\0\0\0\0\0\0\0\0\20\20\20\376\20\20\20\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\b\0300 \0\0\0\0\0\0\0\0\0\376\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\208\20\0\0\0\0\0\2\2\4\b\b\20  @\200\200\0\0\0\0\08DD\202\202\202\202\202\202\202|\0\0\0\0\0p\20\20\20\20\20\20\20\20\20\20\0\0\0\0\0x\4\4\48@\200\200\200\200\376\0\0\0\0\0x\4\4\4x\4\2\2\2\2\374\0\0\0\0\0\f\24$DD\204\204\204\376\4\4\0\0\0\0\0|@@@x\4\2\2\2\2\374\0\0\0\0\08@@\200\270\304\202\202\202\202|\0\0\0\0\0~\2\4\4\b\b\b\20\20\20\20\0\0\0\0\08DDD8D\202\202\202\202|\0\0\0\0\08D\202\202\202\202~\2\4\48\0\0\0\0\0\0\208\20\0\0\0\208\20\0\0\0\0\0\0\0\208\20\0\0\0\0\b\0300 \0\0\0\0\2\4\b\20 @ \20\b\4\2\0\0\0\0\0\0\0\0\376\0\0\0\376\0\0\0\0\0\0\0\0\200@ \20\b\4\b\20 @\200\0\0\0\0\0\374\2\2\2\4\b\20\20\0\20\20\0\0\0\0\08DD\202\236\242\242\242\236\200~\0\0\0\0\08DD\202\376\202\202\202\202\202\202\0\0\0\0\0\370\204\204\204\370\204\202\202\202\202\374\0\0\0\0\0<@@\200\200\200\200\200\200\200~\0\0\0\0\0\370\204\204\202\202\202\202\202\202\202\374\0\0\0\0\0\374\200\200\200\374\200\200\200\200\200\376\0\0\0\0\0\374\200\200\200\374\200\200\200\200\200\200\0\0\0\0\0<@@\200\216\202\202\202\202\202|\0\0\0\0\0\202\202\202\202\376\202\202\202\202\202\202\0\0\0\0\0\20\20\20\20\20\20\20\20\20\20\20\0\0\0\0\0\2\2\2\2\2\2\2\2\202D8\0\0\0\0\0\204\210\220\240\340\220\210\204\204\202\202\0\0\0\0\0\200\200\200\200\200\200\200\200\200\200\376\0\0\0\0\0\202\202\306\306\252\252\252\222\222\222\202\0\0\0\0\0\202\302\242\222\212\212\206\206\202\202\202\0\0\0\0\08DD\202\202\202\252\202\252\202|\0\0\0\0\0\374\202\202\202\374\200\200\200\200\200\200\0\0\0\0\08DD\202\202\202\202\262\312\204z\0\0\0\0\0\370\204\204\204\370\220\210\204\204\202\202\0\0\0\0\0|\200\200\200x\4\2\2\2\2\374\0\0\0\0\0\376\20\20\20\20\20\20\20\20\20\20\0\0\0\0\0\202\202\202\202\202\202\202\202\202\202|\0\0\0\0\0\202\202\202\202DDD((\20\20\0\0\0\0\0\202\222\222\222\252\252\252DDDD\0\0\0\0\0\202D(\20((DD\202\202\202\0\0\0\0\0\202D((\20\20\20\20\20\20\20\0\0\0\0\0|\4\b\b\20  @@\200\376\0\0\0\0\0<         <\0\0\0\0\0DD((\20\376\20\376\20\20\20\0\0\0\0\0x\b\b\b\b\b\b\b\b\bx\0\0\0\0\0\20(D\202\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\376\0\0\0\0 0\30\b\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0|\2\2~\202\202~\0\0\0\0\0\200\200\200\200\270\304\202\202\202\202\374\0\0\0\0\0\0\0\0\0<@\200\200\200\200~\0\0\0\0\0\2\2\2\2:F\202\202\202\202~\0\0\0\0\0\0\0\0\0|\202\202\376\200\200~\0\0\0\0\0\34   \374      \0\0\0\0\0\0\0\0\0:F\202\202\206z\2\4x\0\0\0\200\200\200\200\270\304\202\202\202\202\202\0\0\0\0\0\20\20\0\0\20\20\20\20\20\20\20\0\0\0\0\0\b\b\0\0\b\b\b\b\b\b\b\20\340\0\0\0@@@@FHPhDBB\0\0\0\0\0\20\20\20\20\20\20\20\20\20\20\34\0\0\0\0\0\0\0\0\0\354\222\222\222\222\222\222\0\0\0\0\0\0\0\0\0\274\302\202\202\202\202\202\0\0\0\0\0\0\0\0\08D\202\202\202\202|\0\0\0\0\0\0\0\0\0\270\304\202\202\202\202\374\200\200\200\0\0\0\0\0\0:F\202\202\202\202~\2\2\2\0\0\0\0\0\0^`@@@@@\0\0\0\0\0\0\0\0\0|\200\200|\2\2\374\0\0\0\0\0    \374     \34\0\0\0\0\0\0\0\0\0\202\202\202\202\202\206z\0\0\0\0\0\0\0\0\0\202\202\202DD(\20\0\0\0\0\0\0\0\0\0\202\222\222\252\252DD\0\0\0\0\0\0\0\0\0\306(\20(D\202\202\0\0\0\0\0\0\0\0\0\202\202\202\202\206z\2\4x\0\0\0\0\0\0\0|\4\b\20 @\376\0\0\0\0\30     \300     \30\0\0\0\20\20\20\20\20\20\20\20\20\20\20\20\20\0\0\0000\b\b\b\b\b\6\b\b\b\b\b0\0\0\0\0\376\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\200\200@  \20\b\b\4\2\2\0\0\0\0\0`\222\f\0\0\0\0\0\0\0\0\0\0\0\0\20\20\20\20\20\20\0\20\20\20\20\20\20\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\20\20x L0H\204\200\200~\0\0\0\0\0\0\0  x |\252\222\242D\0\0\0\0\0\0\0\0\0D\202\202\202\202\242@\0\0\0\0\0\0\0000\0x\4\2\2\2\48\0\0\0\0\0\0\0p\0\370\20 P\222\222\214\0\0\0\0\0\0\0 $r |\242\242\242d\0\0\0\0\0\0\0HH\\\342BL   \0\0\0\0\0\0\0\20\210\274\312\212\212\274\b\20\0\0\0\0\0\0\0\20\20\34\20\20x\224\222`\0\0\0\0\0\0\0\0\370\4\2\2\2\4\30`\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0  x |\252\252\222\222\242D\0\0\0\0\0\0DD\202\202\202\202\202\202\242@\0\0\0\0\0p\0\0\370\4\2\2\2\4\30`\0\0\0\0\0p\0\370\20 @`\220\222\222\214\0\0\0\0\0 $r  |\242\242\242\242d\0\0\0\0\0$$\"\362($DDD\204\230\0\0\0\0\0 \20|\b>\4>@\200\200~\0\0\0\0\0\4\4\b\20 @ \20\b\4\4\0\0\0\0\0\204\204\276\204\204\204\204\204\210\210\220\0\0\0\0\0\0|\4\0\0\0\0\0\200\200~\0\0\0\0\0 \20|\b\4>@\200\200\200~\0\0\0\0\0\0\200\200\200\200\202\202\202\202D8\0\0\0\0\0\4\4\376\0044L\204\214t\48\0\0\0\0\0DD\376DDDD@@@<\0\0\0\0\0|\b\20 \376\30 @@@<\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0000HH0\0\0\0\36\20\20\20\20\20\20\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\b\b\b\b\b\bx\0\0\0\0\0\0\0\0\0\0\0\0 0\30\b\0\0\0\0\0\0\0\0\208\20\0\0\0\0\0\0\0\0\0\0\376\2\2\376\2\2\2\2\48\0\0\0\0\0\0\0\0\0\376\2\"$8 \300\0\0\0\0\0\0\0\0\0\2\4\b8\310\b\b\0\0\0\0\0\0\0\20\20\376\202\202\2\2\48\0\0\0\0\0\0\0\0\0|\20\20\20\20\20\376\0\0\0\0\0\0\0\4\4\376\4\f\24$D\204\0\0\0\0\0\0\0  \376\"$(   \0\0\0\0\0\0\0\0\0x\b\b\b\b\b\376\0\0\0\0\0\0\0\0\0~\2\2~\2\2\376\0\0\0\0\0\0\0\0\0\222\222\222\2\2\4\30\0\0\0\0\0\0\0\0\0\0|\0\0\0\0\0\0\0\0\0\0\0\376\2\2\"$8   \300\0\0\0\0\0\0\2\4\b8\310\b\b\b\b\b\0\0\0\0\0\20\20\376\202\202\202\2\2\2\4\30\0\0\0\0\0\0|\20\20\20\20\20\20\20\20\376\0\0\0\0\0\4\4\376\4\4\f\24$D\204\4\0\0\0\0\0  \376\"\"\"BBB\202\236\0\0\0\0\0\20\20\376\20\20\376\20\20\20\20\20\0\0\0\0\0\0~BBBB\202\2\2\4\30\0\0\0\0\0@@~\210\210\210\b\b\b\20`\0\0\0\0\0\0~\2\2\2\2\2\2\2\2\376\0\0\0\0\0DD\376DD\4\4\4\4\bp\0\0\0\0\0\0\360\2\2\362\2\2\2\2\4\370\0\0\0\0\0\0~\2\2\2\4\b\24\"B\202\0\0\0\0\0@@\376BDH@@@@>\0\0\0\0\0\0\202B\"\"\2\4\4\b\20`\0\0\0\0\0\0~BB~B\202\2\2\4\30\0\0\0\0\0\0|\b\b\376\b\b\b\b\20`\0\0\0\0\0\0\222\222\222\222\222\2\2\2\4\30\0\0\0\0\0\0|\0\0\376\b\b\b\b\20`\0\0\0\0\0\0@@@pLB@@@@\0\0\0\0\0\20\20\376\20\20\20\20\20\20 \300\0\0\0\0\0\0\0|\0\0\0\0\0\0\0\376\0\0\0\0\0\0~\2\2\2$\24\b\24\"\300\0\0\0\0\0\20\20\376\4\b\208T\222\20\20\0\0\0\0\0\0\2\2\2\2\4\4\b\20 \300\0\0\0\0\0\0DBBBBBBB\202\202\0\0\0\0\0\200\200\204\230\340\200\200\200\200\200~\0\0\0\0\0\0\376\2\2\2\2\2\4\4\b0\0\0\0\0\0\0  PP\210\210\4\4\2\2\0\0\0\0\0\20\20\376\20\20TT\222\222\222\222\0\0\0\0\0\0\376\2\2\2\2D(\20\b\4\0\0\0\0\0\0|\0\0|\0\0\0\0\0\376\0\0\0\0\0\08@\200\200\202\202\202\202\202\376\0\0\0\0\0\2\2\2\2\"\24\b\24\"@\200\0\0\0\0\0\0|  \376     \36\0\0\0\0\0@@\376BBBDH@@@\0\0\0\0\0\0x\b\b\b\b\b\b\b\b\376\0\0\0\0\0\0~\2\2~\2\2\2\2\2\376\0\0\0\0\0\0|\0\0\376\2\2\2\2\48\0\0\0\0\0\0\202\202\202\202\202\2\2\2\48\0\0\0\0\0\0PPPPPPRR\224\230\0\0\0\0\0\0\200\200\200\200\200\200\202\204\230\340\0\0\0\0\0\0\376\202\202\202\202\202\202\202\202\376\0\0\0\0\0\0\376\202\202\202\2\2\2\2\48\0\0\0\0\0\0\360\0\2\2\2\2\2\2\4\370\0\0\0\0\0\220\220HH\0\0\0\0\0\0\0\0\0\0\0\0000HH0\0\0\0\0\0\0\0\0\0\0\0\0  \374 @\\@\200\240\240\236\0\0\0\0\0  \370 @x\204\2\2\2|\0\0\0\0\0\0\370\4\2\2\2\2\2\4\30`\0\0\0\0\0\0\376\f\20     \20\f\0\0\0\0\0@@&\30 @\200\200\200@>\0\0\0\0\0($\372 H\210<J\210\210p\0\0\0\0\0\0\234\204\200\200\200\200\200\240\240\236\0\0\0\0\0DD|\312\252\252\222\226\252\252L\0\0\0\0\0 ,\362\"bf\252\252\252\252\254\0\0\0\0\0\08T\222\222\222\222\222\222\222d\0\0\0\0\0\204\204\276\204\204\204\236\244\244\244\230\0\0\0\0\0\0\356(DD\202\202\202\202D8\0\0\0\0\08\b\20TT\222\212\212\212\2120\0\0\0\0\0\0\0\0 P\210\4\2\0\0\0\0\0\0\0\0\0\276\204\204\276\204\204\236\244\244\230\0\0\0\0\0\20|\20\20|\20\20x\224\222`\0\0\0\0\0p\20\22\22>R\222\222\222\222d\0\0\0\0\0\20\24z\20p\222\222\342BB<\0\0\0\0\0DD|\312\252\252\222\222\242\242D\0\0\0\0\0\20|\20 \374 @BBB<\0\0\0\0\0\bHH\\\342BBL   \0\0\0\0\0\20\b\274\312\212\212\212\212\274\b\20\0\0\0\0\0\20\20\34\20\20\20x\224\222\220`\0\0\0\0\0000\b@@\200\274\302\202\2\48\0\0\0\0\0\0\234\242\302\202\202\202\2\2\48\0\0\0\0\0x\b\20 x\204\2b\222\222|\0\0\0\0\0 ,\362\"bd\244\250\250\250\246\0\0\0\0\0x\b\20 x\204\2\2\2\2|\0\0\0\0\0  ,\362\"\"bb\242\242,\0\0\0\0\0\20\20  @@`\222\222\222\214\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0".getBytes (XEiJ.ISO_8859_1);

}  //class FontEditor



