//========================================================================================
//  Keyboard.java
//    en:Keyboard
//    ja:キーボード
//  Copyright (C) 2003-2019 Makoto Kamada
//
//  This file is part of the XEiJ (X68000 Emulator in Java).
//  You can use, modify and redistribute the XEiJ if the conditions are met.
//  Read the XEiJ License for more details.
//  https://stdkmd.net/xeij/
//========================================================================================

//----------------------------------------------------------------------------------------
//! SHIFTとCTRLは左ボタンで押されただけでロックするようになっているが、
//  これをプレフィックスにして次に他のキーが離された後にSHIFTとCTRLも自動的に離されるモードが欲しい
//  右ボタンで押されたときのロックは従来通り、ロックしたキーをクリックするまで押されっぱなしにする
//----------------------------------------------------------------------------------------

package xeij;

import java.awt.*;  //BasicStroke,BorderLayout,BoxLayout,Color,Component,Container,Cursor,Desktop,Dimension,Font,Frame,Graphics,Graphics2D,GraphicsDevice,GraphicsEnvironment,GridLayout,Image,Insets,Paint,Point,Rectangle,RenderingHints,Robot,Shape,Stroke,TexturePaint,Toolkit
import java.awt.event.*;  //ActionEvent,ActionListener,ComponentAdapter,ComponentEvent,ComponentListener,FocusAdapter,FocusEvent,FocusListener,InputEvent,KeyAdapter,KeyEvent,KeyListener,MouseAdapter,MouseEvent,MouseListener,MouseMotionAdapter,MouseWheelEvent,WindowAdapter,WindowEvent,WindowListener,WindowStateListener
import java.awt.image.*;  //BufferedImage,DataBuffer,DataBufferByte,DataBufferInt,IndexColorModel
import java.awt.im.*;  //InputContext
import java.io.*;  //BufferedInputStream,BufferedOutputStream,BufferedReader,BufferedWriter,File,FileInputStream,FileNotFoundException,FileReader,InputStream,InputStreamReader,IOException,OutputStreamWriter,RandomAccessFile
import java.lang.*;  //Boolean,Character,Class,Comparable,Double,Exception,Float,IllegalArgumentException,Integer,Long,Math,Number,Object,Runnable,SecurityException,String,StringBuilder,System
import java.lang.reflect.*;  //Field
import java.util.*;  //ArrayList,Arrays,Calendar,GregorianCalendar,HashMap,Map,Map.Entry,Timer,TimerTask,TreeMap
import javax.swing.*;

public class Keyboard {

  //デバッグ
  public static final boolean KBD_DEBUG_LED = false;  //LEDキーが押されたらLEDを反転する

  //種類
  public static final int KBD_STANDARD_TYPE =  0;  //Standard Keyboard
  public static final int KBD_COMPACT_TYPE  =  1;  //Compact Keyboard
  public static final int KBD_TYPES         =  2;  //キーボードの種類の数。Standard KeyboardとCompact Keyboardの2種類

  //サイズ
  //  6+6*94+6=576,2+4*25+2=104  Standard Keyboard
  //  6+6*76+6=468,2+4*25+2=104  Compact Keyboard
  public static final int KBD_COL_WIDTH      =  6;  //列の幅(px)。可変キーの幅の1/4
  public static final int KBD_ROW_HEIGHT     =  4;  //行の高さ(px)。可変キーの高さの1/4
  public static final int KBD_LED_WIDTH      = 11;  //LEDの幅
  public static final int KBD_LED_HEIGHT     =  4;  //LEDのキートップ側の高さ
  public static final int KBD_PADDING_TOP    =  2;  //パディング(px)
  public static final int KBD_PADDING_RIGHT  =  6;
  public static final int KBD_PADDING_BOTTOM =  2;
  public static final int KBD_PADDING_LEFT   =  6;
  public static final int KBD_TOUCHABLE_AREA =  1;  //ポイントできる範囲。0=1段目から,1=2段目から,2=3段目から,3=キートップのみ
  public static final int[] KBD_TYPE_COLS    = { 94, 76 };  //種類→キーボードの幅(列数)
  public static final int[] KBD_TYPE_ROWS    = { 25, 25 };  //種類→キーボードの高さ(行数)
  public static final int KBD_KEY_WIDTH      = KBD_COL_WIDTH * 4 - 1;  //可変キーとLEDキーの幅
  public static final int KBD_KEY_HEIGHT     = KBD_ROW_HEIGHT * 4 - 1;  //可変キーとLEDキーの高さ

  //スタイル
  public static final int KBD_NONE_STYLE     = 0;  //キーなし
  public static final int KBD_NORMAL_STYLE   = 1;  //ノーマル
  public static final int KBD_FUNCTION_STYLE = 2;  //キーの上端を1/2行削る。63～6C:F1～F10
  public static final int KBD_NARROWER_STYLE = 3;  //キートップの左右を1/2列ずつ削る。55:XF1,56:XF2,57:XF3(Compact),58:XF4(Standard),59:XF5(Standard)
  public static final int KBD_NARROW_STYLE   = 4;  //キートップの左右を1列ずつ削る。10:TAB,57:XF3(Standard),71:CTRL,72:OPT.1,73:OPT.2
  public static final int KBD_SHORTER_STYLE  = 5;  //キートップの上下を1行ずつ削る。4E:ENTER(Standard)
  public static final int KBD_SHORT_STYLE    = 6;  //キートップの上下を2行ずつ削る。3B:LEFT(Standard),3D:RIGHT(Standard)
  public static final int KBD_RETURN_STYLE   = 7;  //レクタングルの左下を1列x4行削る。キートップの左右を1列ずつ削る。1D:RETURN
  public static final int KBD_SHIFT_STYLE    = 8;  //レクタングルの左右を9列ずつ残して中央を削る。70:SHIFT

  //文字
  public static final boolean KBD_USE_ROM_FONT = false;  //可変キーの文字をROMフォントにする。ROMフォントが生成済みでなければならない

  //色
  public static final int KBD_LETTER_COLOR       = 14;  //文字
  public static final int KBD_TOP_LEFT_COLOR     = 10;  //ボーダー左上
  public static final int KBD_TOP_COLOR          =  8;  //ボーダー上
  public static final int KBD_LEFT_COLOR         =  8;  //ボーダー左
  public static final int KBD_POINTED_COLOR      =  8;  //キートップ(pointed)
  public static final int KBD_TOP_RIGHT_COLOR    =  6;  //ボーダー右上
  public static final int KBD_BOTTOM_LEFT_COLOR  =  6;  //ボーダー左下
  public static final int KBD_KEYTOP_COLOR       =  6;  //キートップ
  public static final int KBD_RIGHT_COLOR        =  4;  //ボーダー右
  public static final int KBD_BOTTOM_COLOR       =  4;  //ボーダー下
  public static final int KBD_BOTTOM_RIGHT_COLOR =  2;  //ボーダー右下
  public static final int KBD_SHADOW_COLOR       =  2;  //文字の影
  public static final int KBD_BACKGROUND_COLOR   =  0;  //背景
  public static final int KBD_HOLE_COLOR         =  0;  //押し下げられたキーの周囲の隙間
  public static final int KBD_RED_OFF_KEYTOP_COLOR   = Color.HSBtoRGB (0.99F, 0.50F, 0.40F);  //赤色LED,消灯,キートップ側
  public static final int KBD_RED_OFF_BORDER_COLOR   = Color.HSBtoRGB (0.99F, 0.50F, 0.20F);  //赤色LED,消灯,ボーダー側
  public static final int KBD_RED_ON_KEYTOP_COLOR    = Color.HSBtoRGB (0.99F, 1.00F, 1.00F);  //赤色LED,点灯,キートップ側
  public static final int KBD_RED_ON_BORDER_COLOR    = Color.HSBtoRGB (0.99F, 1.00F, 0.50F);  //赤色LED,点灯,ボーダー側
  public static final int KBD_GREEN_OFF_KEYTOP_COLOR = Color.HSBtoRGB (0.24F, 0.50F, 0.40F);  //緑色LED,消灯,キートップ側
  public static final int KBD_GREEN_OFF_BORDER_COLOR = Color.HSBtoRGB (0.24F, 0.50F, 0.20F);  //緑色LED,消灯,ボーダー側
  public static final int KBD_GREEN_ON_KEYTOP_COLOR  = Color.HSBtoRGB (0.24F, 1.00F, 1.00F);  //緑色LED,消灯,キートップ側
  public static final int KBD_GREEN_ON_BORDER_COLOR  = Color.HSBtoRGB (0.24F, 1.00F, 0.50F);  //緑色LED,消灯,ボーダー側

  //状態
  //  stat = KBD_LOCKED_MASK | KBD_KEYDOWN_MASK | KBD_MOUSEDOWN_MASK | KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK
  //  ビット位置に依存しているので変更しないこと
  public static final int KBD_POINTED_BIT   = 0;  //bit0 1=ポイントされている。ポイントされているキーは1つしか存在しない
  public static final int KBD_PRESSED_BIT   = 1;  //bit1 1=押されている。(locked||keydown||mousedown)==pressed
  public static final int KBD_LIGHTED_BIT   = 2;  //bit2 1=LEDが点灯している。LEDキーのみ
  public static final int KBD_MOUSEDOWN_BIT = 3;  //bit3 1=マウスで押されている。pressedを伴う。描画には使わない
  public static final int KBD_KEYDOWN_BIT   = 4;  //bit4 1=キーで押されている。pressedを伴う。描画には使わない
  public static final int KBD_LOCKED_BIT    = 5;  //bit5 1=ロックされている。pressedを伴う。描画には使わない
  public static final int KBD_POINTED_MASK   = 1 << KBD_POINTED_BIT;
  public static final int KBD_PRESSED_MASK   = 1 << KBD_PRESSED_BIT;
  public static final int KBD_LIGHTED_MASK   = 1 << KBD_LIGHTED_BIT;
  public static final int KBD_MOUSEDOWN_MASK = 1 << KBD_MOUSEDOWN_BIT;
  public static final int KBD_KEYDOWN_MASK   = 1 << KBD_KEYDOWN_BIT;
  public static final int KBD_LOCKED_MASK    = 1 << KBD_LOCKED_BIT;
  //  LEDキー
  //    lighted||pointedかどうかで可変キーに表示する文字を変更する
  //    KBD_LED_TEST << (stat & (kbdStat[KBD_NUM_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) < 0
  //  その他のモードキー
  //    pressed||pointedかどうかで可変キーに表示する文字を変更する
  //    KBD_MODE_TEST << (stat & (kbdStat[KBD_NUM_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) < 0
  public static final int KBD_LED_TEST  = 0b01011010_01011010_01011010_01011010;
  public static final int KBD_MODE_TEST = 0b01100110_01100110_01100110_01100110;

  //数
  public static final int KBD_VAR_KEYS = 48;  //可変キーの数
  public static final int KBD_LED_KEYS =  8;  //LEDキーの数
  public static final int KBD_FIX_KEYS = 57;  //固定キーの数
  public static final int KBD_ALL_KEYS = KBD_VAR_KEYS + KBD_LED_KEYS + KBD_FIX_KEYS;  //キーの数

  //番号
  //  可変キー,LEDキー,固定キーの順序でスキャンコード順に割り当てた番号
  //  RETURNキーの上下とSHIFTキーの左右は形を特別扱いにすることで1個の番号で扱う
  public static final int KBD_KANA_INDEX   =  48;  //かなキーの番号
  public static final int KBD_ROMA_INDEX   =  49;  //ローマ字キーの番号
  public static final int KBD_CODE_INDEX   =  50;  //コード入力キーの番号
  public static final int KBD_CAPS_INDEX   =  51;  //CAPSキーの番号
  public static final int KBD_HIRA_INDEX   =  53;  //ひらがなキーの番号
  public static final int KBD_ZEN_INDEX    =  54;  //全角キーの番号
  public static final int KBD_NUM_INDEX    =  55;  //NUMキーの番号
  public static final int KBD_RETURN_INDEX =  59;  //RETURNキーの番号
  public static final int KBD_SHIFT_INDEX  = 109;  //SHIFTキーの番号
  public static final int KBD_CTRL_INDEX   = 110;  //CTRLキーの番号
  public static final int KBD_OPT1_INDEX   = 111;  //OPT.1キーの番号

  //番号→スキャンコード(NUM OFF)
/*
  public static final byte[] KBD_NORMAL_CODE = {
    //              |1| | | | | | | | | | | | |NUM
    //              | |1| | | | | | | | | | | |CTRL
    //              | | |1| | | | | | | | | | |コード入力
    //              | | | |0|0|0|0|0|0| | | | |OPT.1
    //              | | | |1|1|1|1| | | | | | |かな。ひらがな+全角→ひらがな
    //              | | | | | | | |1|1| | | | |ローマ字
    //              | | | |0|0|1|1| | | | | | |五十音
    //              | | | | | | | | | |0|0|1|1|CAPS
    //              | | | |0|1|0|1|0|1|0|1|0|1|SHIFT
    0x02,  //  0  0 | | |1|ﾇ|ﾇ|ｱ|ｧ|1|!|1|!|1|!|
    0x03,  //  1  1 | | |2|ﾌ|ﾌ|ｲ|ｨ|2|"|2|"|2|"|
    0x04,  //  2  2 | | |3|ｱ|ｧ|ｳ|ｩ|3|#|3|#|3|#|
    0x05,  //  3  3 | | |4|ｳ|ｩ|ｴ|ｪ|4|$|4|$|4|$|
    0x06,  //  4  4 | | |5|ｴ|ｪ|ｵ|ｫ|5|%|5|%|5|%|
    0x07,  //  5  5 | | |6|ｵ|ｫ|ﾅ|ﾅ|6|&|6|&|6|&|
    0x08,  //  6  6 |7| |7|ﾔ|ｬ|ﾆ|ﾆ|7|'|7|'|7|'|
    0x09,  //  7  7 |8| |8|ﾕ|ｭ|ﾇ|ﾇ|8|(|8|(|8|(|
    0x0a,  //  8  8 |9| |9|ﾖ|ｮ|ﾈ|ﾈ|9|)|9|)|9|)|
    0x0b,  //  9  9 |/| |0|ﾜ|ｦ|ﾉ|ｦ|0| |0| |0| |
    0x0c,  // 10 10 |*| | |ﾎ|ﾎ|ﾗ|ﾗ|ｰ|=|-|=|-|=|
    0x0d,  // 11 11 | |^| |ﾍ|ﾍ|ﾘ|ﾘ|^|~|^|~|^|~|
    0x0e,  // 12 12 | |\| |ｰ|ｰ|ﾙ|ﾙ|\|||\|||\|||
    0x11,  // 13 13 | |Q| |ﾀ|ﾀ|ｶ|ｶ|Q|Q|q|Q|Q|q|
    0x12,  // 14 14 | |W| |ﾃ|ﾃ|ｷ|ｷ|W|W|w|W|W|w|
    0x13,  // 15 15 | |E|e|ｲ|ｨ|ｸ|ｸ|E|e|e|E|E|e|
    0x14,  // 16 16 | |R| |ｽ|ｽ|ｹ|ｹ|R|R|r|R|R|r|
    0x15,  // 17 17 | |T| |ｶ|ｶ|ｺ|ｺ|T|T|t|T|T|t|
    0x16,  // 18 18 | |Y| |ﾝ|ﾝ|ﾊ|ﾊ|Y|Y|y|Y|Y|y|
    0x17,  // 19 19 |4|U| |ﾅ|ﾅ|ﾋ|ﾋ|U|u|u|U|U|u|
    0x18,  // 20 20 |5|I| |ﾆ|ﾆ|ﾌ|ﾌ|I|i|i|I|I|i|
    0x19,  // 21 21 |6|O| |ﾗ|ﾗ|ﾍ|ﾍ|O|o|o|O|O|o|
    0x1a,  // 22 22 |-|P| |ｾ|ｾ|ﾎ|ﾎ|P|P|p|P|P|p|
    0x1b,  // 23 23 | | | |ﾞ|ﾞ|ﾚ|ﾚ|@|`|@|`|@|`|
    0x1c,  // 24 24 | |[| |ﾟ|｢|ﾛ|｢|｢|{|[|{|[|{|
    0x1e,  // 25 25 | |A|a|ﾁ|ﾁ|ｻ| |A|a|a|A|A|a| かな+五十音+SHIFT→0xa0
    0x1f,  // 26 26 | |S| |ﾄ|ﾄ|ｼ|ｼ|S|S|s|S|S|s|
    0x20,  // 27 27 | |D|d|ｼ|ｼ|ｽ|ｽ|D|D|d|D|D|d|
    0x21,  // 28 28 | |F|f|ﾊ|ﾊ|ｾ|ｾ|F|F|f|F|F|f|
    0x22,  // 29 29 | |G| |ｷ|ｷ|ｿ|ｿ|G|G|g|G|G|g|
    0x23,  // 30 30 | |H| |ｸ|ｸ|ﾏ|ﾏ|H|H|h|H|H|h|
    0x24,  // 31 31 |1|J| |ﾏ|ﾏ|ﾐ|ﾐ|J|J|j|J|J|j|
    0x25,  // 32 32 |2|K| |ﾉ|ﾉ|ﾑ|ﾑ|K|K|k|K|K|k|
    0x26,  // 33 33 |3|L| |ﾘ|ﾘ|ﾒ|ﾒ|L|L|l|L|L|l|
    0x27,  // 34 34 |+| | |ﾚ|ﾚ|ﾓ|ﾓ|;|+|;|+|;|+|
    0x28,  // 35 35 | | | |ｹ|ｹ|ﾞ|ｰ|:|*|:|*|:|*|
    0x29,  // 36 36 | |]| |ﾑ|｣|ﾟ|｣|｣|}|]|}|]|}|
    0x2a,  // 37 37 | |Z| |ﾂ|ｯ|ﾀ|ﾀ|Z|z|z|Z|Z|z|
    0x2b,  // 38 38 | |X| |ｻ|ｻ|ﾁ|ﾁ|X|X|x|X|X|x|
    0x2c,  // 39 39 | |C|c|ｿ|ｿ|ﾂ|ｯ|C|C|c|C|C|c|
    0x2d,  // 40 40 | |V| |ﾋ|ﾋ|ﾃ|ﾃ|V|V|v|V|V|v|
    0x2e,  // 41 41 | |B|b|ｺ|ｺ|ﾄ|ﾄ|B|B|b|B|B|b|
    0x2f,  // 42 42 | |N| |ﾐ|ﾐ|ﾔ|ｬ|N|N|n|N|N|n|
    0x30,  // 43 43 |0|M| |ﾓ|ﾓ|ﾕ|ｭ|M|M|m|M|M|m|
    0x31,  // 44 44 |,| | |ﾈ|､|ﾖ|ｮ|､|<|,|<|,|<|
    0x32,  // 45 45 |.| | |ﾙ|｡|ﾜ|､|｡|>|.|>|.|>|
    0x33,  // 46 46 |=| | |ﾒ|･|ｦ|｡|･|?|/|?|/|?|
    0x34,  // 47 47 |E|_| |ﾛ| |ﾝ|･| |_| |_| |_| NUM→ENTER、かな+SHIFT→0xa0
    0x5a,  // 48  0 かな
    0x5b,  // 49  1 ローマ字
    0x5c,  // 50  2 コード入力
    0x5d,  // 51  3 CAPS
    0x5e,  // 52  4 INS
    0x5f,  // 53  5 ひらがな
    0x60,  // 54  6 全角
    0x74,  // 55  7 NUM
    0x01,  // 56  0 ESC
    0x0f,  // 57  1 BS
    0x10,  // 58  2 TAB
    0x1d,  // 59  3 RETURN
    0x35,  // 60  4 Space
    0x36,  // 61  5 HOME
    0x37,  // 62  6 DEL
    0x38,  // 63  7 ROLLUP
    0x39,  // 64  8 ROLLDOWN
    0x3a,  // 65  9 UNDO
    0x3b,  // 66 10 ←
    0x3c,  // 67 11 ↑
    0x3d,  // 68 12 →
    0x3e,  // 69 13 ↓
    0x3f,  // 70 14 CLR
    0x40,  // 71 15 /
    0x41,  // 72 16 *
    0x42,  // 73 17 -
    0x43,  // 74 18 7
    0x44,  // 75 19 8
    0x45,  // 76 20 9
    0x46,  // 77 21 +
    0x47,  // 78 22 4
    0x48,  // 79 23 5
    0x49,  // 80 24 6
    0x4a,  // 81 25 =
    0x4b,  // 82 26 1
    0x4c,  // 83 27 2
    0x4d,  // 84 28 3
    0x4e,  // 85 29 ENTER
    0x4f,  // 86 30 0
    0x50,  // 87 31 ,
    0x51,  // 88 32 .
    0x52,  // 89 33 記号入力
    0x53,  // 90 34 登録
    0x54,  // 91 35 HELP
    0x55,  // 92 36 XF1
    0x56,  // 93 37 XF2
    0x57,  // 94 38 XF3
    0x58,  // 95 39 XF4
    0x59,  // 96 40 XF5
    0x61,  // 97 41 BREAK
    0x62,  // 98 42 COPY
    0x63,  // 99 43 F1
    0x64,  //100 44 F2
    0x65,  //101 45 F3
    0x66,  //102 46 F4
    0x67,  //103 47 F5
    0x68,  //104 48 F6
    0x69,  //105 49 F7
    0x6a,  //106 50 F8
    0x6b,  //107 51 F9
    0x6c,  //108 52 F10
    0x70,  //109 53 SHIFT
    0x71,  //110 54 CTRL
    0x72,  //111 55 OPT.1
    0x73,  //112 56 OPT.2
  };
*/
  //  perl misc/itob.pl xeij/Keyboard.java KBD_NORMAL_CODE
  public static final byte[] KBD_NORMAL_CODE = "\2\3\4\5\6\7\b\t\n\13\f\r\16\21\22\23\24\25\26\27\30\31\32\33\34\36\37 !\"#$%&\'()*+,-./01234Z[\\]^_`t\1\17\20\03556789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYabcdefghijklpqrs".getBytes (XEiJ.ISO_8859_1);

  //番号→スキャンコード(NUM ON)
/*
  public static final byte[] KBD_NUM_CODE = {
    //              |1| | | | | | | | | | | | |NUM
    //              | |1| | | | | | | | | | | |CTRL
    //              | | |1| | | | | | | | | | |コード入力
    //              | | | |0|0|0|0|0|0| | | | |OPT.1
    //              | | | |1|1|1|1| | | | | | |かな。ひらがな+全角→ひらがな
    //              | | | | | | | |1|1| | | | |ローマ字
    //              | | | |0|0|1|1| | | | | | |五十音
    //              | | | | | | | | | |0|0|1|1|CAPS
    //              | | | |0|1|0|1|0|1|0|1|0|1|SHIFT
    0x02,  //  0  0 | | |1|ﾇ|ﾇ|ｱ|ｧ|1|!|1|!|1|!|
    0x03,  //  1  1 | | |2|ﾌ|ﾌ|ｲ|ｨ|2|"|2|"|2|"|
    0x04,  //  2  2 | | |3|ｱ|ｧ|ｳ|ｩ|3|#|3|#|3|#|
    0x05,  //  3  3 | | |4|ｳ|ｩ|ｴ|ｪ|4|$|4|$|4|$|
    0x06,  //  4  4 | | |5|ｴ|ｪ|ｵ|ｫ|5|%|5|%|5|%|
    0x07,  //  5  5 | | |6|ｵ|ｫ|ﾅ|ﾅ|6|&|6|&|6|&|
    0x43,  //  6  6 |7| |7|ﾔ|ｬ|ﾆ|ﾆ|7|'|7|'|7|'|
    0x44,  //  7  7 |8| |8|ﾕ|ｭ|ﾇ|ﾇ|8|(|8|(|8|(|
    0x45,  //  8  8 |9| |9|ﾖ|ｮ|ﾈ|ﾈ|9|)|9|)|9|)|
    0x40,  //  9  9 |/| |0|ﾜ|ｦ|ﾉ|ｦ|0| |0| |0| |
    0x41,  // 10 10 |*| | |ﾎ|ﾎ|ﾗ|ﾗ|ｰ|=|-|=|-|=|
    0x0d,  // 11 11 | |^| |ﾍ|ﾍ|ﾘ|ﾘ|^|~|^|~|^|~|
    0x0e,  // 12 12 | |\| |ｰ|ｰ|ﾙ|ﾙ|\|||\|||\|||
    0x11,  // 13 13 | |Q| |ﾀ|ﾀ|ｶ|ｶ|Q|Q|q|Q|Q|q|
    0x12,  // 14 14 | |W| |ﾃ|ﾃ|ｷ|ｷ|W|W|w|W|W|w|
    0x13,  // 15 15 | |E|e|ｲ|ｨ|ｸ|ｸ|E|e|e|E|E|e|
    0x14,  // 16 16 | |R| |ｽ|ｽ|ｹ|ｹ|R|R|r|R|R|r|
    0x15,  // 17 17 | |T| |ｶ|ｶ|ｺ|ｺ|T|T|t|T|T|t|
    0x16,  // 18 18 | |Y| |ﾝ|ﾝ|ﾊ|ﾊ|Y|Y|y|Y|Y|y|
    0x47,  // 19 19 |4|U| |ﾅ|ﾅ|ﾋ|ﾋ|U|u|u|U|U|u|
    0x48,  // 20 20 |5|I| |ﾆ|ﾆ|ﾌ|ﾌ|I|i|i|I|I|i|
    0x49,  // 21 21 |6|O| |ﾗ|ﾗ|ﾍ|ﾍ|O|o|o|O|O|o|
    0x42,  // 22 22 |-|P| |ｾ|ｾ|ﾎ|ﾎ|P|P|p|P|P|p|
    0x1b,  // 23 23 | | | |ﾞ|ﾞ|ﾚ|ﾚ|@|`|@|`|@|`|
    0x1c,  // 24 24 | |[| |ﾟ|｢|ﾛ|｢|｢|{|[|{|[|{|
    0x1e,  // 25 25 | |A|a|ﾁ|ﾁ|ｻ| |A|a|a|A|A|a| かな+五十音+SHIFT→0xa0
    0x1f,  // 26 26 | |S| |ﾄ|ﾄ|ｼ|ｼ|S|S|s|S|S|s|
    0x20,  // 27 27 | |D|d|ｼ|ｼ|ｽ|ｽ|D|D|d|D|D|d|
    0x21,  // 28 28 | |F|f|ﾊ|ﾊ|ｾ|ｾ|F|F|f|F|F|f|
    0x22,  // 29 29 | |G| |ｷ|ｷ|ｿ|ｿ|G|G|g|G|G|g|
    0x23,  // 30 30 | |H| |ｸ|ｸ|ﾏ|ﾏ|H|H|h|H|H|h|
    0x4b,  // 31 31 |1|J| |ﾏ|ﾏ|ﾐ|ﾐ|J|J|j|J|J|j|
    0x4c,  // 32 32 |2|K| |ﾉ|ﾉ|ﾑ|ﾑ|K|K|k|K|K|k|
    0x4d,  // 33 33 |3|L| |ﾘ|ﾘ|ﾒ|ﾒ|L|L|l|L|L|l|
    0x46,  // 34 34 |+| | |ﾚ|ﾚ|ﾓ|ﾓ|;|+|;|+|;|+|
    0x28,  // 35 35 | | | |ｹ|ｹ|ﾞ|ｰ|:|*|:|*|:|*|
    0x29,  // 36 36 | |]| |ﾑ|｣|ﾟ|｣|｣|}|]|}|]|}|
    0x2a,  // 37 37 | |Z| |ﾂ|ｯ|ﾀ|ﾀ|Z|z|z|Z|Z|z|
    0x2b,  // 38 38 | |X| |ｻ|ｻ|ﾁ|ﾁ|X|X|x|X|X|x|
    0x2c,  // 39 39 | |C|c|ｿ|ｿ|ﾂ|ｯ|C|C|c|C|C|c|
    0x2d,  // 40 40 | |V| |ﾋ|ﾋ|ﾃ|ﾃ|V|V|v|V|V|v|
    0x2e,  // 41 41 | |B|b|ｺ|ｺ|ﾄ|ﾄ|B|B|b|B|B|b|
    0x2f,  // 42 42 | |N| |ﾐ|ﾐ|ﾔ|ｬ|N|N|n|N|N|n|
    0x4f,  // 43 43 |0|M| |ﾓ|ﾓ|ﾕ|ｭ|M|M|m|M|M|m|
    0x50,  // 44 44 |,| | |ﾈ|､|ﾖ|ｮ|､|<|,|<|,|<|
    0x51,  // 45 45 |.| | |ﾙ|｡|ﾜ|､|｡|>|.|>|.|>|
    0x4a,  // 46 46 |=| | |ﾒ|･|ｦ|｡|･|?|/|?|/|?|
    0x4e,  // 47 47 |E|_| |ﾛ| |ﾝ|･| |_| |_| |_| NUM→ENTER、かな+SHIFT→0xa0
    0x5a,  // 48  0 かな
    0x5b,  // 49  1 ローマ字
    0x5c,  // 50  2 コード入力
    0x5d,  // 51  3 CAPS
    0x5e,  // 52  4 INS
    0x5f,  // 53  5 ひらがな
    0x60,  // 54  6 全角
    0x74,  // 55  7 NUM
    0x01,  // 56  0 ESC
    0x0f,  // 57  1 BS
    0x10,  // 58  2 TAB
    0x1d,  // 59  3 RETURN
    0x35,  // 60  4 Space
    0x36,  // 61  5 HOME
    0x37,  // 62  6 DEL
    0x38,  // 63  7 ROLLUP
    0x39,  // 64  8 ROLLDOWN
    0x3a,  // 65  9 UNDO
    0x3b,  // 66 10 ←
    0x3c,  // 67 11 ↑
    0x3d,  // 68 12 →
    0x3e,  // 69 13 ↓
    0x3f,  // 70 14 CLR
    0x40,  // 71 15 /
    0x41,  // 72 16 *
    0x42,  // 73 17 -
    0x43,  // 74 18 7
    0x44,  // 75 19 8
    0x45,  // 76 20 9
    0x46,  // 77 21 +
    0x47,  // 78 22 4
    0x48,  // 79 23 5
    0x49,  // 80 24 6
    0x4a,  // 81 25 =
    0x4b,  // 82 26 1
    0x4c,  // 83 27 2
    0x4d,  // 84 28 3
    0x4e,  // 85 29 ENTER
    0x4f,  // 86 30 0
    0x50,  // 87 31 ,
    0x51,  // 88 32 .
    0x52,  // 89 33 記号入力
    0x53,  // 90 34 登録
    0x54,  // 91 35 HELP
    0x55,  // 92 36 XF1
    0x56,  // 93 37 XF2
    0x57,  // 94 38 XF3
    0x58,  // 95 39 XF4
    0x59,  // 96 40 XF5
    0x61,  // 97 41 BREAK
    0x62,  // 98 42 COPY
    0x63,  // 99 43 F1
    0x64,  //100 44 F2
    0x65,  //101 45 F3
    0x66,  //102 46 F4
    0x67,  //103 47 F5
    0x68,  //104 48 F6
    0x69,  //105 49 F7
    0x6a,  //106 50 F8
    0x6b,  //107 51 F9
    0x6c,  //108 52 F10
    0x70,  //109 53 SHIFT
    0x71,  //110 54 CTRL
    0x72,  //111 55 OPT.1
    0x73,  //112 56 OPT.2
  };
*/
  //  perl misc/itob.pl xeij/Keyboard.java KBD_NUM_CODE
  public static final byte[] KBD_NUM_CODE = "\2\3\4\5\6\7CDE@A\r\16\21\22\23\24\25\26GHIB\33\34\36\37 !\"#KLMF()*+,-./OPQJNZ[\\]^_`t\1\17\20\03556789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYabcdefghijklpqrs".getBytes (XEiJ.ISO_8859_1);

  //種類→固定キーの番号→固定キーのスタイル
  //  Standard KeyboardとCompact KeyboardではXFnキーやOPT.nキーの形が異なるので種類毎にスタイルデータを用意する
  public static final int[][] KBD_TYPE_FIX_STYLE = {
    //Standard Keyboard
    {
      KBD_NORMAL_STYLE  ,  //0x01  0 ESC
      KBD_NARROW_STYLE  ,  //0x0f  1 BS
      KBD_NARROW_STYLE  ,  //0x10  2 TAB
      KBD_RETURN_STYLE  ,  //0x1d  3 RETURN
      KBD_NORMAL_STYLE  ,  //0x35  4 Space
      KBD_NORMAL_STYLE  ,  //0x36  5 HOME
      KBD_NORMAL_STYLE  ,  //0x37  6 DEL
      KBD_NORMAL_STYLE  ,  //0x38  7 ROLLUP
      KBD_NORMAL_STYLE  ,  //0x39  8 ROLLDOWN
      KBD_NORMAL_STYLE  ,  //0x3a  9 UNDO
      KBD_SHORT_STYLE   ,  //0x3b 10 ←
      KBD_NORMAL_STYLE  ,  //0x3c 11 ↑
      KBD_SHORT_STYLE   ,  //0x3d 12 →
      KBD_NORMAL_STYLE  ,  //0x3e 13 ↓
      KBD_NORMAL_STYLE  ,  //0x3f 14 CLR
      KBD_NORMAL_STYLE  ,  //0x40 15 /
      KBD_NORMAL_STYLE  ,  //0x41 16 *
      KBD_NORMAL_STYLE  ,  //0x42 17 -
      KBD_NORMAL_STYLE  ,  //0x43 18 7
      KBD_NORMAL_STYLE  ,  //0x44 19 8
      KBD_NORMAL_STYLE  ,  //0x45 20 9
      KBD_NORMAL_STYLE  ,  //0x46 21 +
      KBD_NORMAL_STYLE  ,  //0x47 22 4
      KBD_NORMAL_STYLE  ,  //0x48 23 5
      KBD_NORMAL_STYLE  ,  //0x49 24 6
      KBD_NORMAL_STYLE  ,  //0x4a 25 =
      KBD_NORMAL_STYLE  ,  //0x4b 26 1
      KBD_NORMAL_STYLE  ,  //0x4c 27 2
      KBD_NORMAL_STYLE  ,  //0x4d 28 3
      KBD_SHORTER_STYLE ,  //0x4e 29 ENTER
      KBD_NORMAL_STYLE  ,  //0x4f 30 0
      KBD_NORMAL_STYLE  ,  //0x50 31 ,
      KBD_NORMAL_STYLE  ,  //0x51 32 .
      KBD_NORMAL_STYLE  ,  //0x52 33 記号入力
      KBD_NORMAL_STYLE  ,  //0x53 34 登録
      KBD_NORMAL_STYLE  ,  //0x54 35 HELP
      KBD_NARROWER_STYLE,  //0x55 36 XF1
      KBD_NARROWER_STYLE,  //0x56 37 XF2
      KBD_NARROW_STYLE  ,  //0x57 38 XF3
      KBD_NARROWER_STYLE,  //0x58 39 XF4
      KBD_NARROWER_STYLE,  //0x59 40 XF5
      KBD_NORMAL_STYLE  ,  //0x61 41 BREAK
      KBD_NORMAL_STYLE  ,  //0x62 42 COPY
      KBD_FUNCTION_STYLE,  //0x63 43 F1
      KBD_FUNCTION_STYLE,  //0x64 44 F2
      KBD_FUNCTION_STYLE,  //0x65 45 F3
      KBD_FUNCTION_STYLE,  //0x66 46 F4
      KBD_FUNCTION_STYLE,  //0x67 47 F5
      KBD_FUNCTION_STYLE,  //0x68 48 F6
      KBD_FUNCTION_STYLE,  //0x69 49 F7
      KBD_FUNCTION_STYLE,  //0x6a 50 F8
      KBD_FUNCTION_STYLE,  //0x6b 51 F9
      KBD_FUNCTION_STYLE,  //0x6c 52 F10
      KBD_SHIFT_STYLE   ,  //0x70 53 SHIFT
      KBD_NARROW_STYLE  ,  //0x71 54 CTRL
      KBD_NARROW_STYLE  ,  //0x72 55 OPT.1
      KBD_NARROW_STYLE  ,  //0x73 56 OPT.2
    },
    //Compact Keyboard
    {
      KBD_NORMAL_STYLE  ,  //0x01  0 ESC
      KBD_NARROW_STYLE  ,  //0x0f  1 BS
      KBD_NARROW_STYLE  ,  //0x10  2 TAB
      KBD_RETURN_STYLE  ,  //0x1d  3 RETURN
      KBD_NORMAL_STYLE  ,  //0x35  4 Space
      KBD_NORMAL_STYLE  ,  //0x36  5 HOME
      KBD_NORMAL_STYLE  ,  //0x37  6 DEL
      KBD_NORMAL_STYLE  ,  //0x38  7 ROLLUP
      KBD_NORMAL_STYLE  ,  //0x39  8 ROLLDOWN
      KBD_NORMAL_STYLE  ,  //0x3a  9 UNDO
      KBD_NORMAL_STYLE  ,  //0x3b 10 ←
      KBD_NORMAL_STYLE  ,  //0x3c 11 ↑
      KBD_NORMAL_STYLE  ,  //0x3d 12 →
      KBD_NORMAL_STYLE  ,  //0x3e 13 ↓
      KBD_NORMAL_STYLE  ,  //0x3f 14 CLR
      KBD_NONE_STYLE    ,  //0x40 15 /
      KBD_NONE_STYLE    ,  //0x41 16 *
      KBD_NONE_STYLE    ,  //0x42 17 -
      KBD_NONE_STYLE    ,  //0x43 18 7
      KBD_NONE_STYLE    ,  //0x44 19 8
      KBD_NONE_STYLE    ,  //0x45 20 9
      KBD_NONE_STYLE    ,  //0x46 21 +
      KBD_NONE_STYLE    ,  //0x47 22 4
      KBD_NONE_STYLE    ,  //0x48 23 5
      KBD_NONE_STYLE    ,  //0x49 24 6
      KBD_NONE_STYLE    ,  //0x4a 25 =
      KBD_NONE_STYLE    ,  //0x4b 26 1
      KBD_NONE_STYLE    ,  //0x4c 27 2
      KBD_NONE_STYLE    ,  //0x4d 28 3
      KBD_NONE_STYLE    ,  //0x4e 29 ENTER
      KBD_NONE_STYLE    ,  //0x4f 30 0
      KBD_NONE_STYLE    ,  //0x50 31 ,
      KBD_NONE_STYLE    ,  //0x51 32 .
      KBD_NORMAL_STYLE  ,  //0x52 33 記号入力
      KBD_NORMAL_STYLE  ,  //0x53 34 登録
      KBD_NORMAL_STYLE  ,  //0x54 35 HELP
      KBD_NARROWER_STYLE,  //0x55 36 XF1
      KBD_NARROWER_STYLE,  //0x56 37 XF2
      KBD_NARROWER_STYLE,  //0x57 38 XF3
      KBD_NORMAL_STYLE  ,  //0x58 39 XF4
      KBD_NORMAL_STYLE  ,  //0x59 40 XF5
      KBD_NORMAL_STYLE  ,  //0x61 41 BREAK
      KBD_NORMAL_STYLE  ,  //0x62 42 COPY
      KBD_FUNCTION_STYLE,  //0x63 43 F1
      KBD_FUNCTION_STYLE,  //0x64 44 F2
      KBD_FUNCTION_STYLE,  //0x65 45 F3
      KBD_FUNCTION_STYLE,  //0x66 46 F4
      KBD_FUNCTION_STYLE,  //0x67 47 F5
      KBD_FUNCTION_STYLE,  //0x68 48 F6
      KBD_FUNCTION_STYLE,  //0x69 49 F7
      KBD_FUNCTION_STYLE,  //0x6a 50 F8
      KBD_FUNCTION_STYLE,  //0x6b 51 F9
      KBD_FUNCTION_STYLE,  //0x6c 52 F10
      KBD_SHIFT_STYLE   ,  //0x70 53 SHIFT
      KBD_NARROW_STYLE  ,  //0x71 54 CTRL
      KBD_NORMAL_STYLE  ,  //0x72 55 OPT.1
      KBD_NORMAL_STYLE  ,  //0x73 56 OPT.2
    },
  };  //KBD_TYPE_FIX_STYLE

  //赤色LEDの色
  //  5A:かな,5B:ローマ字,5C:コード入力,5D:CAPS,5E:INS,74:NUM
  public static final int[][] KBD_RED_LED_COLOR = {
    { KBD_RED_OFF_KEYTOP_COLOR, KBD_RED_OFF_BORDER_COLOR },
    { KBD_RED_ON_KEYTOP_COLOR, KBD_RED_ON_BORDER_COLOR },
  };

  //緑色LEDの色
  //  5F:ひらがな,60:全角
  public static final int[][] KBD_GREEN_LED_COLOR = {
    { KBD_GREEN_OFF_KEYTOP_COLOR, KBD_GREEN_OFF_BORDER_COLOR },
    { KBD_GREEN_ON_KEYTOP_COLOR, KBD_GREEN_ON_BORDER_COLOR },
  };

  //LEDキーの番号→LEDの色
  public static final int[][][] KBD_LED_COLOR = {
    KBD_RED_LED_COLOR  ,  //0x5a  0 かな
    KBD_RED_LED_COLOR  ,  //0x5b  1 ローマ字
    KBD_RED_LED_COLOR  ,  //0x5c  2 コード入力
    KBD_RED_LED_COLOR  ,  //0x5d  3 CAPS
    KBD_RED_LED_COLOR  ,  //0x5e  4 INS
    KBD_GREEN_LED_COLOR,  //0x5f  5 ひらがな
    KBD_GREEN_LED_COLOR,  //0x60  6 全角
    KBD_RED_LED_COLOR  ,  //0x74  7 NUM
  };

  //Standard Keyboard
  //                       1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 3 3 3 3 3 3 4 4 4 4 4 4 4 4 4 4 5 5 5 5 5 5 5 5 5 5 6 6 6 6 6 6 6 6 6 6 7 7 7 7 7 7 7 7 7 7 8 8 8 8 8 8 8 8 8 8 9 9 9 9 9
  //   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4
  // 0 +-------+ +-------+ . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . +-------+-------+-------+ . +-------+-------+-------+-------+  0
  // 1 |61     | |62     | . +---------+---------+---------+---------+---------+ +---------+---------+---------+---------+---------+ . |5A     |5B     |5C     | . |5D     |52     |53     |54     |  1
  // 2 | BREAK | |  COPY | . |63       |64       |65       |66       |67       | |68       |69       |6A       |6B       |6C       | . | KANA  | ROMA  | CODE  | . | CAPS  | KIGOU |TOUROKU| HELP  |  2
  // 3 |       | |       | . |   F1    |   F2    |   F3    |   F4    |   F5    | |   F6    |   F7    |   F8    |   F9    |  F10    | . |  rrr  |  rrr  |  rrr  | . |  rrr  |       |       |       |  3
  // 4 +-------+ +-------+ . +---------+---------+---------+---------+---------+ +---------+---------+---------+---------+---------+ . +--rrr--+--rrr--+--rrr--+ . +--rrr--+-------+-------+-------+  4
  // 5 +-------+-------+-------+-------+-------+-------+-------+-------+-------+-------+-------+-------+-------+-------+-----------+ . +-------+-------+-------+ . +-------+-------+-------+-------+  5
  // 6 |01     |02     |03     |04     |05     |06     |07     |08     |09     |0A     |0B     |0C     |0D     |0E     | !0F     ! | . |36     |5E     |37     | . |3F     |40     |41     |42     |  6
  // 7 |  ESC  |  1!   |  2"   |  3#   |  4$   |  5%   |  6&   |  7'   |  8(   |  9)   |   0   |  -=   |  ^~   |  \|   | !  BS   ! | . | HOME  |  INS  |  DEL  | . |  CLR  |   /   |   *   |   -   |  7
  // 8 |       |       |       |       |       |       |       |       |       |       |       |       |       |       | !       ! | . |       |  rrr  |       | . |       | ~~~~~ | ~~~~~ | ~~~~~ |  8
  // 9 +-------+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-----------+ . +-------+--rrr--+-------+ . +-------+-------+-------+-------+  9
  //10 | !10     ! |11     |12     |13     |14     |15     |16     |17     |18     |19     |1A     |1B     |1C     | !1D         ! | . |38     |39     |3A     | . |43     |44     |45     |46     | 10
  //11 | !  TAB  ! |   Q   |   W   |   E   |   R   |   T   |   Y   |   U   |   I   |   O   |   P   |  @`   |  [{   | !           ! | . | R-UP  |R-DOWN | UNDO  | . |   7   |   8   |   9   |   +   | 11
  //12 | !       ! |       |       |       |       |       |       |       |       |       |       |       |       | !           ! | . |       |       |       | . | ~~~~~ | ~~~~~ | ~~~~~ | ~~~~~ | 12
  //13 +-----------+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+ ! RETURN  ! | . +-------+-------+-------+ . +-------+-------+-------+-------+ 13
  //14 | !71       ! |1E     |1F     |20     |21     |22     |23     |24     |25     |26     |27     |28     |29     | !         ! | . |       |3C     |       | . |47     |48     |49     |4A     | 14
  //15 | !  CTRL   ! |   A   |   S   |   D   |   F   |   G   |   H   |   J   |   K   |   L   |  ;+   |  :*   |  ]}   | !         ! | . | - - - |  UP   | - - - | . |   4   |   5   |   6   |   =   | 15
  //16 | !         ! |       |       |       |       |       |       |       |       |       |       |       |       | !         ! | . | 3B    |       |3D     | . | ~~~~~ | ~~~~~ | ~~~~~ | ~~~~~ | 16
  //17 +-------------+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-------------+ . | LEFT  +-------+ RIGHT | . +-------+-------+-------+-------+ 17
  //18 | !70           ! |2A     |2B     |2C     |2D     |2E     |2F     |30     |31     |32     |33     |34     | !70           ! | . |       |3E     |       | . |4B     |4C     |4D     | - - - | 18
  //19 | !    SHIFT    ! |   Z   |   X   |   C   |   V   |   B   |   M   |   N   |  ,<   |  .>   |  /?   |   _   | !    SHIFT    ! | . | - - - | DOWN  | - - - | . |   1   |   2   |   3   |4E     | 19
  //20 | !             ! |       |       |       |       |       |       |       |       |       |       |       | !             ! | . |       |       |       | . | ~~~~~ | ~~~~~ | ~~~~~ |       | 20
  //21 +-------------+---+---+---+-----+-+-------+-------+-------+-------+---+---+-------+-------+-+-----+---+---+---+-------------+ . +-------+---+---+-------+ . +-------+-------+-------+ ENTER | 21
  //22 . . . . . . . |5F     |!55     !|!56     !|35                         | !57     ! |!58     !|!59     !|60     | . . . . . . . . | !72     ! | !73     ! | . |4F     |50     |51     | ~~~~~ | 22
  //23 . . . . . . . | HIRA  |!  XF1  !|!  XF2  !|           SPACE           | !  XF3  ! |!  XF4  !|!  XF5  !|ZENKAKU| . . . . . . . . | ! OPT.1 ! | ! OPT.2 ! | . |   0   |   ,   |   .   |       | 23
  //24 . . . . . . . |  ggg  |!       !|!       !|                           | !       ! |!       !|!       !|  ggg  | . . . . . . . . | !       ! | !       ! | . | ~~~~~ | ~~~~~ | ~~~~~ | - - - | 24
  //25 . . . . . . . +--ggg--+---------+---------+---------------------------+-----------+---------+---------+--ggg--+ . . . . . . . . +-----------+-----------+ . +-------+-------+-------+-------+ 25
  //                       1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 3 3 3 3 3 3 4 4 4 4 4 4 4 4 4 4 5 5 5 5 5 5 5 5 5 5 6 6 6 6 6 6 6 6 6 6 7 7 7 7 7 7 7 7 7 7 8 8 8 8 8 8 8 8 8 8 9 9 9 9 9
  //   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4

  //Compact Keyboard
  //                       1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 3 3 3 3 3 3 4 4 4 4 4 4 4 4 4 4 5 5 5 5 5 5 5 5 5 5 6 6 6 6 6 6 6 6 6 6 7 7 7 7 7 7 7
  //   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6
  // 0 +-------+ +-------+ . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . +-------+-------+-------+  0
  // 1 |61     | |62     | . +---------+---------+---------+---------+---------+ +---------+---------+---------+---------+---------+ . |52     |53     |54     |  1
  // 2 | BREAK | |  COPY | . |63       |64       |65       |66       |67       | |68       |69       |6A       |6B       |6C       | . | KIGOU |TOUROKU| HELP  |  2
  // 3 |       | |       | . |   F1    |   F2    |   F3    |   F4    |   F5    | |   F6    |   F7    |   F8    |   F9    |  F10    | . |       |       |       |  3
  // 4 +-------+ +-------+ . +---------+---------+---------+---------+---------+ +---------+---------+---------+---------+---------+ . +-------+-------+-------+  4
  // 5 +-------+-------+-------+-------+-------+-------+-------+-------+-------+-------+-------+-------+-------+-------+-----------+ . |60     |5F     |5C     |  5
  // 6 |01     |02     |03     |04     |05     |06     |07     |08  /  |09  /  |0A  /  |0B  /  |0C  /  |0D     |0E     | !0F     ! | . |ZENKAKU| HIRA  | CODE  |  6
  // 7 |  ESC  |  1!   |  2"   |  3#   |  4$   |  5%   |  6&   | 7'/ 7 | 8(/ 8 | 9)/ 9 | 0 / / | -=/ * |  ^~   |  \|   | !  BS   ! | . |  ggg  |  ggg  |  rrr  |  7
  // 8 |       |       |       |       |       |       |       |  /  43|  /  44|  /  45|  /  40|  /  41|       |       | !       ! | . +--ggg--+--ggg--+--rrr--+  8
  // 9 +-------+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-----------+ . +-------+--rrr--+-------+  9
  //10 | !10     ! |11     |12     |13     |14     |15     |16     |17  /  |18  /  |19  /  |1A  /  |1B     |1C     | !1D         ! | . |36     |3F     |37     | 10
  //11 | !  TAB  ! |   Q   |   W   |   E   |   R   |   T   |   Y   | U / 4 | I / 5 | O / 6 | P / - |  @`   |  [{   | !           ! | . | HOME  |  CLR  |  DEL  | 11
  //12 | !       ! |       |       |       |       |       |       |  /  47|  /  48|  /  49|  /  42|       |       | !           ! | . |       |       |       | 12
  //13 +-----------+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+-----+-+ ! RETURN  ! | . +-------+-------+-------+ 13
  //14 | !71       ! |1E     |1F     |20     |21     |22     |23     |24  /  |25  /  |26  /  |27  /  |28     |29     | !         ! | . |74     |5E     |3A     | 14
  //15 | !  CTRL   ! |   A   |   S   |   D   |   F   |   G   |   H   | J / 1 | K / 2 | L / 3 | ;+/ + |  :*   |  ]}   | !         ! | . |  NUM  |  INS  | UNDO  | 15
  //16 | !         ! |       |       |       |       |       |       |  /  4B|  /  4C|  /  4D|  /  46|       |       | !         ! | . |  rrr  |  rrr  |       | 16
  //17 +-------------+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+-------------+ . +--rrr--+--rrr--+-------+ 17
  //18 | !70           ! |2A     |2B     |2C     |2D     |2E     |2F     |30  /  |31  /  |32  /  |33  /  |34  /  | !70           ! | . |38     |3C     |39     | 18
  //19 | !    SHIFT    ! |   Z   |   X   |   C   |   V   |   B   |   M   | N / 0 | ,</ , | .>/ . | /?/ = | _ /ENT| !    SHIFT    ! | . | R-UP  |  UP   |R-DOWN | 19
  //20 | !             ! |       |       |       |       |       |       |  /  4F|  /  50|  /  51|  /  4A|  /  4E| !             ! | . |       |       |       | 20
  //21 +-----+-------+---+---+---+-----+-+-------+-------+-------+-------+---+---+-----+-+-----+-+-----+-+-----+-+-----+-------+---+ . +-------+-------+-------+ 21
  //22 . . . |72     |5D     |!55     !|!56     !|35                         |! 57    !|58     |59     |5B     |5A     |73     | . . . |3B     |3E     |3D     | 22
  //23 . . . | OPT.1 | CAPS  |!  XF1  !|!  XF2  !|           SPACE           |!  XF3  !|  XF4  |  XF5  | ROMA  | KANA  | OPT.2 | . . . | LEFT  | DOWN  | RIGHT | 23
  //24 . . . |       |  rrr  |!       !|!       !|                           |!       !|       |       |  rrr  |  rrr  |       | . . . |       |       |       | 24
  //25 . . . +-------+--rrr--+---------+---------+---------------------------+---------+-------+-------+--rrr--+--rrr--+-------+ . . . +-------+-------+-------+ 25
  //                       1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2 2 3 3 3 3 3 3 3 3 3 3 4 4 4 4 4 4 4 4 4 4 5 5 5 5 5 5 5 5 5 5 6 6 6 6 6 6 6 6 6 6 7 7 7 7 7 7 7
  //   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6

  //可変キーの番号→ボックス
  //  Standard KeyboardとCompact Keyboardで共通
  //  欠番なし
  public static final int[][] KBD_VAR_BOX = {
    {  4,  5,  4, 4 },  //0x02  0 1
    {  8,  5,  4, 4 },  //0x03  1 2
    { 12,  5,  4, 4 },  //0x04  2 3
    { 16,  5,  4, 4 },  //0x05  3 4
    { 20,  5,  4, 4 },  //0x06  4 5
    { 24,  5,  4, 4 },  //0x07  5 6
    { 28,  5,  4, 4 },  //0x08  6 7
    { 32,  5,  4, 4 },  //0x09  7 8
    { 36,  5,  4, 4 },  //0x0a  8 9
    { 40,  5,  4, 4 },  //0x0b  9 0
    { 44,  5,  4, 4 },  //0x0c 10 -
    { 48,  5,  4, 4 },  //0x0d 11 ^
    { 52,  5,  4, 4 },  //0x0e 12 \
    {  6,  9,  4, 4 },  //0x11 13 q
    { 10,  9,  4, 4 },  //0x12 14 w
    { 14,  9,  4, 4 },  //0x13 15 e
    { 18,  9,  4, 4 },  //0x14 16 r
    { 22,  9,  4, 4 },  //0x15 17 t
    { 26,  9,  4, 4 },  //0x16 18 y
    { 30,  9,  4, 4 },  //0x17 19 u
    { 34,  9,  4, 4 },  //0x18 20 i
    { 38,  9,  4, 4 },  //0x19 21 o
    { 42,  9,  4, 4 },  //0x1a 22 p
    { 46,  9,  4, 4 },  //0x1b 23 @
    { 50,  9,  4, 4 },  //0x1c 24 [
    {  7, 13,  4, 4 },  //0x1e 25 a
    { 11, 13,  4, 4 },  //0x1f 26 s
    { 15, 13,  4, 4 },  //0x20 27 d
    { 19, 13,  4, 4 },  //0x21 28 f
    { 23, 13,  4, 4 },  //0x22 29 g
    { 27, 13,  4, 4 },  //0x23 30 h
    { 31, 13,  4, 4 },  //0x24 31 j
    { 35, 13,  4, 4 },  //0x25 32 k
    { 39, 13,  4, 4 },  //0x26 33 l
    { 43, 13,  4, 4 },  //0x27 34 ;
    { 47, 13,  4, 4 },  //0x28 35 :
    { 51, 13,  4, 4 },  //0x29 36 ]
    {  9, 17,  4, 4 },  //0x2a 37 z
    { 13, 17,  4, 4 },  //0x2b 38 x
    { 17, 17,  4, 4 },  //0x2c 39 c
    { 21, 17,  4, 4 },  //0x2d 40 v
    { 25, 17,  4, 4 },  //0x2e 41 b
    { 29, 17,  4, 4 },  //0x2f 42 n
    { 33, 17,  4, 4 },  //0x30 43 m
    { 37, 17,  4, 4 },  //0x31 44 ,
    { 41, 17,  4, 4 },  //0x32 45 .
    { 45, 17,  4, 4 },  //0x33 46 /
    { 49, 17,  4, 4 },  //0x34 47
  };  //KBD_VAR_BOX

  //種類→LEDキーの番号→ボックス
  //  欠番あり
  //    Standard KeyboardのNUMキー
  public static final int[][][] KBD_TYPE_LED_BOX = {
    //Standard Keyboard
    {
      { 64,  0,  4, 4 },  //0x5a  0 かな
      { 68,  0,  4, 4 },  //0x5b  1 ローマ字
      { 72,  0,  4, 4 },  //0x5c  2 コード入力
      { 78,  0,  4, 4 },  //0x5d  3 CAPS
      { 68,  5,  4, 4 },  //0x5e  4 INS
      {  7, 21,  4, 4 },  //0x5f  5 ひらがな
      { 51, 21,  4, 4 },  //0x60  6 全角
      null             ,  //0x74  7 NUM
    },
    //Compact Keyboard
    {
      { 52, 21,  4, 4 },  //0x5a  0 かな
      { 48, 21,  4, 4 },  //0x5b  1 ローマ字
      { 72,  4,  4, 4 },  //0x5c  2 コード入力
      {  7, 21,  4, 4 },  //0x5d  3 CAPS
      { 68, 13,  4, 4 },  //0x5e  4 INS
      { 68,  4,  4, 4 },  //0x5f  5 ひらがな
      { 64,  4,  4, 4 },  //0x60  6 全角
      { 64, 13,  4, 4 },  //0x74  7 NUM
    },
  };  //KBD_TYPE_LED_BOX

  //種類→固定キーのインデックス→ボックス
  //  欠番あり
  //    Compact Keyboardのテンキー
  public static final int[][][] KBD_TYPE_FIX_BOX = {
    //Standard Keyboard
    {
      {  0,  5,  4, 4 },  //0x01  0 ESC
      { 56,  5,  6, 4 },  //0x0f  1 BS
      {  0,  9,  6, 4 },  //0x10  2 TAB
      { 54,  9,  8, 8 },  //0x1d  3 RETURN
      { 21, 21, 14, 4 },  //0x35  4 Space
      { 64,  5,  4, 4 },  //0x36  5 HOME
      { 72,  5,  4, 4 },  //0x37  6 DEL
      { 64,  9,  4, 4 },  //0x38  7 ROLLUP
      { 68,  9,  4, 4 },  //0x39  8 ROLLDOWN
      { 72,  9,  4, 4 },  //0x3a  9 UNDO
      { 64, 13,  4, 8 },  //0x3b 10 ←
      { 68, 13,  4, 4 },  //0x3c 11 ↑
      { 72, 13,  4, 8 },  //0x3d 12 →
      { 68, 17,  4, 4 },  //0x3e 13 ↓
      { 78,  5,  4, 4 },  //0x3f 14 CLR
      { 82,  5,  4, 4 },  //0x40 15 /
      { 86,  5,  4, 4 },  //0x41 16 *
      { 90,  5,  4, 4 },  //0x42 17 -
      { 78,  9,  4, 4 },  //0x43 18 7
      { 82,  9,  4, 4 },  //0x44 19 8
      { 86,  9,  4, 4 },  //0x45 20 9
      { 90,  9,  4, 4 },  //0x46 21 +
      { 78, 13,  4, 4 },  //0x47 22 4
      { 82, 13,  4, 4 },  //0x48 23 5
      { 86, 13,  4, 4 },  //0x49 24 6
      { 90, 13,  4, 4 },  //0x4a 25 =
      { 78, 17,  4, 4 },  //0x4b 26 1
      { 82, 17,  4, 4 },  //0x4c 27 2
      { 86, 17,  4, 4 },  //0x4d 28 3
      { 90, 17,  4, 8 },  //0x4e 29 ENTER
      { 78, 21,  4, 4 },  //0x4f 30 0
      { 82, 21,  4, 4 },  //0x50 31 ,
      { 86, 21,  4, 4 },  //0x51 32 .
      { 82,  0,  4, 4 },  //0x52 33 記号入力
      { 86,  0,  4, 4 },  //0x53 34 登録
      { 90,  0,  4, 4 },  //0x54 35 HELP
      { 11, 21,  5, 4 },  //0x55 36 XF1
      { 16, 21,  5, 4 },  //0x56 37 XF2
      { 35, 21,  6, 4 },  //0x57 38 XF3
      { 41, 21,  5, 4 },  //0x58 39 XF4
      { 46, 21,  5, 4 },  //0x59 40 XF5
      {  0,  0,  4, 4 },  //0x61 41 BREAK
      {  5,  0,  4, 4 },  //0x62 42 COPY
      { 11,  0,  5, 4 },  //0x63 43 F1
      { 16,  0,  5, 4 },  //0x64 44 F2
      { 21,  0,  5, 4 },  //0x65 45 F3
      { 26,  0,  5, 4 },  //0x66 46 F4
      { 31,  0,  5, 4 },  //0x67 47 F5
      { 37,  0,  5, 4 },  //0x68 48 F6
      { 42,  0,  5, 4 },  //0x69 49 F7
      { 47,  0,  5, 4 },  //0x6a 50 F8
      { 52,  0,  5, 4 },  //0x6b 51 F9
      { 57,  0,  5, 4 },  //0x6c 52 F10
      {  0, 17, 62, 4 },  //0x70 53 SHIFT
      {  0, 13,  7, 4 },  //0x71 54 CTRL
      { 64, 21,  6, 4 },  //0x72 55 OPT.1
      { 70, 21,  6, 4 },  //0x73 56 OPT.2
    },
    //Compact Keyboard
    {
      {  0,  5,  4, 4 },  //0x01  0 ESC
      { 56,  5,  6, 4 },  //0x0f  1 BS
      {  0,  9,  6, 4 },  //0x10  2 TAB
      { 54,  9,  8, 8 },  //0x1d  3 RETURN
      { 21, 21, 14, 4 },  //0x35  4 Space
      { 64,  9,  4, 4 },  //0x36  5 HOME
      { 72,  9,  4, 4 },  //0x37  6 DEL
      { 64, 17,  4, 4 },  //0x38  7 ROLLUP
      { 72, 17,  4, 4 },  //0x39  8 ROLLDOWN
      { 72, 13,  4, 4 },  //0x3a  9 UNDO
      { 64, 21,  4, 4 },  //0x3b 10 ←
      { 68, 17,  4, 4 },  //0x3c 11 ↑
      { 72, 21,  4, 4 },  //0x3d 12 →
      { 68, 21,  4, 4 },  //0x3e 13 ↓
      { 68,  9,  4, 4 },  //0x3f 14 CLR
      null             ,  //0x40 15 /
      null             ,  //0x41 16 *
      null             ,  //0x42 17 -
      null             ,  //0x43 18 7
      null             ,  //0x44 19 8
      null             ,  //0x45 20 9
      null             ,  //0x46 21 +
      null             ,  //0x47 22 4
      null             ,  //0x48 23 5
      null             ,  //0x49 24 6
      null             ,  //0x4a 25 =
      null             ,  //0x4b 26 1
      null             ,  //0x4c 27 2
      null             ,  //0x4d 28 3
      null             ,  //0x4e 29 ENTER
      null             ,  //0x4f 30 0
      null             ,  //0x50 31 ,
      null             ,  //0x51 32 .
      { 64,  0,  4, 4 },  //0x52 33 記号入力
      { 68,  0,  4, 4 },  //0x53 34 登録
      { 72,  0,  4, 4 },  //0x54 35 HELP
      { 11, 21,  5, 4 },  //0x55 36 XF1
      { 16, 21,  5, 4 },  //0x56 37 XF2
      { 35, 21,  5, 4 },  //0x57 38 XF3
      { 40, 21,  4, 4 },  //0x58 39 XF4
      { 44, 21,  4, 4 },  //0x59 40 XF5
      {  0,  0,  4, 4 },  //0x61 41 BREAK
      {  5,  0,  4, 4 },  //0x62 42 COPY
      { 11,  0,  5, 4 },  //0x63 43 F1
      { 16,  0,  5, 4 },  //0x64 44 F2
      { 21,  0,  5, 4 },  //0x65 45 F3
      { 26,  0,  5, 4 },  //0x66 46 F4
      { 31,  0,  5, 4 },  //0x67 47 F5
      { 37,  0,  5, 4 },  //0x68 48 F6
      { 42,  0,  5, 4 },  //0x69 49 F7
      { 47,  0,  5, 4 },  //0x6a 50 F8
      { 52,  0,  5, 4 },  //0x6b 51 F9
      { 57,  0,  5, 4 },  //0x6c 52 F10
      {  0, 17, 62, 4 },  //0x70 53 SHIFT
      {  0, 13,  7, 4 },  //0x71 54 CTRL
      {  3, 21,  4, 4 },  //0x72 55 OPT.1
      { 56, 21,  4, 4 },  //0x73 56 OPT.2
    },
  };  //KBD_TYPE_FIX_BOX

  //可変キーの番号→可変キーに表示する文字の文字コード
  //  可変キーに表示する文字はモードキーの状態によって変化する
  //  かなの配列はSRAMの設定によって変化する
  //    0x00ed002b  キーボードのかなの配列。0=旧JIS,0以外=五十音
  //  文字コード0x5c,0x7e,0x7cの字体はSRAMの設定によって変化する
  //    0x00ed0059  bit0  文字コード0x5cの字体。0=YEN SIGN(0x5c,U+00A5),1=REVERSE SOLIDUS(0x80,U+005C)
  //                bit1  文字コード0x7eの字体。0=MACRON(0x7e,U+00AF),1=TILDE(0x81,U+007E)
  //                bit2  文字コード0x7cの字体。0=VERTICAL LINE(0x7c,U+007C),1=BROKEN BAR(0x82,U+00A6)
  //  文字の字体は半角と全角を区別しない
  //  文字の字体はカタカナとひらがなを区別する。ひらがな+全角のときだけひらがなにする
  //  テレビコントロール(SHIFT|OPT.2+←|↑|→|↓|CLR|/|*|-|7|8|9|+|4|5|6|=|1|2|3|0|,|.)は区別しない
  public static final char[] KBD_CAPS_BASE           = "1234567890-^\\QWERTYUIOP@[ASDFGHJKL;:]ZXCVBNM,./\u0000".toCharArray ();
  public static final char[] KBD_CAPS_SHIFT_BASE     = "!\"#$%&'()\u0000=~|qwertyuiop`{asdfghjkl+*}zxcvbnm<>?_".toCharArray ();
  public static final char[] KBD_KANA_BASE           = "ﾇﾌｱｳｴｵﾔﾕﾖﾜﾎﾍｰﾀﾃｲｽｶﾝﾅﾆﾗｾﾞﾟﾁﾄｼﾊｷｸﾏﾉﾘﾚｹﾑﾂｻｿﾋｺﾐﾓﾈﾙﾒﾛ".toCharArray ();
  public static final char[] KBD_KANA_SHIFT_BASE     = "ﾇﾌｧｩｪｫｬｭｮｦﾎﾍｰﾀﾃｨｽｶﾝﾅﾆﾗｾﾞ｢ﾁﾄｼﾊｷｸﾏﾉﾘﾚｹ｣ｯｻｿﾋｺﾐﾓ､｡･\u00a0".toCharArray ();
  public static final char[] KBD_KANA_AIU_BASE       = "ｱｲｳｴｵﾅﾆﾇﾈﾉﾗﾘﾙｶｷｸｹｺﾊﾋﾌﾍﾎﾚﾛｻｼｽｾｿﾏﾐﾑﾒﾓﾞﾟﾀﾁﾂﾃﾄﾔﾕﾖﾜｦﾝ".toCharArray ();
  public static final char[] KBD_KANA_AIU_SHIFT_BASE = "ｧｨｩｪｫﾅﾆﾇﾈｦﾗﾘﾙｶｷｸｹｺﾊﾋﾌﾍﾎﾚ｢\u00a0ｼｽｾｿﾏﾐﾑﾒﾓｰ｣ﾀﾁｯﾃﾄｬｭｮ､｡･".toCharArray ();
  public static final char[] KBD_ROMA_BASE           = "1234567890ｰ^\\QWERTYUIOP@｢ASDFGHJKL;:｣ZXCVBNM､｡･\u0000".toCharArray ();
  public static final char[] KBD_ROMA_SHIFT_BASE     = "!\"#$%&'()\u0000=~|QWeRTYuioP`{aSDFGHJKL+*}zXCVBNM<>?_".toCharArray ();
  public static final char[] KBD_NUM_BASE            = ("      789/" + "*        456-        123+        0,.=\u0084").toCharArray ();
  public static final byte[] KBD_CAPS_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NORMAL_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_CODE_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_CAPS_SHIFT_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_SHIFT_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_CTRL_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_KANA_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_KANA_SHIFT_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_KANA_AIU_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_KANA_AIU_SHIFT_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_ROMA_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_ROMA_SHIFT_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_CAPS_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_CODE_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_CAPS_SHIFT_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_SHIFT_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_CTRL_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_KANA_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_KANA_SHIFT_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_KANA_AIU_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_KANA_AIU_SHIFT_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_ROMA_CHR = new byte[KBD_VAR_KEYS];
  public static final byte[] KBD_NUM_ROMA_SHIFT_CHR = new byte[KBD_VAR_KEYS];

  //文字コード→可変キーに表示する文字のパターン
  //
  //       +0 +1 +2 +3 +4 +5 +6 +7 +8 +9 +a +b +c +d +e +f
  //  0x00 　 ^A ^B ^C ^D ^E ^F ^G ^H ^I ^J ^K ^L ^M ^N ^O
  //  0x10 ^P ^Q ^R ^S ^T ^U ^V ^W ^X ^Y ^Z ^[ ^¥ ^] ^^ ^_
  //  0x20 　 ！ ” ＃ ＄ ％ ＆ ’ （ ） ＊ ＋ ， － ． ／
  //  0x30 ０ １ ２ ３ ４ ５ ６ ７ ８ ９ ： ； ＜ ＝ ＞ ？
  //  0x40 ＠ Ａ Ｂ Ｃ Ｄ Ｅ Ｆ Ｇ Ｈ Ｉ Ｊ Ｋ Ｌ Ｍ Ｎ Ｏ
  //  0x50 Ｐ Ｑ Ｒ Ｓ Ｔ Ｕ Ｖ Ｗ Ｘ Ｙ Ｚ ［ ￥ ］ ＾ ＿
  //  0x60 ｀ ａ ｂ ｃ ｄ ｅ ｆ ｇ ｈ ｉ ｊ ｋ ｌ ｍ ｎ ｏ
  //  0x70 ｐ ｑ ｒ ｓ ｔ ｕ ｖ ｗ ｘ ｙ ｚ ｛ ｜ ｝ ￣ ^?
  //  0x80 ＼ ～ ￤ ^\ EN 　 を ぁ ぃ ぅ ぇ ぉ ゃ ゅ ょ っ
  //  0x90 　 あ い う え お か き く け こ さ し す せ そ
  //  0xa0 □ 。 「 」 、 ・ ヲ ァ ィ ゥ ェ ォ ャ ュ ョ ッ
  //  0xb0 ー ア イ ウ エ オ カ キ ク ケ コ サ シ ス セ ソ
  //  0xc0 タ チ ツ テ ト ナ ニ ヌ ネ ノ ハ ヒ フ ヘ ホ マ
  //  0xd0 ミ ム メ モ ヤ ユ ヨ ラ リ ル レ ロ ワ ン ゛ ゜
  //  0xe0 た ち つ て と な に ぬ ね の は ひ ふ へ ほ ま
  //  0xf0 み む め も や ゆ よ ら り る れ ろ わ ん 　 　
  //
  //    0x00         U+0020         SPACE。文字が割り当てられていないキーを表示するときに使うので^@ではなくて空白
  //    0x1c  ^¥     U+005E,U+00A5  CIRCUMFLEX ACCENT,YEN SIGN
  //    0x5c  ¥      U+00A5         YEN SIGN
  //    0x7c  |      U+007C         VERTICAL LINE
  //    0x7e  ¯      U+00AF         MACRON
  //    0x7f  ^?     U+005E,U+003F  CIRCUMFLEX ACCENT,QUESTION MARK
  //    0x80  \      U+005C         REVERSE SOLIDUS。0x5cの代わり
  //    0x81  ~      U+007E         TILDE。0x7eの代わり
  //    0x82  ¦      U+00A6         BROKEN BAR。0x7cの代わり
  //    0x83  ^\     U+005E,U+005C  CIRCUMFLEX ACCENT,REVERSE SOLIDUS。0x1cの代わり
  //    0x84  ENTER                 NUM ONのときに使う
  //    0xa0  □     U+00A0         NO-BREAK SPACE。空白と見分けがつかないと困るのでキーボードには市松模様で表示する
  //
  public static final int[][] KBD_VAR_LETTER = {
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x00
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000001000000,
      0b000001010000010100000,
      0b000010001000010100000,
      0b000100000100100010000,
      0b000000000000100010000,
      0b000000000001000001000,
      0b000000000001111111000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x01 ^A
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001111100000,
      0b000001010001000010000,
      0b000010001001000001000,
      0b000100000101000010000,
      0b000000000001111100000,
      0b000000000001000010000,
      0b000000000001000001000,
      0b000000000001000010000,
      0b000000000001111100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x02 ^B
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000011100000,
      0b000001010000100010000,
      0b000010001001000001000,
      0b000100000101000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000001000,
      0b000000000000100010000,
      0b000000000000011100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x03 ^C
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001111100000,
      0b000001010001000010000,
      0b000010001001000001000,
      0b000100000101000001000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000001000010000,
      0b000000000001111100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x04 ^D
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001111111000,
      0b000001010001000000000,
      0b000010001001000000000,
      0b000100000101000000000,
      0b000000000001111110000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001111111000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x05 ^E
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001111111000,
      0b000001010001000000000,
      0b000010001001000000000,
      0b000100000101000000000,
      0b000000000001111110000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x06 ^F
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000011110000,
      0b000001010000100000000,
      0b000010001001000000000,
      0b000100000101000000000,
      0b000000000001001111000,
      0b000000000001000010000,
      0b000000000001000010000,
      0b000000000000100010000,
      0b000000000000011100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x07 ^G
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000001000,
      0b000001010001000001000,
      0b000010001001000001000,
      0b000100000101000001000,
      0b000000000001111111000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x08 ^H
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000011100000,
      0b000001010000001000000,
      0b000010001000001000000,
      0b000100000100001000000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000011100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x09 ^I
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000000001000,
      0b000001010000000001000,
      0b000010001000000001000,
      0b000100000100000001000,
      0b000000000000000001000,
      0b000000000000000001000,
      0b000000000001000001000,
      0b000000000000100010000,
      0b000000000000011100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x0a ^J
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000001000,
      0b000001010001000010000,
      0b000010001001000100000,
      0b000100000101001000000,
      0b000000000001010000000,
      0b000000000001101000000,
      0b000000000001000100000,
      0b000000000001000010000,
      0b000000000001000001000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x0b ^K
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000000000,
      0b000001010001000000000,
      0b000010001001000000000,
      0b000100000101000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001111111000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x0c ^L
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000001000,
      0b000001010001100011000,
      0b000010001001010101000,
      0b000100000101010101000,
      0b000000000001001001000,
      0b000000000001001001000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x0d ^M
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000001000,
      0b000001010001000001000,
      0b000010001001100001000,
      0b000100000101010001000,
      0b000000000001001001000,
      0b000000000001000101000,
      0b000000000001000011000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x0e ^N
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000011100000,
      0b000001010000100010000,
      0b000010001001000001000,
      0b000100000101000001000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000000100010000,
      0b000000000000011100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x0f ^O
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001111100000,
      0b000001010001000010000,
      0b000010001001000001000,
      0b000100000101000001000,
      0b000000000001000010000,
      0b000000000001111100000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x10 ^P
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000011100000,
      0b000001010000100010000,
      0b000010001001000001000,
      0b000100000101000001000,
      0b000000000001000001000,
      0b000000000001001001000,
      0b000000000001000101000,
      0b000000000000100010000,
      0b000000000000011101000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x11 ^Q
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001111100000,
      0b000001010001000010000,
      0b000010001001000001000,
      0b000100000101000001000,
      0b000000000001000010000,
      0b000000000001111100000,
      0b000000000001000010000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x12 ^R
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000011110000,
      0b000001010000100000000,
      0b000010001001000000000,
      0b000100000100100000000,
      0b000000000000011100000,
      0b000000000000000010000,
      0b000000000000000001000,
      0b000000000000000010000,
      0b000000000000111100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x13 ^S
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001111111000,
      0b000001010000001000000,
      0b000010001000001000000,
      0b000100000100001000000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x14 ^T
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000001000,
      0b000001010001000001000,
      0b000010001001000001000,
      0b000100000101000001000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000000100010000,
      0b000000000000011100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x15 ^U
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000001000,
      0b000001010001000001000,
      0b000010001001000001000,
      0b000100000101000001000,
      0b000000000000100010000,
      0b000000000000100010000,
      0b000000000000010100000,
      0b000000000000010100000,
      0b000000000000001000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x16 ^V
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100010001000100,
      0b000001010010001000100,
      0b000010001010001000100,
      0b000100000101010101000,
      0b000000000001010101000,
      0b000000000001010101000,
      0b000000000000100010000,
      0b000000000000100010000,
      0b000000000000100010000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x17 ^W
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000001000,
      0b000001010001000001000,
      0b000010001000100010000,
      0b000100000100010100000,
      0b000000000000001000000,
      0b000000000000010100000,
      0b000000000000100010000,
      0b000000000001000001000,
      0b000000000001000001000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x18 ^X
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000001000,
      0b000001010001000001000,
      0b000010001001000001000,
      0b000100000100100010000,
      0b000000000000010100000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x19 ^Y
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001111111000,
      0b000001010000000001000,
      0b000010001000000010000,
      0b000100000100000100000,
      0b000000000000001000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000001111111000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x1a ^Z
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000011100000,
      0b000001010000010000000,
      0b000010001000010000000,
      0b000100000100010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000011100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x1b ^[
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000001000,
      0b000001010000100010000,
      0b000010001000010100000,
      0b000100000100001000000,
      0b000000000001111111000,
      0b000000000000001000000,
      0b000000000001111111000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x1c ^¥
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000011100000,
      0b000001010000000100000,
      0b000010001000000100000,
      0b000100000100000100000,
      0b000000000000000100000,
      0b000000000000000100000,
      0b000000000000000100000,
      0b000000000000000100000,
      0b000000000000011100000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x1d ^]
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000001000000,
      0b000001010000010100000,
      0b000010001000100010000,
      0b000100000101000001000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x1e ^^
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000000000000,
      0b000001010000000000000,
      0b000010001000000000000,
      0b000100000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001111111000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x1f ^_
    null,  //0x20
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x21 !
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x22 "
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000011111110000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000011111110000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x23 #
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000001111110000000,
      0b000000010010000000000,
      0b000000010010000000000,
      0b000000001111100000000,
      0b000000000010010000000,
      0b000000000010010000000,
      0b000000011111100000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x24 $
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011100000000000,
      0b000000010100010000000,
      0b000000011100100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001001110000000,
      0b000000010001010000000,
      0b000000000001110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x25 %
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001110000000000,
      0b000000010001000000000,
      0b000000010001000000000,
      0b000000001010000000000,
      0b000000000100000000000,
      0b000000001010000000000,
      0b000000010001010000000,
      0b000000010000100000000,
      0b000000001111010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x26 &
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x27 '
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x28 (
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x29 )
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000010010010000000,
      0b000000001010100000000,
      0b000000000111000000000,
      0b000000001010100000000,
      0b000000010010010000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x2a *
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x2b +
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000000110000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x2c ,
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x2d -
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x2e .
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x2f /
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x30 0
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001000000000,
      0b000000000011000000000,
      0b000000000101000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x31 1
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000001100000000,
      0b000000000110000000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x32 2
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000011100000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x33 3
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001100000000,
      0b000000000010100000000,
      0b000000000100100000000,
      0b000000001000100000000,
      0b000000010000100000000,
      0b000000010000100000000,
      0b000000011111110000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x34 4
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000011111100000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x35 5
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111100000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000011111100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x36 6
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x37 7
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x38 8
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001111110000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000001110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x39 9
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3a :
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000000110000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3b ;
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3c <
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3d =
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3e >
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000001000100000000,
      0b000000010000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3f ?
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000001000100000000,
      0b000000010010010000000,
      0b000000010101010000000,
      0b000000010101010000000,
      0b000000010101010000000,
      0b000000010011110000000,
      0b000000001000000000000,
      0b000000000111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x40 @
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000010000010000000,
      0b000000011111110000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x41 A
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111000000000,
      0b000000010000100000000,
      0b000000010000010000000,
      0b000000010000100000000,
      0b000000011111000000000,
      0b000000010000100000000,
      0b000000010000010000000,
      0b000000010000100000000,
      0b000000011111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x42 B
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000001000100000000,
      0b000000010000010000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000010000000,
      0b000000001000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x43 C
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111000000000,
      0b000000010000100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000100000000,
      0b000000011111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x44 D
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000011111100000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x45 E
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000011111100000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x46 F
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111100000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010011110000000,
      0b000000010000100000000,
      0b000000010000100000000,
      0b000000001000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x47 G
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000011111110000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x48 H
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x49 I
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000010000010000000,
      0b000000001000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4a J
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000010000000,
      0b000000010000100000000,
      0b000000010001000000000,
      0b000000010010000000000,
      0b000000010100000000000,
      0b000000011010000000000,
      0b000000010001000000000,
      0b000000010000100000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4b K
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4c L
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000010000000,
      0b000000011000110000000,
      0b000000010101010000000,
      0b000000010101010000000,
      0b000000010010010000000,
      0b000000010010010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4d M
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000011000010000000,
      0b000000010100010000000,
      0b000000010010010000000,
      0b000000010001010000000,
      0b000000010000110000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4e N
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000001000100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4f O
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111000000000,
      0b000000010000100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000100000000,
      0b000000011111000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x50 P
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000001000100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010010010000000,
      0b000000010001010000000,
      0b000000001000100000000,
      0b000000000111010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x51 Q
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111000000000,
      0b000000010000100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000100000000,
      0b000000011111000000000,
      0b000000010000100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x52 R
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111100000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000001000000000000,
      0b000000000111000000000,
      0b000000000000100000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000001111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x53 S
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x54 T
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x55 U
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x56 V
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100010001000000,
      0b000000100010001000000,
      0b000000100010001000000,
      0b000000010101010000000,
      0b000000010101010000000,
      0b000000010101010000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x57 W
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001000100000000,
      0b000000000101000000000,
      0b000000000010000000000,
      0b000000000101000000000,
      0b000000001000100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x58 X
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001000100000000,
      0b000000000101000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x59 Y
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5a Z
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5b [
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000010000000,
      0b000000001000100000000,
      0b000000000101000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5c ¥
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5d ]
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000101000000000,
      0b000000001000100000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5e ^
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5f _
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x60 `
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001001100000000,
      0b000000000110100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x61 a
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001011000000000,
      0b000000001100100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x62 b
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x63 c
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000111100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001001100000000,
      0b000000000110100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x64 d
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000001000100000000,
      0b000000001111100000000,
      0b000000001000000000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x65 e
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001100000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000001111100000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x66 f
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000000111100000000,
      0b000000000000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x67 g
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001111000000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x68 h
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x69 i
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000001000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x6a j
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001001100000000,
      0b000000001010000000000,
      0b000000001110000000000,
      0b000000001001000000000,
      0b000000001000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x6b k
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x6c l
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011101100000000,
      0b000000010010010000000,
      0b000000010010010000000,
      0b000000010010010000000,
      0b000000010010010000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x6d m
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001011000000000,
      0b000000001100100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x6e n
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x6f o
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001011000000000,
      0b000000001100100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001111000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x70 p
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110100000000,
      0b000000001001100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000000111100000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x71 q
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001011000000000,
      0b000000001100100000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x72 r
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111100000000,
      0b000000001000000000000,
      0b000000000111000000000,
      0b000000000000100000000,
      0b000000001111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x73 s
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000001111100000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000001100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x74 t
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001001100000000,
      0b000000000110100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x75 u
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x76 v
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010010010000000,
      0b000000010010010000000,
      0b000000010010010000000,
      0b000000010010010000000,
      0b000000001101110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x77 w
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000100000000,
      0b000000000101000000000,
      0b000000000010000000000,
      0b000000000101000000000,
      0b000000001000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x78 x
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000000111100000000,
      0b000000000000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x79 y
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x7a z
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x7b {
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x7c |
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x7d }
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x7e ¯
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000011100000,
      0b000001010000100010000,
      0b000010001001000001000,
      0b000100000100000001000,
      0b000000000000000010000,
      0b000000000000000100000,
      0b000000000000001000000,
      0b000000000000000000000,
      0b000000000000001000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x7f ^?
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000000000000,
      0b000000001000000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000000100000000,
      0b000000000000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x80 \
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001100000000000,
      0b000000010010010000000,
      0b000000000001100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x81 ~
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x82 ¦
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000000000,
      0b000001010000100000000,
      0b000010001000010000000,
      0b000100000100010000000,
      0b000000000000001000000,
      0b000000000000000100000,
      0b000000000000000100000,
      0b000000000000000010000,
      0b000000000000000001000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x83 ^\
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b011101010111011101100,
      0b010001010010010001010,
      0b010001110010010001010,
      0b011101110010011101100,
      0b010001110010010001010,
      0b010001010010010001010,
      0b011101010010011101010,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x84 ENTER
    null,  //0x85
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000011111100000000,
      0b000000000100000000000,
      0b000000001110000000000,
      0b000000010011100000000,
      0b000000000101000000000,
      0b000000001001000000000,
      0b000000001000000000000,
      0b000000000111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x86 を
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000011111110000000,
      0b000000000100000000000,
      0b000000001111100000000,
      0b000000010101010000000,
      0b000000010010010000000,
      0b000000001100100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x87 ぁ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010001000000000,
      0b000000010000100000000,
      0b000000010000100000000,
      0b000000010000010000000,
      0b000000010100010000000,
      0b000000001000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x88 ぃ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001110000000000,
      0b000000000000000000000,
      0b000000001111000000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x89 ぅ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000110000000000,
      0b000000001001100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x8a ぇ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000011110100000000,
      0b000000000100100000000,
      0b000000001111000000000,
      0b000000010100100000000,
      0b000000010100100000000,
      0b000000001101000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x8b ぉ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001010000000000,
      0b000000001111100000000,
      0b000000011001010000000,
      0b000000000101010000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x8c ゃ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010010000000000,
      0b000000010111000000000,
      0b000000011010100000000,
      0b000000010010100000000,
      0b000000010100100000000,
      0b000000000101000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x8d ゅ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000011100000000,
      0b000000000010000000000,
      0b000000000111000000000,
      0b000000001010100000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x8e ょ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111000000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x8f っ
    null,  //0x90
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000011111110000000,
      0b000000000100000000000,
      0b000000001111100000000,
      0b000000010101010000000,
      0b000000100101001000000,
      0b000000100101001000000,
      0b000000100010001000000,
      0b000000011100010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x91 あ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100000100000000,
      0b000000100000010000000,
      0b000000100000010000000,
      0b000000100000010000000,
      0b000000100000001000000,
      0b000000100000001000000,
      0b000000101000001000000,
      0b000000010000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x92 い
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111000000000,
      0b000000000000000000000,
      0b000000001111000000000,
      0b000000110000100000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x93 う
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000101000000000,
      0b000000001001000000000,
      0b000000010000110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x94 え
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000111110100000000,
      0b000000001000010000000,
      0b000000001111001000000,
      0b000000011000100000000,
      0b000000101000010000000,
      0b000001001000010000000,
      0b000000101000100000000,
      0b000000011001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x95 お
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100100000000,
      0b000000000100010000000,
      0b000000111110010000000,
      0b000000001001001000000,
      0b000000001001001000000,
      0b000000010001000000000,
      0b000000010001000000000,
      0b000000100001000000000,
      0b000000100110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x96 か
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000001000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000010000000000000,
      0b000000001111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x97 き
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x98 く
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000000000,
      0b000000100001000000000,
      0b000000101111110000000,
      0b000000100001000000000,
      0b000000100001000000000,
      0b000000100001000000000,
      0b000000100001000000000,
      0b000000100010000000000,
      0b000000100100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x99 け
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000001111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x9a こ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000000100100000000,
      0b000000011111000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000001111100000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000001111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x9b さ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x9c し
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001000000000,
      0b000000111111110000000,
      0b000000000001000000000,
      0b000000001101000000000,
      0b000000010011000000000,
      0b000000010011000000000,
      0b000000001101000000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x9d す
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000111111111000000,
      0b000000001000100000000,
      0b000000001001100000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000000111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x9e せ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000001100000000000,
      0b000000111111111000000,
      0b000000000110000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000000111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x9f そ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010101010000000,
      0b000000000000000000000,
      0b000000010101010000000,
      0b000000000000000000000,
      0b000000010101010000000,
      0b000000000000000000000,
      0b000000010101010000000,
      0b000000000000000000000,
      0b000000010101010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xa0 □
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000001001000000000,
      0b000000001001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xa1 。
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xa2 「
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xa3 」
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xa4 、
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xa5 ・
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xa6 ヲ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000000100000000,
      0b000000000101000000000,
      0b000000000110000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xa7 ァ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000011000000000,
      0b000000000101000000000,
      0b000000001001000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xa8 ィ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000001111100000000,
      0b000000001000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xa9 ゥ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xaa ェ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001000000000,
      0b000000001111100000000,
      0b000000000001000000000,
      0b000000000011000000000,
      0b000000000101000000000,
      0b000000001001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xab ォ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000001111100000000,
      0b000000000100100000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xac ャ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xad ュ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000000100000000,
      0b000000001111100000000,
      0b000000000000100000000,
      0b000000001111100000000,
      0b000000000000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xae ョ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001010100000000,
      0b000000001010100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xaf ッ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000000000000,
      0b000000001111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xb0 ー
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000100100000000,
      0b000000000111000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xb1 ア
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000011000000000,
      0b000000000101000000000,
      0b000000001001000000000,
      0b000000010001000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xb2 イ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xb3 ウ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xb4 エ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000011111110000000,
      0b000000000001000000000,
      0b000000000011000000000,
      0b000000000101000000000,
      0b000000001001000000000,
      0b000000010001000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xb5 オ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000010010000000,
      0b000000000100010000000,
      0b000000000100010000000,
      0b000000001000010000000,
      0b000000001000010000000,
      0b000000010001100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xb6 カ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100100000000,
      0b000000000101000000000,
      0b000000011110010000000,
      0b000000000010100000000,
      0b000000000011000000000,
      0b000000011101000000000,
      0b000000000001000000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xb7 キ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000011110000000,
      0b000000000100010000000,
      0b000000000100010000000,
      0b000000001000100000000,
      0b000000010000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xb8 ク
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000111110000000,
      0b000000001001000000000,
      0b000000010001000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xb9 ケ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xba コ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000011111110000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xbb サ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011000000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000011000010000000,
      0b000000000100010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000011100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xbc シ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000101000000000,
      0b000000001000100000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xbd ス
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001011110000000,
      0b000000011100010000000,
      0b000000001000100000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000000111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xbe セ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000010000000,
      0b000000010000010000000,
      0b000000001000010000000,
      0b000000000100010000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xbf ソ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000011110000000,
      0b000000000100010000000,
      0b000000000110010000000,
      0b000000001001100000000,
      0b000000010000100000000,
      0b000000000001010000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xc0 タ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000100000000,
      0b000000000011000000000,
      0b000000001110000000000,
      0b000000000001000000000,
      0b000000011111110000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xc1 チ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100010000000,
      0b000000000010010000000,
      0b000000010010010000000,
      0b000000001000010000000,
      0b000000001000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xc2 ツ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xc3 テ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001110000000000,
      0b000000001001000000000,
      0b000000001000100000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xc4 ト
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xc5 ナ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xc6 ニ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000001000100000000,
      0b000000000101000000000,
      0b000000000010000000000,
      0b000000000101000000000,
      0b000000001000100000000,
      0b000000010000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xc7 ヌ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000110100000000,
      0b000000011010010000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xc8 ネ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000011000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xc9 ノ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000001001000000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xca ハ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000110000000,
      0b000000011111000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000001111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xcb ヒ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xcc フ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000001010000000000,
      0b000000001001000000000,
      0b000000010001000000000,
      0b000000010000100000000,
      0b000000000000100000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xcd ヘ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000001010100000000,
      0b000000010010010000000,
      0b000000010010010000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xce ホ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000010001000000000,
      0b000000001010000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xcf マ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001110000000000,
      0b000000000001100000000,
      0b000000000000010000000,
      0b000000001110000000000,
      0b000000000001100000000,
      0b000000000000010000000,
      0b000000011110000000000,
      0b000000000001100000000,
      0b000000000000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xd0 ミ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000001000100000000,
      0b000000001000100000000,
      0b000000010011010000000,
      0b000000011100010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xd1 ム
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000011001000000000,
      0b000000000110000000000,
      0b000000000011000000000,
      0b000000000100100000000,
      0b000000001000010000000,
      0b000000010000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xd2 メ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000001110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xd3 モ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000001001110000000,
      0b000000000110010000000,
      0b000000011100100000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xd4 ヤ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xd5 ユ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xd6 ヨ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xd7 ラ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xd8 リ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001010000000000,
      0b000000001010000000000,
      0b000000001010000000000,
      0b000000001010000000000,
      0b000000001010000000000,
      0b000000001010010000000,
      0b000000001010010000000,
      0b000000010010100000000,
      0b000000010011000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xd9 ル
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000100000000,
      0b000000010011000000000,
      0b000000011100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xda レ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000011111110000000,
      0b000000010000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xdb ロ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xdc ワ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010000000000000,
      0b000000001000000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000011000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xdd ン
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010010000000000,
      0b000000001001000000000,
      0b000000000100100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xde ゛
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000001001000000000,
      0b000000001001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xdf ゜
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000011111110000000,
      0b000000000100000000000,
      0b000000001001110000000,
      0b000000001000010000000,
      0b000000001000000000000,
      0b000000010010000000000,
      0b000000010001110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xe0 た
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000111111111000000,
      0b000000001000000000000,
      0b000000001011100000000,
      0b000000011100010000000,
      0b000000010000001000000,
      0b000000000000001000000,
      0b000000000000010000000,
      0b000000000111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xe1 ち
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000110000010000000,
      0b000000000000001000000,
      0b000000000000001000000,
      0b000000000000010000000,
      0b000000000001100000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xe2 つ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000111111110000000,
      0b000000000001100000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000000100000000000,
      0b000000000011100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xe3 て
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000001001110000000,
      0b000000000110000000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000001111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xe4 と
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001001000000000,
      0b000000001000100000000,
      0b000000111100010000000,
      0b000000010001010000000,
      0b000000010001000000000,
      0b000000100111000000000,
      0b000000001001100000000,
      0b000000001001010000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xe5 な
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010011100000000,
      0b000000010000010000000,
      0b000000010000100000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000010100000000000,
      0b000000010100000000000,
      0b000000010011110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xe6 に
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001000000000,
      0b000000010001000000000,
      0b000000010111110000000,
      0b000000001010001000000,
      0b000000011010001000000,
      0b000000101010001000000,
      0b000000100100111000000,
      0b000000100101001100000,
      0b000000011000110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xe7 ぬ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000111011100000000,
      0b000000001100010000000,
      0b000000001000010000000,
      0b000000011000010000000,
      0b000000011001110000000,
      0b000000101010011000000,
      0b000000101010010000000,
      0b000000001001100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xe8 ね
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010010010000000,
      0b000000100010001000000,
      0b000000100010001000000,
      0b000000100010001000000,
      0b000000100100001000000,
      0b000000011000010000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xe9 の
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000100001000000000,
      0b000000100001000000000,
      0b000000101111110000000,
      0b000000100001000000000,
      0b000000100001000000000,
      0b000000100111000000000,
      0b000000101001100000000,
      0b000000101001010000000,
      0b000000100110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xea は
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000111001000000000,
      0b000000001001100000000,
      0b000000010001010000000,
      0b000000100001000000000,
      0b000000100001000000000,
      0b000000100001000000000,
      0b000000100010000000000,
      0b000000011100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xeb ひ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000100100010000000,
      0b000000100010010000000,
      0b000001000001001000000,
      0b000001000001001000000,
      0b000001010001001000000,
      0b000000001110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xec ふ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000001001000000000,
      0b000000010000100000000,
      0b000000100000010000000,
      0b000000000000001000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xed へ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000101111110000000,
      0b000000100001000000000,
      0b000000101111110000000,
      0b000000100001000000000,
      0b000000100001000000000,
      0b000000100111100000000,
      0b000000101001010000000,
      0b000000101001000000000,
      0b000000100110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xee ほ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000001111000000000,
      0b000000010010100000000,
      0b000000010010010000000,
      0b000000001100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xef ま
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011110000000000,
      0b000000000010000000000,
      0b000000000010010000000,
      0b000000001111110000000,
      0b000000010100011000000,
      0b000000100100010000000,
      0b000000100100010000000,
      0b000000011001100000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xf0 み
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000111111110000000,
      0b000000000100000000000,
      0b000000000100110000000,
      0b000000011100001000000,
      0b000000100100000000000,
      0b000000100100010000000,
      0b000000011100010000000,
      0b000000000011100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xf1 む
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001000000000,
      0b000000010001000000000,
      0b000000010111100000000,
      0b000000001010010000000,
      0b000000011010001000000,
      0b000000101010001000000,
      0b000000100100001000000,
      0b000000100100001000000,
      0b000000011000110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xf2 め
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000011111100000000,
      0b000000000100000000000,
      0b000000011111100000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000100010000000,
      0b000000000100010000000,
      0b000000000011100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xf3 も
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000010010000000000,
      0b000000010011110000000,
      0b000000001101001000000,
      0b000000011001001000000,
      0b000000101000010000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000010000000000,
      0b000000000001000000000,
      0b000000000000000000000,
    },  //0xf4 や
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000100001000000000,
      0b000000100111110000000,
      0b000000101001001000000,
      0b000000010001001000000,
      0b000000010001001000000,
      0b000000010010001000000,
      0b000000000010010000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xf5 ゆ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000011110000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000001111000000000,
      0b000000010010100000000,
      0b000000010010010000000,
      0b000000001100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xf6 よ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011000000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000010000000000000,
      0b000000010111100000000,
      0b000000011000010000000,
      0b000000010000010000000,
      0b000000000000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xf7 ら
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000010011100000000,
      0b000000010100010000000,
      0b000000011000010000000,
      0b000000011000010000000,
      0b000000010000010000000,
      0b000000010000100000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xf8 り
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000000111010000000,
      0b000000001000110000000,
      0b000000000111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xf9 る
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000111011100000000,
      0b000000001100010000000,
      0b000000001000010000000,
      0b000000011000010000000,
      0b000000011000100000000,
      0b000000101000100000000,
      0b000000101001000000000,
      0b000000001001110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xfa れ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xfb ろ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000111011100000000,
      0b000000001100010000000,
      0b000000011000010000000,
      0b000000011000010000000,
      0b000000111000010000000,
      0b000000101000100000000,
      0b000000001001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xfc わ
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000001000000000000,
      0b000000011100000000000,
      0b000000010010010000000,
      0b000000100010010000000,
      0b000000100001100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0xfd ん
    null,  //0xfe
    null,  //0xff
  };

  //LEDキーの番号→LEDキーに表示する文字のパターン
  public static final int[][] KBD_LED_LETTER = {
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000001001000001001000,
      0b000001000100001000100,
      0b001111100100111100010,
      0b000010010010010001010,
      0b000010010010010001000,
      0b000100010000100111000,
      0b000100010000001001100,
      0b001000010000001001010,
      0b001001100000000110000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5a  0 かな
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001111100000011111000,
      0b001000100000000001000,
      0b001000100000000001000,
      0b001000101111010010000,
      0b001000100000001100000,
      0b001000100000000100000,
      0b001111100000000010000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5b  1 ローマ字
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000010100,
      0b001111100000001010100,
      0b000000100000001000000,
      0b000000100000001000000,
      0b000000101111001110000,
      0b000000100000001001000,
      0b000000100000001000000,
      0b001111100000001000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5c  2 コード入力
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001100011001110001100,
      0b010010100101001010010,
      0b010000100101001010000,
      0b010000111101110001100,
      0b010000100101000000010,
      0b010010100101000010010,
      0b001100100101000001100,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5d  3 CAPS
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000111001000100111000,
      0b000010001000101000100,
      0b000010001100101000000,
      0b000010001010100111000,
      0b000010001001100000100,
      0b000010001000101000100,
      0b000111001000100111000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5e  4 INS
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000110000000,
      0b000111001000001100000,
      0b000001001100000000000,
      0b000010001010100000000,
      0b000100001000101111000,
      0b000100001000110000100,
      0b000100001000100000100,
      0b000100010000000001000,
      0b000011100000001110000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x5f  5 ひらがな
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000001000000011110000,
      0b000010100000100010000,
      0b000100010001111111100,
      0b011000001100100100100,
      0b000111110000111111100,
      0b000001000000100100100,
      0b000111110000111111100,
      0b000001000000100000100,
      0b001111111001000001100,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x60  6 全角
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001000101000101000100,
      0b001000101000101101100,
      0b001100101000101010100,
      0b001010101000101000100,
      0b001001101000101000100,
      0b001000101000101000100,
      0b001000100111001000100,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x74  7 NUM
  };

  //固定キーの番号→固定キーに表示する文字のパターン
  public static final int[][] KBD_FIX_LETTER = {
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001111100111000111000,
      0b001000001000101000100,
      0b001000001000001000000,
      0b001111000111001000000,
      0b001000000000101000000,
      0b001000001000101000100,
      0b001111100111000111000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x01  0 ESC
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000001111000111000000,
      0b000001000101000100000,
      0b000001000101000000000,
      0b000001111000111000000,
      0b000001000100000100000,
      0b000001000101000100000,
      0b000001111000111000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x0f  1 BS
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001111100010001111000,
      0b000010000101001000100,
      0b000010000101001000100,
      0b000010001000101111000,
      0b000010001111101000100,
      0b000010001000101000100,
      0b000010001000101111000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x10  2 TAB
    {
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000001110000000000,
      0b0,0b00000000000000000001010000000000,
      0b0,0b00000000000000000001010000000000,
      0b0,0b00000000000000000001010000000000,
      0b0,0b00000000000000000001010000000000,
      0b0,0b00000000000000000001010000000000,
      0b0,0b00000000000000000001010000000000,
      0b0,0b00000000000000000001010000000000,
      0b0,0b00000000000000000001010000000000,
      0b0,0b00000000000000000001010000000000,
      0b0,0b00000000000000001001010000000000,
      0b0,0b00000000000000011001010000000000,
      0b0,0b00000000000000101111010000000000,
      0b0,0b00000000000001000000010000000000,
      0b0,0b00000000000000101111110000000000,
      0b0,0b00000000000000011000000000000000,
      0b0,0b00000000000000001000000000000000,
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000000000000000000,
      0b0,0b00000000000000000000000000000000,
    },  //0x1d  3 RETURN
    null,  //0x35  4 Space
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b010010011001001011110,
      0b010010100101111010000,
      0b010010100101111010000,
      0b011110100101001011100,
      0b010010100101001010000,
      0b010010100101001010000,
      0b010010011001001011110,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x36  5 HOME
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001110001111101000000,
      0b001001001000001000000,
      0b001000101000001000000,
      0b001000101111001000000,
      0b001000101000001000000,
      0b001001001000001000000,
      0b001110001111101111100,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x37  6 DEL
    {
      0b000000000000000000000,
      0b011100011001000010000,
      0b010010100101000010000,
      0b011100100101000010000,
      0b010010100101000010000,
      0b010010011001111011110,
      0b000000000000000000000,
      0b000000100101110000000,
      0b000000100101001000000,
      0b000000100101110000000,
      0b000000100101000000000,
      0b000000011001000000000,
      0b000000000000000000000,
    },  //0x38  7 ROLLUP
    {
      0b000000000000000000000,
      0b011100011001000010000,
      0b010010100101000010000,
      0b011100100101000010000,
      0b010010100101000010000,
      0b010010011001111011110,
      0b000000000000000000000,
      0b011100011001001010010,
      0b010010100101001011010,
      0b010010100101111011110,
      0b010010100101111010110,
      0b011100011001001010010,
      0b000000000000000000000,
    },  //0x39  8 ROLLDOWN
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b010010100101100001100,
      0b010010100101010010010,
      0b010010110101001010010,
      0b010010111101001010010,
      0b010010101101001010010,
      0b010010100101010010010,
      0b001100100101100001100,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3a  9 UNDO
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000100000000000,
      0b000000001100000000000,
      0b000000010111111000000,
      0b000000100000001000000,
      0b000000010111111000000,
      0b000000001100000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3b 10 ←
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000101000000000,
      0b000000001000100000000,
      0b000000011101110000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3c 11 ↑
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001000000000,
      0b000000000001100000000,
      0b000000111111010000000,
      0b000000100000001000000,
      0b000000111111010000000,
      0b000000000001100000000,
      0b000000000001000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3d 12 →
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111000000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000000101000000000,
      0b000000011101110000000,
      0b000000001000100000000,
      0b000000000101000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3e 13 ↓
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000111001000001111000,
      0b001000101000001000100,
      0b001000001000001000100,
      0b001000001000001111000,
      0b001000001000001000100,
      0b001000101000001000100,
      0b000111001111101000100,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x3f 14 CLR
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x40 15 /
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000010010010000000,
      0b000000001010100000000,
      0b000000000111000000000,
      0b000000001010100000000,
      0b000000010010010000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x41 16 *
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x42 17 -
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000000001000000000,
      0b000000000001000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x43 18 7
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x44 19 8
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001111110000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000100000000,
      0b000000001111000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x45 20 9
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000011111110000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x46 21 +
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000001100000000,
      0b000000000010100000000,
      0b000000000100100000000,
      0b000000001000100000000,
      0b000000010000100000000,
      0b000000010000100000000,
      0b000000011111110000000,
      0b000000000000100000000,
      0b000000000000100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x47 22 4
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000011111100000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x48 23 5
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000111100000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000010000000000000,
      0b000000011111100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x49 24 6
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4a 25 =
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000010000000000,
      0b000000000110000000000,
      0b000000001010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000010000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4b 26 1
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000001100000000,
      0b000000000110000000000,
      0b000000001000000000000,
      0b000000010000000000000,
      0b000000011111110000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4c 27 2
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000000011100000000,
      0b000000000000010000000,
      0b000000000000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4d 28 3
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b011101010111011101100,
      0b010001010010010001010,
      0b010001110010010001010,
      0b011101110010011101100,
      0b010001110010010001010,
      0b010001010010010001010,
      0b011101010010011101010,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4e 29 ENTER
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000001111100000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000010000010000000,
      0b000000001111100000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x4f 30 0
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000000110000000000,
      0b000000000010000000000,
      0b000000000100000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x50 31 ,
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000110000000000,
      0b000000000110000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x51 32 .
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000111011110011111000,
      0b000000000010010001000,
      0b001111000010011111000,
      0b000000011110000000000,
      0b000111010000111111100,
      0b000000010000001000000,
      0b000111010000001111000,
      0b000101010010000001000,
      0b000111011110011111000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x52 33 記号入力
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b010111000100110011100,
      0b001010101001001000100,
      0b000111110010111011100,
      0b001000001001100000100,
      0b010111110100110111110,
      0b000100010001100001000,
      0b000111110010101011110,
      0b000010100001110011100,
      0b001111111011000101010,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x53 34 登録
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b010010111101000011100,
      0b010010100001000010010,
      0b010010100001000010010,
      0b011110111001000011100,
      0b010010100001000010000,
      0b010010100001000010000,
      0b010010111101111010000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x54 35 HELP
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001000101111100010000,
      0b001000101000000110000,
      0b000101001000000010000,
      0b000010001111000010000,
      0b000101001000000010000,
      0b001000101000000010000,
      0b001000101000000010000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x55 36 XF1
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001000101111100111000,
      0b001000101000001000100,
      0b000101001000000000100,
      0b000010001111000011000,
      0b000101001000000100000,
      0b001000101000001000000,
      0b001000101000001111100,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x56 37 XF2
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001000101111100111000,
      0b001000101000001000100,
      0b000101001000000000100,
      0b000010001111000011000,
      0b000101001000000000100,
      0b001000101000001000100,
      0b001000101000000111000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x57 38 XF3
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001000101111100001000,
      0b001000101000000011000,
      0b000101001000000101000,
      0b000010001111001001000,
      0b000101001000001111100,
      0b001000101000000001000,
      0b001000101000000001000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x58 39 XF4
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001000101111101111100,
      0b001000101000001000000,
      0b000101001000001111000,
      0b000010001111000000100,
      0b000101001000000000100,
      0b001000101000001000100,
      0b001000101000000111000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x59 40 XF5
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b011001100111001001010,
      0b010101010100010101010,
      0b010101010100010101010,
      0b011001100111011101100,
      0b010101010100010101010,
      0b010101010100010101010,
      0b011001010111010101010,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x61 41 BREAK
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001100011001110100010,
      0b010010100101001100010,
      0b010000100101001010100,
      0b010000100101110001000,
      0b010000100101000001000,
      0b010010100101000001000,
      0b001100011001000001000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x62 42 COPY
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000001111100010000000000,
      0b000000001000000110000000000,
      0b000000001000000010000000000,
      0b000000001111000010000000000,
      0b000000001000000010000000000,
      0b000000001000000010000000000,
      0b000000001000000010000000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x63 43 F1
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000001111100111000000000,
      0b000000001000001000100000000,
      0b000000001000000000100000000,
      0b000000001111000011000000000,
      0b000000001000000100000000000,
      0b000000001000001000000000000,
      0b000000001000001111100000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x64 44 F2
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000001111100111000000000,
      0b000000001000001000100000000,
      0b000000001000000000100000000,
      0b000000001111000011000000000,
      0b000000001000000000100000000,
      0b000000001000001000100000000,
      0b000000001000000111000000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x65 45 F3
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000001111100001000000000,
      0b000000001000000011000000000,
      0b000000001000000101000000000,
      0b000000001111001001000000000,
      0b000000001000001111100000000,
      0b000000001000000001000000000,
      0b000000001000000001000000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x66 46 F4
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000001111101111100000000,
      0b000000001000001000000000000,
      0b000000001000001111000000000,
      0b000000001111000000100000000,
      0b000000001000000000100000000,
      0b000000001000001000100000000,
      0b000000001000000111000000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x67 47 F5
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000001111100011000000000,
      0b000000001000000100000000000,
      0b000000001000001000000000000,
      0b000000001111001111000000000,
      0b000000001000001000100000000,
      0b000000001000001000100000000,
      0b000000001000000111000000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x68 48 F6
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000001111101111100000000,
      0b000000001000000000100000000,
      0b000000001000000000100000000,
      0b000000001111000001000000000,
      0b000000001000000010000000000,
      0b000000001000000010000000000,
      0b000000001000000010000000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x69 49 F7
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000001111100111000000000,
      0b000000001000001000100000000,
      0b000000001000001000100000000,
      0b000000001111000111000000000,
      0b000000001000001000100000000,
      0b000000001000001000100000000,
      0b000000001000000111000000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x6a 50 F8
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000001111100111000000000,
      0b000000001000001000100000000,
      0b000000001000001000100000000,
      0b000000001111000111100000000,
      0b000000001000000000100000000,
      0b000000001000000001000000000,
      0b000000001000000110000000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x6b 51 F9
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000001111100010000111000000,
      0b000001000000110001000100000,
      0b000001000000010001000100000,
      0b000001111000010001000100000,
      0b000001000000010001000100000,
      0b000001000000010001000100000,
      0b000001000000010000111000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x6c 52 F10
    {
      0b0000000,0b00000000000000000000000000000000,
      0b0000000,0b00000000000000000000000000000000,
      0b0000000,0b00000000000000000000000000000000,
      0b0000001,0b11001000100111001111101111100000,
      0b0000010,0b00101000100010001000000010000000,
      0b0000010,0b00001000100010001000000010000000,
      0b0000001,0b11001111100010001111000010000000,
      0b0000000,0b00101000100010001000000010000000,
      0b0000010,0b00101000100010001000000010000000,
      0b0000001,0b11001000100111001000000010000000,
      0b0000000,0b00000000000000000000000000000000,
      0b0000000,0b00000000000000000000000000000000,
      0b0000000,0b00000000000000000000000000000000,
    },  //0x70 53 SHIFT
    {
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000111001111101111001000000,
      0b001000100010001000101000000,
      0b001000000010001000101000000,
      0b001000000010001111001000000,
      0b001000000010001000101000000,
      0b001000100010001000101000000,
      0b000111000010001000101111100,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
      0b000000000000000000000000000,
    },  //0x71 54 CTRL
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000110011100111000100,
      0b001001010010010001100,
      0b001001010010010000100,
      0b001001011100010000100,
      0b001001010000010000100,
      0b001001010000010000100,
      0b000110010000010100100,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x72 55 OPT.1
    {
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
      0b001100111001110001100,
      0b010010100100100010010,
      0b010010100100100000010,
      0b010010111000100000100,
      0b010010100000100001000,
      0b010010100000100010000,
      0b001100100000101011110,
      0b000000000000000000000,
      0b000000000000000000000,
      0b000000000000000000000,
    },  //0x73 56 OPT.2
  };

  //キーのビットマップ
  //  可変キーとLEDキーはキー毎にビットマップを用意する
  //  可変キーのビットマップの幅はキーの幅
  public static final int[][][] kbdVarStatBitmap = new int[256][][];  //文字コード→状態→可変キーのビットマップ
  public static final int[][][] kbdLedStatBitmap = new int[KBD_LED_KEYS][][];  //LEDキーの番号→状態→LEDキーのビットマップ

  //種類毎に別々に用意するもの
  //  ビットマップ
  //    ピクセルオフセット→RGBコード
  //    固定キーのビットマップの幅はキーボードの幅
  //  番号マップ
  //    ピクセルオフセット→番号。-1=キーがない
  //    マウスの座標にどのキーがあるか瞬時に判断するための配列
  //  レクタングル
  //    {}または{x,y,w,h}または{x1,y1,w1,h1,x2,y2,w2,h2}
  //    1段目を囲む0個～2個の長方形。欠番は0個、RETURNキーとSHIFTキーは2個、その他は1個
  //    キーを描き変えるときに使う
  public static final BufferedImage[] kbdDataTypeImage = new BufferedImage[KBD_TYPES];  //種類→イメージ
  public static final int[] kbdDataTypeWidth = new int[KBD_TYPES];  //種類→キーボードの幅
  public static final int[] kbdDataTypeHeight = new int[KBD_TYPES];  //種類→キーボードの高さ
  public static final int[][] kbdDataTypeBitmap = new int[KBD_TYPES][];  //種類→ビットマップ
  public static final byte[][] kbdDataTypeIndexMap = new byte[KBD_TYPES][];  //種類→番号マップ
  public static final int[][][] kbdDataTypeStatFixBitmap = new int[KBD_TYPES][][];  //種類→状態→固定キーのビットマップ
  public static final int[][][] kbdDataTypeIndexRect = new int[KBD_TYPES][][];  //種類→番号→レクタングル
  public static final int[][] kbdDataTypeIndexCharacterStat = new int[KBD_TYPES][];  //種類→番号→最後に描画したときの文字<<8|状態。-1=まだ描画されていない

  //現在の状態
  //  表示位置のXEiJ.pnlKeyboardX,XEiJ.pnlKeyboardYはパネル側で制御するのでここにはない
  public static boolean kbdOn;  //true=表示する,false=表示しない
  public static int kbdType;  //キーボードの種類。KBD_STANDARD_TYPEまたはKBD_COMPACT_TYPE
  public static int kbdWidth;  //キーボードの幅
  public static int kbdHeight;  //キーボードの高さ
  public static BufferedImage kbdImage;  //イメージ
  public static int[] kbdBitmap;  //ビットマップ
  public static byte[] kbdIndexMap;  //番号マップ
  public static final int[] kbdStat = new int[KBD_ALL_KEYS];  //番号→状態
  public static int kbdPointedIndex;  //ポイントされているキーの番号。-1=どのキーもポイントされていない
  public static TimerTask kbdRepeatTask;  //リピート中のキーコードをMFPに転送するタスク。null=リピート中ではない
  public static int kbdRepeatDelay;  //リピートディレイ(ms)
  public static int kbdRepeatInterval;  //リピートインターバル(ms)

  public static final int KBD_F11_FULLSCREEN = 0;
  public static final int KBD_F11_SCREENSHOT = 1;
  public static final int KBD_F11_STOPANDSTART = 2;
  public static int kbdF11Mode;  //F11キーが押されたとき何をするか

  //  eo  emulator order。可変キー48個、LEDキー8個、固定キー57個の順
  //  xo  X68000 order。X68000(NUM OFF)のキーコードの昇順

  //  keyLocation,locationMask
  //    1   2  standard
  //    2   4  left
  //    3   8  right
  //    4  16  numpad

  //  X68000標準キーボード⇔日本語106キーボード変換テーブル
  //    F11=0x1007aは全画面モードなどの切り替えに使う
  //    F12=0x1007bはマウスモードの切り替えに使う
  //  xo→(keyLocation1,keyLocation2,rawCode,keyCode)
  //    keyLocation1==0のとき割り当てなし
  //    keyLocation1==0のときkeyLocation2==0,rawCode==0,keyCode==0
  //    キーを2個登録できるのはkeyLocationだけが異なる場合に限る
  public static final int[] KBD_DEFAULT_MAP = {
    //                                       |1| | | | | | | | | | | | |NUM
    //                                       | |1| | | | | | | | | | | |CTRL
    //                                       | | |1| | | | | | | | | | |コード入力
    //                                       | | | |0|0|0|0|0|0| | | | |OPT.1
    //                                       | | | |1|1|1|1| | | | | | |かな。ひらがな+全角→ひらがな
    //                                       | | | | | | | |1|1| | | | |ローマ字
    //                                       | | | |0|0|1|1| | | | | | |五十音
    //locationMask                           | | | | | | | | | |0|0|1|1|CAPS
    //  rawCode                              | | | |0|1|0|1|0|1|0|1|0|1|SHIFT       キートップ          
    //  keyCode         eo   xo   x68kCode                                X68000            106         備考
     2,  27,  27,  // 56    0  0x01  0x01                               ESC       Esc                 
     2,  49,  49,  //  0    1  0x02  0x02  | | |1|ﾇ|ﾇ|ｱ|ｧ|1|!|1|!|1|!|  1!ぬ      1!ぬ                
     2,  50,  50,  //  1    2  0x03  0x03  | | |2|ﾌ|ﾌ|ｲ|ｨ|2|"|2|"|2|"|  2"ふ      2"ふ                
     2,  51,  51,  //  2    3  0x04  0x04  | | |3|ｱ|ｧ|ｳ|ｩ|3|#|3|#|3|#|  3#あぁ    3#あぁ              
     2,  52,  52,  //  3    4  0x05  0x05  | | |4|ｳ|ｩ|ｴ|ｪ|4|$|4|$|4|$|  4$うぅ    4$うぅ              
     2,  53,  53,  //  4    5  0x06  0x06  | | |5|ｴ|ｪ|ｵ|ｫ|5|%|5|%|5|%|  5%えぇ    5%えぇ              
     2,  54,  54,  //  5    6  0x07  0x07  | | |6|ｵ|ｫ|ﾅ|ﾅ|6|&|6|&|6|&|  6&おぉ    6&おぉ              
     2,  55,  55,  //  6    7  0x08  0x43  |7| |7|ﾔ|ｬ|ﾆ|ﾆ|7|'|7|'|7|'|  7'やゃ    7'やゃ              
     2,  56,  56,  //  7    8  0x09  0x44  |8| |8|ﾕ|ｭ|ﾇ|ﾇ|8|(|8|(|8|(|  8(ゆゅ    8(ゆゅ              
     2,  57,  57,  //  8    9  0x0a  0x45  |9| |9|ﾖ|ｮ|ﾈ|ﾈ|9|)|9|)|9|)|  9)よょ    9)よょ              
     2,  48,  48,  //  9   10  0x0b  0x40  |/| |0|ﾜ|ｦ|ﾉ|ｦ|0| |0| |0| |  0 わを    0 わを              
     2, 189,  45,  // 10   11  0x0c  0x41  |*| | |ﾎ|ﾎ|ﾗ|ﾗ|ｰ|=|-|=|-|=|  -=ほ      -=ほ                
     2, 222, 514,  // 11   12  0x0d  0x0d  | |^| |ﾍ|ﾍ|ﾘ|ﾘ|^|~|^|~|^|~|  ^~へ      ^~へ                
     2, 220,  92,  // 12   13  0x0e  0x0e  | |\| |ｰ|ｰ|ﾙ|ﾙ|\|||\|||\|||  \|ー      \|ー                keyCode=92では[\_ろ]と[\|ー]を見分けられない
     2,   8,   8,  // 57   14  0x0f  0x0f                               BS        BackSpace           
     2,   9,   9,  // 58   15  0x10  0x10                               TAB       Tab                 
     2,  81,  81,  // 13   16  0x11  0x11  | |Q| |ﾀ|ﾀ|ｶ|ｶ|Q|Q|q|Q|Q|q|  Q た      Q た                
     2,  87,  87,  // 14   17  0x12  0x12  | |W| |ﾃ|ﾃ|ｷ|ｷ|W|W|w|W|W|w|  W て      W て                
     2,  69,  69,  // 15   18  0x13  0x13  | |E|e|ｲ|ｨ|ｸ|ｸ|E|e|e|E|E|e|  E いぃ    E いぃ              
     2,  82,  82,  // 16   19  0x14  0x14  | |R| |ｽ|ｽ|ｹ|ｹ|R|R|r|R|R|r|  R す      R す                
     2,  84,  84,  // 17   20  0x15  0x15  | |T| |ｶ|ｶ|ｺ|ｺ|T|T|t|T|T|t|  T か      T か                
     2,  89,  89,  // 18   21  0x16  0x16  | |Y| |ﾝ|ﾝ|ﾊ|ﾊ|Y|Y|y|Y|Y|y|  Y ん      Y ん                
     2,  85,  85,  // 19   22  0x17  0x47  |4|U| |ﾅ|ﾅ|ﾋ|ﾋ|U|u|u|U|U|u|  U な      U な                
     2,  73,  73,  // 20   23  0x18  0x48  |5|I| |ﾆ|ﾆ|ﾌ|ﾌ|I|i|i|I|I|i|  I に      I に                
     2,  79,  79,  // 21   24  0x19  0x49  |6|O| |ﾗ|ﾗ|ﾍ|ﾍ|O|o|o|O|O|o|  O ら      O ら                
     2,  80,  80,  // 22   25  0x1a  0x42  |-|P| |ｾ|ｾ|ﾎ|ﾎ|P|P|p|P|P|p|  P せ      P せ                
     2, 192, 512,  // 23   26  0x1b  0x1b  | | | |ﾞ|ﾞ|ﾚ|ﾚ|@|`|@|`|@|`|  @`゛      @`゛                
     2, 219,  91,  // 24   27  0x1c  0x1c  | |[| |ﾟ|｢|ﾛ|｢|｢|{|[|{|[|{|  [{゜「    [{゜「              
     2,  13,  10,  // 59   28  0x1d  0x1d                               リターン  Enter               
     2,  65,  65,  // 25   29  0x1e  0x1e  | |A|a|ﾁ|ﾁ|ｻ| |A|a|a|A|A|a|  A ち      A ち                かな+五十音+SHIFT→0xa0
     2,  83,  83,  // 26   30  0x1f  0x1f  | |S| |ﾄ|ﾄ|ｼ|ｼ|S|S|s|S|S|s|  S と      S と                
     2,  68,  68,  // 27   31  0x20  0x20  | |D|d|ｼ|ｼ|ｽ|ｽ|D|D|d|D|D|d|  D し      D し                
     2,  70,  70,  // 28   32  0x21  0x21  | |F|f|ﾊ|ﾊ|ｾ|ｾ|F|F|f|F|F|f|  F は      F は                
     2,  71,  71,  // 29   33  0x22  0x22  | |G| |ｷ|ｷ|ｿ|ｿ|G|G|g|G|G|g|  G き      G き                
     2,  72,  72,  // 30   34  0x23  0x23  | |H| |ｸ|ｸ|ﾏ|ﾏ|H|H|h|H|H|h|  H く      H く                
     2,  74,  74,  // 31   35  0x24  0x4b  |1|J| |ﾏ|ﾏ|ﾐ|ﾐ|J|J|j|J|J|j|  J ま      J ま                
     2,  75,  75,  // 32   36  0x25  0x4c  |2|K| |ﾉ|ﾉ|ﾑ|ﾑ|K|K|k|K|K|k|  K の      K の                
     2,  76,  76,  // 33   37  0x26  0x4d  |3|L| |ﾘ|ﾘ|ﾒ|ﾒ|L|L|l|L|L|l|  L り      L り                
     2, 187,  59,  // 34   38  0x27  0x46  |+| | |ﾚ|ﾚ|ﾓ|ﾓ|;|+|;|+|;|+|  ;+れ      ;+れ                
     2, 186, 513,  // 35   39  0x28  0x28  | | | |ｹ|ｹ|ﾞ|ｰ|:|*|:|*|:|*|  :*け      :*け                
     2, 221,  93,  // 36   40  0x29  0x29  | |]| |ﾑ|｣|ﾟ|｣|｣|}|]|}|]|}|  ]}む」    ]}む」              
     2,  90,  90,  // 37   41  0x2a  0x2a  | |Z| |ﾂ|ｯ|ﾀ|ﾀ|Z|z|z|Z|Z|z|  Z つっ    Z つっ              
     2,  88,  88,  // 38   42  0x2b  0x2b  | |X| |ｻ|ｻ|ﾁ|ﾁ|X|X|x|X|X|x|  X さ      X さ                
     2,  67,  67,  // 39   43  0x2c  0x2c  | |C|c|ｿ|ｿ|ﾂ|ｯ|C|C|c|C|C|c|  C そ      C そ                
     2,  86,  86,  // 40   44  0x2d  0x2d  | |V| |ﾋ|ﾋ|ﾃ|ﾃ|V|V|v|V|V|v|  V ひ      V ひ                
     2,  66,  66,  // 41   45  0x2e  0x2e  | |B|b|ｺ|ｺ|ﾄ|ﾄ|B|B|b|B|B|b|  B こ      B こ                
     2,  78,  78,  // 42   46  0x2f  0x2f  | |N| |ﾐ|ﾐ|ﾔ|ｬ|N|N|n|N|N|n|  N み      N み                
     2,  77,  77,  // 43   47  0x30  0x4f  |0|M| |ﾓ|ﾓ|ﾕ|ｭ|M|M|m|M|M|m|  M も      M も                
     2, 188,  44,  // 44   48  0x31  0x50  |,| | |ﾈ|､|ﾖ|ｮ|､|<|,|<|,|<|  ,<ね、    ,<ね、              
     2, 190,  46,  // 45   49  0x32  0x51  |.| | |ﾙ|｡|ﾜ|､|｡|>|.|>|.|>|  .>る。    .>る。              
     2, 191,  47,  // 46   50  0x33  0x4a  |=| | |ﾒ|･|ｦ|｡|･|?|/|?|/|?|  /?め・    /?め・              
     2, 226,  92,  // 47   51  0x34  0x4e  |E|_| |ﾛ| |ﾝ|･| |_| |_| |_|   _ろ□    \_ろ                keyCode=92では[\_ろ]と[\|ー]を見分けられない。NUM→ENTER、かな+SHIFT→0xa0
     2,  32,  32,  // 60   52  0x35  0x35                               スペース  スペース            
    18,  36,  36,  // 61   53  0x36  0x36                               HOME      Home      7Home     NumLock=OFF
    18,  46, 127,  // 62   54  0x37  0x37                               DEL       Delete    .Del      NumLock=OFF
    18,  33,  33,  // 63   55  0x38  0x38                               ROLLUP    PageUp    9PgUp     NumLock=OFF
    18,  34,  34,  // 64   56  0x39  0x39                               ROLLDOWN  PageDown  3PgDn     NumLock=OFF
    18,  35,  35,  // 65   57  0x3a  0x3a                               UNDO      End       1End      NumLock=OFF
    18,  37,  37,  // 66   58  0x3b  0x3b                               ←        ←        4←       NumLock=OFF
    18,  38,  38,  // 67   59  0x3c  0x3c                               ↑        ↑        8↑       NumLock=OFF
    18,  39,  39,  // 68   60  0x3d  0x3d                               →        →        6→       NumLock=OFF
    18,  40,  40,  // 69   61  0x3e  0x3e                               ↓        ↓        2↓       NumLock=OFF
    16,  12,  12,  // 70   62  0x3f  0x3f                               CLR       5                   NumLock=OFF
    16, 111, 111,  // 71   63  0x40  0x40                               /         /                   NumLock=ON/OFF
    16, 106, 106,  // 72   64  0x41  0x41                               *         *                   NumLock=ON/OFF
    16, 109, 109,  // 73   65  0x42  0x42                               -         -                   NumLock=ON/OFF
    16, 103, 103,  // 74   66  0x43  0x43                               7         7Home               NumLock=ON
    16, 104, 104,  // 75   67  0x44  0x44                               8         8↑                 NumLock=ON
    16, 105, 105,  // 76   68  0x45  0x45                               9         9PgUp               NumLock=ON
    16, 107, 107,  // 77   69  0x46  0x46                               +         +                   NumLock=ON/OFF
    16, 100, 100,  // 78   70  0x47  0x47                               4         4←                 NumLock=ON
    16, 101, 101,  // 79   71  0x48  0x48                               5         5                   NumLock=ON
    16, 102, 102,  // 80   72  0x49  0x49                               6         6→                 NumLock=ON
     0,   0,   0,  // 81   73  0x4a  0x4a                               =                             なし
    16,  97,  97,  // 82   74  0x4b  0x4b                               1         1End                NumLock=ON
    16,  98,  98,  // 83   75  0x4c  0x4c                               2         2↓                 NumLock=ON
    16,  99,  99,  // 84   76  0x4d  0x4d                               3         3PgDn               NumLock=ON
    16,  13,  10,  // 85   77  0x4e  0x4e                               ENTER     Enter               NumLock=ON/OFF
    16,  96,  96,  // 86   78  0x4f  0x4f                               0         0Ins                NumLock=ON
     0,   0,   0,  // 87   79  0x50  0x50                               ,                             なし
    16, 110, 110,  // 88   80  0x51  0x51                               .         .Del                NumLock=ON
     0,   0,   0,  // 89   81  0x52  0x52                               記号入力                      なし
     0,   0,   0,  // 90   82  0x53  0x53                               登録                          なし
     0,   0,   0,  // 91   83  0x54  0x54                               HELP                          なし
     4,  18,  18,  // 92   84  0x55  0x55                               XF1       左Alt               窓キー=(2,91,524)はconsumeしてもスタートメニューが開いてしまう
     2,  29,  29,  // 93   85  0x56  0x56                               XF2       無変換              
     8,  18,  18,  // 94   86  0x57  0x57                               XF3       右Alt               変換とカタカナひらがなはKeyEventが発生しない
     2,  93, 525,  // 95   87  0x58  0x58                               XF4       メニュー            
     8,  17,  17,  // 96   88  0x59  0x59                               XF5       右Ctrl              
     0,   0,   0,  // 48   89  0x5a  0x5a                               かな                          なし
     0,   0,   0,  // 49   90  0x5b  0x5b                               ローマ字                      なし
     0,   0,   0,  // 50   91  0x5c  0x5c                               コート入力                    なし
     0,   0,   0,  // 51   92  0x5d  0x5d                               CAPS                          なし
    18,  45, 155,  // 52   93  0x5e  0x5e                               INS       Insert    0Ins      NumLock=OFF
     0,   0,   0,  // 53   94  0x5f  0x5f                               ひらがな                      なし
     0,   0,   0,  // 54   95  0x60  0x60                               全角                          なし。半角全角漢字はKeyEventが発生しない
     2,  19,  19,  // 97   96  0x61  0x61                               BREAK     Pause/Break         
     0,   0,   0,  // 98   97  0x62  0x62                               COPY                          なし。ScrollLock=(1,145,145)はconsumeしてもScrollLockが切り替わってしまう
     2, 112, 112,  // 99   98  0x63  0x63                               F1        F1                  
     2, 113, 113,  //100   99  0x64  0x64                               F2        F2                  
     2, 114, 114,  //101  100  0x65  0x65                               F3        F3                  
     2, 115, 115,  //102  101  0x66  0x66                               F4        F4                  
     2, 116, 116,  //103  102  0x67  0x67                               F5        F5                  
     2, 117, 117,  //104  103  0x68  0x68                               F6        F6                  
     2, 118, 118,  //105  104  0x69  0x69                               F7        F7                  
     2, 119, 119,  //106  105  0x6a  0x6a                               F8        F8                  
     2, 120, 120,  //107  106  0x6b  0x6b                               F9        F9                  
     2, 121, 121,  //108  107  0x6c  0x6c                               F10       F10                 
    12,  16,  16,  //109  108  0x70  0x70                               SHIFT     左Shift   右Shift   
     4,  17,  17,  //110  109  0x71  0x71                               CTRL      左Ctrl              右CtrlはXF5
     0,   0,   0,  //111  110  0x72  0x72                               OPT.1                         なし
     0,   0,   0,  //112  111  0x73  0x73                               OPT.2                         なし
     0,   0,   0,  // 55  112  0x74  0x74                               NUM                           なし。NumLock=(4,144,144)はconsumeしてもNumLockが切り替わってしまう
  };
  public static int[] kbdCurrentMap;

  //  xo→キートップの文字列
  public static final String[] KBD_KEYTOP_TEXT = (
    //                  eo   xo   x68kCode
    "ESC,"        +  // 56    0  0x01  0x01
    "１！ぬ　,"   +  //  0    1  0x02  0x02
    "２＂ふ　,"   +  //  1    2  0x03  0x03
    "３＃あぁ,"   +  //  2    3  0x04  0x04
    "４＄うぅ,"   +  //  3    4  0x05  0x05
    "５％えぇ,"   +  //  4    5  0x06  0x06
    "６＆おぉ,"   +  //  5    6  0x07  0x07
    "７＇やゃ,"   +  //  6    7  0x08  0x43
    "８（ゆゅ,"   +  //  7    8  0x09  0x44
    "９）よょ,"   +  //  8    9  0x0a  0x45
    "０　わを,"   +  //  9   10  0x0b  0x40
    "－＝ほ　,"   +  // 10   11  0x0c  0x41
    "＾～へ　,"   +  // 11   12  0x0d  0x0d
    "￥｜ー　,"   +  // 12   13  0x0e  0x0e
    "BS,"         +  // 57   14  0x0f  0x0f
    "TAB,"        +  // 58   15  0x10  0x10
    "Ｑ　た　,"   +  // 13   16  0x11  0x11
    "Ｗ　て　,"   +  // 14   17  0x12  0x12
    "Ｅ　いぃ,"   +  // 15   18  0x13  0x13
    "Ｒ　す　,"   +  // 16   19  0x14  0x14
    "Ｔ　か　,"   +  // 17   20  0x15  0x15
    "Ｙ　ん　,"   +  // 18   21  0x16  0x16
    "Ｕ　な　,"   +  // 19   22  0x17  0x47
    "Ｉ　に　,"   +  // 20   23  0x18  0x48
    "Ｏ　ら　,"   +  // 21   24  0x19  0x49
    "Ｐ　せ　,"   +  // 22   25  0x1a  0x42
    "＠｀゛　,"   +  // 23   26  0x1b  0x1b
    "［｛゜「,"   +  // 24   27  0x1c  0x1c
    "リターン,"   +  // 59   28  0x1d  0x1d
    "Ａ　ち　,"   +  // 25   29  0x1e  0x1e
    "Ｓ　と　,"   +  // 26   30  0x1f  0x1f
    "Ｄ　し　,"   +  // 27   31  0x20  0x20
    "Ｆ　は　,"   +  // 28   32  0x21  0x21
    "Ｇ　き　,"   +  // 29   33  0x22  0x22
    "Ｈ　く　,"   +  // 30   34  0x23  0x23
    "Ｊ　ま　,"   +  // 31   35  0x24  0x4b
    "Ｋ　の　,"   +  // 32   36  0x25  0x4c
    "Ｌ　り　,"   +  // 33   37  0x26  0x4d
    "；＋れ　,"   +  // 34   38  0x27  0x46
    "：＊け　,"   +  // 35   39  0x28  0x28
    "］｝む」,"   +  // 36   40  0x29  0x29
    "Ｚ　つっ,"   +  // 37   41  0x2a  0x2a
    "Ｘ　さ　,"   +  // 38   42  0x2b  0x2b
    "Ｃ　そ　,"   +  // 39   43  0x2c  0x2c
    "Ｖ　ひ　,"   +  // 40   44  0x2d  0x2d
    "Ｂ　こ　,"   +  // 41   45  0x2e  0x2e
    "Ｎ　み　,"   +  // 42   46  0x2f  0x2f
    "Ｍ　も　,"   +  // 43   47  0x30  0x4f
    "，＜ね、,"   +  // 44   48  0x31  0x50
    "．＞る。,"   +  // 45   49  0x32  0x51
    "／？め・,"   +  // 46   50  0x33  0x4a
    "　＿ろ□,"   +  // 47   51  0x34  0x4e
    "スペース,"   +  // 60   52  0x35  0x35
    "HOME,"       +  // 61   53  0x36  0x36
    "DEL,"        +  // 62   54  0x37  0x37
    "ROLLUP,"     +  // 63   55  0x38  0x38
    "ROLLDOWN,"   +  // 64   56  0x39  0x39
    "UNDO,"       +  // 65   57  0x3a  0x3a
    "←,"         +  // 66   58  0x3b  0x3b
    "↑,"         +  // 67   59  0x3c  0x3c
    "→,"         +  // 68   60  0x3d  0x3d
    "↓,"         +  // 69   61  0x3e  0x3e
    "CLR,"        +  // 70   62  0x3f  0x3f
    "／,"         +  // 71   63  0x40  0x40
    "＊,"         +  // 72   64  0x41  0x41
    "－,"         +  // 73   65  0x42  0x42
    "７,"         +  // 74   66  0x43  0x43
    "８,"         +  // 75   67  0x44  0x44
    "９,"         +  // 76   68  0x45  0x45
    "＋,"         +  // 77   69  0x46  0x46
    "４,"         +  // 78   70  0x47  0x47
    "５,"         +  // 79   71  0x48  0x48
    "６,"         +  // 80   72  0x49  0x49
    "＝,"         +  // 81   73  0x4a  0x4a
    "１,"         +  // 82   74  0x4b  0x4b
    "２,"         +  // 83   75  0x4c  0x4c
    "３,"         +  // 84   76  0x4d  0x4d
    "ENTER,"      +  // 85   77  0x4e  0x4e
    "０,"         +  // 86   78  0x4f  0x4f
    "，,"         +  // 87   79  0x50  0x50
    "．,"         +  // 88   80  0x51  0x51
    "記号入力,"   +  // 89   81  0x52  0x52
    "登録,"       +  // 90   82  0x53  0x53
    "HELP,"       +  // 91   83  0x54  0x54
    "XF1,"        +  // 92   84  0x55  0x55
    "XF2,"        +  // 93   85  0x56  0x56
    "XF3,"        +  // 94   86  0x57  0x57
    "XF4,"        +  // 95   87  0x58  0x58
    "XF5,"        +  // 96   88  0x59  0x59
    "かな,"       +  // 48   89  0x5a  0x5a
    "ローマ字,"   +  // 49   90  0x5b  0x5b
    "コード入力," +  // 50   91  0x5c  0x5c
    "CAPS,"       +  // 51   92  0x5d  0x5d
    "INS,"        +  // 52   93  0x5e  0x5e
    "ひらがな,"   +  // 53   94  0x5f  0x5f
    "全角,"       +  // 54   95  0x60  0x60
    "BREAK,"      +  // 97   96  0x61  0x61
    "COPY,"       +  // 98   97  0x62  0x62
    "F1,"         +  // 99   98  0x63  0x63
    "F2,"         +  //100   99  0x64  0x64
    "F3,"         +  //101  100  0x65  0x65
    "F4,"         +  //102  101  0x66  0x66
    "F5,"         +  //103  102  0x67  0x67
    "F6,"         +  //104  103  0x68  0x68
    "F7,"         +  //105  104  0x69  0x69
    "F8,"         +  //106  105  0x6a  0x6a
    "F9,"         +  //107  106  0x6b  0x6b
    "F10,"        +  //108  107  0x6c  0x6c
    "SHIFT,"      +  //109  108  0x70  0x70
    "CTRL,"       +  //110  109  0x71  0x71
    "OPT.1,"      +  //111  110  0x72  0x72
    "OPT.2,"      +  //112  111  0x73  0x73
    "NUM"            // 55  112  0x74  0x74
    ).split (",");

  //  eo→xo
  public static final int[] KBD_EO_TO_XO = {
    //eo     xo   x68kCode
      1,  //  0  0x02  0x02  "１！ぬ　"
      2,  //  1  0x03  0x03  "２＂ふ　"
      3,  //  2  0x04  0x04  "３＃あぁ"
      4,  //  3  0x05  0x05  "４＄うぅ"
      5,  //  4  0x06  0x06  "５％えぇ"
      6,  //  5  0x07  0x07  "６＆おぉ"
      7,  //  6  0x08  0x43  "７＇やゃ"
      8,  //  7  0x09  0x44  "８（ゆゅ"
      9,  //  8  0x0a  0x45  "９）よょ"
     10,  //  9  0x0b  0x40  "０　わを"
     11,  // 10  0x0c  0x41  "－＝ほ　"
     12,  // 11  0x0d  0x0d  "＾～へ　"
     13,  // 12  0x0e  0x0e  "￥｜ー　"
     16,  // 13  0x11  0x11  "Ｑ　た　"
     17,  // 14  0x12  0x12  "Ｗ　て　"
     18,  // 15  0x13  0x13  "Ｅ　いぃ"
     19,  // 16  0x14  0x14  "Ｒ　す　"
     20,  // 17  0x15  0x15  "Ｔ　か　"
     21,  // 18  0x16  0x16  "Ｙ　ん　"
     22,  // 19  0x17  0x47  "Ｕ　な　"
     23,  // 20  0x18  0x48  "Ｉ　に　"
     24,  // 21  0x19  0x49  "Ｏ　ら　"
     25,  // 22  0x1a  0x42  "Ｐ　せ　"
     26,  // 23  0x1b  0x1b  "＠｀゛　"
     27,  // 24  0x1c  0x1c  "［｛゜「"
     29,  // 25  0x1e  0x1e  "Ａ　ち　"
     30,  // 26  0x1f  0x1f  "Ｓ　と　"
     31,  // 27  0x20  0x20  "Ｄ　し　"
     32,  // 28  0x21  0x21  "Ｆ　は　"
     33,  // 29  0x22  0x22  "Ｇ　き　"
     34,  // 30  0x23  0x23  "Ｈ　く　"
     35,  // 31  0x24  0x4b  "Ｊ　ま　"
     36,  // 32  0x25  0x4c  "Ｋ　の　"
     37,  // 33  0x26  0x4d  "Ｌ　り　"
     38,  // 34  0x27  0x46  "；＋れ　"
     39,  // 35  0x28  0x28  "：＊け　"
     40,  // 36  0x29  0x29  "］｝む」"
     41,  // 37  0x2a  0x2a  "Ｚ　つっ"
     42,  // 38  0x2b  0x2b  "Ｘ　さ　"
     43,  // 39  0x2c  0x2c  "Ｃ　そ　"
     44,  // 40  0x2d  0x2d  "Ｖ　ひ　"
     45,  // 41  0x2e  0x2e  "Ｂ　こ　"
     46,  // 42  0x2f  0x2f  "Ｎ　み　"
     47,  // 43  0x30  0x4f  "Ｍ　も　"
     48,  // 44  0x31  0x50  "，＜ね、"
     49,  // 45  0x32  0x51  "．＞る。"
     50,  // 46  0x33  0x4a  "／？め・"
     51,  // 47  0x34  0x4e  "　＿ろ□"
     89,  // 48  0x5a  0x5a  "かな"
     90,  // 49  0x5b  0x5b  "ローマ字"
     91,  // 50  0x5c  0x5c  "コード入力"
     92,  // 51  0x5d  0x5d  "CAPS"
     93,  // 52  0x5e  0x5e  "INS"
     94,  // 53  0x5f  0x5f  "ひらがな"
     95,  // 54  0x60  0x60  "全角"
    112,  // 55  0x74  0x74  "NUM"
      0,  // 56  0x01  0x01  "ESC"
     14,  // 57  0x0f  0x0f  "BS"
     15,  // 58  0x10  0x10  "TAB"
     28,  // 59  0x1d  0x1d  "リターン"
     52,  // 60  0x35  0x35  "スペース"
     53,  // 61  0x36  0x36  "HOME"
     54,  // 62  0x37  0x37  "DEL"
     55,  // 63  0x38  0x38  "ROLLUP"
     56,  // 64  0x39  0x39  "ROLLDOWN"
     57,  // 65  0x3a  0x3a  "UNDO"
     58,  // 66  0x3b  0x3b  "←"
     59,  // 67  0x3c  0x3c  "↑"
     60,  // 68  0x3d  0x3d  "→"
     61,  // 69  0x3e  0x3e  "↓"
     62,  // 70  0x3f  0x3f  "CLR"
     63,  // 71  0x40  0x40  "／"
     64,  // 72  0x41  0x41  "＊"
     65,  // 73  0x42  0x42  "－"
     66,  // 74  0x43  0x43  "７"
     67,  // 75  0x44  0x44  "８"
     68,  // 76  0x45  0x45  "９"
     69,  // 77  0x46  0x46  "＋"
     70,  // 78  0x47  0x47  "４"
     71,  // 79  0x48  0x48  "５"
     72,  // 80  0x49  0x49  "６"
     73,  // 81  0x4a  0x4a  "＝"
     74,  // 82  0x4b  0x4b  "１"
     75,  // 83  0x4c  0x4c  "２"
     76,  // 84  0x4d  0x4d  "３"
     77,  // 85  0x4e  0x4e  "ENTER"
     78,  // 86  0x4f  0x4f  "０"
     79,  // 87  0x50  0x50  "，"
     80,  // 88  0x51  0x51  "．"
     81,  // 89  0x52  0x52  "記号入力"
     82,  // 90  0x53  0x53  "登録"
     83,  // 91  0x54  0x54  "HELP"
     84,  // 92  0x55  0x55  "XF1"
     85,  // 93  0x56  0x56  "XF2"
     86,  // 94  0x57  0x57  "XF3"
     87,  // 95  0x58  0x58  "XF4"
     88,  // 96  0x59  0x59  "XF5"
     96,  // 97  0x61  0x61  "BREAK"
     97,  // 98  0x62  0x62  "COPY"
     98,  // 99  0x63  0x63  "F1"
     99,  //100  0x64  0x64  "F2"
    100,  //101  0x65  0x65  "F3"
    101,  //102  0x66  0x66  "F4"
    102,  //103  0x67  0x67  "F5"
    103,  //104  0x68  0x68  "F6"
    104,  //105  0x69  0x69  "F7"
    105,  //106  0x6a  0x6a  "F8"
    106,  //107  0x6b  0x6b  "F9"
    107,  //108  0x6c  0x6c  "F10"
    108,  //109  0x70  0x70  "SHIFT"
    109,  //110  0x71  0x71  "CTRL"
    110,  //111  0x72  0x72  "OPT.1"
    111,  //112  0x73  0x73  "OPT.2"
  };

  //  xo→eo
  public static final int[] KBD_XO_TO_EO = {
    //eo     xo   x68kCode
     56,  //  0  0x01  0x01  "ESC"
      0,  //  1  0x02  0x02  "１！ぬ　"
      1,  //  2  0x03  0x03  "２＂ふ　"
      2,  //  3  0x04  0x04  "３＃あぁ"
      3,  //  4  0x05  0x05  "４＄うぅ"
      4,  //  5  0x06  0x06  "５％えぇ"
      5,  //  6  0x07  0x07  "６＆おぉ"
      6,  //  7  0x08  0x43  "７＇やゃ"
      7,  //  8  0x09  0x44  "８（ゆゅ"
      8,  //  9  0x0a  0x45  "９）よょ"
      9,  // 10  0x0b  0x40  "０　わを"
     10,  // 11  0x0c  0x41  "－＝ほ　"
     11,  // 12  0x0d  0x0d  "＾～へ　"
     12,  // 13  0x0e  0x0e  "￥｜ー　"
     57,  // 14  0x0f  0x0f  "BS"
     58,  // 15  0x10  0x10  "TAB"
     13,  // 16  0x11  0x11  "Ｑ　た　"
     14,  // 17  0x12  0x12  "Ｗ　て　"
     15,  // 18  0x13  0x13  "Ｅ　いぃ"
     16,  // 19  0x14  0x14  "Ｒ　す　"
     17,  // 20  0x15  0x15  "Ｔ　か　"
     18,  // 21  0x16  0x16  "Ｙ　ん　"
     19,  // 22  0x17  0x47  "Ｕ　な　"
     20,  // 23  0x18  0x48  "Ｉ　に　"
     21,  // 24  0x19  0x49  "Ｏ　ら　"
     22,  // 25  0x1a  0x42  "Ｐ　せ　"
     23,  // 26  0x1b  0x1b  "＠｀゛　"
     24,  // 27  0x1c  0x1c  "［｛゜「"
     59,  // 28  0x1d  0x1d  "リターン"
     25,  // 29  0x1e  0x1e  "Ａ　ち　"
     26,  // 30  0x1f  0x1f  "Ｓ　と　"
     27,  // 31  0x20  0x20  "Ｄ　し　"
     28,  // 32  0x21  0x21  "Ｆ　は　"
     29,  // 33  0x22  0x22  "Ｇ　き　"
     30,  // 34  0x23  0x23  "Ｈ　く　"
     31,  // 35  0x24  0x4b  "Ｊ　ま　"
     32,  // 36  0x25  0x4c  "Ｋ　の　"
     33,  // 37  0x26  0x4d  "Ｌ　り　"
     34,  // 38  0x27  0x46  "；＋れ　"
     35,  // 39  0x28  0x28  "：＊け　"
     36,  // 40  0x29  0x29  "］｝む」"
     37,  // 41  0x2a  0x2a  "Ｚ　つっ"
     38,  // 42  0x2b  0x2b  "Ｘ　さ　"
     39,  // 43  0x2c  0x2c  "Ｃ　そ　"
     40,  // 44  0x2d  0x2d  "Ｖ　ひ　"
     41,  // 45  0x2e  0x2e  "Ｂ　こ　"
     42,  // 46  0x2f  0x2f  "Ｎ　み　"
     43,  // 47  0x30  0x4f  "Ｍ　も　"
     44,  // 48  0x31  0x50  "，＜ね、"
     45,  // 49  0x32  0x51  "．＞る。"
     46,  // 50  0x33  0x4a  "／？め・"
     47,  // 51  0x34  0x4e  "　＿ろ□"
     60,  // 52  0x35  0x35  "スペース"
     61,  // 53  0x36  0x36  "HOME"
     62,  // 54  0x37  0x37  "DEL"
     63,  // 55  0x38  0x38  "ROLLUP"
     64,  // 56  0x39  0x39  "ROLLDOWN"
     65,  // 57  0x3a  0x3a  "UNDO"
     66,  // 58  0x3b  0x3b  "←"
     67,  // 59  0x3c  0x3c  "↑"
     68,  // 60  0x3d  0x3d  "→"
     69,  // 61  0x3e  0x3e  "↓"
     70,  // 62  0x3f  0x3f  "CLR"
     71,  // 63  0x40  0x40  "／"
     72,  // 64  0x41  0x41  "＊"
     73,  // 65  0x42  0x42  "－"
     74,  // 66  0x43  0x43  "７"
     75,  // 67  0x44  0x44  "８"
     76,  // 68  0x45  0x45  "９"
     77,  // 69  0x46  0x46  "＋"
     78,  // 70  0x47  0x47  "４"
     79,  // 71  0x48  0x48  "５"
     80,  // 72  0x49  0x49  "６"
     81,  // 73  0x4a  0x4a  "＝"
     82,  // 74  0x4b  0x4b  "１"
     83,  // 75  0x4c  0x4c  "２"
     84,  // 76  0x4d  0x4d  "３"
     85,  // 77  0x4e  0x4e  "ENTER"
     86,  // 78  0x4f  0x4f  "０"
     87,  // 79  0x50  0x50  "，"
     88,  // 80  0x51  0x51  "．"
     89,  // 81  0x52  0x52  "記号入力"
     90,  // 82  0x53  0x53  "登録"
     91,  // 83  0x54  0x54  "HELP"
     92,  // 84  0x55  0x55  "XF1"
     93,  // 85  0x56  0x56  "XF2"
     94,  // 86  0x57  0x57  "XF3"
     95,  // 87  0x58  0x58  "XF4"
     96,  // 88  0x59  0x59  "XF5"
     48,  // 89  0x5a  0x5a  "かな"
     49,  // 90  0x5b  0x5b  "ローマ字"
     50,  // 91  0x5c  0x5c  "コード入力"
     51,  // 92  0x5d  0x5d  "CAPS"
     52,  // 93  0x5e  0x5e  "INS"
     53,  // 94  0x5f  0x5f  "ひらがな"
     54,  // 95  0x60  0x60  "全角"
     97,  // 96  0x61  0x61  "BREAK"
     98,  // 97  0x62  0x62  "COPY"
     99,  // 98  0x63  0x63  "F1"
    100,  // 99  0x64  0x64  "F2"
    101,  //100  0x65  0x65  "F3"
    102,  //101  0x66  0x66  "F4"
    103,  //102  0x67  0x67  "F5"
    104,  //103  0x68  0x68  "F6"
    105,  //104  0x69  0x69  "F7"
    106,  //105  0x6a  0x6a  "F8"
    107,  //106  0x6b  0x6b  "F9"
    108,  //107  0x6c  0x6c  "F10"
    109,  //108  0x70  0x70  "SHIFT"
    110,  //109  0x71  0x71  "CTRL"
    111,  //110  0x72  0x72  "OPT.1"
    112,  //111  0x73  0x73  "OPT.2"
     55,  //112  0x74  0x74  "NUM"
  };

  public static JFrame kbdFrame;
  public static JTextField[] kbdTextFieldArray;

  //KeyEvent.rawCode
  public static final boolean KBD_ACCESS_RAW_CODE = true;  //KeyEvent.rawCodeにアクセスを試みる
  public static boolean kbdRawCodeAccessible;  //true=KeyEvent.rawCodeにアクセスできる
  public static Field kbdRawCodeField;  //KeyEvent.rawCodeのフィールド

  //kbdInit ()
  //  キーボードを初期化する
  public static void kbdInit () {

    int[] map = Settings.sgsGetIntArray ("keymap");
    kbdCurrentMap = new int[KBD_DEFAULT_MAP.length];
    if (map.length == 0) {
      System.arraycopy (KBD_DEFAULT_MAP, 0, kbdCurrentMap, 0, KBD_DEFAULT_MAP.length);
    } else {
      System.arraycopy (map, 0, kbdCurrentMap, 0, Math.min (map.length, KBD_DEFAULT_MAP.length));
      if (map.length < KBD_DEFAULT_MAP.length) {
        Arrays.fill (kbdCurrentMap, map.length, KBD_DEFAULT_MAP.length, 0);
      }
    }

    kbdRawCodeAccessible = false;  //KeyEvent.rawCodeにアクセスできない
    kbdRawCodeField = null;
    if (KBD_ACCESS_RAW_CODE && Settings.sgsGetOnOff ("accessrawcode")) {  //KeyEvent.rawCodeにアクセスを試みる
      System.out.println (Multilingual.mlnJapanese ?
                          "KeyEvent.rawCode にアクセスを試みます。警告が表示されます" :
                          "Trying to access KeyEvent.rawCode. Warning messages will appear.");
      System.out.println ("--------");
      kbdRawCodeField = null;
      try {
        kbdRawCodeField = KeyEvent.class.getDeclaredField ("rawCode");  //KeyEvent.rawCodeのフィールドを
        kbdRawCodeField.setAccessible (true);  //アクセスできるようにして
        kbdRawCodeField.getLong (new KeyEvent (new JPanel (), 0, 0L, 0, 0, '\0', 0));  //ダミーのKeyEventを作ってrawCodeを読み出してみる
        kbdRawCodeAccessible = true;  //ここまで動けばKeyEvent.rawCodeにアクセスできる
      } catch (Exception e) {
      }
      System.out.println ("--------");
      if (kbdRawCodeAccessible) {  //KeyEvent.rawCodeにアクセスできる
        System.out.println (Multilingual.mlnJapanese ?
                            "KeyEvent.rawCode にアクセスできます" :
                            "KeyEvent.rawCode is accessible");
      } else {  //KeyEvent.rawCodeにアクセスできない
        System.out.println (Multilingual.mlnJapanese ?
                            "KeyEvent.rawCode にアクセスできません" :
                            "KeyEvent.rawCode is not accessible");
      }
    }

    //  キーボードの種類
    String paramKeyboard = Settings.sgsGetString ("keyboard");
    switch (paramKeyboard.toLowerCase ()) {
    case "none":
      kbdOn = false;
      kbdType = KBD_STANDARD_TYPE;
      break;
    case "standard":
      kbdOn = true;
      kbdType = KBD_STANDARD_TYPE;
      break;
    case "compact":
      kbdOn = true;
      kbdType = KBD_COMPACT_TYPE;
      break;
    default:
      kbdOn = true;
      kbdType = KBD_STANDARD_TYPE;
    }

    //配列を初期化する
    for (int i = 0; i < KBD_VAR_KEYS; i++) {
      int t = KBD_CAPS_BASE[i];
      KBD_CAPS_CHR[i] = (byte) t;
      KBD_NORMAL_CHR[i] = (byte) (t >= 'A' && t <= 'Z' ? t ^ 0x20 : t);
      KBD_CODE_CHR[i] = (byte) (t >= '0' && t <= '9' || t >= 'A' && t <= 'F' ? t : 0);
      t = KBD_CAPS_SHIFT_BASE[i];
      KBD_CAPS_SHIFT_CHR[i] = (byte) t;
      KBD_SHIFT_CHR[i] = (byte) (t >= 'a' && t <= 'z' ? t ^ 0x20 : t);
      KBD_CTRL_CHR[i] = (byte) (t < 0x20 ? t : t >= 0x40 ? t & 0x1f : 0);
      t = KBD_KANA_BASE[i];
      KBD_KANA_CHR[i] = (byte) (t <= 0xff ? t : t - 0xfec0);
      t = KBD_KANA_SHIFT_BASE[i];
      KBD_KANA_SHIFT_CHR[i] = (byte) (t <= 0xff ? t : t - 0xfec0);
      t = KBD_KANA_AIU_BASE[i];
      KBD_KANA_AIU_CHR[i] = (byte) (t <= 0xff ? t : t - 0xfec0);
      t = KBD_KANA_AIU_SHIFT_BASE[i];
      KBD_KANA_AIU_SHIFT_CHR[i] = (byte) (t <= 0xff ? t : t - 0xfec0);
      t = KBD_ROMA_BASE[i];
      KBD_ROMA_CHR[i] = (byte) (t <= 0xff ? t : t - 0xfec0);
      t = KBD_ROMA_SHIFT_BASE[i];
      KBD_ROMA_SHIFT_CHR[i] = (byte) (t <= 0xff ? t : t - 0xfec0);
      t = KBD_NUM_BASE[i];
      if (t != ' ') {
        KBD_NUM_CAPS_CHR[i] =
          KBD_NUM_CHR[i] =
            KBD_NUM_CODE_CHR[i] =
              KBD_NUM_CAPS_SHIFT_CHR[i] =
                KBD_NUM_SHIFT_CHR[i] =
                  KBD_NUM_CTRL_CHR[i] =
                    KBD_NUM_KANA_CHR[i] =
                      KBD_NUM_KANA_SHIFT_CHR[i] =
                        KBD_NUM_KANA_AIU_CHR[i] =
                          KBD_NUM_KANA_AIU_SHIFT_CHR[i] =
                            KBD_NUM_ROMA_CHR[i] =
                              KBD_NUM_ROMA_SHIFT_CHR[i] = (byte) t;
      } else {
        KBD_NUM_CAPS_CHR[i] = KBD_CAPS_CHR[i];
        KBD_NUM_CHR[i] = KBD_NORMAL_CHR[i];
        KBD_NUM_CODE_CHR[i] = KBD_CODE_CHR[i];
        KBD_NUM_CAPS_SHIFT_CHR[i] = KBD_CAPS_SHIFT_CHR[i];
        KBD_NUM_SHIFT_CHR[i] = KBD_SHIFT_CHR[i];
        KBD_NUM_CTRL_CHR[i] = KBD_CTRL_CHR[i];
        KBD_NUM_KANA_CHR[i] = KBD_KANA_CHR[i];
        KBD_NUM_KANA_SHIFT_CHR[i] = KBD_KANA_SHIFT_CHR[i];
        KBD_NUM_KANA_AIU_CHR[i] = KBD_KANA_AIU_CHR[i];
        KBD_NUM_KANA_AIU_SHIFT_CHR[i] = KBD_KANA_AIU_SHIFT_CHR[i];
        KBD_NUM_ROMA_CHR[i] = KBD_ROMA_CHR[i];
        KBD_NUM_ROMA_SHIFT_CHR[i] = KBD_ROMA_SHIFT_CHR[i];
      }
    }

    //可変キーを描く
    //kbdVarStatBitmap = new int[256][][];
    for (int i = 0; i < 256; i++) {
      int[] letter = KBD_VAR_LETTER[i];
      if (letter == null) {
        kbdVarStatBitmap[i] = null;
        continue;
      }
      int[][] statBitmap = kbdVarStatBitmap[i] = new int[4][];
      for (int stat = 0; stat < 4; stat++) {
        if (KBD_USE_ROM_FONT) {  //ROMフォントを使う
          kbdDrawKey2 (statBitmap[stat] = new int[KBD_KEY_WIDTH * KBD_KEY_HEIGHT],
                       KBD_KEY_WIDTH, 0, 0, KBD_KEY_WIDTH, KBD_KEY_HEIGHT, 0, 0, null, i, stat);
        } else {
          kbdDrawKey (statBitmap[stat] = new int[KBD_KEY_WIDTH * KBD_KEY_HEIGHT],
                      KBD_KEY_WIDTH, 0, 0, KBD_KEY_WIDTH, KBD_KEY_HEIGHT, 0, 0, null, letter, stat);
        }
      }
    }

    //LEDキーを描く
    //kbdLedStatBitmap = new int[KBD_LED_KEYS][][];
    for (int i = 0; i < KBD_LED_KEYS; i++) {
      int[] letter = KBD_LED_LETTER[i];
      if (letter == null) {
        kbdLedStatBitmap[i] = null;
        continue;
      }
      int[][] statBitmap = kbdLedStatBitmap[i] = new int[8][];
      for (int stat = 0; stat < 8; stat++) {
        kbdDrawKey (statBitmap[stat] = new int[KBD_KEY_WIDTH * KBD_KEY_HEIGHT],
                    KBD_KEY_WIDTH, 0, 0, KBD_KEY_WIDTH, KBD_KEY_HEIGHT, 0, 0, KBD_LED_COLOR[i][stat >> 2], letter, stat & 0x03);
      }
    }

    //種類毎に別々に用意するもの
    //kbdDataTypeImage = new BufferedImage[KBD_TYPES];
    //kbdDataTypeWidth = new int[KBD_TYPES];
    //kbdDataTypeHeight = new int[KBD_TYPES];
    //kbdDataTypeBitmap = new int[KBD_TYPES][];
    //kbdDataTypeIndexMap = new byte[KBD_TYPES][];
    //kbdDataTypeStatFixBitmap = new int[KBD_TYPES][][];
    //kbdDataTypeIndexRect = new int[KBD_TYPES][][];
    //kbdDataTypeIndexCharacterStat = new int[KBD_TYPES][];
    for (int type = 0; type < KBD_TYPES; type++) {
      int cols = KBD_TYPE_COLS[type];
      int rows = KBD_TYPE_ROWS[type];
      int width = kbdDataTypeWidth[type] = KBD_PADDING_LEFT + KBD_COL_WIDTH * cols + KBD_PADDING_RIGHT;
      int height = kbdDataTypeHeight[type] = KBD_PADDING_TOP + KBD_ROW_HEIGHT * rows + KBD_PADDING_BOTTOM;
      BufferedImage image = kbdDataTypeImage[type] = new BufferedImage (width, height, BufferedImage.TYPE_INT_RGB);
      kbdDataTypeBitmap[type] = ((DataBufferInt) image.getRaster ().getDataBuffer ()).getData ();
      Arrays.fill (kbdDataTypeBitmap[type], LnF.LNF_RGB[KBD_BACKGROUND_COLOR]);
      byte[] indexMap = kbdDataTypeIndexMap[type] = new byte[width * height];
      Arrays.fill (kbdDataTypeIndexMap[type], (byte) -1);
      kbdDataTypeStatFixBitmap[type] = new int[4][];
      int[][] indexRect = kbdDataTypeIndexRect[type] = new int[KBD_ALL_KEYS][];
      int[] indexCharacterStat = kbdDataTypeIndexCharacterStat[type] = new int[KBD_ALL_KEYS];
      Arrays.fill (indexCharacterStat, -1);  //初回はすべてのキーを描画する

      //可変キーとLEDキーのレクタングルを作る
      int[][] ledBox = KBD_TYPE_LED_BOX[type];
      for (int i = 0; i < KBD_VAR_KEYS + KBD_LED_KEYS; i++) {
        int[] box = i < KBD_VAR_KEYS ? KBD_VAR_BOX[i] : ledBox[i - KBD_VAR_KEYS];
        if (box == null) {  //欠番。Standard KeyboardのNUMキー
          indexRect[i] = new int[0];
          continue;
        }
        int x = KBD_PADDING_LEFT + KBD_COL_WIDTH * box[0];
        int y = KBD_PADDING_TOP + KBD_ROW_HEIGHT * box[1];
        int w = KBD_COL_WIDTH * box[2] - 1;
        int h = KBD_ROW_HEIGHT * box[3] - 1;
        indexRect[i] = box == null ? new int[0] : new int[] { x, y, w, h };
        int tx = x + KBD_TOUCHABLE_AREA;
        int ty = y + KBD_TOUCHABLE_AREA;
        int tw = w - KBD_TOUCHABLE_AREA * 2;
        int th = h - KBD_TOUCHABLE_AREA * 2;
        int p = tx + width * ty;
        for (int v = 0; v < th; v++) {
          for (int u = 0; u < tw; u++) {
            indexMap[p + u] = (byte) i;
          }
          p += width;
        }
      }

      //固定キーを描く
      for (int stat = 0; stat < 4; stat++) {
        int[] bitmap = kbdDataTypeStatFixBitmap[type][stat] = new int[width * height];
        Arrays.fill (bitmap, LnF.LNF_RGB[KBD_BACKGROUND_COLOR]);
      }
      int[][] fixBox = KBD_TYPE_FIX_BOX[type];
      int[] fixStyle = KBD_TYPE_FIX_STYLE[type];
      for (int i = KBD_VAR_KEYS + KBD_LED_KEYS; i < KBD_ALL_KEYS; i++) {
        int[] box = fixBox[i - (KBD_VAR_KEYS + KBD_LED_KEYS)];
        if (box == null) {  //欠番。Compact Keyboardのテンキー
          indexRect[i] = new int[0];
          continue;
        }
        int col = box[0];  //列
        int row = box[1];  //行
        int wide = box[2];  //幅
        int tall = box[3];  //高さ
        int x = KBD_PADDING_LEFT + KBD_COL_WIDTH * col;
        int y = KBD_PADDING_TOP + KBD_ROW_HEIGHT * row;
        int w = KBD_COL_WIDTH * wide - 1;
        int h = KBD_ROW_HEIGHT * tall - 1;
        int style = fixStyle[i - (KBD_VAR_KEYS + KBD_LED_KEYS)];
        if (style == KBD_RETURN_STYLE) {  //RETURNキー
          //レクタングルの左下を1列x4行削る。キートップの左右を1列ずつ削る
          int wl = KBD_COL_WIDTH;  //左側
          int hb = KBD_ROW_HEIGHT * 4;  //下側
          int wr = w - wl;  //右側
          int ht = h - hb;  //上側
          indexRect[i] = new int[] { x, y, w, ht, x + wl, y + ht, wr, hb };
          int tx = x + (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + KBD_COL_WIDTH);
          int ty = y + KBD_TOUCHABLE_AREA;
          int tw = w - (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + KBD_COL_WIDTH) * 2;
          int th = ht - KBD_TOUCHABLE_AREA * 2;
          int p = tx + width * ty;
          for (int v = 0; v < th; v++) {
            for (int u = 0; u < tw; u++) {
              indexMap[p + u] = (byte) i;
            }
            p += width;
          }
          tx = x + wl + (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + KBD_COL_WIDTH);
          ty = y + ht - KBD_TOUCHABLE_AREA;
          tw = wr - (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + KBD_COL_WIDTH) * 2;
          th = hb;
          p = tx + width * ty;
          for (int v = 0; v < th; v++) {
            for (int u = 0; u < tw; u++) {
              indexMap[p + u] = (byte) i;
            }
            p += width;
          }
          //1段目
          for (int stat = 0; stat < 2; stat++) {  //!pressedのみ
            int[] bitmap = kbdDataTypeStatFixBitmap[type][stat];
            kbdDrawLine (bitmap, x + width * y,
                         1, 1, LnF.LNF_RGB[KBD_TOP_LEFT_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_TOP_COLOR],  //→→
                         width, 1, LnF.LNF_RGB[KBD_TOP_RIGHT_COLOR], width, h - 2, LnF.LNF_RGB[KBD_RIGHT_COLOR],  //↓↓
                         -1, 1, LnF.LNF_RGB[KBD_BOTTOM_RIGHT_COLOR], -1, wr - 2, LnF.LNF_RGB[KBD_BOTTOM_COLOR],  //←
                         -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, hb - 1, LnF.LNF_RGB[KBD_LEFT_COLOR],  //↑
                         -1, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR],  -1, wl - 1, LnF.LNF_RGB[KBD_BOTTOM_COLOR],  //←
                         -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, ht - 2, LnF.LNF_RGB[KBD_LEFT_COLOR]);  //↑
          }
          for (int stat = 2; stat < 4; stat++) {  //pressedのみ
            int[] bitmap = kbdDataTypeStatFixBitmap[type][stat];
            kbdDrawLine (bitmap, x + width * y,
                         1, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_HOLE_COLOR],  //→→
                         width, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], width, h - 2, LnF.LNF_RGB[KBD_HOLE_COLOR],  //↓↓
                         -1, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], -1, wr - 2, LnF.LNF_RGB[KBD_HOLE_COLOR],  //←
                         -width, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], -width, hb - 1, LnF.LNF_RGB[KBD_HOLE_COLOR],  //↑
                         -1, 1, LnF.LNF_RGB[KBD_HOLE_COLOR],  -1, wl - 1, LnF.LNF_RGB[KBD_HOLE_COLOR],  //←
                         -width, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], -width, ht - 2, LnF.LNF_RGB[KBD_HOLE_COLOR]);  //↑
          }
          int u = KBD_COL_WIDTH;
          x += 1 + u;
          y += 1;
          w -= 2 + u * 2;
          h -= 2;
          wr -= 2 + u * 2;
          ht -= 2;
          //島
          for (int stat = 0; stat < 4; stat++) {
            int[] bitmap = kbdDataTypeStatFixBitmap[type][stat];
            kbdFillRect (bitmap, width, x - u, y, u, ht, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //左側の上半分
            kbdFillRect (bitmap, width, x + wl - u, y + ht, u, h - ht, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //左側の下半分
            kbdFillRect (bitmap, width, x + w, y, u, h, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //右側
          }
          //2段目
          int lx = x;
          int ly = y;
          int lw = w;
          int lh = h;
          for (int stat = 0; stat < 4; stat++) {
            int[] bitmap = kbdDataTypeStatFixBitmap[type][stat];
            kbdDrawLine (bitmap, x + width * y,
                         1, 1, LnF.LNF_RGB[KBD_TOP_LEFT_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_TOP_COLOR],  //→→
                         width, 1, LnF.LNF_RGB[KBD_TOP_RIGHT_COLOR], width, h - 2, LnF.LNF_RGB[KBD_RIGHT_COLOR],  //↓↓
                         -1, 1, LnF.LNF_RGB[KBD_BOTTOM_RIGHT_COLOR], -1, wr - 2, LnF.LNF_RGB[KBD_BOTTOM_COLOR],  //←
                         -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, hb - 1, LnF.LNF_RGB[KBD_LEFT_COLOR],  //↑
                         -1, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -1, wl - 1, LnF.LNF_RGB[KBD_BOTTOM_COLOR],  //←
                         -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, ht - 2, LnF.LNF_RGB[KBD_LEFT_COLOR]);  //↑
          }
          x++;
          y++;
          w -= 2;
          h -= 2;
          wr -= 2;
          ht -= 2;
          //3段目
          for (int stat = 0; stat < 4; stat++) {
            int[] bitmap = kbdDataTypeStatFixBitmap[type][stat];
            kbdDrawLine (bitmap, x + width * y,
                         1, 1, LnF.LNF_RGB[KBD_TOP_LEFT_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_TOP_COLOR],  //→→
                         width, 1, LnF.LNF_RGB[KBD_TOP_RIGHT_COLOR], width, h - 2, LnF.LNF_RGB[KBD_RIGHT_COLOR],  //↓↓
                         -1, 1, LnF.LNF_RGB[KBD_BOTTOM_RIGHT_COLOR], -1, wr - 2, LnF.LNF_RGB[KBD_BOTTOM_COLOR],  //←
                         -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, hb - 1, LnF.LNF_RGB[KBD_LEFT_COLOR],  //↑
                         -1, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -1, wl - 1, LnF.LNF_RGB[KBD_BOTTOM_COLOR],  //←
                         -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, ht - 2, LnF.LNF_RGB[KBD_LEFT_COLOR]);  //↑
          }
          x++;
          y++;
          w -= 2;
          h -= 2;
          wr -= 2;
          ht -= 2;
          //キートップと文字
          int[] letter = KBD_FIX_LETTER[KBD_RETURN_INDEX - (KBD_VAR_KEYS + KBD_LED_KEYS)];
          for (int stat = 0; stat < 4; stat++) {
            int[] bitmap = kbdDataTypeStatFixBitmap[type][stat];
            boolean pointed = (stat & KBD_POINTED_MASK) != 0;
            int color = pointed ? LnF.LNF_RGB[KBD_POINTED_COLOR] : LnF.LNF_RGB[KBD_KEYTOP_COLOR];
            kbdFillRect (bitmap, width, x, y, w, ht, color);  //上半分
            kbdFillRect (bitmap, width, x + wl, y + ht, w - wl, h - ht, color);  //下半分
            kbdDrawLetter (bitmap, width, lx, ly, lw, lh, letter);
          }
        } else if (style == KBD_SHIFT_STYLE) {  //SHIFTキー
          int xr = x + KBD_COL_WIDTH * (wide - 9);  //右
          w = KBD_COL_WIDTH * 9 - 1;  //幅
          indexRect[i] = new int[] { x, y, w, h, xr, y, w, h };
          int tx = x + (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + KBD_COL_WIDTH);
          int ty = y + KBD_TOUCHABLE_AREA;
          int tw = w - (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + KBD_COL_WIDTH) * 2;
          int th = h - KBD_TOUCHABLE_AREA * 2;
          int p = tx + width * ty;
          for (int v = 0; v < th; v++) {
            for (int u = 0; u < tw; u++) {
              indexMap[p + u] = (byte) i;
            }
            p += width;
          }
          tx = xr + (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + KBD_COL_WIDTH);
          p = tx + width * ty;
          for (int v = 0; v < th; v++) {
            for (int u = 0; u < tw; u++) {
              indexMap[p + u] = (byte) i;
            }
            p += width;
          }
          int[] letter = KBD_FIX_LETTER[KBD_SHIFT_INDEX - (KBD_VAR_KEYS + KBD_LED_KEYS)];
          for (int stat = 0; stat < 4; stat++) {
            int[] bitmap = kbdDataTypeStatFixBitmap[type][stat];
            kbdDrawKey (bitmap, width, x, y, w, h, KBD_COL_WIDTH, 0, null, letter, stat);
            kbdDrawKey (bitmap, width, xr, y, w, h, KBD_COL_WIDTH, 0, null, letter, stat);
          }
        } else {  //RETURNキーとSHIFTキー以外
          if (style == KBD_FUNCTION_STYLE) {  //F1～F10
            y += KBD_ROW_HEIGHT >> 1;  //キーの上端を1/2行削る
            h -= KBD_ROW_HEIGHT >> 1;
          }
          int pw = (style == KBD_NARROWER_STYLE ? KBD_COL_WIDTH >> 1 :  //キートップの左右を1/2列ずつ削る
                    style == KBD_NARROW_STYLE ? KBD_COL_WIDTH :  //キートップの左右を1列ずつ削る
                    0);
          int ph = (style == KBD_SHORTER_STYLE ? KBD_ROW_HEIGHT :  //キートップの上下を1行ずつ削る
                    style == KBD_SHORT_STYLE ? KBD_ROW_HEIGHT << 1 :  //キートップの上下を2行ずつ削る
                    0);
          indexRect[i] = new int[] { x, y, w, h };
          int tx = x + (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + pw);
          int ty = y + (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + ph);
          int tw = w - (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + pw) * 2;
          int th = h - (KBD_TOUCHABLE_AREA == 0 ? 0 : KBD_TOUCHABLE_AREA + ph) * 2;
          int p = tx + width * ty;
          for (int v = 0; v < th; v++) {
            for (int u = 0; u < tw; u++) {
              indexMap[p + u] = (byte) i;
            }
            p += width;
          }
          for (int stat = 0; stat < 4; stat++) {
            int[] bitmap = kbdDataTypeStatFixBitmap[type][stat];
            kbdDrawKey (bitmap, width, x, y, w, h, pw, ph, null, KBD_FIX_LETTER[i - (KBD_VAR_KEYS + KBD_LED_KEYS)], stat);
          }
        }  //RETURNキー以外
      }  //for i
    }  //for type

    //現在の状態
    if (kbdOn) {
      kbdWidth = kbdDataTypeWidth[kbdType];
      kbdHeight = kbdDataTypeHeight[kbdType];
      kbdImage = kbdDataTypeImage[kbdType];
      kbdBitmap = kbdDataTypeBitmap[kbdType];
      kbdIndexMap = kbdDataTypeIndexMap[kbdType];
      kbdFlush ();
    } else {
      kbdWidth = 0;
      kbdHeight = 0;
      kbdImage = null;
      kbdBitmap = null;
      kbdIndexMap = null;
    }
    //kbdStat = new int[KBD_ALL_KEYS];
    Arrays.fill (kbdStat, 0);
    kbdPointedIndex = -1;
    if (!MC68901.MFP_KBD_ON) {
      kbdRepeatTask = null;
    }

    //kbdRepeatDelay = 200 + 100 * 3;
    //kbdRepeatInterval = 30 + 5 * 4 * 4;
    kbdSetRepeatDelay (MainMemory.mmrRbs (0x00ed003a));
    kbdSetRepeatInterval (MainMemory.mmrRbs (0x00ed003b));

    //kbdF11Mode = KBD_F11_FULLSCREEN;
    switch (Settings.sgsGetString ("f11key").toLowerCase ()) {
      //case "fullscreen":
      //  kbdF11Mode = KBD_F11_FULLSCREEN;
      //  break;
    case "screenshot":
      kbdF11Mode = KBD_F11_SCREENSHOT;
      break;
    case "stopandstart":
      kbdF11Mode = KBD_F11_STOPANDSTART;
      break;
    default:
      kbdF11Mode = KBD_F11_FULLSCREEN;
    }

  }  //kbdInit()

  //kbdTini ()
  //  後始末
  public static void kbdTini () {
    Settings.sgsPutIntArray ("keymap", Arrays.equals (kbdCurrentMap, KBD_DEFAULT_MAP) ? new int[0] : kbdCurrentMap);
    Settings.sgsPutString ("f11key",
                           //kbdF11Mode == KBD_F11_FULLSCREEN ? "fullscreen" :
                           kbdF11Mode == KBD_F11_SCREENSHOT ? "screenshot" :
                           kbdF11Mode == KBD_F11_STOPANDSTART ? "stopandstart" :
                           "fullscreen");
  }

  //kbdDrawKey (bitmap, width, x, y, w, h, u, v, ledColors, letter, stat)
  //  キーを描く
  //  bitmap     ビットマップ
  //  width      bitmapの幅(px)
  //  x          キーのx座標(px)
  //  y          キーのy座標(px)
  //  w          キーの幅(px)
  //  h          キーの高さ(px)
  //  u          キートップの左右を削る幅(px)
  //  v          キートップの上下を削る幅(px)
  //  ledColors  LEDの色。[0]=キートップ側の色(rgb),[1]=ボーダー側の色(rgb)
  //  letter     文字のパターン
  //  stat       キーの状態。bit0=pointed,bit1=pressed
  public static void kbdDrawKey (int[] bitmap, int width, int x, int y, int w, int h, int u, int v, int[] ledColors, int[] letter, int stat) {
    boolean pointed = (stat & KBD_POINTED_MASK) != 0;
    boolean pressed = (stat & KBD_PRESSED_MASK) != 0;
    //1段目
    if (!pressed) {
      kbdDrawLine (bitmap, x + width * y,
                   1, 1, LnF.LNF_RGB[KBD_TOP_LEFT_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_TOP_COLOR],
                   width, 1, LnF.LNF_RGB[KBD_TOP_RIGHT_COLOR], width, h - 2, LnF.LNF_RGB[KBD_RIGHT_COLOR],
                   -1, 1, LnF.LNF_RGB[KBD_BOTTOM_RIGHT_COLOR], -1, w - 2, LnF.LNF_RGB[KBD_BOTTOM_COLOR],
                   -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, h - 2, LnF.LNF_RGB[KBD_LEFT_COLOR]);
    } else {
      kbdDrawLine (bitmap, x + width * y,
                   1, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_HOLE_COLOR],
                   width, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], width, h - 2, LnF.LNF_RGB[KBD_HOLE_COLOR],
                   -1, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], -1, w - 2, LnF.LNF_RGB[KBD_HOLE_COLOR],
                   -width, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], -width, h - 2, LnF.LNF_RGB[KBD_HOLE_COLOR]);
    }
    x += 1 + u;
    y += 1 + v;
    w -= 2 + u * 2;
    h -= 2 + v * 2;
    //島
    if (u > 0) {
      kbdFillRect (bitmap, width, x - u, y, u, h, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //左側
      kbdFillRect (bitmap, width, x + w, y, u, h, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //右側
    }
    if (v > 0) {
      kbdFillRect (bitmap, width, x, y - v, w, v, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //上側
      kbdFillRect (bitmap, width, x, y + h, w, v, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //下側
    }
    //2段目
    int lx = x;
    int ly = y;
    int lw = w;
    int lh = h;
    kbdDrawLine (bitmap, x + width * y,
                 1, 1, LnF.LNF_RGB[KBD_TOP_LEFT_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_TOP_COLOR],
                 width, 1, LnF.LNF_RGB[KBD_TOP_RIGHT_COLOR], width, h - 2, LnF.LNF_RGB[KBD_RIGHT_COLOR],
                 -1, 1, LnF.LNF_RGB[KBD_BOTTOM_RIGHT_COLOR], -1, w - 2, LnF.LNF_RGB[KBD_BOTTOM_COLOR],
                 -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, h - 2, LnF.LNF_RGB[KBD_LEFT_COLOR]);
    x++;
    y++;
    w -= 2;
    h -= 2;
    //3段目
    kbdDrawLine (bitmap, x + width * y,
                 1, 1, LnF.LNF_RGB[KBD_TOP_LEFT_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_TOP_COLOR],
                 width, 1, LnF.LNF_RGB[KBD_TOP_RIGHT_COLOR], width, h - 2, LnF.LNF_RGB[KBD_RIGHT_COLOR],
                 -1, 1, LnF.LNF_RGB[KBD_BOTTOM_RIGHT_COLOR], -1, w - 2, LnF.LNF_RGB[KBD_BOTTOM_COLOR],
                 -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, h - 2, LnF.LNF_RGB[KBD_LEFT_COLOR]);
    x++;
    y++;
    w -= 2;
    h -= 2;
    //キートップ
    kbdFillRect (bitmap, width, x, y, w, h, pointed ? LnF.LNF_RGB[KBD_POINTED_COLOR] : LnF.LNF_RGB[KBD_KEYTOP_COLOR]);
    //LED
    if (ledColors != null) {
      x += w - KBD_LED_WIDTH >> 1;
      y += h;
      kbdFillRect (bitmap, width, x, y - KBD_LED_HEIGHT, KBD_LED_WIDTH, KBD_LED_HEIGHT, ledColors[0]);  //上半分
      kbdFillRect (bitmap, width, x, y, KBD_LED_WIDTH, pressed ? 2 : 3, ledColors[1]);  //下半分の高さはボーダーの幅
    }
    //文字
    if (letter != null) {  //文字あり
      kbdDrawLetter (bitmap, width, lx, ly, lw, lh, letter);
    }
  }  //kbdDrawKey

  //kbdDrawKey2 (bitmap, width, x, y, w, h, u, v, ledColors, code, stat)
  //  キーを描く
  //  bitmap     ビットマップ
  //  width      bitmapの幅(px)
  //  x          キーのx座標(px)
  //  y          キーのy座標(px)
  //  w          キーの幅(px)
  //  h          キーの高さ(px)
  //  u          キートップの左右を削る幅(px)
  //  v          キートップの上下を削る幅(px)
  //  ledColors  LEDの色。[0]=キートップ側の色(rgb),[1]=ボーダー側の色(rgb)
  //  code       文字コード
  //  stat       キーの状態。bit0=pointed,bit1=pressed
  public static void kbdDrawKey2 (int[] bitmap, int width, int x, int y, int w, int h, int u, int v, int[] ledColors, int code, int stat) {
    if (KBD_USE_ROM_FONT) {  //ROMフォントを使う
      boolean pointed = (stat & KBD_POINTED_MASK) != 0;
      boolean pressed = (stat & KBD_PRESSED_MASK) != 0;
      //1段目
      if (!pressed) {
        kbdDrawLine (bitmap, x + width * y,
                     1, 1, LnF.LNF_RGB[KBD_TOP_LEFT_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_TOP_COLOR],
                     width, 1, LnF.LNF_RGB[KBD_TOP_RIGHT_COLOR], width, h - 2, LnF.LNF_RGB[KBD_RIGHT_COLOR],
                     -1, 1, LnF.LNF_RGB[KBD_BOTTOM_RIGHT_COLOR], -1, w - 2, LnF.LNF_RGB[KBD_BOTTOM_COLOR],
                     -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, h - 2, LnF.LNF_RGB[KBD_LEFT_COLOR]);
      } else {
        kbdDrawLine (bitmap, x + width * y,
                     1, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_HOLE_COLOR],
                     width, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], width, h - 2, LnF.LNF_RGB[KBD_HOLE_COLOR],
                     -1, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], -1, w - 2, LnF.LNF_RGB[KBD_HOLE_COLOR],
                     -width, 1, LnF.LNF_RGB[KBD_HOLE_COLOR], -width, h - 2, LnF.LNF_RGB[KBD_HOLE_COLOR]);
      }
      x += 1 + u;
      y += 1 + v;
      w -= 2 + u * 2;
      h -= 2 + v * 2;
      //島
      if (u > 0) {
        kbdFillRect (bitmap, width, x - u, y, u, h, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //左側
        kbdFillRect (bitmap, width, x + w, y, u, h, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //右側
      }
      if (v > 0) {
        kbdFillRect (bitmap, width, x, y - v, w, v, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //上側
        kbdFillRect (bitmap, width, x, y + h, w, v, LnF.LNF_RGB[KBD_KEYTOP_COLOR]);  //下側
      }
      //2段目
      int lx = x;
      int ly = y;
      int lw = w;
      int lh = h;
      kbdDrawLine (bitmap, x + width * y,
                   1, 1, LnF.LNF_RGB[KBD_TOP_LEFT_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_TOP_COLOR],
                   width, 1, LnF.LNF_RGB[KBD_TOP_RIGHT_COLOR], width, h - 2, LnF.LNF_RGB[KBD_RIGHT_COLOR],
                   -1, 1, LnF.LNF_RGB[KBD_BOTTOM_RIGHT_COLOR], -1, w - 2, LnF.LNF_RGB[KBD_BOTTOM_COLOR],
                   -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, h - 2, LnF.LNF_RGB[KBD_LEFT_COLOR]);
      x++;
      y++;
      w -= 2;
      h -= 2;
      //3段目
      kbdDrawLine (bitmap, x + width * y,
                   1, 1, LnF.LNF_RGB[KBD_TOP_LEFT_COLOR], 1, w - 2, LnF.LNF_RGB[KBD_TOP_COLOR],
                   width, 1, LnF.LNF_RGB[KBD_TOP_RIGHT_COLOR], width, h - 2, LnF.LNF_RGB[KBD_RIGHT_COLOR],
                   -1, 1, LnF.LNF_RGB[KBD_BOTTOM_RIGHT_COLOR], -1, w - 2, LnF.LNF_RGB[KBD_BOTTOM_COLOR],
                   -width, 1, LnF.LNF_RGB[KBD_BOTTOM_LEFT_COLOR], -width, h - 2, LnF.LNF_RGB[KBD_LEFT_COLOR]);
      x++;
      y++;
      w -= 2;
      h -= 2;
      //キートップ
      kbdFillRect (bitmap, width, x, y, w, h, pointed ? LnF.LNF_RGB[KBD_POINTED_COLOR] : LnF.LNF_RGB[KBD_KEYTOP_COLOR]);
      //LED
      if (ledColors != null) {
        x += w - KBD_LED_WIDTH >> 1;
        y += h;
        kbdFillRect (bitmap, width, x, y - KBD_LED_HEIGHT, KBD_LED_WIDTH, KBD_LED_HEIGHT, ledColors[0]);  //上半分
        kbdFillRect (bitmap, width, x, y, KBD_LED_WIDTH, pressed ? 2 : 3, ledColors[1]);  //下半分の高さはボーダーの幅
      }
      //文字
      if (code != 0) {  //文字あり
        kbdDrawLetter2 (bitmap, width, lx, ly, lw, lh, code);
      }
    }  //if KBD_USE_ROM_FONT
  }  //kbdDrawKey2

  //kbdFillRect (bitmap, width, x, y, w, h, color)
  //  矩形を塗り潰す
  //  bitmap  ビットマップ
  //  width   bitmapの幅(px)
  //  x       x座標(px)
  //  y       y座標(px)
  //  w       幅(px)
  //  h       高さ(px)
  //  color   色(rgb)
  public static void kbdFillRect (int[] bitmap, int width, int x, int y, int w, int h, int color) {
    for (int j = 0; j < h; j++) {
      int dst = x + width * (y + j);
      for (int i = 0; i < w; i++) {
        bitmap[dst + i] = color;
      }
    }
  }  //kbdFillRect(int[],int,int,int,int,int,int)

  //kbdDrawLetter (bitmap, width, lx, ly, lw, lh, letter)
  //  文字のパターンを描く
  //  影を付ける
  //  bitmap  ビットマップ
  //  width   bitmapの幅(px)
  //  lx      x座標(px)
  //  ly      y座標(px)
  //  lw      幅(px)
  //  lh      高さ(px)
  //  letter  文字のパターン
  public static void kbdDrawLetter (int[] bitmap, int width, int lx, int ly, int lw, int lh, int[] letter) {
    if (lw <= 32) {
      for (int j = 0; j < lh; j++) {
        int dst = lx + width * (ly + j);
        for (int i = 0; i < lw; i++) {
          if ((letter[j] >> lw - 1 - i & 1) != 0) {
            bitmap[dst + i] = LnF.LNF_RGB[KBD_LETTER_COLOR];
          } else if (i > 0 && j > 0 && (letter[j - 1] >> lw - 1 - (i - 1) & 1) != 0) {
            bitmap[dst + i] = LnF.LNF_RGB[KBD_SHADOW_COLOR];
            //bitmap[dst + i] = 0xff000000 | bitmap[dst + i] >> 2 & 0x003f3f3f;
            //bitmap[dst + i] = 0xff000000 | bitmap[dst + i] >> 1 & 0x007f7f7f;
          }
        }
      }
    } else {
      int n = lw + 31 >> 5;  //1ラスタあたりの文字データの数
      for (int j = 0; j < lh; j++) {
        int src = n * j + n - 1;  //文字データの右端
        int dst = lx + width * (ly + j) + lw - 1;  //キートップの右端
        for (int i = 0; i < lw; i++) {  //右端が0,左端がlw-1
          if ((letter[src - (i >> 5)] >> (i & 0x1f) & 1) != 0) {
            bitmap[dst - i] = LnF.LNF_RGB[KBD_LETTER_COLOR];
          } else if (i > 0 && j > 0 && (letter[src - n - (i + 1 >> 5)] >> (i + 1 & 0x1f) & 1) != 0) {
            bitmap[dst - i] = LnF.LNF_RGB[KBD_SHADOW_COLOR];
            //bitmap[dst - i] = 0xff000000 | bitmap[dst - i] >> 2 & 0x003f3f3f;
            //bitmap[dst - i] = 0xff000000 | bitmap[dst - i] >> 1 & 0x007f7f7f;
          }
        }
      }
    }
  }  //kbdDrawLetter(int[],int,int,int,int,int,int[])

  //kbdDrawLetter2 (bitmap, width, lx, ly, lw, lh, code)
  //  文字のパターンを描く
  //  影を付ける
  //  bitmap  ビットマップ
  //  width   bitmapの幅(px)
  //  lx      x座標(px)
  //  ly      y座標(px)
  //  lw      幅(px)
  //  lh      高さ(px)
  //  code    文字コード
  public static void kbdDrawLetter2 (int[] bitmap, int width, int lx, int ly, int lw, int lh, int code) {
    final int fw = 8;
    final int fh = 8;
    final int fa = FontEditor.FNT_ADDRESS_Yon8x8;
    int dst = lx + (lw - fw >> 1) + width * (ly + (lh - fh >> 1));
    int src = fa + (fw + 7 >> 3) * fh * code;
    for (int y = 0; y < fh; y++) {
      for (int x = 0; x < fw; x++) {
        if ((MainMemory.mmrM8[src + (fw + 7 >> 3) * y + (x >> 3)] & 1 << (~x & 0x07)) != 0) {
          bitmap[dst + width * y + x] = LnF.LNF_RGB[KBD_LETTER_COLOR];
        } else if (x > 0 && y > 0 &&
                   (MainMemory.mmrM8[src + (fw + 7 >> 3) * (y - 1) + (x - 1 >> 3)] & 1 << (~(x - 1) & 0x07)) != 0) {
          bitmap[dst + width * y + x] = LnF.LNF_RGB[KBD_SHADOW_COLOR];
        }
      }
    }
  }  //kbdDrawLetter2(int[],int,int,int,int,int,int)

  //kbdDrawLine (bitmap, pos, vect0, step0, color0, vect1, step1, color1, ...)
  //  bitmap  ビットマップ
  //  pos     開始位置のオフセット
  //  vect    方向。オフセットの差分
  //  step    長さ(px)
  //  color   色(rgb)
  public static void kbdDrawLine (int[] bitmap, int pos, int... vect_step_color) {
    for (int i = 0; i < vect_step_color.length; i += 3) {
      int vect = vect_step_color[i];
      int step = vect_step_color[i + 1];
      int color = vect_step_color[i + 2];
      for (; step > 0; step--) {
        bitmap[pos] = color;
        pos += vect;
      }
    }
  }  //kbdDrawLine(int[],int,int...)

  //kbdStart ()
  //  キーボードのイベントリスナーを設定して動作を開始する
  //  MFPが初期化されている必要がある
  public static void kbdStart () {

    //Tabキーによるフォーカスの移動を無効にする
    //  setFocusTraversalKeysにCollections.EMPTY_SETを指定すると無検査変換の警告が出る
    //XEiJ.pnlPanel.setFocusTraversalKeys (KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS, Collections.EMPTY_SET);
    //XEiJ.pnlPanel.setFocusTraversalKeys (KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS, Collections.EMPTY_SET);
    //XEiJ.pnlPanel.setFocusTraversalKeys (KeyboardFocusManager.UP_CYCLE_TRAVERSAL_KEYS, Collections.EMPTY_SET);
    XEiJ.pnlPanel.setFocusTraversalKeysEnabled (false);  //Component.setFocusTraversalKeysEnabled()は1.4から

    //フォーカスリスナーとキーリスナー
    XEiJ.pnlPanel.setFocusable (true);  //普通のコンポーネントはクリックしてもフォーカスされないのでキーイベントが発生しない
    ComponentFactory.addListener (
      XEiJ.pnlPanel,
      new FocusAdapter () {
        @Override public void focusGained (FocusEvent fe) {
          kbdCloseIME ();  //IMEが開いていたら閉じる
        }
        //@Override public void focusLost (FocusEvent fe) {
        //}
      });
    XEiJ.pnlPanel.requestFocusInWindow ();  //フォーカスを要求する。Component.requestFocusInWindow()は1.4から
    ComponentFactory.addListener (
      XEiJ.pnlPanel,
      new KeyAdapter () {
        @Override public void keyPressed (KeyEvent ke) {
          if (XEiJ.musOutputButtonStatus) {
            System.out.println (String.format ("keyPressed keyChar=0x%04x keyCode=0x%08x extendedKeyCode=0x%08x keyLocation=%d",
                                               (int) ke.getKeyChar (), ke.getKeyCode (), ke.getExtendedKeyCode (), ke.getKeyLocation ()));
          }
          kbdCloseIME ();  //IMEが開いていたら閉じる
          kbdInput (ke, true);
          ke.consume ();
        }
        @Override public void keyReleased (KeyEvent ke) {
          if (XEiJ.musOutputButtonStatus) {
            System.out.println (String.format ("keyReleased keyChar=0x%04x keyCode=0x%08x extendedKeyCode=0x%08x keyLocation=%d",
                                               (int) ke.getKeyChar (), ke.getKeyCode (), ke.getExtendedKeyCode (), ke.getKeyLocation ()));
          }
          kbdCloseIME ();  //IMEが開いていたら閉じる
          kbdInput (ke, false);
          ke.consume ();
        }
        @Override public void keyTyped (KeyEvent ke) {
          if (XEiJ.musOutputButtonStatus) {
            System.out.println (String.format ("keyTyped keyChar=0x%04x keyCode=0x%08x extendedKeyCode=0x%08x keyLocation=%d",
                                               (int) ke.getKeyChar (), ke.getKeyCode (), ke.getExtendedKeyCode (), ke.getKeyLocation ()));
          }
          kbdCloseIME ();  //IMEが開いていたら閉じる
          ke.consume ();
        }
      });

    //IMEが開いていたら閉じる
    kbdCloseIME ();

    if (kbdOn) {
      //点灯要求コードを送出する
      if (MC68901.MFP_KBD_ON) {
        MC68901.mfpKbdInput (0xff, false);
      } else {
        kbdTransfer (0xff, false);
      }
    }

    //キーボード設定ウィンドウの準備
    if (RestorableFrame.rfmGetOpened (Settings.SGS_KBS_FRAME_KEY)) {
      kbdOpen ();
    }

  }  //kbdStart()

  //kbdOpen ()
  //  キーボード設定ウィンドウを開く
  public static void kbdOpen () {
    if (kbdFrame == null) {
      kbdMakeFrame ();
    }
    kbdFrame.setVisible (true);
  }  //kbdOpen()

  //kbdMakeFrame ()
  //  キーボード設定ウィンドウを作る
  //  ここでは開かない
  public static void kbdMakeFrame () {

    kbdTextFieldArray = new JTextField[KBD_ALL_KEYS];

    //キーリスナー
    KeyListener keyListener = new KeyAdapter () {
      @Override public void keyPressed (KeyEvent ke) {
        int keyLocation = ke.getKeyLocation ();
        if (keyLocation == 0) {
          keyLocation = 1;
        }
        int locationMask = 1 << keyLocation;
        int rawCode = kbdGetRawCode (ke);
        int keyCode = ke.getKeyCode ();
        int xo = Integer.parseInt (((JTextField) ke.getSource ()).getName ());
        if (keyCode == KeyEvent.VK_ESCAPE) {  //EscはESC以外解除。ESCは再配置できない
          if (kbdCurrentMap[3 * xo + 1] != 27) {  //ESC以外
            kbdCurrentMap[3 * xo] = 0;  //keyLocationを削除
            kbdCurrentMap[3 * xo + 1] = 0;  //rawCodeを削除
            kbdCurrentMap[3 * xo + 2] = 0;  //keyCodeを削除
            kbdTextFieldArray[xo].setText (kbdGetMapText (xo));
          }
        } else {
          //既に割り当てられているキーがあれば取り除く
          for (int xp = 0; xp < KBD_ALL_KEYS; xp++) {
            if (kbdCurrentMap[3 * xp + 1] == rawCode &&  //rawCodeが一致
                (kbdCurrentMap[3 * xp] & locationMask) != 0) {  //keyLocationが一致
              kbdCurrentMap[3 * xp] &= ~locationMask;  //keyLocationを削除
              if (kbdCurrentMap[3 * xp] == 0) {  //keyLocationがなくなった
                kbdCurrentMap[3 * xp + 1] = 0;  //rawCodeを削除
                kbdCurrentMap[3 * xp + 2] = 0;  //keyCodeを削除
              }
              kbdTextFieldArray[xp].setText (kbdGetMapText (xp));
              break;
            }
          }
          if (kbdCurrentMap[3 * xo + 1] == rawCode) {  //rawCodeが同じ
            kbdCurrentMap[3 * xo] |= locationMask;  //keyLocationを追加
          } else {  //rawCodeが違う
            kbdCurrentMap[3 * xo] = locationMask;  //keyLocationを変更
            kbdCurrentMap[3 * xo + 1] = rawCode;  //rawCodeを変更
          }
          kbdCurrentMap[3 * xo + 2] = keyCode;  //keyCodeを変更
          kbdTextFieldArray[xo].setText (kbdGetMapText (xo));
        }
        ke.consume ();
      }
      @Override public void keyReleased (KeyEvent ke) {
        ke.consume ();
      }
      @Override public void keyTyped (KeyEvent ke) {
        ke.consume ();
      }
    };

    //             0             1
    //    0  Key of X68000  Assignment
    //    1  ---------------------------
    //    2        L            TF
    //  114        L            TF
    ArrayList<Object> cellList = new ArrayList<Object> ();
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Key of X68000"), "ja", "X68000 のキー"));  //(0,0)
    cellList.add (Multilingual.mlnText (ComponentFactory.createLabel ("Assignment"), "ja", "割り当て"));  //(1,0)
    cellList.add (ComponentFactory.createHorizontalSeparator ());  //(0,1)-(1,1)
    for (int xo = 0; xo < KBD_ALL_KEYS; xo++) {
      kbdTextFieldArray[xo] = ComponentFactory.addListener (
        ComponentFactory.setHorizontalAlignment (
          ComponentFactory.setName (
            ComponentFactory.createTextField (kbdGetMapText (xo), 28),
            String.valueOf (xo)),
          JTextField.CENTER),
        keyListener);
      cellList.add (KBD_KEYTOP_TEXT[xo]);  //(0,2+xo)
      cellList.add (kbdTextFieldArray[xo]);  //(1,2+xo)
    }
    JPanel gridPanel = ComponentFactory.createGridPanel (
      2,  //colCount
      2 + KBD_ALL_KEYS,  //rowCount
      "paddingLeft=3,paddingRight=3,center",   //gridStyles
      "",  //colStyles
      ";colSpan=2,widen",  //rowStyles
      "",  //cellStyles
      cellList.toArray (new Object[0]));

    //ウインドウ
    kbdFrame = Multilingual.mlnTitle (
      ComponentFactory.createRestorableSubFrame (
        Settings.SGS_KBS_FRAME_KEY,
        "Key assignment",
        null,
        ComponentFactory.createVerticalBox (
          ComponentFactory.setPreferredSize (
            new JScrollPane (gridPanel),
            400, 600))),
      "ja", "キー割り当て");

  }  //kbdMakeFrame()

  public static String kbdGetMapText (int xo) {
    int locationMask = kbdCurrentMap[3 * xo];
    int keyCode = kbdCurrentMap[3 * xo + 2];
    StringBuilder sb = new StringBuilder ();
    if (locationMask == 0) {
      sb.append (Multilingual.mlnJapanese ? "なし" : "none");
    } else {
      while (locationMask != 0) {
        int m = locationMask & -locationMask;  //下位から
        switch (m) {
        case 2:  //standard
          break;
        case 4:  //left
          sb.append (Multilingual.mlnJapanese ? "左 " : "Left ");
          break;
        case 8:  //right
          sb.append (Multilingual.mlnJapanese ? "右 " : "Right ");
          break;
        case 16:  //numpad
          sb.append (Multilingual.mlnJapanese ? "テンキー " : "Numpad ");
          break;
        }
        locationMask &= ~m;
        sb.append (KeyEvent.getKeyText (keyCode));  //!!! 標準のテキストが残念な感じなので自前で用意したい
        if (locationMask != 0) {
          sb.append (Multilingual.mlnJapanese ? " または " : " or ");
        }
      }
    }
    return sb.toString ();
  }


  //kbdCloseIME ()
  //  IMEが開いていたら閉じる
  //  継続的にIMEを無効化する方法が見当たらないのでIMEを開くキーが押されたらすぐに閉じることで擬似的にIMEを無効化する
  public static void kbdCloseIME () {
    try {
      InputContext context = XEiJ.pnlPanel.getInputContext ();
      if (context.isCompositionEnabled ()) {  //InputContext.isCompositionEnabled()は1.3から
        context.setCompositionEnabled (false);  //InputContext.setCompositionEnabled()は1.3から
        //context.setCharacterSubsets (null);
      }
    } catch (UnsupportedOperationException uoe) {
    }
  }  //kbdCloseIME()

  //kbdSetOn (on)
  //  キーボードを表示する/表示しない
  public static void kbdSetOn (boolean on) {
    if (kbdOn != on) {
      kbdOn = on;
      if (on) {
        if (kbdType == KBD_STANDARD_TYPE) {
          System.out.println (Multilingual.mlnJapanese ?
                              "標準キーボードを接続します" :
                              "Connecting standard keyboard");
        } else if (kbdType == KBD_COMPACT_TYPE) {
          System.out.println (Multilingual.mlnJapanese ?
                              "コンパクトキーボードを接続します" :
                              "Connecting compact keyboard");
        }
        kbdWidth = kbdDataTypeWidth[kbdType];
        kbdHeight = kbdDataTypeHeight[kbdType];
        kbdImage = kbdDataTypeImage[kbdType];
        kbdBitmap = kbdDataTypeBitmap[kbdType];
        kbdIndexMap = kbdDataTypeIndexMap[kbdType];
        kbdFlush ();
        //点灯要求コードを送出する
        if (MC68901.MFP_KBD_ON) {
          MC68901.mfpKbdInput (0xff, false);
        } else {
          kbdTransfer (0xff, false);
        }
      } else {
        System.out.println (Multilingual.mlnJapanese ?
                            "キーボードを切り離します" :
                            "Disconnecting keyboard");
        //ポイントされているキーがあれば解除する
        if (kbdPointedIndex >= 0) {
          kbdHover (0, 0);
        }
        kbdWidth = 0;
        kbdHeight = 0;
        kbdImage = null;
        kbdBitmap = null;
        kbdIndexMap = null;
      }
      XEiJ.pnlUpdateArrangement ();
    }
  }  //kbdSetOn(boolean)

  //kbdSetType (type)
  //  キーボードの種類を変更する
  //  type  キーボードの種類。KEY_STANDARD_TYPEまたはKBD_COMPACT_TYPE
  public static void kbdSetType (int type) {
    if (kbdType != type) {
      kbdType = type;
      if (kbdOn) {
        System.out.println (Multilingual.mlnJapanese ?
                            "キーボードを切り離します" :
                            "Disconnecting keyboard");
        //ポイントされているキーがあれば解除する
        if (kbdPointedIndex >= 0) {
          kbdHover (0, 0);
        }
        if (kbdType == KBD_STANDARD_TYPE) {
          System.out.println (Multilingual.mlnJapanese ?
                              "標準キーボードを接続します" :
                              "Connecting standard keyboard");
        } else if (kbdType == KBD_COMPACT_TYPE) {
          System.out.println (Multilingual.mlnJapanese ?
                              "コンパクトキーボードを接続します" :
                              "Connecting compact keyboard");
        }
        kbdWidth = kbdDataTypeWidth[kbdType];
        kbdHeight = kbdDataTypeHeight[kbdType];
        kbdImage = kbdDataTypeImage[kbdType];
        kbdBitmap = kbdDataTypeBitmap[kbdType];
        kbdIndexMap = kbdDataTypeIndexMap[kbdType];
        kbdFlush ();
        //点灯要求コードを送出する
        if (MC68901.MFP_KBD_ON) {
          MC68901.mfpKbdInput (0xff, false);
        } else {
          kbdTransfer (0xff, false);
        }
        XEiJ.pnlUpdateArrangement ();
      }
    }
  }  //kbdSetType(int)

  //kbdRepaint ()
  //  キーボードが表示されているときkbdImageを作り直して再描画する
  public static void kbdRepaint () {
    if (kbdOn) {  //キーボードが表示されているとき
      kbdFlush ();  //kbdImageを作り直して
      XEiJ.pnlPanel.repaint (XEiJ.pnlKeyboardX, XEiJ.pnlKeyboardY, kbdWidth, kbdHeight);  //再描画する
    }
  }  //kbdRepaint()

  //kbdFlush ()
  //  kbdImageを作り直す
  public static void kbdFlush () {
    int width = kbdDataTypeWidth[kbdType];
    int[][] indexRect = kbdDataTypeIndexRect[kbdType];
    int[] indexCharacterStat = kbdDataTypeIndexCharacterStat[kbdType];
    int[][] statFixBitmap = kbdDataTypeStatFixBitmap[kbdType];
    //LEDキーとモードキーの状態によってキートップに表示する文字の配列を選択する
    boolean numOff = KBD_LED_TEST << (kbdStat[KBD_NUM_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) >= 0;
    boolean ctrlOn = KBD_MODE_TEST << (kbdStat[KBD_CTRL_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) < 0;
    boolean codeOn = KBD_LED_TEST << (kbdStat[KBD_CODE_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) < 0;
    boolean opt1Off = KBD_MODE_TEST << (kbdStat[KBD_OPT1_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) >= 0;
    boolean kanaOn = KBD_LED_TEST << (kbdStat[KBD_KANA_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) < 0;
    boolean shiftOff = KBD_MODE_TEST << (kbdStat[KBD_SHIFT_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) >= 0;
    boolean romaOn = KBD_LED_TEST << (kbdStat[KBD_ROMA_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) < 0;
    boolean capsOff = KBD_LED_TEST << (kbdStat[KBD_CAPS_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) >= 0;
    boolean zenOn = KBD_LED_TEST << (kbdStat[KBD_ZEN_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) < 0;
    boolean hiraOn = KBD_LED_TEST << (kbdStat[KBD_HIRA_INDEX] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK)) < 0;
    byte[] varChr =
      (numOff
       ?  //-NUM
       (ctrlOn
        ?  //-NUM+CTRL
        KBD_CTRL_CHR
        :  //-NUM-CTRL
        (codeOn
         ?  //-NUM-CTRL+CODE
         KBD_CODE_CHR
         :  //-NUM-CTRL-CODE
         (opt1Off && kanaOn
          ?  //-NUM-CTRL-CODE+KANA
          (XEiJ.busPbs (0x00ed002b) == 0
           ?  //-NUM-CTRL-CODE+KANA-AIU
           (shiftOff
            ?  //-NUM-CTRL-CODE+KANA-AIU-SHIFT
            KBD_KANA_CHR
            :  //-NUM-CTRL-CODE+KANA-AIU+SHIFT
            KBD_KANA_SHIFT_CHR
            )  //!SHIFT
           :  //-NUM-CTRL-CODE+KANA+AIU
           (shiftOff
            ?  //-NUM-CTRL-CODE+KANA+AIU-SHIFT
            KBD_KANA_AIU_CHR
            :  //-NUM-CTRL-CODE+KANA+AIU+SHIFT
            KBD_KANA_AIU_SHIFT_CHR
            )  //!SHIFT
           )  //!AIU
          :  //-NUM-CTRL-CODE-KANA
          (opt1Off && romaOn
           ?  //-NUM-CTRL-CODE-KANA+ROMA
           (shiftOff
            ?  //-NUM-CTRL-CODE-KANA+ROMA-SHIFT
            KBD_ROMA_CHR
            :  //-NUM-CTRL-CODE-KANA+ROMA+SHIFT
            KBD_ROMA_SHIFT_CHR
            )  //!SHIFT
           :  //-NUM-CTRL-CODE-KANA-ROMA
           (capsOff
            ?  //-NUM-CTRL-CODE-KANA-ROMA-CAPS
            (shiftOff
             ?  //-NUM-CTRL-CODE-KANA-ROMA-CAPS-SHIFT
             KBD_NORMAL_CHR
             :  //-NUM-CTRL-CODE-KANA-ROMA-CAPS+SHIFT
             KBD_SHIFT_CHR
             )  //!SHIFT
            :  //-NUM-CTRL-CODE-KANA-ROMA+CAPS
            (shiftOff
             ?  //-NUM-CTRL-CODE-KANA-ROMA+CAPS-SHIFT
             KBD_CAPS_CHR
             :  //-NUM-CTRL-CODE-KANA-ROMA+CAPS+SHIFT
             KBD_CAPS_SHIFT_CHR
             )  //!SHIFT
            )  //!CAPS
           )  //!OPT.1&&ROMA
          )  //!OPT.1&&KANA
         )  //CODE
        )  //CTRL
       :  //+NUM
       (ctrlOn
        ?  //+NUM+CTRL
        KBD_NUM_CTRL_CHR
        :  //+NUM-CTRL
        (codeOn
         ?  //+NUM-CTRL+CODE
         KBD_NUM_CODE_CHR
         :  //+NUM-CTRL-CODE
         (opt1Off && kanaOn
          ?  //+NUM-CTRL-CODE+KANA
          (MainMemory.mmrM8[0x00ed002b] == 0
           ?  //+NUM-CTRL-CODE+KANA-AIU
           (shiftOff
            ?  //+NUM-CTRL-CODE+KANA-AIU-SHIFT
            KBD_NUM_KANA_CHR
            :  //+NUM-CTRL-CODE+KANA-AIU+SHIFT
            KBD_NUM_KANA_SHIFT_CHR
            )  //!SHIFT
           :  //+NUM-CTRL-CODE+KANA+AIU
           (shiftOff
            ?  //+NUM-CTRL-CODE+KANA+AIU-SHIFT
            KBD_NUM_KANA_AIU_CHR
            :  //+NUM-CTRL-CODE+KANA+AIU+SHIFT
            KBD_NUM_KANA_AIU_SHIFT_CHR
            )  //!SHIFT
           )  //!AIU
          :  //+NUM-CTRL-CODE-KANA
          (opt1Off && romaOn
           ?  //+NUM-CTRL-CODE-KANA+ROMA
           (shiftOff
            ?  //+NUM-CTRL-CODE-KANA+ROMA-SHIFT
            KBD_NUM_ROMA_CHR
            :  //+NUM-CTRL-CODE-KANA+ROMA+SHIFT
            KBD_NUM_ROMA_SHIFT_CHR
            )  //!SHIFT
           :  //+NUM-CTRL-CODE-KANA-ROMA
           (capsOff
            ?  //+NUM-CTRL-CODE-KANA-ROMA-CAPS
            (shiftOff
             ?  //+NUM-CTRL-CODE-KANA-ROMA-CAPS-SHIFT
             KBD_NUM_CHR
             :  //+NUM-CTRL-CODE-KANA-ROMA-CAPS+SHIFT
             KBD_NUM_SHIFT_CHR
             )  //!SHIFT
            :  //+NUM-CTRL-CODE-KANA-ROMA+CAPS
            (shiftOff
             ?  //+NUM-CTRL-CODE-KANA-ROMA+CAPS-SHIFT
             KBD_NUM_CAPS_CHR
             :  //+NUM-CTRL-CODE-KANA-ROMA+CAPS+SHIFT
             KBD_NUM_CAPS_SHIFT_CHR
             )  //!SHIFT
            )  //!CAPS
           )  //!OPT.1&&ROMA
          )  //!OPT.1&&KANA
         )  //CODE
        )  //CTRL
       );  //!NUM
    for (int i = 0; i < KBD_ALL_KEYS; i++) {
      int chr = 0;  //可変キーの文字。固定キーの文字は変化しないので0のままでよい
      if (i < KBD_VAR_KEYS) {  //可変キー
        chr = varChr[i] & 255;
        if (0xa6 <= chr && chr <= 0xdd && chr != 0xb0 &&  //記号を除くカタカナで
            zenOn && hiraOn) {  //全角とひらがなが両方点灯しているとき
          chr ^= 0x20;  //ひらがなにする
        } else if (chr == 0x5c && (MainMemory.mmrM8[0x00ed0059] & 0b00000001) != 0) {
          chr = 0x80;
        } else if (chr == 0x7e && (MainMemory.mmrM8[0x00ed0059] & 0b00000010) != 0) {
          chr = 0x81;
        } else if (chr == 0x7c && (MainMemory.mmrM8[0x00ed0059] & 0b00000100) != 0) {
          chr = 0x82;
        }
      }
      int stat = kbdStat[i] & (KBD_LIGHTED_MASK | KBD_PRESSED_MASK | KBD_POINTED_MASK);
      int characterStat = chr << 8 | stat;
      if (indexCharacterStat[i] != characterStat) {  //最後に描画した状態から変化した
        indexCharacterStat[i] = characterStat;
        int[] rect = indexRect[i];
        if (i < KBD_VAR_KEYS) {  //可変キー
          int[] bitmap = kbdVarStatBitmap[chr][stat];
          int p = rect[0] + width * rect[1];
          int w = rect[2];
          int h = rect[3];
          int q = 0;
          for (int dy = 0; dy < h; dy++) {
            for (int dx = 0; dx < w; dx++) {
              kbdBitmap[p++] = bitmap[q++];
            }
            p += width - w;
          }
        } else if (i < KBD_VAR_KEYS + KBD_LED_KEYS) {  //LEDキー
          int[] bitmap = kbdLedStatBitmap[i - KBD_VAR_KEYS][stat];
          if (rect.length >= 4) {  //Standard KeyboardはNUMキーがない
            int p = rect[0] + width * rect[1];
            int w = rect[2];
            int h = rect[3];
            int q = 0;
            for (int dy = 0; dy < h; dy++) {
              for (int dx = 0; dx < w; dx++) {
                kbdBitmap[p++] = bitmap[q++];
              }
              p += width - w;
            }
          }
        } else {  //固定キー
          int[] bitmap = statFixBitmap[stat];
          if (rect.length >= 4) {  //Compact Keyboardはテンキーがない
            int p = rect[0] + width * rect[1];
            int w = rect[2];
            int h = rect[3];
            for (int dy = 0; dy < h; dy++) {
              for (int dx = 0; dx < w; dx++) {
                kbdBitmap[p] = bitmap[p];
                p++;
              }
              p += width - w;
            }
            if (rect.length >= 8) {  //RETURNキーとSHIFTキー
              p = rect[4] + width * rect[5];
              w = rect[6];
              h = rect[7];
              for (int dy = 0; dy < h; dy++) {
                for (int dx = 0; dx < w; dx++) {
                  kbdBitmap[p] = bitmap[p];
                  p++;
                }
                p += width - w;
              }
            }
          }
        }
      }
    }
  }  //kbdFlush()

  //kbdHover (x, y)
  //  ポイントされているキーを変更する
  //  座標はキーボードのローカル座標で範囲内であること
  public static void kbdHover (int x, int y) {
    boolean flag = false;
    int i = kbdIndexMap[x + kbdWidth * y];  //今回ポイントされたキー。-1=キーがポイントされていない
    if (kbdPointedIndex != i) {  //ポイントされているキーが変わった
      if (kbdPointedIndex >= 0) {  //前回ポイントされていたキーがあった
        int stat = kbdStat[kbdPointedIndex];
        kbdStat[kbdPointedIndex] = stat &= ~KBD_POINTED_MASK;  //ポイントを外す
        if ((stat & KBD_MOUSEDOWN_MASK) != 0) {  //マウスで押されていた
          kbdStat[kbdPointedIndex] = stat &= ~KBD_MOUSEDOWN_MASK;  //マウスを離す
          if ((stat & (KBD_LOCKED_MASK | KBD_KEYDOWN_MASK | KBD_MOUSEDOWN_MASK)) == 0) {  //ロックされておらずキーもマウスも押されていない
            kbdStat[kbdPointedIndex] = stat &= ~KBD_PRESSED_MASK;  //離す
            kbdRelease (kbdPointedIndex);
          }
        }
      }
      kbdPointedIndex = i;
      flag = true;
    }
    if (kbdPointedIndex >= 0) {  //ポイントされているキーがある
      int stat = kbdStat[kbdPointedIndex];
      kbdStat[kbdPointedIndex] = stat |= KBD_POINTED_MASK;  //ポイントする。前回もポイントされていた場合もある
      if (((stat & KBD_MOUSEDOWN_MASK) != 0) != (XEiJ.musButtonLeft || XEiJ.musButtonRight)) {  //マウスの押し下げ状態が変わった
        if (XEiJ.musButtonLeft && (kbdPointedIndex == KBD_SHIFT_INDEX ||  //マウスで押されていなかったSHIFTキーが左ボタンで押された
                                   kbdPointedIndex == KBD_CTRL_INDEX) ||  //マウスで押されていなかったCTRLキーが左ボタンで押された
            XEiJ.musButtonRight) {  //マウスで押されていなかったキーが右ボタンで押された
          stat ^= KBD_LOCKED_MASK;  //ロックを反転する
          kbdStat[kbdPointedIndex] = stat |= KBD_MOUSEDOWN_MASK;  //マウスで押す
          if ((stat & KBD_PRESSED_MASK) == 0) {  //押されていなかった
            kbdStat[kbdPointedIndex] = stat |= KBD_PRESSED_MASK;  //押す
            kbdPress (kbdPointedIndex);
          }
        } else if (XEiJ.musButtonLeft) {  //マウスで押されていなかったキーが左ボタンで押された
          stat &= ~KBD_LOCKED_MASK;  //ロックを解除する
          kbdStat[kbdPointedIndex] = stat |= KBD_MOUSEDOWN_MASK;  //マウスで押す
          if ((stat & KBD_PRESSED_MASK) == 0) {  //押されていなかった
            kbdStat[kbdPointedIndex] = stat |= KBD_PRESSED_MASK;  //押す
            kbdPress (kbdPointedIndex);
          }
        } else {  //マウスで押されていたキーが離された
          kbdStat[kbdPointedIndex] = stat &= ~KBD_MOUSEDOWN_MASK;  //マウスを離す
          if ((stat & (KBD_LOCKED_MASK | KBD_KEYDOWN_MASK | KBD_MOUSEDOWN_MASK)) == 0) {  //ロックされておらずキーもマウスも押されていない
            kbdStat[kbdPointedIndex] = stat &= ~KBD_PRESSED_MASK;  //離す
            kbdRelease (kbdPointedIndex);
          }
        }
        flag = true;
      }
    }
    if (flag) {
      kbdRepaint ();
    }
  }  //kbdHover(int,int)

  //kbdInput (ke, pressed)
  //  リアルキーボードから入力する
  public static void kbdInput (KeyEvent ke, boolean pressed) {
    int keyCode = ke.getKeyCode ();
    if (pressed) {
      if (keyCode == KeyEvent.VK_F11) {  //F11
        //if ((ke.getModifiersEx () & InputEvent.SHIFT_DOWN_MASK) == InputEvent.SHIFT_DOWN_MASK) {  //Shift+F11
        //}
        if (kbdF11Mode == KBD_F11_FULLSCREEN) {
          //全画面表示の切り替え
          if (XEiJ.mnbFullscreenMenuItem.isEnabled ()) {  //全画面表示に移行できて全画面表示の切り替えがJavaScriptに渡っていない
            if (!XEiJ.mnbFullscreenMenuItem.isSelected ()) {  //全画面表示になっていない
              XEiJ.mnbFullscreenMenuItem.doClick ();
            } else if (XEiJ.pnlPrevFitInWindowOn) {  //全画面表示になっていて全画面表示に移行する前にウインドウに合わせていた
              XEiJ.mnbFitInWindowMenuItem.doClick ();
            } else {  //全画面表示になっていて全画面表示に移行する前にウインドウに合わせていなかった
              XEiJ.mnbFixedScaleMenuItem.doClick ();
            }
          }
        } else if (kbdF11Mode == KBD_F11_SCREENSHOT) {
          kbdDoCapture ();
        } else if (kbdF11Mode == KBD_F11_STOPANDSTART) {
          if (XEiJ.mpuTask == null) {  //Run
            XEiJ.mpuStart ();
          } else {  //Stop
            if (RootPointerList.RTL_ON) {
              if (RootPointerList.rtlCurrentSupervisorTaskIsStoppable ||
                  RootPointerList.rtlCurrentUserTaskIsStoppable) {
                XEiJ.mpuStop (null);
              }
            } else {
              XEiJ.mpuStop (null);
            }
          }
        }
        return;
      } else if (keyCode == KeyEvent.VK_F12) {  //F12
        XEiJ.musSetSeamlessOn (!XEiJ.musSeamlessOn);  //シームレス/エクスクルーシブを切り替える
        return;
      } else if ((ke.getModifiersEx () & XEiJ.MNB_MODIFIERS) == XEiJ.MNB_MODIFIERS) {  //Alt+Shift
        switch (keyCode) {
        case KeyEvent.VK_C:  //Alt+Shift-C
          XEiJ.mnbConfigMenu.doClick ();
          return;
        case KeyEvent.VK_D:  //Alt+Shift-D
          XEiJ.mnbDisplayMenu.doClick ();
          return;
        case KeyEvent.VK_F:  //Alt+Shift-F
          XEiJ.mnbFileMenu.doClick ();
          return;
        case KeyEvent.VK_I:  //Alt+Shift-I
          XEiJ.mnbInputMenu.doClick ();
          return;
        case KeyEvent.VK_K:  //Alt+Shift-K
          if (kbdOn) {
            XEiJ.mnbNoKeyboardMenuItem.doClick ();
          } else if (kbdType == KBD_COMPACT_TYPE) {
            XEiJ.mnbCompactKeyboardMenuItem.doClick ();
          } else {
            XEiJ.mnbStandardKeyboardMenuItem.doClick ();
          }
          return;
        case KeyEvent.VK_L:  //Alt+Shift-L
          XEiJ.mnbLanguageMenu.doClick ();
          return;
        case KeyEvent.VK_M:  //Alt+Shift-M
          XEiJ.mpuMenu.doClick ();
          return;
        case KeyEvent.VK_O:  //Alt+Shift-O
          XEiJ.mpuOpt1ResetMenuItem.doClick ();
          return;
        case KeyEvent.VK_P:  //Alt+Shift-P
          if (XEiJ.mnbPlayMenuItem.isEnabled ()) {
            XEiJ.mnbPlayMenuItem.doClick ();
          }
          return;
        case KeyEvent.VK_Q:  //Alt+Shift-Q
          if (XEiJ.mnbQuitMenuItem.isEnabled ()) {
            XEiJ.mnbQuitMenuItem.doClick ();
          }
          return;
        case KeyEvent.VK_R:  //Alt+Shift-R
          XEiJ.mpuResetMenuItem.doClick ();
          return;
        case KeyEvent.VK_S:  //Alt+Shift-S
          if (XEiJ.mnbSoundMenu.isEnabled ()) {
            XEiJ.mnbSoundMenu.doClick ();
          }
          return;
        case KeyEvent.VK_V:  //Alt+Shift-V
          if (XEiJ.mnbPasteMenuItem.isEnabled ()) {
            XEiJ.mnbPasteMenuItem.doClick ();
          }
          return;
        case KeyEvent.VK_W:  //Alt+Shift-W
          if (XEiJ.mnbFitInWindowMenuItem.isEnabled ()) {  //ウインドウに合わせられるとき
            XEiJ.mnbFitInWindowMenuItem.doClick ();
          }
          return;
        case KeyEvent.VK_X:  //Alt+Shift-X
          XEiJ.mnbFixedScaleMenuItem.doClick ();
          return;
        }
      }
    }
    if (PPI.ppiInput (ke, pressed)) {  //JOYKEY
      return;
    }
    {
      int keyLocation = ke.getKeyLocation ();
      int rawCode = kbdGetRawCode (ke);
      //int keyCode = ke.getKeyCode ();
      int locationMask = 1 << keyLocation;
      for (int xo = 0; xo < KBD_ALL_KEYS; xo++) {
        if (kbdCurrentMap[3 * xo + 1] == rawCode &&  //rawCodeが一致
            (kbdCurrentMap[3 * xo] & locationMask) != 0) {  //keyLocationが一致
          int eo = KBD_XO_TO_EO[xo];
          int stat = kbdStat[eo];
          if (pressed) {  //キーが押された
            if ((stat & KBD_KEYDOWN_MASK) == 0) {  //キーで押されていなかった
              kbdStat[eo] = stat |= KBD_KEYDOWN_MASK;  //キーで押す
              if ((stat & KBD_PRESSED_MASK) == 0) {  //押されていなかった
                kbdStat[eo] = stat |= KBD_PRESSED_MASK;  //押す
                kbdPress (eo);
                kbdRepaint ();
              }
            }
          } else {  //キーが離された
            if ((stat & KBD_KEYDOWN_MASK) != 0) {  //キーで押されていた
              kbdStat[eo] = stat &= ~KBD_KEYDOWN_MASK;  //キーを離す
              if ((stat & (KBD_LOCKED_MASK | KBD_KEYDOWN_MASK | KBD_MOUSEDOWN_MASK)) == 0) {  //ロックされておらずキーもマウスも押されていない
                kbdStat[eo] = stat &= ~KBD_PRESSED_MASK;  //離す
                kbdRelease (eo);
                kbdRepaint ();
              }
            }
          }
          break;
        }
      }
    }
  }  //kbdInput(KeyEvent,boolean)

  //rawCode = kbdGetRawCode (ke)
  //  KeyEventからrawCodeを取り出す
  public static int kbdGetRawCode (KeyEvent ke) {
    int rawCode = 0;
    if (KBD_ACCESS_RAW_CODE && kbdRawCodeAccessible) {  //KeyEvent.rawCodeにアクセスできるとき
      //KeyEvent.rawCodeを直接読み出す
      try {
        rawCode = (int) kbdRawCodeField.getLong (ke);
      } catch (Exception e) {
      }
    } else {  //KeyEvent.rawCodeにアクセスできないとき
      //KeyEvent.toString()で出力される文字列の中からrawCode=～を取り出す
      String s = ke.toString ();
      int i = s.indexOf ("rawCode=");
      if (0 <= i) {
        i += 8;
        for (int k = s.length (); i < k; i++) {
          char c = s.charAt (i);
          if (c < '0' || '9' < c) {
            break;
          }
          rawCode = rawCode * 10 + (c - '0');
        }
      }
    }
    return rawCode;
  }

  //kbdRePress()
  //  押されているキーを再入力する
  //  リセット後キー入力割り込みが許可されたら直ちに押されているキーを入力しないとOPT.2キーをロックしていてもROMデバッガが起動しない
  public static void kbdRePress () {
    for (int i = 0; i < KBD_ALL_KEYS; i++) {
      if (i != KBD_NUM_INDEX &&  //NUMキーは除く
          (kbdStat[i] & KBD_PRESSED_MASK) != 0) {  //押されている
        //キーコードを求める
        int code = ((kbdStat[KBD_NUM_INDEX] & KBD_LIGHTED_MASK) != 0 ? KBD_NUM_CODE : KBD_NORMAL_CODE)[i] & 0x7f;
        //MFPに転送する
        if (MC68901.MFP_KBD_ON) {
          MC68901.mfpKbdInput (code, true);
        } else {
          kbdTransfer (code, true);
        }
      }
    }
  }  //kbdRePress()

  //kbdPress (i)
  //  番号iのキーが押された
  public static void kbdPress (int i) {
    //キーコードを求める
    int code = ((kbdStat[KBD_NUM_INDEX] & KBD_LIGHTED_MASK) != 0 ? KBD_NUM_CODE : KBD_NORMAL_CODE)[i] & 0x7f;
    //NUMキーが押されたらLEDを反転する
    if (KBD_DEBUG_LED ? KBD_KANA_INDEX <= i && i <= KBD_NUM_INDEX : i == KBD_NUM_INDEX) {
      kbdStat[i] ^= KBD_LIGHTED_MASK;
      kbdRepaint ();
    }
    //MFPに転送する
    if (MC68901.MFP_KBD_ON) {
      MC68901.mfpKbdInput (code, true);
    } else {
      kbdTransfer (code, true);
    }
  }  //kbdPress(int)

  //kbdRelease (i)
  //  番号iのキーが離された
  public static void kbdRelease (int i) {
    //キーコードを求める
    int code = ((kbdStat[KBD_NUM_INDEX] & KBD_LIGHTED_MASK) != 0 ? KBD_NUM_CODE : KBD_NORMAL_CODE)[i] & 0x7f | 0x80;
    //MFPに転送する
    if (MC68901.MFP_KBD_ON) {
      MC68901.mfpKbdInput (code, false);
    } else {
      kbdTransfer (code, false);
    }
  }  //kbdRelease(int)

  //kbdTransfer (code, repeat)
  //  キーコードをMFPに転送する
  //  割り込み要求が競合しないようにコアのスレッドで行う
  public static void kbdTransfer (int code, boolean repeat) {
    XEiJ.tmrTimer.schedule (new KBDTask (code, repeat), 0L);
  }  //kbdTransfer(int)

  //kbdSetRepeatDelay (n)
  //  リピートディレイ
  //  200+100*n(ms)
  public static void kbdSetRepeatDelay (int n) {
    n &= 15;
    kbdRepeatDelay = 200 + 100 * n;
  }  //kbdSetRepeatDelay(int)

  //kbdSetRepeatInterval (n)
  //  リピートインターバル
  //  30+5*n^2(ms)
  public static void kbdSetRepeatInterval (int n) {
    n &= 15;
    kbdRepeatInterval = 30 + 5 * n * n;
  }  //kbdSetRepeatInterval(int)

  //kbdSetLedStatus (n)
  //  LEDキーのLEDの状態をまとめて設定する。各ビット0=点灯,1=消灯
  //  bit0  かな
  //  bit1  ローマ字
  //  bit2  コード入力
  //  bit3  CAPS
  //  bit4  INS
  //  bit5  ひらがな
  //  bit6  全角
  public static void kbdSetLedStatus (int n) {
    boolean flag = false;
    n = ~n;
    for (int i = KBD_KANA_INDEX; i <= KBD_ZEN_INDEX; i++) {
      int mask = (n & 1) << KBD_LIGHTED_BIT;
      int stat = kbdStat[i];
      if ((stat & KBD_LIGHTED_MASK) != mask) {
        kbdStat[i] = stat ^ KBD_LIGHTED_MASK;
        flag = true;
      }
      n >>= 1;
    }
    if (flag) {
      kbdRepaint ();
    }
  }  //kbdSetLedStatus(int)

  //kbdDoCapture ()
  //  キャプチャ実行
  //  カレントディレクトリのcaptureというサブディレクトリの中に1.png,2.png,...を作るだけ
  public static void kbdDoCapture () {
    try {
      File dir = new File ("capture");
      if (dir.exists ()) {  //captureがある
        if (!dir.isDirectory ()) {  //captureがあるがディレクトリでない
          return;
        }
      } else {  //captureがない
        if (!dir.mkdir ()) {  //captureがなくてディレクトリが作れない
          return;
        }
      }
      HashSet<String> nameSet = new HashSet<String> ();
      for (String name : dir.list ()) {
        nameSet.add (name);
      }
      int number = 0;
      for (String key : Settings.SGS_FRAME_KEYS) {
        BufferedImage image = RestorableFrame.rfmCapture (key);
        if (image != null) {  //ウインドウが開いていてキャプチャできた
          String name;
          do {
            number++;
            name = number + ".png";
          } while (!nameSet.add (name));  //セットに追加できるまで繰り返す
          name = "capture/" + name;
          if (XEiJ.saveImage (image, name)) {  //成功
            System.out.println (Multilingual.mlnJapanese ? name + " を更新しました" : name + " was updated");
          }
        }
      }
    } catch (Exception e) {
    }
  }  //kbdDoCapture()



  //$$KBT キー入力タスク
  public static class KBDTask extends TimerTask {
    public int code;
    public boolean repeat;
    public KBDTask (int c, boolean r) {
      code = c;
      repeat = r;
    }  //KBDTask(int,boolean)
    @Override public void run () {
      if (kbdRepeatTask != null) {
        //リピートを終了する
        //  開始と終了は同じスレッドで行わなければならない
        kbdRepeatTask.cancel ();
        kbdRepeatTask = null;
      }
      //MFPに転送する
      MC68901.mfpKeyboardInput (code);
      if (repeat) {
        //リピートを開始する
        XEiJ.tmrTimer.schedule (kbdRepeatTask = new KBDRepeatTask (code), (long) kbdRepeatDelay, (long) kbdRepeatInterval);
      }
    }  //run()
  }  //class KBDTask



  //$$KBR キーリピートタスク
  public static class KBDRepeatTask extends TimerTask {
    public int code;
    public KBDRepeatTask (int c) {
      code = c;
    }  //KBDRepeatTask(int)
    @Override public void run () {
      //MFPに転送する
      MC68901.mfpKeyboardInput (code);
    }  //run()
  }  //class KBDRepeatTask



}  //class Keyboard



